// Copyright (c) 2017 Intel Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef __VM_SHARED_OBJECT_H__
#define __VM_SHARED_OBJECT_H__

#include "vm_types.h"
#include "vm_strings.h"

typedef void* vm_so_handle;
typedef void (*vm_so_func)(void);

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

vm_so_handle vm_so_load(const vm_char* so_file_name);
vm_so_func   vm_so_get_addr(vm_so_handle so_handle, const char* so_func_name);
void         vm_so_free(vm_so_handle so_handle);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#ifndef NULL
#define NULL (void*)0L
#endif

#endif /* __VM_SHARED_OBJECT_H__ */
