// Copyright (c) 2018 Intel Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
#include "../include/tree.h"
#include "../include/asc_defs.h"

bool SCDetect1(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffTSC < 7) {
        if (diffMVdiffVal < 53) {
            if (gchDC < 14) {
                if (ssDCval < 21) {
                    if (RsCsDiff < 251) {
                        if (diffRsCsdiff < 5) {
                            if (diffMVdiffVal < 2) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 2) {
                                if (MVDiff < 10) {
                                    return 1;
                                }
                                else if (MVDiff >= 10) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 5) {
                            if (MVDiff < 7) {
                                return 1;
                            }
                            else if (MVDiff >= 7) {
                                return 0;
                            }
                        }
                    }
                    else if (RsCsDiff >= 251) {
                        return 1;
                    }
                }
                else if (ssDCval >= 21) {
                    if (gchDC < 5) {
                        if (MVDiff < 55) {
                            return 0;
                        }
                        else if (MVDiff >= 55) {
                            if (refDCval < 24) {
                                if (ssDCval < 24) {
                                    return 0;
                                }
                                else if (ssDCval >= 24) {
                                    if (diffTSC < -4) {
                                        return 0;
                                    }
                                    else if (diffTSC >= -4) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 24) {
                                if (RsCsDiff < 2455) {
                                    if (TSC < 11) {
                                        return 0;
                                    }
                                    else if (TSC >= 11) {
                                        if (refDCval < 43) {
                                            if (Rs < 645) {
                                                return 1;
                                            }
                                            else if (Rs >= 645) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 43) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 2455) {
                                    if (Rs < 2357) {
                                        return 1;
                                    }
                                    else if (Rs >= 2357) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (gchDC >= 5) {
                        if (Rs < 94) {
                            if (Rs < 93) {
                                if (refDCval < 227) {
                                    if (diffAFD < 8) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 8) {
                                        if (Rs < 16) {
                                            return 1;
                                        }
                                        else if (Rs >= 16) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 227) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 93) {
                                if (SC < 293) {
                                    return 0;
                                }
                                else if (SC >= 293) {
                                    return 1;
                                }
                            }
                        }
                        else if (Rs >= 94) {
                            if (RsDiff < 32) {
                                if (diffTSC < 6) {
                                    return 0;
                                }
                                else if (diffTSC >= 6) {
                                    if (diffMVdiffVal < 46) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 46) {
                                        if (RsCsDiff < 30) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 30) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 32) {
                                if (posBalance < 64) {
                                    return 0;
                                }
                                else if (posBalance >= 64) {
                                    if (Cs < 2814) {
                                        return 1;
                                    }
                                    else if (Cs >= 2814) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (gchDC >= 14) {
                if (AFD < 62) {
                    if (ssDCval < 40) {
                        if (diffAFD < 13) {
                            if (negBalance < 86) {
                                if (TSC < 9) {
                                    return 0;
                                }
                                else if (TSC >= 9) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 86) {
                                return 0;
                            }
                        }
                        else if (diffAFD >= 13) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 40) {
                        if (RsDiff < 75) {
                            if (Rs < 408) {
                                if (diffMVdiffVal < -233) {
                                    if (CsDiff < 6) {
                                        if (negBalance < 125) {
                                            return 0;
                                        }
                                        else if (negBalance >= 125) {
                                            if (AFD < 1) {
                                                return 1;
                                            }
                                            else if (AFD >= 1) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 6) {
                                        if (diffTSC < -4) {
                                            return 0;
                                        }
                                        else if (diffTSC >= -4) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= -233) {
                                    if (AFD < 38) {
                                        return 0;
                                    }
                                    else if (AFD >= 38) {
                                        if (SC < 169) {
                                            return 1;
                                        }
                                        else if (SC >= 169) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 408) {
                                if (AFD < 59) {
                                    return 0;
                                }
                                else if (AFD >= 59) {
                                    if (Rs < 938) {
                                        return 0;
                                    }
                                    else if (Rs >= 938) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 75) {
                            return 1;
                        }
                    }
                }
                else if (AFD >= 62) {
                    if (Cs < 1324) {
                        return 0;
                    }
                    else if (Cs >= 1324) {
                        return 1;
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 53) {
            if (diffAFD < 8) {
                if (gchDC < 23) {
                    if (ssDCval < 38) {
                        if (diffMVdiffVal < 187) {
                            if (diffRsCsdiff < 31) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 31) {
                                if (diffMVdiffVal < 129) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= 129) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 187) {
                            if (AFD < 10) {
                                if (diffTSC < 4) {
                                    return 0;
                                }
                                else if (diffTSC >= 4) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 10) {
                                if (Cs < 376) {
                                    if (AFD < 14) {
                                        return 1;
                                    }
                                    else if (AFD >= 14) {
                                        if (CsDiff < 4) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 4) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Cs >= 376) {
                                    if (CsDiff < 7) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 7) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 38) {
                        if (AFD < 34) {
                            if (CsDiff < 41) {
                                if (ssDCval < 41) {
                                    if (diffTSC < 4) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 4) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 41) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 41) {
                                return 1;
                            }
                        }
                        else if (AFD >= 34) {
                            if (diffAFD < -1) {
                                if (MVDiff < 270) {
                                    if (SC < 3607) {
                                        return 1;
                                    }
                                    else if (SC >= 3607) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 270) {
                                    if (diffTSC < 4) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 4) {
                                        if (Rs < 667) {
                                            return 0;
                                        }
                                        else if (Rs >= 667) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= -1) {
                                return 0;
                            }
                        }
                    }
                }
                else if (gchDC >= 23) {
                    if (AFD < 58) {
                        if (SC < 2) {
                            if (diffAFD < -1) {
                                return 0;
                            }
                            else if (diffAFD >= -1) {
                                return 1;
                            }
                        }
                        else if (SC >= 2) {
                            if (diffTSC < -24) {
                                if (Rs < 204) {
                                    return 1;
                                }
                                else if (Rs >= 204) {
                                    if (Cs < 1574) {
                                        return 0;
                                    }
                                    else if (Cs >= 1574) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffTSC >= -24) {
                                if (RsCsDiff < 181) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 181) {
                                    if (RsCsDiff < 183) {
                                        return 1;
                                    }
                                    else if (RsCsDiff >= 183) {
                                        if (refDCval < 73) {
                                            if (Rs < 1248) {
                                                return 1;
                                            }
                                            else if (Rs >= 1248) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 73) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 58) {
                        if (Cs < 910) {
                            return 0;
                        }
                        else if (Cs >= 910) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffAFD >= 8) {
                if (ssDCval < 42) {
                    if (Cs < 162) {
                        if (diffMVdiffVal < 348) {
                            if (SC < 58) {
                                if (SC < 42) {
                                    if (diffMVdiffVal < 201) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 201) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 42) {
                                    return 0;
                                }
                            }
                            else if (SC >= 58) {
                                if (TSC < 11) {
                                    return 1;
                                }
                                else if (TSC >= 11) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 348) {
                            return 1;
                        }
                    }
                    else if (Cs >= 162) {
                        if (SC < 368) {
                            return 0;
                        }
                        else if (SC >= 368) {
                            if (diffAFD < 13) {
                                if (refDCval < 36) {
                                    return 1;
                                }
                                else if (refDCval >= 36) {
                                    if (MVDiff < 188) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 188) {
                                        if (MVDiff < 359) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 359) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 13) {
                                return 1;
                            }
                        }
                    }
                }
                else if (ssDCval >= 42) {
                    if (Rs < 106) {
                        if (RsCsDiff < 17) {
                            if (refDCval < 48) {
                                if (diffAFD < 10) {
                                    return 0;
                                }
                                else if (diffAFD >= 10) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 48) {
                                return 0;
                            }
                        }
                        else if (RsCsDiff >= 17) {
                            if (diffRsCsdiff < -1) {
                                if (Cs < 188) {
                                    return 1;
                                }
                                else if (Cs >= 188) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= -1) {
                                return 1;
                            }
                        }
                    }
                    else if (Rs >= 106) {
                        if (MVDiff < 426) {
                            if (SC < 1355) {
                                if (diffRsCsdiff < 17) {
                                    if (MVDiff < 396) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 396) {
                                        if (negBalance < 60) {
                                            return 0;
                                        }
                                        else if (negBalance >= 60) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 17) {
                                    if (diffTSC < 6) {
                                        if (negBalance < 73) {
                                            if (negBalance < 41) {
                                                return 0;
                                            }
                                            else if (negBalance >= 41) {
                                                if (CsDiff < 5) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 5) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 73) {
                                            if (Cs < 227) {
                                                return 0;
                                            }
                                            else if (Cs >= 227) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 6) {
                                        return 1;
                                    }
                                }
                            }
                            else if (SC >= 1355) {
                                return 0;
                            }
                        }
                        else if (MVDiff >= 426) {
                            if (diffRsCsdiff < 61) {
                                if (refDCval < 165) {
                                    return 0;
                                }
                                else if (refDCval >= 165) {
                                    if (SC < 684) {
                                        return 1;
                                    }
                                    else if (SC >= 684) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 61) {
                                if (SC < 2551) {
                                    if (AFD < 64) {
                                        return 1;
                                    }
                                    else if (AFD >= 64) {
                                        if (Rs < 766) {
                                            return 1;
                                        }
                                        else if (Rs >= 766) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (SC >= 2551) {
                                    if (ssDCval < 92) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 92) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffTSC >= 7) {
        if (MVDiff < 52) {
            if (refDCval < 34) {
                if (diffRsCsdiff < 73) {
                    if (TSC < 17) {
                        if (diffMVdiffVal < -70) {
                            if (Rs < 402) {
                                return 1;
                            }
                            else if (Rs >= 402) {
                                return 0;
                            }
                        }
                        else if (diffMVdiffVal >= -70) {
                            return 0;
                        }
                    }
                    else if (TSC >= 17) {
                        return 1;
                    }
                }
                else if (diffRsCsdiff >= 73) {
                    return 1;
                }
            }
            else if (refDCval >= 34) {
                if (diffTSC < 13) {
                    if (AFD < 12) {
                        if (RsDiff < 24) {
                            return 0;
                        }
                        else if (RsDiff >= 24) {
                            return 1;
                        }
                    }
                    else if (AFD >= 12) {
                        return 0;
                    }
                }
                else if (diffTSC >= 13) {
                    if (CsDiff < 5) {
                        return 1;
                    }
                    else if (CsDiff >= 5) {
                        if (MVDiff < 42) {
                            if (RsDiff < 178) {
                                return 0;
                            }
                            else if (RsDiff >= 178) {
                                return 1;
                            }
                        }
                        else if (MVDiff >= 42) {
                            if (negBalance < 32) {
                                return 0;
                            }
                            else if (negBalance >= 32) {
                                if (Cs < 5540) {
                                    return 1;
                                }
                                else if (Cs >= 5540) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 52) {
            if (diffRsCsdiff < 72) {
                if (diffMVdiffVal < 88) {
                    if (AFD < 66) {
                        if (Rs < 478) {
                            if (MVDiff < 344) {
                                if (gchDC < 5) {
                                    if (negBalance < 61) {
                                        if (posBalance < 49) {
                                            return 1;
                                        }
                                        else if (posBalance >= 49) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 61) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 5) {
                                    if (diffAFD < 20) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 20) {
                                        if (ssDCval < 158) {
                                            if (RsDiff < 7) {
                                                if (Rs < 208) {
                                                    if (Cs < 191) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 191) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 208) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsDiff >= 7) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 158) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 344) {
                                if (SC < 412) {
                                    return 0;
                                }
                                else if (SC >= 412) {
                                    if (AFD < 31) {
                                        return 1;
                                    }
                                    else if (AFD >= 31) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 478) {
                            return 0;
                        }
                    }
                    else if (AFD >= 66) {
                        return 1;
                    }
                }
                else if (diffMVdiffVal >= 88) {
                    if (gchDC < 15) {
                        if (diffAFD < 13) {
                            if (gchDC < 10) {
                                if (MVDiff < 210) {
                                    return 1;
                                }
                                else if (MVDiff >= 210) {
                                    if (gchDC < 4) {
                                        return 1;
                                    }
                                    else if (gchDC >= 4) {
                                        return 0;
                                    }
                                }
                            }
                            else if (gchDC >= 10) {
                                return 0;
                            }
                        }
                        else if (diffAFD >= 13) {
                            if (RsCsDiff < 31) {
                                return 1;
                            }
                            else if (RsCsDiff >= 31) {
                                if (diffRsCsdiff < 43) {
                                    if (refDCval < 40) {
                                        return 0;
                                    }
                                    else if (refDCval >= 40) {
                                        if (Cs < 638) {
                                            return 1;
                                        }
                                        else if (Cs >= 638) {
                                            if (AFD < 35) {
                                                return 1;
                                            }
                                            else if (AFD >= 35) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 43) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 15) {
                        if (diffAFD < 23) {
                            if (TSC < 11) {
                                if (Cs < 263) {
                                    return 1;
                                }
                                else if (Cs >= 263) {
                                    if (MVDiff < 112) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 112) {
                                        return 0;
                                    }
                                }
                            }
                            else if (TSC >= 11) {
                                if (ssDCval < 54) {
                                    if (RsCsDiff < 17) {
                                        if (negBalance < 18) {
                                            return 1;
                                        }
                                        else if (negBalance >= 18) {
                                            return 0;
                                        }
                                    }
                                    else if (RsCsDiff >= 17) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 54) {
                                    if (AFD < 63) {
                                        if (refDCval < 34) {
                                            return 1;
                                        }
                                        else if (refDCval >= 34) {
                                            if (diffRsCsdiff < 24) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 24) {
                                                if (posBalance < 101) {
                                                    if (SC < 1213) {
                                                        if (MVDiff < 257) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 257) {
                                                            if (Cs < 404) {
                                                                return 0;
                                                            }
                                                            else if (Cs >= 404) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (SC >= 1213) {
                                                        if (MVDiff < 208) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 208) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 101) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 63) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 23) {
                            if (negBalance < 25) {
                                if (posBalance < 99) {
                                    return 0;
                                }
                                else if (posBalance >= 99) {
                                    if (TSC < 32) {
                                        if (negBalance < 7) {
                                            if (Rs < 68) {
                                                if (RsCsDiff < 13) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 13) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 68) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 7) {
                                            if (MVDiff < 199) {
                                                if (SC < 606) {
                                                    return 0;
                                                }
                                                else if (SC >= 606) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 199) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (TSC >= 32) {
                                        if (AFD < 69) {
                                            return 0;
                                        }
                                        else if (AFD >= 69) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 25) {
                                if (gchDC < 24) {
                                    return 1;
                                }
                                else if (gchDC >= 24) {
                                    if (TSC < 20) {
                                        return 1;
                                    }
                                    else if (TSC >= 20) {
                                        if (CsDiff < 9) {
                                            if (Cs < 76) {
                                                return 0;
                                            }
                                            else if (Cs >= 76) {
                                                if (TSC < 27) {
                                                    return 1;
                                                }
                                                else if (TSC >= 27) {
                                                    if (Cs < 597) {
                                                        if (TSC < 46) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 46) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (Cs >= 597) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 9) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 72) {
                if (diffMVdiffVal < 58) {
                    if (posBalance < 84) {
                        if (ssDCval < 102) {
                            if (diffTSC < 20) {
                                if (refDCval < 96) {
                                    if (TSC < 19) {
                                        if (diffRsCsdiff < 151) {
                                            if (Rs < 336) {
                                                return 1;
                                            }
                                            else if (Rs >= 336) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 151) {
                                            if (ssDCval < 81) {
                                                if (diffRsCsdiff < 314) {
                                                    if (RsCsDiff < 191) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 191) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 314) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 81) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (TSC >= 19) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 96) {
                                    if (RsCsDiff < 146) {
                                        return 1;
                                    }
                                    else if (RsCsDiff >= 146) {
                                        if (Rs < 2539) {
                                            return 0;
                                        }
                                        else if (Rs >= 2539) {
                                            if (Cs < 2987) {
                                                return 1;
                                            }
                                            else if (Cs >= 2987) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 20) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 102) {
                            if (posBalance < 60) {
                                if (Rs < 954) {
                                    return 1;
                                }
                                else if (Rs >= 954) {
                                    return 0;
                                }
                            }
                            else if (posBalance >= 60) {
                                if (Rs < 1737) {
                                    if (Cs < 907) {
                                        return 1;
                                    }
                                    else if (Cs >= 907) {
                                        return 0;
                                    }
                                }
                                else if (Rs >= 1737) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (posBalance >= 84) {
                        if (diffRsCsdiff < 501) {
                            if (Cs < 411) {
                                return 1;
                            }
                            else if (Cs >= 411) {
                                if (ssDCval < 101) {
                                    return 0;
                                }
                                else if (ssDCval >= 101) {
                                    if (ssDCval < 119) {
                                        if (MVDiff < 70) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 70) {
                                            if (SC < 1875) {
                                                return 0;
                                            }
                                            else if (SC >= 1875) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 119) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 501) {
                            if (refDCval < 68) {
                                return 0;
                            }
                            else if (refDCval >= 68) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 58) {
                    if (diffAFD < 17) {
                        if (Rs < 1321) {
                            if (diffMVdiffVal < 70) {
                                if (AFD < 18) {
                                    return 1;
                                }
                                else if (AFD >= 18) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 70) {
                                if (posBalance < 100) {
                                    if (ssDCval < 150) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 150) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 100) {
                                    return 0;
                                }
                            }
                        }
                        else if (Rs >= 1321) {
                            if (Cs < 365) {
                                return 1;
                            }
                            else if (Cs >= 365) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 17) {
                        if (negBalance < 17) {
                            if (SC < 1982) {
                                if (TSC < 12) {
                                    return 0;
                                }
                                else if (TSC >= 12) {
                                    return 1;
                                }
                            }
                            else if (SC >= 1982) {
                                if (RsDiff < 10) {
                                    return 0;
                                }
                                else if (RsDiff >= 10) {
                                    if (diffAFD < 20) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 20) {
                                        if (TSC < 29) {
                                            if (TSC < 17) {
                                                return 0;
                                            }
                                            else if (TSC >= 17) {
                                                if (SC < 2643) {
                                                    if (SC < 2477) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 2477) {
                                                        return 0;
                                                    }
                                                }
                                                else if (SC >= 2643) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (TSC >= 29) {
                                            if (diffAFD < 39) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 39) {
                                                if (AFD < 58) {
                                                    return 1;
                                                }
                                                else if (AFD >= 58) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 17) {
                            if (diffAFD < 26) {
                                if (Rs < 619) {
                                    return 1;
                                }
                                else if (Rs >= 619) {
                                    if (CsDiff < 10) {
                                        if (MVDiff < 145) {
                                            if (diffTSC < 20) {
                                                return 1;
                                            }
                                            else if (diffTSC >= 20) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 145) {
                                            if (AFD < 30) {
                                                return 1;
                                            }
                                            else if (AFD >= 30) {
                                                if (diffAFD < 22) {
                                                    if (Cs < 626) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 626) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffAFD >= 22) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 10) {
                                        if (gchDC < 20) {
                                            if (ssDCval < 146) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 146) {
                                                if (Cs < 3727) {
                                                    return 0;
                                                }
                                                else if (Cs >= 3727) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 20) {
                                            if (Rs < 716) {
                                                return 0;
                                            }
                                            else if (Rs >= 716) {
                                                if (AFD < 42) {
                                                    if (Rs < 1849) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1849) {
                                                        return 0;
                                                    }
                                                }
                                                else if (AFD >= 42) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 26) {
                                if (diffMVdiffVal < 80) {
                                    if (RsDiff < 16) {
                                        if (Cs < 635) {
                                            return 1;
                                        }
                                        else if (Cs >= 635) {
                                            if (MVDiff < 74) {
                                                if (Rs < 1680) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1680) {
                                                    return 0;
                                                }
                                            }
                                            else if (MVDiff >= 74) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 16) {
                                        if (ssDCval < 155) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 155) {
                                            if (Rs < 7496) {
                                                return 1;
                                            }
                                            else if (Rs >= 7496) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 80) {
                                    if (diffAFD < 38) {
                                        if (AFD < 49) {
                                            if (TSC < 10) {
                                                return 0;
                                            }
                                            else if (TSC >= 10) {
                                                if (negBalance < 123) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 123) {
                                                    if (TSC < 27) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 27) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (AFD >= 49) {
                                            if (CsDiff < 7) {
                                                if (Cs < 575) {
                                                    return 1;
                                                }
                                                else if (Cs >= 575) {
                                                    return 0;
                                                }
                                            }
                                            else if (CsDiff >= 7) {
                                                if (MVDiff < 373) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 373) {
                                                    if (gchDC < 28) {
                                                        if (AFD < 54) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 54) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (gchDC >= 28) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 38) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect2(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffMVdiffVal < 65) {
        if (diffAFD < 10) {
            if (gchDC < 19) {
                if (ssDCval < 21) {
                    if (RsCsDiff < 263) {
                        if (diffRsCsdiff < 7) {
                            if (RsCsDiff < 12) {
                                return 0;
                            }
                            else if (RsCsDiff >= 12) {
                                if (negBalance < 5) {
                                    if (Rs < 1277) {
                                        return 1;
                                    }
                                    else if (Rs >= 1277) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 5) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 7) {
                            if (TSC < 2) {
                                return 1;
                            }
                            else if (TSC >= 2) {
                                return 0;
                            }
                        }
                    }
                    else if (RsCsDiff >= 263) {
                        return 1;
                    }
                }
                else if (ssDCval >= 21) {
                    if (diffMVdiffVal < -65) {
                        if (MVDiff < 145) {
                            return 0;
                        }
                        else if (MVDiff >= 145) {
                            if (diffRsCsdiff < 2) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 2) {
                                if (Rs < 109) {
                                    return 1;
                                }
                                else if (Rs >= 109) {
                                    if (refDCval < 47) {
                                        if (ssDCval < 61) {
                                            if (diffMVdiffVal < -108) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= -108) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 61) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 47) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= -65) {
                        if (RsDiff < 67) {
                            if (ssDCval < 221) {
                                if (diffRsCsdiff < 400) {
                                    if (diffMVdiffVal < 54) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 54) {
                                        if (refDCval < 29) {
                                            if (Rs < 19) {
                                                return 0;
                                            }
                                            else if (Rs >= 19) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 29) {
                                            if (TSC < 15) {
                                                return 0;
                                            }
                                            else if (TSC >= 15) {
                                                if (Cs < 272) {
                                                    return 1;
                                                }
                                                else if (Cs >= 272) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 400) {
                                    if (Scindex < 5) {
                                        if (Rs < 1868) {
                                            return 0;
                                        }
                                        else if (Rs >= 1868) {
                                            if (SC < 2838) {
                                                return 0;
                                            }
                                            else if (SC >= 2838) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Scindex >= 5) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 221) {
                                if (TSC < 7) {
                                    return 0;
                                }
                                else if (TSC >= 7) {
                                    return 1;
                                }
                            }
                        }
                        else if (RsDiff >= 67) {
                            if (Rs < 2688) {
                                if (Rs < 1118) {
                                    return 0;
                                }
                                else if (Rs >= 1118) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 2688) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (gchDC >= 19) {
                if (AFD < 62) {
                    if (refDCval < 25) {
                        return 1;
                    }
                    else if (refDCval >= 25) {
                        if (ssDCval < 91) {
                            if (AFD < 30) {
                                if (RsCsDiff < 644) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 644) {
                                    if (Cs < 571) {
                                        return 1;
                                    }
                                    else if (Cs >= 571) {
                                        return 0;
                                    }
                                }
                            }
                            else if (AFD >= 30) {
                                if (diffAFD < -8) {
                                    if (ssDCval < 27) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 27) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= -8) {
                                    if (Rs < 75) {
                                        return 1;
                                    }
                                    else if (Rs >= 75) {
                                        if (diffRsCsdiff < 58) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 58) {
                                            if (Cs < 1417) {
                                                return 1;
                                            }
                                            else if (Cs >= 1417) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 91) {
                            return 0;
                        }
                    }
                }
                else if (AFD >= 62) {
                    return 1;
                }
            }
        }
        else if (diffAFD >= 10) {
            if (diffAFD < 30) {
                if (refDCval < 34) {
                    if (RsCsDiff < 254) {
                        if (refDCval < 24) {
                            if (diffRsCsdiff < -62) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= -62) {
                                if (AFD < 21) {
                                    if (Cs < 27) {
                                        return 0;
                                    }
                                    else if (Cs >= 27) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 21) {
                                    return 0;
                                }
                            }
                        }
                        else if (refDCval >= 24) {
                            return 0;
                        }
                    }
                    else if (RsCsDiff >= 254) {
                        return 1;
                    }
                }
                else if (refDCval >= 34) {
                    if (MVDiff < 59) {
                        if (diffTSC < 10) {
                            return 0;
                        }
                        else if (diffTSC >= 10) {
                            if (Cs < 1909) {
                                if (RsCsDiff < 338) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 338) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 1909) {
                                return 0;
                            }
                        }
                    }
                    else if (MVDiff >= 59) {
                        if (negBalance < 22) {
                            return 0;
                        }
                        else if (negBalance >= 22) {
                            if (diffAFD < 21) {
                                if (MVDiff < 329) {
                                    if (diffMVdiffVal < 54) {
                                        if (refDCval < 45) {
                                            if (diffRsCsdiff < 184) {
                                                if (diffRsCsdiff < 90) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= 90) {
                                                    if (Cs < 1006) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1006) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 184) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 45) {
                                            if (diffMVdiffVal < -1164) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= -1164) {
                                                if (SC < 6709) {
                                                    return 0;
                                                }
                                                else if (SC >= 6709) {
                                                    if (Rs < 5217) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 5217) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 54) {
                                        if (Rs < 41) {
                                            return 1;
                                        }
                                        else if (Rs >= 41) {
                                            if (Cs < 624) {
                                                return 0;
                                            }
                                            else if (Cs >= 624) {
                                                if (MVDiff < 71) {
                                                    if (TSC < 11) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 11) {
                                                        return 1;
                                                    }
                                                }
                                                else if (MVDiff >= 71) {
                                                    if (TSC < 15) {
                                                        if (Cs < 1301) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 1301) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (TSC >= 15) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 329) {
                                    if (diffTSC < 9) {
                                        if (ssDCval < 108) {
                                            if (RsDiff < 3) {
                                                if (Rs < 61) {
                                                    return 0;
                                                }
                                                else if (Rs >= 61) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsDiff >= 3) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 108) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 9) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffAFD >= 21) {
                                if (diffRsCsdiff < 15) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 15) {
                                    if (AFD < 50) {
                                        if (Rs < 328) {
                                            return 1;
                                        }
                                        else if (Rs >= 328) {
                                            if (MVDiff < 133) {
                                                if (Rs < 1343) {
                                                    if (AFD < 45) {
                                                        return 0;
                                                    }
                                                    else if (AFD >= 45) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 1343) {
                                                    if (RsCsDiff < 201) {
                                                        return 0;
                                                    }
                                                    else if (RsCsDiff >= 201) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 133) {
                                                if (refDCval < 44) {
                                                    return 0;
                                                }
                                                else if (refDCval >= 44) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 50) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 30) {
                if (TSC < 16) {
                    if (CsDiff < 11) {
                        if (RsCsDiff < 235) {
                            return 0;
                        }
                        else if (RsCsDiff >= 235) {
                            return 1;
                        }
                    }
                    else if (CsDiff >= 11) {
                        return 0;
                    }
                }
                else if (TSC >= 16) {
                    if (MVDiff < 64) {
                        if (Cs < 8222) {
                            return 1;
                        }
                        else if (Cs >= 8222) {
                            return 0;
                        }
                    }
                    else if (MVDiff >= 64) {
                        if (diffTSC < 16) {
                            return 1;
                        }
                        else if (diffTSC >= 16) {
                            if (RsDiff < 26) {
                                if (AFD < 39) {
                                    return 1;
                                }
                                else if (AFD >= 39) {
                                    if (diffMVdiffVal < -94) {
                                        if (Rs < 620) {
                                            return 0;
                                        }
                                        else if (Rs >= 620) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= -94) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsDiff >= 26) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffMVdiffVal >= 65) {
        if (diffRsCsdiff < 2) {
            if (diffAFD < 7) {
                if (AFD < 61) {
                    if (MVDiff < 378) {
                        return 0;
                    }
                    else if (MVDiff >= 378) {
                        if (gchDC < 167) {
                            if (MVDiff < 393) {
                                if (diffTSC < 3) {
                                    return 0;
                                }
                                else if (diffTSC >= 3) {
                                    if (Rs < 29) {
                                        return 0;
                                    }
                                    else if (Rs >= 29) {
                                        return 1;
                                    }
                                }
                            }
                            else if (MVDiff >= 393) {
                                if (Rs < 7) {
                                    if (Rs < 5) {
                                        if (gchDC < 102) {
                                            return 0;
                                        }
                                        else if (gchDC >= 102) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 5) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 7) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 167) {
                            return 1;
                        }
                    }
                }
                else if (AFD >= 61) {
                    return 1;
                }
            }
            else if (diffAFD >= 7) {
                if (ssDCval < 43) {
                    if (diffMVdiffVal < 116) {
                        return 0;
                    }
                    else if (diffMVdiffVal >= 116) {
                        if (posBalance < 26) {
                            if (refDCval < 72) {
                                return 0;
                            }
                            else if (refDCval >= 72) {
                                return 1;
                            }
                        }
                        else if (posBalance >= 26) {
                            return 1;
                        }
                    }
                }
                else if (ssDCval >= 43) {
                    if (SC < 211) {
                        if (diffTSC < 9) {
                            if (RsCsDiff < 26) {
                                if (posBalance < 112) {
                                    return 0;
                                }
                                else if (posBalance >= 112) {
                                    if (CsDiff < 1) {
                                        return 1;
                                    }
                                    else if (CsDiff >= 1) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 26) {
                                if (diffTSC < 5) {
                                    return 0;
                                }
                                else if (diffTSC >= 5) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffTSC >= 9) {
                            if (ssDCval < 112) {
                                if (MVDiff < 183) {
                                    return 0;
                                }
                                else if (MVDiff >= 183) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 112) {
                                return 0;
                            }
                        }
                    }
                    else if (SC >= 211) {
                        if (RsDiff < 46) {
                            if (diffMVdiffVal < 221) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 221) {
                                if (MVDiff < 314) {
                                    return 1;
                                }
                                else if (MVDiff >= 314) {
                                    return 0;
                                }
                            }
                        }
                        else if (RsDiff >= 46) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (diffRsCsdiff >= 2) {
            if (diffAFD < 12) {
                if (ssDCval < 39) {
                    if (refDCval < 31) {
                        if (diffTSC < 5) {
                            return 1;
                        }
                        else if (diffTSC >= 5) {
                            if (SC < 499) {
                                return 1;
                            }
                            else if (SC >= 499) {
                                if (Rs < 951) {
                                    return 0;
                                }
                                else if (Rs >= 951) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 31) {
                        if (diffAFD < 10) {
                            if (diffAFD < 4) {
                                return 0;
                            }
                            else if (diffAFD >= 4) {
                                if (ssDCval < 37) {
                                    if (AFD < 31) {
                                        return 0;
                                    }
                                    else if (AFD >= 31) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 37) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffAFD >= 10) {
                            if (Rs < 17) {
                                return 0;
                            }
                            else if (Rs >= 17) {
                                return 1;
                            }
                        }
                    }
                }
                else if (ssDCval >= 39) {
                    if (gchDC < 83) {
                        if (diffAFD < 9) {
                            if (TSC < 32) {
                                if (RsDiff < 5) {
                                    if (diffMVdiffVal < 102) {
                                        if (diffRsCsdiff < 13) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 13) {
                                            if (ssDCval < 71) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 71) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 102) {
                                        return 0;
                                    }
                                }
                                else if (RsDiff >= 5) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 32) {
                                if (RsDiff < 10) {
                                    return 0;
                                }
                                else if (RsDiff >= 10) {
                                    if (Cs < 214) {
                                        return 0;
                                    }
                                    else if (Cs >= 214) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 9) {
                            if (diffMVdiffVal < 103) {
                                if (MVDiff < 418) {
                                    return 0;
                                }
                                else if (MVDiff >= 418) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 103) {
                                if (SC < 337) {
                                    return 1;
                                }
                                else if (SC >= 337) {
                                    if (gchDC < 8) {
                                        if (ssDCval < 51) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 51) {
                                            if (diffRsCsdiff < 54) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 54) {
                                                if (Rs < 1716) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1716) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (gchDC >= 8) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (gchDC >= 83) {
                        return 1;
                    }
                }
            }
            else if (diffAFD >= 12) {
                if (negBalance < 33) {
                    if (diffAFD < 28) {
                        if (AFD < 30) {
                            if (ssDCval < 48) {
                                return 1;
                            }
                            else if (ssDCval >= 48) {
                                if (SC < 341) {
                                    if (CsDiff < 3) {
                                        if (AFD < 23) {
                                            if (Cs < 91) {
                                                return 1;
                                            }
                                            else if (Cs >= 91) {
                                                return 0;
                                            }
                                        }
                                        else if (AFD >= 23) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 3) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 341) {
                                    if (diffRsCsdiff < 23) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 23) {
                                        if (diffMVdiffVal < 77) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 77) {
                                            if (diffAFD < 15) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 15) {
                                                if (refDCval < 41) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 41) {
                                                    if (negBalance < 5) {
                                                        return 0;
                                                    }
                                                    else if (negBalance >= 5) {
                                                        if (negBalance < 23) {
                                                            if (MVDiff < 319) {
                                                                if (Rs < 335) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 335) {
                                                                    return 0;
                                                                }
                                                            }
                                                            else if (MVDiff >= 319) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (negBalance >= 23) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 30) {
                            if (CsDiff < 8) {
                                if (gchDC < 18) {
                                    return 1;
                                }
                                else if (gchDC >= 18) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 8) {
                                if (diffAFD < 18) {
                                    return 0;
                                }
                                else if (diffAFD >= 18) {
                                    if (Cs < 1037) {
                                        return 1;
                                    }
                                    else if (Cs >= 1037) {
                                        if (diffTSC < 21) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 21) {
                                            if (AFD < 48) {
                                                return 1;
                                            }
                                            else if (AFD >= 48) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 28) {
                        if (posBalance < 86) {
                            return 1;
                        }
                        else if (posBalance >= 86) {
                            if (negBalance < 13) {
                                if (TSC < 22) {
                                    return 1;
                                }
                                else if (TSC >= 22) {
                                    if (ssDCval < 131) {
                                        if (Cs < 230) {
                                            return 1;
                                        }
                                        else if (Cs >= 230) {
                                            if (Rs < 3620) {
                                                if (negBalance < 12) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 12) {
                                                    if (Rs < 1064) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1064) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (Rs >= 3620) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 131) {
                                        if (Rs < 2698) {
                                            return 1;
                                        }
                                        else if (Rs >= 2698) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 13) {
                                if (TSC < 33) {
                                    if (posBalance < 92) {
                                        if (diffTSC < 15) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 15) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 92) {
                                        if (diffTSC < 26) {
                                            return 1;
                                        }
                                        else if (diffTSC >= 26) {
                                            if (refDCval < 50) {
                                                if (TSC < 31) {
                                                    return 0;
                                                }
                                                else if (TSC >= 31) {
                                                    return 1;
                                                }
                                            }
                                            else if (refDCval >= 50) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 33) {
                                    if (CsDiff < 19) {
                                        if (diffAFD < 43) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 43) {
                                            return 1;
                                        }
                                    }
                                    else if (CsDiff >= 19) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 33) {
                    if (diffRsCsdiff < 74) {
                        if (Cs < 601) {
                            if (AFD < 29) {
                                if (refDCval < 66) {
                                    return 1;
                                }
                                else if (refDCval >= 66) {
                                    if (Rs < 259) {
                                        return 1;
                                    }
                                    else if (Rs >= 259) {
                                        if (diffTSC < 8) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 8) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 29) {
                                if (diffAFD < 24) {
                                    if (ssDCval < 89) {
                                        if (diffRsCsdiff < 12) {
                                            if (diffTSC < 8) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 8) {
                                                if (Cs < 264) {
                                                    return 1;
                                                }
                                                else if (Cs >= 264) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 12) {
                                            if (Rs < 61) {
                                                return 0;
                                            }
                                            else if (Rs >= 61) {
                                                if (AFD < 31) {
                                                    return 0;
                                                }
                                                else if (AFD >= 31) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 89) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 24) {
                                    if (CsDiff < 4) {
                                        if (diffMVdiffVal < 138) {
                                            if (AFD < 35) {
                                                return 1;
                                            }
                                            else if (AFD >= 35) {
                                                return 0;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 138) {
                                            if (AFD < 74) {
                                                return 1;
                                            }
                                            else if (AFD >= 74) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 4) {
                                        if (ssDCval < 46) {
                                            if (ssDCval < 45) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 45) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 46) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Cs >= 601) {
                            if (diffTSC < 8) {
                                return 0;
                            }
                            else if (diffTSC >= 8) {
                                if (gchDC < 24) {
                                    if (TSC < 18) {
                                        return 1;
                                    }
                                    else if (TSC >= 18) {
                                        if (Rs < 375) {
                                            return 1;
                                        }
                                        else if (Rs >= 375) {
                                            if (Rs < 921) {
                                                return 0;
                                            }
                                            else if (Rs >= 921) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (gchDC >= 24) {
                                    if (ssDCval < 54) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 54) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 74) {
                        if (diffTSC < 10) {
                            if (SC < 2379) {
                                if (diffTSC < 5) {
                                    if (ssDCval < 74) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 74) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 5) {
                                    if (diffMVdiffVal < 67) {
                                        if (Rs < 430) {
                                            return 0;
                                        }
                                        else if (Rs >= 430) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 67) {
                                        if (refDCval < 104) {
                                            return 1;
                                        }
                                        else if (refDCval >= 104) {
                                            if (Rs < 1116) {
                                                return 1;
                                            }
                                            else if (Rs >= 1116) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (SC >= 2379) {
                                if (ssDCval < 78) {
                                    if (refDCval < 77) {
                                        if (diffRsCsdiff < 532) {
                                            if (Rs < 1826) {
                                                return 1;
                                            }
                                            else if (Rs >= 1826) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 532) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 77) {
                                        if (TSC < 19) {
                                            return 0;
                                        }
                                        else if (TSC >= 19) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (ssDCval >= 78) {
                                    if (diffMVdiffVal < 392) {
                                        if (posBalance < 86) {
                                            return 0;
                                        }
                                        else if (posBalance >= 86) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 392) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 10) {
                            if (posBalance < 29) {
                                if (negBalance < 71) {
                                    if (TSC < 26) {
                                        return 1;
                                    }
                                    else if (TSC >= 26) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 71) {
                                    if (diffAFD < 24) {
                                        if (AFD < 35) {
                                            return 1;
                                        }
                                        else if (AFD >= 35) {
                                            if (Rs < 514) {
                                                return 1;
                                            }
                                            else if (Rs >= 514) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 24) {
                                        if (gchDC < 86) {
                                            return 1;
                                        }
                                        else if (gchDC >= 86) {
                                            if (Rs < 1778) {
                                                return 1;
                                            }
                                            else if (Rs >= 1778) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 29) {
                                if (diffTSC < 18) {
                                    if (AFD < 65) {
                                        if (diffMVdiffVal < 77) {
                                            if (diffRsCsdiff < 223) {
                                                if (SC < 1225) {
                                                    return 1;
                                                }
                                                else if (SC >= 1225) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffRsCsdiff >= 223) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 77) {
                                            if (Rs < 2001) {
                                                return 1;
                                            }
                                            else if (Rs >= 2001) {
                                                if (Cs < 1052) {
                                                    if (Cs < 944) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 944) {
                                                        return 0;
                                                    }
                                                }
                                                else if (Cs >= 1052) {
                                                    if (TSC < 30) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 30) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 65) {
                                        if (posBalance < 60) {
                                            return 1;
                                        }
                                        else if (posBalance >= 60) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffTSC >= 18) {
                                    if (Rs < 9666) {
                                        if (RsCsDiff < 183) {
                                            if (ssDCval < 146) {
                                                if (diffMVdiffVal < 84) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 84) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 146) {
                                                return 0;
                                            }
                                        }
                                        else if (RsCsDiff >= 183) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 9666) {
                                        if (Rs < 10564) {
                                            return 0;
                                        }
                                        else if (Rs >= 10564) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect3(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 13) {
        if (negBalance < 68) {
            if (diffMVdiffVal < 66) {
                if (refDCval < 24) {
                    if (diffRsCsdiff < 109) {
                        if (CsDiff < 2) {
                            if (RsDiff < 4) {
                                return 0;
                            }
                            else if (RsDiff >= 4) {
                                return 1;
                            }
                        }
                        else if (CsDiff >= 2) {
                            if (refDCval < 18) {
                                if (MVDiff < 6) {
                                    return 1;
                                }
                                else if (MVDiff >= 6) {
                                    if (RsDiff < 3) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 3) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 18) {
                                if (AFD < 9) {
                                    return 0;
                                }
                                else if (AFD >= 9) {
                                    if (Rs < 659) {
                                        return 1;
                                    }
                                    else if (Rs >= 659) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 109) {
                        return 1;
                    }
                }
                else if (refDCval >= 24) {
                    if (TSC < 32) {
                        if (diffMVdiffVal < 19) {
                            if (MVDiff < 370) {
                                if (diffTSC < 10) {
                                    return 0;
                                }
                                else if (diffTSC >= 10) {
                                    if (ssDCval < 44) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 44) {
                                        return 0;
                                    }
                                }
                            }
                            else if (MVDiff >= 370) {
                                if (AFD < 20) {
                                    return 0;
                                }
                                else if (AFD >= 20) {
                                    if (ssDCval < 63) {
                                        if (refDCval < 49) {
                                            return 1;
                                        }
                                        else if (refDCval >= 49) {
                                            if (Cs < 83) {
                                                return 1;
                                            }
                                            else if (Cs >= 83) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 63) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 19) {
                            if (RsCsDiff < 2305) {
                                if (diffTSC < -24) {
                                    if (CsDiff < 15) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 15) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= -24) {
                                    if (diffTSC < 6) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 6) {
                                        if (diffRsCsdiff < 247) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 247) {
                                            if (AFD < 14) {
                                                return 1;
                                            }
                                            else if (AFD >= 14) {
                                                if (TSC < 22) {
                                                    return 0;
                                                }
                                                else if (TSC >= 22) {
                                                    if (Rs < 2699) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 2699) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 2305) {
                                if (AFD < 26) {
                                    if (Cs < 3880) {
                                        return 1;
                                    }
                                    else if (Cs >= 3880) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 26) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (TSC >= 32) {
                        if (Rs < 1639) {
                            return 0;
                        }
                        else if (Rs >= 1639) {
                            if (Cs < 4784) {
                                return 1;
                            }
                            else if (Cs >= 4784) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 66) {
                if (diffTSC < 4) {
                    if (ssDCval < 40) {
                        if (diffAFD < 8) {
                            if (CsDiff < 3) {
                                return 0;
                            }
                            else if (CsDiff >= 3) {
                                if (diffMVdiffVal < 139) {
                                    if (RsCsDiff < 30) {
                                        return 1;
                                    }
                                    else if (RsCsDiff >= 30) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 139) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffAFD >= 8) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 40) {
                        if (AFD < 41) {
                            if (negBalance < 1) {
                                if (RsDiff < 3) {
                                    return 0;
                                }
                                else if (RsDiff >= 3) {
                                    if (Rs < 633) {
                                        return 1;
                                    }
                                    else if (Rs >= 633) {
                                        return 0;
                                    }
                                }
                            }
                            else if (negBalance >= 1) {
                                return 0;
                            }
                        }
                        else if (AFD >= 41) {
                            if (Rs < 950) {
                                return 0;
                            }
                            else if (Rs >= 950) {
                                if (SC < 3451) {
                                    if (MVDiff < 335) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 335) {
                                        if (Rs < 1670) {
                                            return 1;
                                        }
                                        else if (Rs >= 1670) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (SC >= 3451) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffTSC >= 4) {
                    if (ssDCval < 48) {
                        if (diffAFD < 7) {
                            if (TSCindex < 8) {
                                if (Rs < 42) {
                                    return 0;
                                }
                                else if (Rs >= 42) {
                                    return 1;
                                }
                            }
                            else if (TSCindex >= 8) {
                                return 0;
                            }
                        }
                        else if (diffAFD >= 7) {
                            if (posBalance < 57) {
                                if (RsCsDiff < 9) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 9) {
                                    if (MVDiff < 130) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 130) {
                                        return 0;
                                    }
                                }
                            }
                            else if (posBalance >= 57) {
                                return 1;
                            }
                        }
                    }
                    else if (ssDCval >= 48) {
                        if (Rs < 11) {
                            if (diffAFD < 5) {
                                return 0;
                            }
                            else if (diffAFD >= 5) {
                                return 1;
                            }
                        }
                        else if (Rs >= 11) {
                            if (RsDiff < 25) {
                                if (diffMVdiffVal < 141) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 141) {
                                    if (MVDiff < 178) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 178) {
                                        if (diffTSC < 12) {
                                            if (MVDiff < 828) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 828) {
                                                return 1;
                                            }
                                        }
                                        else if (diffTSC >= 12) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 25) {
                                if (AFD < 37) {
                                    return 0;
                                }
                                else if (AFD >= 37) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (negBalance >= 68) {
            if (MVDiff < 217) {
                if (diffMVdiffVal < 105) {
                    if (refDCval < 227) {
                        if (ssDCval < 24) {
                            if (RsDiff < 1) {
                                return 0;
                            }
                            else if (RsDiff >= 1) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 24) {
                            if (AFD < 40) {
                                if (refDCval < 43) {
                                    if (posBalance < 43) {
                                        return 0;
                                    }
                                    else if (posBalance >= 43) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 43) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 40) {
                                if (diffAFD < -1) {
                                    if (posBalance < 46) {
                                        return 0;
                                    }
                                    else if (posBalance >= 46) {
                                        if (RsDiff < 10) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 10) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= -1) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 227) {
                        if (AFD < 10) {
                            return 1;
                        }
                        else if (AFD >= 10) {
                            return 0;
                        }
                    }
                }
                else if (diffMVdiffVal >= 105) {
                    if (diffAFD < 7) {
                        return 0;
                    }
                    else if (diffAFD >= 7) {
                        if (Rs < 803) {
                            if (diffAFD < 12) {
                                return 1;
                            }
                            else if (diffAFD >= 12) {
                                return 0;
                            }
                        }
                        else if (Rs >= 803) {
                            return 0;
                        }
                    }
                }
            }
            else if (MVDiff >= 217) {
                if (diffAFD < 9) {
                    if (CsDiff < 20) {
                        if (CsDiff < 5) {
                            if (diffTSC < -43) {
                                return 1;
                            }
                            else if (diffTSC >= -43) {
                                if (refDCval < 31) {
                                    if (Cs < 132) {
                                        return 0;
                                    }
                                    else if (Cs >= 132) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 31) {
                                    return 0;
                                }
                            }
                        }
                        else if (CsDiff >= 5) {
                            if (diffAFD < -5) {
                                return 0;
                            }
                            else if (diffAFD >= -5) {
                                if (posBalance < 56) {
                                    if (Rs < 948) {
                                        if (diffRsCsdiff < 38) {
                                            if (diffMVdiffVal < 336) {
                                                if (negBalance < 73) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 73) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 336) {
                                                if (refDCval < 97) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 97) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 38) {
                                            if (ssDCval < 56) {
                                                if (posBalance < 35) {
                                                    if (TSC < 21) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 21) {
                                                        return 1;
                                                    }
                                                }
                                                else if (posBalance >= 35) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 56) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 948) {
                                        if (CsDiff < 15) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 15) {
                                            if (Rs < 1272) {
                                                return 1;
                                            }
                                            else if (Rs >= 1272) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 56) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 20) {
                        if (ssDCval < 95) {
                            if (RsDiff < 25) {
                                return 1;
                            }
                            else if (RsDiff >= 25) {
                                if (Rs < 2062) {
                                    return 1;
                                }
                                else if (Rs >= 2062) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 95) {
                            return 0;
                        }
                    }
                }
                else if (diffAFD >= 9) {
                    if (refDCval < 68) {
                        if (TSC < 5) {
                            return 0;
                        }
                        else if (TSC >= 5) {
                            if (Rs < 223) {
                                if (posBalance < 13) {
                                    if (gchDC < 19) {
                                        return 0;
                                    }
                                    else if (gchDC >= 19) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 13) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 223) {
                                if (Cs < 864) {
                                    return 0;
                                }
                                else if (Cs >= 864) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 68) {
                        if (ssDCval < 45) {
                            return 1;
                        }
                        else if (ssDCval >= 45) {
                            if (diffRsCsdiff < -32) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= -32) {
                                if (Cs < 1167) {
                                    return 0;
                                }
                                else if (Cs >= 1167) {
                                    if (Cs < 1185) {
                                        return 1;
                                    }
                                    else if (Cs >= 1185) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 13) {
        if (diffMVdiffVal < 65) {
            if (TSC < 24) {
                if (ssDCval < 70) {
                    if (negBalance < 52) {
                        if (diffRsCsdiff < 553) {
                            if (diffTSC < 17) {
                                if (diffMVdiffVal < 53) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 53) {
                                    if (Cs < 722) {
                                        return 0;
                                    }
                                    else if (Cs >= 722) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffTSC >= 17) {
                                if (Rs < 711) {
                                    return 1;
                                }
                                else if (Rs >= 711) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 553) {
                            return 1;
                        }
                    }
                    else if (negBalance >= 52) {
                        if (posBalance < 18) {
                            return 0;
                        }
                        else if (posBalance >= 18) {
                            if (negBalance < 65) {
                                if (refDCval < 61) {
                                    return 0;
                                }
                                else if (refDCval >= 61) {
                                    if (AFD < 30) {
                                        if (Rs < 8896) {
                                            return 0;
                                        }
                                        else if (Rs >= 8896) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 30) {
                                        return 1;
                                    }
                                }
                            }
                            else if (negBalance >= 65) {
                                return 1;
                            }
                        }
                    }
                }
                else if (ssDCval >= 70) {
                    if (gchDC < 8) {
                        if (MVDiff < 59) {
                            return 0;
                        }
                        else if (MVDiff >= 59) {
                            if (diffTSC < 13) {
                                return 0;
                            }
                            else if (diffTSC >= 13) {
                                return 1;
                            }
                        }
                    }
                    else if (gchDC >= 8) {
                        if (RsDiff < 12) {
                            if (Cs < 305) {
                                if (negBalance < 51) {
                                    if (Cs < 250) {
                                        return 0;
                                    }
                                    else if (Cs >= 250) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 51) {
                                    if (RsDiff < 2) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 2) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 305) {
                                return 0;
                            }
                        }
                        else if (RsDiff >= 12) {
                            if (SC < 2529) {
                                if (TSC < 19) {
                                    if (MVDiff < 46) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 46) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 19) {
                                    return 0;
                                }
                            }
                            else if (SC >= 2529) {
                                if (refDCval < 120) {
                                    return 0;
                                }
                                else if (refDCval >= 120) {
                                    if (Rs < 2360) {
                                        return 0;
                                    }
                                    else if (Rs >= 2360) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (TSC >= 24) {
                if (CsDiff < 13) {
                    if (MVDiff < 65) {
                        if (diffAFD < 22) {
                            return 0;
                        }
                        else if (diffAFD >= 22) {
                            return 1;
                        }
                    }
                    else if (MVDiff >= 65) {
                        if (refDCval < 95) {
                            if (diffRsCsdiff < -114) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= -114) {
                                if (diffTSC < 22) {
                                    return 0;
                                }
                                else if (diffTSC >= 22) {
                                    if (diffRsCsdiff < 250) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 250) {
                                        if (ssDCval < 131) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 131) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 95) {
                            if (posBalance < 69) {
                                if (RsDiff < 12) {
                                    if (ssDCval < 178) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 178) {
                                        if (Rs < 174) {
                                            return 1;
                                        }
                                        else if (Rs >= 174) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsDiff >= 12) {
                                    if (SC < 2895) {
                                        if (diffMVdiffVal < 5) {
                                            return 1;
                                        }
                                        else if (diffMVdiffVal >= 5) {
                                            if (Rs < 1372) {
                                                return 0;
                                            }
                                            else if (Rs >= 1372) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 2895) {
                                        return 0;
                                    }
                                }
                            }
                            else if (posBalance >= 69) {
                                return 1;
                            }
                        }
                    }
                }
                else if (CsDiff >= 13) {
                    if (ssDCval < 105) {
                        if (MVDiff < 91) {
                            return 1;
                        }
                        else if (MVDiff >= 91) {
                            if (SC < 4714) {
                                return 1;
                            }
                            else if (SC >= 4714) {
                                return 0;
                            }
                        }
                    }
                    else if (ssDCval >= 105) {
                        if (diffRsCsdiff < 1809) {
                            if (AFD < 56) {
                                return 0;
                            }
                            else if (AFD >= 56) {
                                return 1;
                            }
                        }
                        else if (diffRsCsdiff >= 1809) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 65) {
            if (negBalance < 26) {
                if (negBalance < 6) {
                    if (refDCval < 17) {
                        return 1;
                    }
                    else if (refDCval >= 17) {
                        if (MVDiff < 2733) {
                            if (gchDC < 75) {
                                if (TSC < 11) {
                                    return 1;
                                }
                                else if (TSC >= 11) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 75) {
                                if (SC < 6991) {
                                    if (Cs < 636) {
                                        if (TSC < 16) {
                                            if (Rs < 297) {
                                                return 0;
                                            }
                                            else if (Rs >= 297) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 16) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 636) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 6991) {
                                    return 1;
                                }
                            }
                        }
                        else if (MVDiff >= 2733) {
                            return 1;
                        }
                    }
                }
                else if (negBalance >= 6) {
                    if (diffAFD < 29) {
                        if (Rs < 804) {
                            if (TSC < 13) {
                                return 1;
                            }
                            else if (TSC >= 13) {
                                if (RsDiff < 5) {
                                    if (MVDiff < 328) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 328) {
                                        if (negBalance < 17) {
                                            if (MVDiff < 826) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 826) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 17) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (RsDiff >= 5) {
                                    if (RsDiff < 15) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 15) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 804) {
                            if (posBalance < 114) {
                                return 0;
                            }
                            else if (posBalance >= 114) {
                                if (Cs < 420) {
                                    return 0;
                                }
                                else if (Cs >= 420) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 29) {
                        if (diffRsCsdiff < 163) {
                            if (SC < 1668) {
                                if (posBalance < 101) {
                                    if (Rs < 872) {
                                        return 0;
                                    }
                                    else if (Rs >= 872) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 101) {
                                    if (TSC < 28) {
                                        return 1;
                                    }
                                    else if (TSC >= 28) {
                                        if (AFD < 60) {
                                            if (TSC < 30) {
                                                return 0;
                                            }
                                            else if (TSC >= 30) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 60) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (SC >= 1668) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 163) {
                            if (AFD < 59) {
                                return 1;
                            }
                            else if (AFD >= 59) {
                                if (diffAFD < 40) {
                                    return 0;
                                }
                                else if (diffAFD >= 40) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 26) {
                if (diffRsCsdiff < 64) {
                    if (SC < 593) {
                        if (diffMVdiffVal < 129) {
                            if (diffRsCsdiff < 1) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 1) {
                                if (MVDiff < 189) {
                                    if (posBalance < 54) {
                                        return 1;
                                    }
                                    else if (posBalance >= 54) {
                                        if (TSC < 11) {
                                            if (SC < 480) {
                                                return 0;
                                            }
                                            else if (SC >= 480) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 11) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (MVDiff >= 189) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 129) {
                            if (MVDiff < 480) {
                                return 1;
                            }
                            else if (MVDiff >= 480) {
                                if (AFD < 37) {
                                    return 1;
                                }
                                else if (AFD >= 37) {
                                    if (Rs < 179) {
                                        if (diffMVdiffVal < 890) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 890) {
                                            if (Rs < 39) {
                                                return 1;
                                            }
                                            else if (Rs >= 39) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Rs >= 179) {
                                        if (RsDiff < 8) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 8) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (SC >= 593) {
                        if (AFD < 42) {
                            if (TSC < 12) {
                                if (ssDCval < 41) {
                                    return 1;
                                }
                                else if (ssDCval >= 41) {
                                    if (gchDC < 2) {
                                        return 1;
                                    }
                                    else if (gchDC >= 2) {
                                        return 0;
                                    }
                                }
                            }
                            else if (TSC >= 12) {
                                if (diffTSC < 9) {
                                    if (refDCval < 56) {
                                        return 1;
                                    }
                                    else if (refDCval >= 56) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 9) {
                                    if (diffRsCsdiff < 10) {
                                        if (SC < 912) {
                                            return 0;
                                        }
                                        else if (SC >= 912) {
                                            return 1;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 10) {
                                        if (diffMVdiffVal < 75) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 75) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 42) {
                            if (diffRsCsdiff < 20) {
                                if (AFD < 70) {
                                    return 0;
                                }
                                else if (AFD >= 70) {
                                    return 1;
                                }
                            }
                            else if (diffRsCsdiff >= 20) {
                                if (Cs < 584) {
                                    if (RsDiff < 6) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 6) {
                                        if (diffMVdiffVal < 195) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 195) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Cs >= 584) {
                                    if (diffTSC < 19) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 19) {
                                        if (TSC < 29) {
                                            return 1;
                                        }
                                        else if (TSC >= 29) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 64) {
                    if (diffMVdiffVal < 92) {
                        if (diffTSC < 10) {
                            if (Rs < 1345) {
                                if (diffRsCsdiff < 106) {
                                    if (MVDiff < 72) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 72) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 106) {
                                    if (Cs < 1367) {
                                        return 1;
                                    }
                                    else if (Cs >= 1367) {
                                        return 0;
                                    }
                                }
                            }
                            else if (Rs >= 1345) {
                                if (RsDiff < 36) {
                                    return 0;
                                }
                                else if (RsDiff >= 36) {
                                    if (Rs < 5582) {
                                        return 1;
                                    }
                                    else if (Rs >= 5582) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 10) {
                            if (diffRsCsdiff < 225) {
                                if (RsCsDiff < 221) {
                                    if (SC < 2140) {
                                        return 1;
                                    }
                                    else if (SC >= 2140) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 221) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 225) {
                                if (AFD < 24) {
                                    if (SC < 2819) {
                                        return 1;
                                    }
                                    else if (SC >= 2819) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 24) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 92) {
                        if (diffAFD < 26) {
                            if (AFD < 34) {
                                if (Rs < 4750) {
                                    if (diffTSC < 9) {
                                        if (negBalance < 37) {
                                            if (AFD < 26) {
                                                return 1;
                                            }
                                            else if (AFD >= 26) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 37) {
                                            if (Scindex < 5) {
                                                return 1;
                                            }
                                            else if (Scindex >= 5) {
                                                if (Cs < 1999) {
                                                    return 0;
                                                }
                                                else if (Cs >= 1999) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 9) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 4750) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 34) {
                                if (MVDiff < 297) {
                                    if (ssDCval < 92) {
                                        if (TSC < 15) {
                                            return 0;
                                        }
                                        else if (TSC >= 15) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 92) {
                                        if (Rs < 1098) {
                                            return 1;
                                        }
                                        else if (Rs >= 1098) {
                                            if (MVDiff < 133) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 133) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 297) {
                                    if (MVDiff < 369) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 369) {
                                        if (diffRsCsdiff < 180) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 180) {
                                            if (AFD < 37) {
                                                return 0;
                                            }
                                            else if (AFD >= 37) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 26) {
                            if (diffAFD < 38) {
                                if (AFD < 59) {
                                    if (negBalance < 123) {
                                        if (RsCsDiff < 1439) {
                                            return 1;
                                        }
                                        else if (RsCsDiff >= 1439) {
                                            if (RsCsDiff < 1475) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 1475) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 123) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 59) {
                                    if (MVDiff < 285) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 285) {
                                        if (Rs < 1210) {
                                            return 1;
                                        }
                                        else if (Rs >= 1210) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 38) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect4(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffMVdiffVal < 52) {
        if (gchDC < 11) {
            if (diffTSC < 6) {
                if (refDCval < 21) {
                    if (RsCsDiff < 263) {
                        if (RsCsDiff < 7) {
                            return 0;
                        }
                        else if (RsCsDiff >= 7) {
                            if (MVDiff < 9) {
                                if (negBalance < 2) {
                                    return 1;
                                }
                                else if (negBalance >= 2) {
                                    return 0;
                                }
                            }
                            else if (MVDiff >= 9) {
                                return 0;
                            }
                        }
                    }
                    else if (RsCsDiff >= 263) {
                        return 1;
                    }
                }
                else if (refDCval >= 21) {
                    if (MVDiff < 25) {
                        return 0;
                    }
                    else if (MVDiff >= 25) {
                        if (ssDCval < 221) {
                            if (diffRsCsdiff < 259) {
                                if (MVDiff < 224) {
                                    return 0;
                                }
                                else if (MVDiff >= 224) {
                                    if (gchDC < 6) {
                                        return 0;
                                    }
                                    else if (gchDC >= 6) {
                                        if (MVDiff < 225) {
                                            if (Rs < 531) {
                                                return 0;
                                            }
                                            else if (Rs >= 531) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 225) {
                                            if (diffRsCsdiff < 16) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 16) {
                                                if (Rs < 278) {
                                                    return 1;
                                                }
                                                else if (Rs >= 278) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 259) {
                                if (ssDCval < 46) {
                                    if (RsDiff < 15) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 15) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 46) {
                                    if (TSCindex < 5) {
                                        if (RsDiff < 34) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 34) {
                                            return 1;
                                        }
                                    }
                                    else if (TSCindex >= 5) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 221) {
                            if (gchDC < 5) {
                                return 0;
                            }
                            else if (gchDC >= 5) {
                                if (refDCval < 232) {
                                    return 1;
                                }
                                else if (refDCval >= 232) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 6) {
                if (Cs < 1563) {
                    if (diffTSC < 13) {
                        if (ssDCval < 66) {
                            if (MVDiff < 149) {
                                if (CsDiff < 12) {
                                    return 0;
                                }
                                else if (CsDiff >= 12) {
                                    return 1;
                                }
                            }
                            else if (MVDiff >= 149) {
                                if (AFD < 15) {
                                    return 0;
                                }
                                else if (AFD >= 15) {
                                    if (MVDiff < 623) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 623) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 66) {
                            return 0;
                        }
                    }
                    else if (diffTSC >= 13) {
                        if (posBalance < 56) {
                            return 1;
                        }
                        else if (posBalance >= 56) {
                            if (Rs < 1967) {
                                if (posBalance < 62) {
                                    return 0;
                                }
                                else if (posBalance >= 62) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 1967) {
                                return 0;
                            }
                        }
                    }
                }
                else if (Cs >= 1563) {
                    if (diffTSC < 16) {
                        return 0;
                    }
                    else if (diffTSC >= 16) {
                        if (Cs < 4506) {
                            return 1;
                        }
                        else if (Cs >= 4506) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (gchDC >= 11) {
            if (TSC < 22) {
                if (refDCval < 28) {
                    if (gchDC < 18) {
                        if (Rs < 453) {
                            return 0;
                        }
                        else if (Rs >= 453) {
                            return 1;
                        }
                    }
                    else if (gchDC >= 18) {
                        return 1;
                    }
                }
                else if (refDCval >= 28) {
                    if (diffAFD < 14) {
                        if (ssDCval < 26) {
                            if (CsDiff < 9) {
                                return 0;
                            }
                            else if (CsDiff >= 9) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 26) {
                            if (refDCval < 228) {
                                return 0;
                            }
                            else if (refDCval >= 228) {
                                if (TSC < 4) {
                                    return 1;
                                }
                                else if (TSC >= 4) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 14) {
                        if (diffMVdiffVal < -112) {
                            if (Cs < 1352) {
                                return 1;
                            }
                            else if (Cs >= 1352) {
                                return 0;
                            }
                        }
                        else if (diffMVdiffVal >= -112) {
                            if (diffRsCsdiff < 161) {
                                if (TSC < 8) {
                                    if (TSC < 7) {
                                        return 0;
                                    }
                                    else if (TSC >= 7) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 8) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 161) {
                                if (MVDiff < 206) {
                                    if (Cs < 423) {
                                        if (Rs < 1755) {
                                            return 1;
                                        }
                                        else if (Rs >= 1755) {
                                            return 0;
                                        }
                                    }
                                    else if (Cs >= 423) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 206) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (TSC >= 22) {
                if (RsCsDiff < 297) {
                    if (refDCval < 25) {
                        return 1;
                    }
                    else if (refDCval >= 25) {
                        if (gchDC < 23) {
                            if (ssDCval < 64) {
                                if (Rs < 987) {
                                    return 1;
                                }
                                else if (Rs >= 987) {
                                    return 0;
                                }
                            }
                            else if (ssDCval >= 64) {
                                if (RsDiff < 13) {
                                    return 0;
                                }
                                else if (RsDiff >= 13) {
                                    return 1;
                                }
                            }
                        }
                        else if (gchDC >= 23) {
                            if (gchDC < 142) {
                                if (Rs < 133) {
                                    if (MVDiff < 344) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 344) {
                                        return 0;
                                    }
                                }
                                else if (Rs >= 133) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 142) {
                                return 1;
                            }
                        }
                    }
                }
                else if (RsCsDiff >= 297) {
                    if (Rs < 2190) {
                        if (Cs < 1626) {
                            if (ssDCval < 104) {
                                if (diffTSC < 11) {
                                    if (MVDiff < 402) {
                                        if (RsDiff < 57) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 57) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 402) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= 11) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 104) {
                                return 0;
                            }
                        }
                        else if (Cs >= 1626) {
                            return 1;
                        }
                    }
                    else if (Rs >= 2190) {
                        if (diffTSC < 17) {
                            return 0;
                        }
                        else if (diffTSC >= 17) {
                            if (posBalance < 111) {
                                return 1;
                            }
                            else if (posBalance >= 111) {
                                if (AFD < 35) {
                                    return 1;
                                }
                                else if (AFD >= 35) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffMVdiffVal >= 52) {
        if (diffAFD < 14) {
            if (diffTSC < 4) {
                if (diffAFD < 7) {
                    if (diffTSC < -43) {
                        return 1;
                    }
                    else if (diffTSC >= -43) {
                        if (AFD < 34) {
                            if (ssDCval < 28) {
                                if (diffTSC < -10) {
                                    return 1;
                                }
                                else if (diffTSC >= -10) {
                                    if (RsCsDiff < 5) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 5) {
                                        if (AFD < 13) {
                                            if (TSC < 3) {
                                                return 0;
                                            }
                                            else if (TSC >= 3) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 13) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 28) {
                                if (negBalance < 30) {
                                    if (negBalance < 29) {
                                        if (gchDC < 130) {
                                            return 0;
                                        }
                                        else if (gchDC >= 130) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 29) {
                                        if (diffAFD < 2) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 2) {
                                            if (SC < 870) {
                                                return 1;
                                            }
                                            else if (SC >= 870) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 30) {
                                    return 0;
                                }
                            }
                        }
                        else if (AFD >= 34) {
                            if (RsCsDiff < 99) {
                                return 0;
                            }
                            else if (RsCsDiff >= 99) {
                                if (diffAFD < 1) {
                                    if (SC < 1561) {
                                        if (TSC < 33) {
                                            return 1;
                                        }
                                        else if (TSC >= 33) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 1561) {
                                        if (MVDiff < 516) {
                                            if (ssDCval < 37) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 37) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 516) {
                                            if (ssDCval < 144) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 144) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 1) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 7) {
                    if (AFD < 16) {
                        if (gchDC < 6) {
                            if (diffTSC < 2) {
                                return 0;
                            }
                            else if (diffTSC >= 2) {
                                if (Cs < 9) {
                                    if (SC < 12) {
                                        return 1;
                                    }
                                    else if (SC >= 12) {
                                        return 0;
                                    }
                                }
                                else if (Cs >= 9) {
                                    return 1;
                                }
                            }
                        }
                        else if (gchDC >= 6) {
                            return 0;
                        }
                    }
                    else if (AFD >= 16) {
                        if (ssDCval < 41) {
                            if (RsDiff < 6) {
                                if (gchDC < 11) {
                                    return 1;
                                }
                                else if (gchDC >= 11) {
                                    return 0;
                                }
                            }
                            else if (RsDiff >= 6) {
                                if (MVDiff < 106) {
                                    return 0;
                                }
                                else if (MVDiff >= 106) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 41) {
                            if (MVDiff < 1426) {
                                if (ssDCval < 76) {
                                    if (TSC < 14) {
                                        return 0;
                                    }
                                    else if (TSC >= 14) {
                                        if (gchDC < 15) {
                                            if (RsCsDiff < 517) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 517) {
                                                return 0;
                                            }
                                        }
                                        else if (gchDC >= 15) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (ssDCval >= 76) {
                                    return 0;
                                }
                            }
                            else if (MVDiff >= 1426) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 4) {
                if (ssDCval < 54) {
                    if (gchDC < 5) {
                        if (CsDiff < 4) {
                            if (AFD < 23) {
                                return 1;
                            }
                            else if (AFD >= 23) {
                                return 0;
                            }
                        }
                        else if (CsDiff >= 4) {
                            if (gchDC < 2) {
                                return 0;
                            }
                            else if (gchDC >= 2) {
                                if (AFD < 31) {
                                    return 1;
                                }
                                else if (AFD >= 31) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 5) {
                        if (RsCsDiff < 19) {
                            if (AFD < 16) {
                                if (gchDC < 17) {
                                    if (ssDCval < 38) {
                                        if (ssDCval < 30) {
                                            if (TSC < 8) {
                                                return 0;
                                            }
                                            else if (TSC >= 8) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 30) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 38) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 17) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 16) {
                                return 0;
                            }
                        }
                        else if (RsCsDiff >= 19) {
                            if (negBalance < 103) {
                                if (refDCval < 42) {
                                    if (SC < 840) {
                                        return 1;
                                    }
                                    else if (SC >= 840) {
                                        if (AFD < 15) {
                                            return 1;
                                        }
                                        else if (AFD >= 15) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 42) {
                                    if (diffMVdiffVal < 83) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 83) {
                                        return 1;
                                    }
                                }
                            }
                            else if (negBalance >= 103) {
                                return 0;
                            }
                        }
                    }
                }
                else if (ssDCval >= 54) {
                    if (RsCsDiff < 55) {
                        if (refDCval < 38) {
                            return 1;
                        }
                        else if (refDCval >= 38) {
                            if (TSC < 11) {
                                if (RsCsDiff < 17) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 17) {
                                    if (Rs < 210) {
                                        return 1;
                                    }
                                    else if (Rs >= 210) {
                                        return 0;
                                    }
                                }
                            }
                            else if (TSC >= 11) {
                                return 0;
                            }
                        }
                    }
                    else if (RsCsDiff >= 55) {
                        if (TSC < 36) {
                            if (Cs < 681) {
                                if (AFD < 43) {
                                    if (AFD < 34) {
                                        if (diffTSC < 8) {
                                            if (negBalance < 57) {
                                                return 0;
                                            }
                                            else if (negBalance >= 57) {
                                                if (Rs < 765) {
                                                    return 1;
                                                }
                                                else if (Rs >= 765) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffTSC >= 8) {
                                            if (AFD < 30) {
                                                return 1;
                                            }
                                            else if (AFD >= 30) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (AFD >= 34) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 43) {
                                    return 0;
                                }
                            }
                            else if (Cs >= 681) {
                                if (MVDiff < 510) {
                                    return 0;
                                }
                                else if (MVDiff >= 510) {
                                    if (Rs < 1931) {
                                        return 1;
                                    }
                                    else if (Rs >= 1931) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 36) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 14) {
            if (negBalance < 32) {
                if (diffAFD < 29) {
                    if (diffMVdiffVal < 73) {
                        return 0;
                    }
                    else if (diffMVdiffVal >= 73) {
                        if (TSC < 11) {
                            if (diffMVdiffVal < 82) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 82) {
                                return 1;
                            }
                        }
                        else if (TSC >= 11) {
                            if (diffRsCsdiff < 35) {
                                if (gchDC < 28) {
                                    if (Rs < 368) {
                                        if (diffRsCsdiff < 15) {
                                            if (negBalance < 22) {
                                                return 0;
                                            }
                                            else if (negBalance >= 22) {
                                                if (CsDiff < 1) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 1) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 15) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 368) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 28) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 35) {
                                if (diffAFD < 17) {
                                    return 0;
                                }
                                else if (diffAFD >= 17) {
                                    if (TSC < 22) {
                                        if (ssDCval < 75) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 75) {
                                            if (MVDiff < 146) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 146) {
                                                if (posBalance < 95) {
                                                    if (diffTSC < 14) {
                                                        return 0;
                                                    }
                                                    else if (diffTSC >= 14) {
                                                        return 1;
                                                    }
                                                }
                                                else if (posBalance >= 95) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 22) {
                                        if (CsDiff < 8) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 8) {
                                            if (Cs < 1345) {
                                                return 1;
                                            }
                                            else if (Cs >= 1345) {
                                                if (SC < 7548) {
                                                    return 0;
                                                }
                                                else if (SC >= 7548) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 29) {
                    if (TSC < 34) {
                        if (negBalance < 4) {
                            return 0;
                        }
                        else if (negBalance >= 4) {
                            if (ssDCval < 110) {
                                if (diffTSC < 15) {
                                    return 0;
                                }
                                else if (diffTSC >= 15) {
                                    if (diffRsCsdiff < 70) {
                                        if (gchDC < 57) {
                                            return 0;
                                        }
                                        else if (gchDC >= 57) {
                                            return 1;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 70) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 110) {
                                return 1;
                            }
                        }
                    }
                    else if (TSC >= 34) {
                        if (AFD < 48) {
                            return 0;
                        }
                        else if (AFD >= 48) {
                            if (diffAFD < 40) {
                                return 0;
                            }
                            else if (diffAFD >= 40) {
                                if (CsDiff < 14) {
                                    if (ssDCval < 116) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 116) {
                                        if (Cs < 1540) {
                                            return 1;
                                        }
                                        else if (Cs >= 1540) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (CsDiff >= 14) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 32) {
                if (diffTSC < 10) {
                    if (Rs < 2011) {
                        if (refDCval < 92) {
                            if (ssDCval < 74) {
                                if (diffTSC < -1) {
                                    if (TSC < 11) {
                                        return 0;
                                    }
                                    else if (TSC >= 11) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= -1) {
                                    if (negBalance < 35) {
                                        return 0;
                                    }
                                    else if (negBalance >= 35) {
                                        if (diffRsCsdiff < 97) {
                                            if (diffMVdiffVal < 108) {
                                                if (posBalance < 48) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 48) {
                                                    if (TSC < 10) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 10) {
                                                        if (Rs < 473) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 473) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= 108) {
                                                if (TSC < 20) {
                                                    if (diffRsCsdiff < -11) {
                                                        return 0;
                                                    }
                                                    else if (diffRsCsdiff >= -11) {
                                                        if (refDCval < 37) {
                                                            if (Rs < 339) {
                                                                return 1;
                                                            }
                                                            else if (Rs >= 339) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (refDCval >= 37) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (TSC >= 20) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffRsCsdiff >= 97) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 74) {
                                if (diffTSC < 9) {
                                    if (gchDC < 5) {
                                        return 1;
                                    }
                                    else if (gchDC >= 5) {
                                        if (diffAFD < 27) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 27) {
                                            if (Rs < 196) {
                                                return 0;
                                            }
                                            else if (Rs >= 196) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (diffTSC >= 9) {
                                    return 1;
                                }
                            }
                        }
                        else if (refDCval >= 92) {
                            if (Cs < 947) {
                                if (RsDiff < 11) {
                                    if (negBalance < 111) {
                                        if (diffAFD < 24) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 24) {
                                            if (Cs < 561) {
                                                return 1;
                                            }
                                            else if (Cs >= 561) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 111) {
                                        if (TSC < 17) {
                                            return 1;
                                        }
                                        else if (TSC >= 17) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsDiff >= 11) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 947) {
                                return 0;
                            }
                        }
                    }
                    else if (Rs >= 2011) {
                        return 0;
                    }
                }
                else if (diffTSC >= 10) {
                    if (diffMVdiffVal < 73) {
                        if (ssDCval < 123) {
                            if (refDCval < 102) {
                                if (posBalance < 18) {
                                    if (Rs < 324) {
                                        return 1;
                                    }
                                    else if (Rs >= 324) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 18) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 102) {
                                if (diffMVdiffVal < 60) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 60) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 123) {
                            if (AFD < 40) {
                                return 1;
                            }
                            else if (AFD >= 40) {
                                if (TSC < 38) {
                                    return 0;
                                }
                                else if (TSC >= 38) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 73) {
                        if (diffAFD < 25) {
                            if (TSC < 29) {
                                if (gchDC < 20) {
                                    return 1;
                                }
                                else if (gchDC >= 20) {
                                    if (ssDCval < 90) {
                                        if (Cs < 1134) {
                                            return 1;
                                        }
                                        else if (Cs >= 1134) {
                                            if (Cs < 3594) {
                                                return 0;
                                            }
                                            else if (Cs >= 3594) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 90) {
                                        if (refDCval < 86) {
                                            return 0;
                                        }
                                        else if (refDCval >= 86) {
                                            if (Cs < 2888) {
                                                return 1;
                                            }
                                            else if (Cs >= 2888) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 29) {
                                if (MVDiff < 121) {
                                    return 1;
                                }
                                else if (MVDiff >= 121) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffAFD >= 25) {
                            if (MVDiff < 271) {
                                if (diffMVdiffVal < 80) {
                                    if (MVDiff < 90) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 90) {
                                        if (Cs < 635) {
                                            return 1;
                                        }
                                        else if (Cs >= 635) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 80) {
                                    if (diffRsCsdiff < -17) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= -17) {
                                        return 1;
                                    }
                                }
                            }
                            else if (MVDiff >= 271) {
                                if (negBalance < 54) {
                                    if (diffMVdiffVal < 214) {
                                        if (Cs < 3455) {
                                            return 0;
                                        }
                                        else if (Cs >= 3455) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 214) {
                                        if (MVDiff < 509) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 509) {
                                            if (gchDC < 21) {
                                                return 1;
                                            }
                                            else if (gchDC >= 21) {
                                                if (diffMVdiffVal < 516) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 516) {
                                                    if (MVDiff < 1055) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 1055) {
                                                        if (Rs < 1557) {
                                                            return 0;
                                                        }
                                                        else if (Rs >= 1557) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 54) {
                                    if (diffAFD < 27) {
                                        if (AFD < 58) {
                                            if (diffMVdiffVal < 270) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 270) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 58) {
                                            return 0;
                                        }
                                    }
                                    else if (diffAFD >= 27) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect5(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffTSC < 7) {
        if (diffAFD < 8) {
            if (MVDiff < 252) {
                if (ssDCval < 20) {
                    if (diffRsCsdiff < 4) {
                        if (CsDiff < 3) {
                            return 0;
                        }
                        else if (CsDiff >= 3) {
                            if (refDCval < 18) {
                                if (Rs < 1027) {
                                    return 1;
                                }
                                else if (Rs >= 1027) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 18) {
                                return 0;
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 4) {
                        if (SC < 1040) {
                            if (SC < 326) {
                                return 1;
                            }
                            else if (SC >= 326) {
                                return 0;
                            }
                        }
                        else if (SC >= 1040) {
                            return 1;
                        }
                    }
                }
                else if (ssDCval >= 20) {
                    if (RsCsDiff < 4482) {
                        if (negBalance < 106) {
                            if (MVDiff < 40) {
                                if (gchDC < 129) {
                                    if (ssDCval < 21) {
                                        if (AFD < 5) {
                                            return 0;
                                        }
                                        else if (AFD >= 5) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 21) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 129) {
                                    return 1;
                                }
                            }
                            else if (MVDiff >= 40) {
                                if (ssDCval < 28) {
                                    if (TSC < 4) {
                                        if (diffRsCsdiff < 31) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 31) {
                                            if (AFD < 1) {
                                                return 1;
                                            }
                                            else if (AFD >= 1) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (TSC >= 4) {
                                        if (diffRsCsdiff < 42) {
                                            if (ssDCval < 24) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 24) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 42) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (ssDCval >= 28) {
                                    if (RsDiff < 37) {
                                        if (diffMVdiffVal < 100) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 100) {
                                            if (diffMVdiffVal < 101) {
                                                if (Rs < 295) {
                                                    if (Cs < 145) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 145) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 295) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 101) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 37) {
                                        if (TSC < 4) {
                                            return 1;
                                        }
                                        else if (TSC >= 4) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 106) {
                            if (refDCval < 226) {
                                if (diffMVdiffVal < 108) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 108) {
                                    if (diffAFD < 5) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 5) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 226) {
                                if (MVDiff < 135) {
                                    return 1;
                                }
                                else if (MVDiff >= 135) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 4482) {
                        if (diffAFD < -1) {
                            if (CsDiff < 34) {
                                return 1;
                            }
                            else if (CsDiff >= 34) {
                                return 0;
                            }
                        }
                        else if (diffAFD >= -1) {
                            return 0;
                        }
                    }
                }
            }
            else if (MVDiff >= 252) {
                if (TSC < 32) {
                    if (gchDC < 2) {
                        if (refDCval < 18) {
                            if (Rs < 215) {
                                return 0;
                            }
                            else if (Rs >= 215) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 18) {
                            return 0;
                        }
                    }
                    else if (gchDC >= 2) {
                        if (ssDCval < 38) {
                            if (diffRsCsdiff < 10) {
                                if (diffTSC < 1) {
                                    if (CsDiff < 7) {
                                        if (AFD < 15) {
                                            return 0;
                                        }
                                        else if (AFD >= 15) {
                                            if (AFD < 16) {
                                                return 1;
                                            }
                                            else if (AFD >= 16) {
                                                if (posBalance < 33) {
                                                    if (MVDiff < 1653) {
                                                        return 0;
                                                    }
                                                    else if (MVDiff >= 1653) {
                                                        return 1;
                                                    }
                                                }
                                                else if (posBalance >= 33) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 7) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= 1) {
                                    if (posBalance < 61) {
                                        if (ssDCval < 24) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 24) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 61) {
                                        if (MVDiff < 1197) {
                                            if (TSC < 5) {
                                                return 1;
                                            }
                                            else if (TSC >= 5) {
                                                if (Cs < 89) {
                                                    return 0;
                                                }
                                                else if (Cs >= 89) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (MVDiff >= 1197) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 10) {
                                if (ssDCval < 30) {
                                    return 1;
                                }
                                else if (ssDCval >= 30) {
                                    if (Rs < 498) {
                                        if (ssDCval < 34) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 34) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 498) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 38) {
                            if (TSC < 13) {
                                if (gchDC < 132) {
                                    if (refDCval < 228) {
                                        return 0;
                                    }
                                    else if (refDCval >= 228) {
                                        if (ssDCval < 226) {
                                            if (Rs < 1) {
                                                return 1;
                                            }
                                            else if (Rs >= 1) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 226) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (gchDC >= 132) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 13) {
                                if (posBalance < 50) {
                                    if (gchDC < 85) {
                                        return 0;
                                    }
                                    else if (gchDC >= 85) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 50) {
                                    if (posBalance < 78) {
                                        if (Cs < 860) {
                                            if (ssDCval < 55) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 55) {
                                                if (CsDiff < 9) {
                                                    if (AFD < 35) {
                                                        if (MVDiff < 425) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 425) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (AFD >= 35) {
                                                        return 0;
                                                    }
                                                }
                                                else if (CsDiff >= 9) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 860) {
                                            if (AFD < 58) {
                                                return 0;
                                            }
                                            else if (AFD >= 58) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 78) {
                                        if (Rs < 1) {
                                            if (AFD < 41) {
                                                return 1;
                                            }
                                            else if (AFD >= 41) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 1) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (TSC >= 32) {
                    if (refDCval < 116) {
                        return 1;
                    }
                    else if (refDCval >= 116) {
                        return 0;
                    }
                }
            }
        }
        else if (diffAFD >= 8) {
            if (diffMVdiffVal < 59) {
                if (negBalance < 82) {
                    if (diffMVdiffVal < 33) {
                        return 0;
                    }
                    else if (diffMVdiffVal >= 33) {
                        if (posBalance < 51) {
                            if (ssDCval < 36) {
                                if (Rs < 574) {
                                    return 0;
                                }
                                else if (Rs >= 574) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 36) {
                                return 0;
                            }
                        }
                        else if (posBalance >= 51) {
                            return 0;
                        }
                    }
                }
                else if (negBalance >= 82) {
                    if (refDCval < 56) {
                        return 1;
                    }
                    else if (refDCval >= 56) {
                        if (gchDC < 42) {
                            return 0;
                        }
                        else if (gchDC >= 42) {
                            if (Cs < 753) {
                                return 1;
                            }
                            else if (Cs >= 753) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 59) {
                if (ssDCval < 62) {
                    if (ssDCval < 42) {
                        if (refDCval < 53) {
                            if (SC < 360) {
                                return 1;
                            }
                            else if (SC >= 360) {
                                if (RsCsDiff < 35) {
                                    if (diffMVdiffVal < 94) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 94) {
                                        if (Rs < 390) {
                                            return 1;
                                        }
                                        else if (Rs >= 390) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 35) {
                                    return 1;
                                }
                            }
                        }
                        else if (refDCval >= 53) {
                            if (CsDiff < 6) {
                                if (gchDC < 20) {
                                    return 0;
                                }
                                else if (gchDC >= 20) {
                                    if (TSC < 9) {
                                        return 1;
                                    }
                                    else if (TSC >= 9) {
                                        if (diffRsCsdiff < -10) {
                                            return 1;
                                        }
                                        else if (diffRsCsdiff >= -10) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 6) {
                                return 1;
                            }
                        }
                    }
                    else if (ssDCval >= 42) {
                        if (posBalance < 22) {
                            return 0;
                        }
                        else if (posBalance >= 22) {
                            if (diffRsCsdiff < 68) {
                                if (Rs < 82) {
                                    if (diffTSC < 5) {
                                        if (Rs < 62) {
                                            return 0;
                                        }
                                        else if (Rs >= 62) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= 5) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 82) {
                                    if (ssDCval < 50) {
                                        if (Cs < 462) {
                                            if (RsCsDiff < 29) {
                                                if (Rs < 323) {
                                                    return 0;
                                                }
                                                else if (Rs >= 323) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 29) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 462) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 50) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 68) {
                                return 1;
                            }
                        }
                    }
                }
                else if (ssDCval >= 62) {
                    if (diffAFD < 24) {
                        if (Rs < 106) {
                            if (CsDiff < 4) {
                                if (ssDCval < 65) {
                                    if (Rs < 51) {
                                        return 0;
                                    }
                                    else if (Rs >= 51) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 65) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 4) {
                                if (RsDiff < 1) {
                                    return 0;
                                }
                                else if (RsDiff >= 1) {
                                    return 1;
                                }
                            }
                        }
                        else if (Rs >= 106) {
                            if (diffAFD < 21) {
                                return 0;
                            }
                            else if (diffAFD >= 21) {
                                if (TSC < 23) {
                                    return 1;
                                }
                                else if (TSC >= 23) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 24) {
                        if (refDCval < 83) {
                            return 0;
                        }
                        else if (refDCval >= 83) {
                            if (MVDiff < 291) {
                                return 0;
                            }
                            else if (MVDiff >= 291) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffTSC >= 7) {
        if (diffMVdiffVal < 55) {
            if (diffTSC < 16) {
                if (MVDiff < 130) {
                    if (ssDCval < 69) {
                        if (RsDiff < 24) {
                            if (diffAFD < 15) {
                                if (gchDC < 36) {
                                    return 0;
                                }
                                else if (gchDC >= 36) {
                                    if (Cs < 49) {
                                        return 0;
                                    }
                                    else if (Cs >= 49) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffAFD >= 15) {
                                if (diffMVdiffVal < 21) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 21) {
                                    if (RsDiff < 8) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 8) {
                                        if (AFD < 31) {
                                            return 1;
                                        }
                                        else if (AFD >= 31) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 24) {
                            if (Cs < 1889) {
                                return 1;
                            }
                            else if (Cs >= 1889) {
                                return 0;
                            }
                        }
                    }
                    else if (ssDCval >= 69) {
                        if (Cs < 1) {
                            return 1;
                        }
                        else if (Cs >= 1) {
                            return 0;
                        }
                    }
                }
                else if (MVDiff >= 130) {
                    if (gchDC < 8) {
                        if (CsDiff < 5) {
                            if (Rs < 157) {
                                return 1;
                            }
                            else if (Rs >= 157) {
                                return 0;
                            }
                        }
                        else if (CsDiff >= 5) {
                            return 1;
                        }
                    }
                    else if (gchDC >= 8) {
                        if (diffRsCsdiff < 36) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 36) {
                            if (ssDCval < 99) {
                                if (Cs < 574) {
                                    if (CsDiff < 8) {
                                        return 1;
                                    }
                                    else if (CsDiff >= 8) {
                                        return 0;
                                    }
                                }
                                else if (Cs >= 574) {
                                    if (posBalance < 22) {
                                        return 1;
                                    }
                                    else if (posBalance >= 22) {
                                        if (ssDCval < 94) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 94) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 99) {
                                if (gchDC < 56) {
                                    return 0;
                                }
                                else if (gchDC >= 56) {
                                    if (Rs < 1080) {
                                        return 0;
                                    }
                                    else if (Rs >= 1080) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 16) {
                if (RsDiff < 13) {
                    if (Rs < 217) {
                        if (CsDiff < 4) {
                            if (Cs < 46) {
                                return 0;
                            }
                            else if (Cs >= 46) {
                                return 1;
                            }
                        }
                        else if (CsDiff >= 4) {
                            return 0;
                        }
                    }
                    else if (Rs >= 217) {
                        if (MVDiff < 44) {
                            if (Rs < 1439) {
                                return 1;
                            }
                            else if (Rs >= 1439) {
                                return 0;
                            }
                        }
                        else if (MVDiff >= 44) {
                            if (ssDCval < 76) {
                                if (gchDC < 28) {
                                    if (negBalance < 40) {
                                        return 0;
                                    }
                                    else if (negBalance >= 40) {
                                        if (Cs < 161) {
                                            return 0;
                                        }
                                        else if (Cs >= 161) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (gchDC >= 28) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 76) {
                                return 0;
                            }
                        }
                    }
                }
                else if (RsDiff >= 13) {
                    if (Cs < 3893) {
                        if (diffTSC < 30) {
                            if (posBalance < 92) {
                                if (Cs < 2268) {
                                    return 1;
                                }
                                else if (Cs >= 2268) {
                                    return 0;
                                }
                            }
                            else if (posBalance >= 92) {
                                if (Rs < 592) {
                                    return 1;
                                }
                                else if (Rs >= 592) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffTSC >= 30) {
                            return 1;
                        }
                    }
                    else if (Cs >= 3893) {
                        if (Rs < 4866) {
                            if (RsDiff < 18) {
                                return 0;
                            }
                            else if (RsDiff >= 18) {
                                return 1;
                            }
                        }
                        else if (Rs >= 4866) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 55) {
            if (diffRsCsdiff < 69) {
                if (negBalance < 39) {
                    if (SC < 673) {
                        if (diffRsCsdiff < 15) {
                            if (TSC < 11) {
                                if (SC < 299) {
                                    return 1;
                                }
                                else if (SC >= 299) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 11) {
                                if (negBalance < 7) {
                                    return 0;
                                }
                                else if (negBalance >= 7) {
                                    if (diffRsCsdiff < 4) {
                                        if (AFD < 17) {
                                            return 1;
                                        }
                                        else if (AFD >= 17) {
                                            if (posBalance < 114) {
                                                return 0;
                                            }
                                            else if (posBalance >= 114) {
                                                if (Cs < 36) {
                                                    return 0;
                                                }
                                                else if (Cs >= 36) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 4) {
                                        if (Cs < 205) {
                                            if (refDCval < 37) {
                                                return 1;
                                            }
                                            else if (refDCval >= 37) {
                                                if (Rs < 105) {
                                                    return 1;
                                                }
                                                else if (Rs >= 105) {
                                                    if (Rs < 204) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 204) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (Cs >= 205) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 15) {
                            if (diffMVdiffVal < 106) {
                                if (ssDCval < 154) {
                                    return 1;
                                }
                                else if (ssDCval >= 154) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 106) {
                                return 1;
                            }
                        }
                    }
                    else if (SC >= 673) {
                        if (CsDiff < 4) {
                            return 0;
                        }
                        else if (CsDiff >= 4) {
                            if (Cs < 652) {
                                if (RsCsDiff < 63) {
                                    if (Rs < 348) {
                                        if (TSC < 24) {
                                            return 0;
                                        }
                                        else if (TSC >= 24) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 348) {
                                        if (diffAFD < 37) {
                                            return 1;
                                        }
                                        else if (diffAFD >= 37) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 63) {
                                    return 0;
                                }
                            }
                            else if (Cs >= 652) {
                                if (Rs < 5818) {
                                    return 0;
                                }
                                else if (Rs >= 5818) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 39) {
                    if (refDCval < 139) {
                        if (diffAFD < 12) {
                            if (refDCval < 59) {
                                if (Rs < 239) {
                                    return 1;
                                }
                                else if (Rs >= 239) {
                                    if (Cs < 455) {
                                        return 0;
                                    }
                                    else if (Cs >= 455) {
                                        if (SC < 1571) {
                                            return 1;
                                        }
                                        else if (SC >= 1571) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 59) {
                                if (diffRsCsdiff < 38) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 38) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffAFD >= 12) {
                            if (AFD < 39) {
                                if (diffMVdiffVal < 847) {
                                    if (negBalance < 40) {
                                        if (AFD < 30) {
                                            return 1;
                                        }
                                        else if (AFD >= 30) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 40) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 847) {
                                    if (MVDiff < 1145) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 1145) {
                                        return 1;
                                    }
                                }
                            }
                            else if (AFD >= 39) {
                                if (diffRsCsdiff < 45) {
                                    if (diffMVdiffVal < 134) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 134) {
                                        if (MVDiff < 305) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 305) {
                                            if (diffRsCsdiff < -1) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= -1) {
                                                if (CsDiff < 8) {
                                                    if (SC < 110) {
                                                        return 0;
                                                    }
                                                    else if (SC >= 110) {
                                                        if (MVDiff < 314) {
                                                            return 0;
                                                        }
                                                        else if (MVDiff >= 314) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                                else if (CsDiff >= 8) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 45) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 139) {
                        if (diffRsCsdiff < 26) {
                            if (Rs < 68) {
                                return 1;
                            }
                            else if (Rs >= 68) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 26) {
                            if (ssDCval < 149) {
                                if (TSC < 51) {
                                    return 0;
                                }
                                else if (TSC >= 51) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 149) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 69) {
                if (diffAFD < 26) {
                    if (SC < 2150) {
                        if (diffMVdiffVal < 111) {
                            if (diffRsCsdiff < 98) {
                                if (Cs < 664) {
                                    if (Cs < 375) {
                                        return 0;
                                    }
                                    else if (Cs >= 375) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 664) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 98) {
                                if (posBalance < 21) {
                                    if (TSC < 13) {
                                        return 1;
                                    }
                                    else if (TSC >= 13) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 21) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 111) {
                            if (ssDCval < 102) {
                                return 1;
                            }
                            else if (ssDCval >= 102) {
                                if (ssDCval < 113) {
                                    return 0;
                                }
                                else if (ssDCval >= 113) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (SC >= 2150) {
                        if (diffAFD < 17) {
                            if (gchDC < 36) {
                                if (AFD < 27) {
                                    if (gchDC < 10) {
                                        return 1;
                                    }
                                    else if (gchDC >= 10) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 27) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 36) {
                                return 1;
                            }
                        }
                        else if (diffAFD >= 17) {
                            if (gchDC < 17) {
                                if (ssDCval < 137) {
                                    return 1;
                                }
                                else if (ssDCval >= 137) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 17) {
                                if (ssDCval < 64) {
                                    return 1;
                                }
                                else if (ssDCval >= 64) {
                                    if (diffRsCsdiff < 830) {
                                        if (Cs < 1324) {
                                            if (posBalance < 93) {
                                                return 0;
                                            }
                                            else if (posBalance >= 93) {
                                                if (diffAFD < 20) {
                                                    return 0;
                                                }
                                                else if (diffAFD >= 20) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 1324) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 830) {
                                        if (AFD < 26) {
                                            return 0;
                                        }
                                        else if (AFD >= 26) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 26) {
                    if (negBalance < 3) {
                        if (Rs < 8601) {
                            if (Rs < 1090) {
                                if (Rs < 520) {
                                    return 0;
                                }
                                else if (Rs >= 520) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 1090) {
                                return 0;
                            }
                        }
                        else if (Rs >= 8601) {
                            return 1;
                        }
                    }
                    else if (negBalance >= 3) {
                        if (posBalance < 82) {
                            if (negBalance < 121) {
                                if (diffMVdiffVal < 73) {
                                    if (MVDiff < 73) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 73) {
                                        if (ssDCval < 93) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 93) {
                                            if (Rs < 8571) {
                                                return 1;
                                            }
                                            else if (Rs >= 8571) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 73) {
                                    if (diffMVdiffVal < 781) {
                                        if (posBalance < 28) {
                                            if (posBalance < 27) {
                                                return 1;
                                            }
                                            else if (posBalance >= 27) {
                                                if (CsDiff < 7) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 7) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 28) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 781) {
                                        if (posBalance < 71) {
                                            return 1;
                                        }
                                        else if (posBalance >= 71) {
                                            if (RsDiff < 10) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 10) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 121) {
                                return 0;
                            }
                        }
                        else if (posBalance >= 82) {
                            if (diffRsCsdiff < 218) {
                                if (Cs < 920) {
                                    if (CsDiff < 5) {
                                        if (Rs < 1707) {
                                            return 0;
                                        }
                                        else if (Rs >= 1707) {
                                            return 1;
                                        }
                                    }
                                    else if (CsDiff >= 5) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 920) {
                                    if (diffAFD < 43) {
                                        if (Rs < 1020) {
                                            return 1;
                                        }
                                        else if (Rs >= 1020) {
                                            return 0;
                                        }
                                    }
                                    else if (diffAFD >= 43) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 218) {
                                if (refDCval < 74) {
                                    return 1;
                                }
                                else if (refDCval >= 74) {
                                    if (refDCval < 75) {
                                        return 0;
                                    }
                                    else if (refDCval >= 75) {
                                        if (MVDiff < 130) {
                                            if (Rs < 2476) {
                                                return 0;
                                            }
                                            else if (Rs >= 2476) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 130) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect6(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffMVdiffVal < 65) {
        if (diffAFD < 13) {
            if (gchDC < 22) {
                if (refDCval < 24) {
                    if (CsDiff < 12) {
                        if (MVDiff < 1517) {
                            if (RsCsDiff < 7) {
                                return 0;
                            }
                            else if (RsCsDiff >= 7) {
                                if (posBalance < 2) {
                                    if (SC < 1346) {
                                        return 1;
                                    }
                                    else if (SC >= 1346) {
                                        if (CsDiff < 4) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 4) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (posBalance >= 2) {
                                    return 0;
                                }
                            }
                        }
                        else if (MVDiff >= 1517) {
                            if (AFD < 1) {
                                return 0;
                            }
                            else if (AFD >= 1) {
                                return 1;
                            }
                        }
                    }
                    else if (CsDiff >= 12) {
                        return 1;
                    }
                }
                else if (refDCval >= 24) {
                    if (MVDiff < 55) {
                        if (refDCval < 230) {
                            return 0;
                        }
                        else if (refDCval >= 230) {
                            if (TSC < 4) {
                                return 0;
                            }
                            else if (TSC >= 4) {
                                return 1;
                            }
                        }
                    }
                    else if (MVDiff >= 55) {
                        if (RsDiff < 32) {
                            if (diffTSC < 5) {
                                if (MVDiff < 224) {
                                    return 0;
                                }
                                else if (MVDiff >= 224) {
                                    if (AFD < 22) {
                                        return 0;
                                    }
                                    else if (AFD >= 22) {
                                        if (refDCval < 52) {
                                            if (gchDC < 3) {
                                                return 0;
                                            }
                                            else if (gchDC >= 3) {
                                                if (gchDC < 10) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 10) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 52) {
                                            if (diffMVdiffVal < 53) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 53) {
                                                if (Rs < 859) {
                                                    return 0;
                                                }
                                                else if (Rs >= 859) {
                                                    if (Cs < 478) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 478) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 5) {
                                if (ssDCval < 46) {
                                    if (negBalance < 46) {
                                        return 0;
                                    }
                                    else if (negBalance >= 46) {
                                        if (Rs < 813) {
                                            if (Rs < 354) {
                                                if (SC < 320) {
                                                    return 1;
                                                }
                                                else if (SC >= 320) {
                                                    return 0;
                                                }
                                            }
                                            else if (Rs >= 354) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 813) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (ssDCval >= 46) {
                                    return 0;
                                }
                            }
                        }
                        else if (RsDiff >= 32) {
                            if (posBalance < 71) {
                                if (SC < 3912) {
                                    if (ssDCval < 101) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 101) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 3912) {
                                    return 0;
                                }
                            }
                            else if (posBalance >= 71) {
                                if (AFD < 37) {
                                    if (Rs < 3555) {
                                        return 0;
                                    }
                                    else if (Rs >= 3555) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 37) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (gchDC >= 22) {
                if (RsCsDiff < 767) {
                    if (refDCval < 25) {
                        return 1;
                    }
                    else if (refDCval >= 25) {
                        if (diffRsCsdiff < 63) {
                            if (refDCval < 226) {
                                if (refDCval < 33) {
                                    if (SC < 1321) {
                                        return 1;
                                    }
                                    else if (SC >= 1321) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 33) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 226) {
                                if (SC < 289) {
                                    return 1;
                                }
                                else if (SC >= 289) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 63) {
                            if (MVDiff < 226) {
                                return 0;
                            }
                            else if (MVDiff >= 226) {
                                if (ssDCval < 95) {
                                    if (TSC < 18) {
                                        return 0;
                                    }
                                    else if (TSC >= 18) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 95) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (RsCsDiff >= 767) {
                    if (negBalance < 92) {
                        if (SC < 4758) {
                            if (RsDiff < 13) {
                                return 0;
                            }
                            else if (RsDiff >= 13) {
                                return 1;
                            }
                        }
                        else if (SC >= 4758) {
                            return 0;
                        }
                    }
                    else if (negBalance >= 92) {
                        return 0;
                    }
                }
            }
        }
        else if (diffAFD >= 13) {
            if (diffTSC < 13) {
                if (MVDiff < 56) {
                    if (Cs < 180) {
                        if (Cs < 158) {
                            return 0;
                        }
                        else if (Cs >= 158) {
                            return 1;
                        }
                    }
                    else if (Cs >= 180) {
                        return 0;
                    }
                }
                else if (MVDiff >= 56) {
                    if (negBalance < 62) {
                        if (diffMVdiffVal < 50) {
                            if (ssDCval < 61) {
                                if (MVDiff < 93) {
                                    return 0;
                                }
                                else if (MVDiff >= 93) {
                                    if (gchDC < 9) {
                                        return 1;
                                    }
                                    else if (gchDC >= 9) {
                                        if (Rs < 1544) {
                                            return 0;
                                        }
                                        else if (Rs >= 1544) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 61) {
                                return 0;
                            }
                        }
                        else if (diffMVdiffVal >= 50) {
                            if (CsDiff < 8) {
                                return 0;
                            }
                            else if (CsDiff >= 8) {
                                if (SC < 3913) {
                                    if (diffTSC < 6) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 6) {
                                        if (RsCsDiff < 127) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 127) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (SC >= 3913) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 62) {
                        if (diffRsCsdiff < 15) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 15) {
                            if (CsDiff < 8) {
                                if (Rs < 95) {
                                    return 0;
                                }
                                else if (Rs >= 95) {
                                    if (SC < 6043) {
                                        return 1;
                                    }
                                    else if (SC >= 6043) {
                                        return 0;
                                    }
                                }
                            }
                            else if (CsDiff >= 8) {
                                if (refDCval < 96) {
                                    return 1;
                                }
                                else if (refDCval >= 96) {
                                    if (AFD < 60) {
                                        return 0;
                                    }
                                    else if (AFD >= 60) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 13) {
                if (RsCsDiff < 483) {
                    if (posBalance < 84) {
                        if (diffMVdiffVal < -44) {
                            if (RsDiff < 6) {
                                return 0;
                            }
                            else if (RsDiff >= 6) {
                                return 1;
                            }
                        }
                        else if (diffMVdiffVal >= -44) {
                            if (RsDiff < 9) {
                                if (MVDiff < 16) {
                                    return 1;
                                }
                                else if (MVDiff >= 16) {
                                    return 0;
                                }
                            }
                            else if (RsDiff >= 9) {
                                if (Rs < 2096) {
                                    if (TSC < 24) {
                                        if (gchDC < 17) {
                                            return 1;
                                        }
                                        else if (gchDC >= 17) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 24) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 2096) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (posBalance >= 84) {
                        if (refDCval < 37) {
                            if (diffMVdiffVal < 4) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 4) {
                                if (posBalance < 117) {
                                    if (MVDiff < 134) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 134) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 117) {
                                    return 1;
                                }
                            }
                        }
                        else if (refDCval >= 37) {
                            return 0;
                        }
                    }
                }
                else if (RsCsDiff >= 483) {
                    if (Cs < 3040) {
                        if (RsCsDiff < 642) {
                            if (SC < 3874) {
                                if (ssDCval < 133) {
                                    return 1;
                                }
                                else if (ssDCval >= 133) {
                                    return 0;
                                }
                            }
                            else if (SC >= 3874) {
                                if (AFD < 63) {
                                    return 0;
                                }
                                else if (AFD >= 63) {
                                    return 1;
                                }
                            }
                        }
                        else if (RsCsDiff >= 642) {
                            return 1;
                        }
                    }
                    else if (Cs >= 3040) {
                        if (gchDC < 9) {
                            if (Cs < 12050) {
                                return 0;
                            }
                            else if (Cs >= 12050) {
                                return 1;
                            }
                        }
                        else if (gchDC >= 9) {
                            if (RsDiff < 17) {
                                return 0;
                            }
                            else if (RsDiff >= 17) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffMVdiffVal >= 65) {
        if (diffAFD < 13) {
            if (diffAFD < 8) {
                if (TSC < 32) {
                    if (ssDCval < 30) {
                        if (diffTSC < 3) {
                            if (diffTSC < -16) {
                                return 1;
                            }
                            else if (diffTSC >= -16) {
                                if (diffAFD < 2) {
                                    if (Rs < 420) {
                                        return 0;
                                    }
                                    else if (Rs >= 420) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= 2) {
                                    if (gchDC < 72) {
                                        if (SC < 131) {
                                            return 0;
                                        }
                                        else if (SC >= 131) {
                                            if (Rs < 85) {
                                                return 1;
                                            }
                                            else if (Rs >= 85) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 72) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 3) {
                            if (Rs < 181) {
                                return 0;
                            }
                            else if (Rs >= 181) {
                                return 1;
                            }
                        }
                    }
                    else if (ssDCval >= 30) {
                        if (gchDC < 93) {
                            if (CsDiff < 5) {
                                if (refDCval < 35) {
                                    if (diffMVdiffVal < 268) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 268) {
                                        if (SC < 119) {
                                            return 1;
                                        }
                                        else if (SC >= 119) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 35) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 5) {
                                if (AFD < 36) {
                                    if (ssDCval < 35) {
                                        if (TSC < 11) {
                                            return 0;
                                        }
                                        else if (TSC >= 11) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 35) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 36) {
                                    if (SC < 1612) {
                                        if (Rs < 999) {
                                            if (posBalance < 42) {
                                                return 0;
                                            }
                                            else if (posBalance >= 42) {
                                                if (Rs < 188) {
                                                    return 1;
                                                }
                                                else if (Rs >= 188) {
                                                    if (posBalance < 49) {
                                                        return 1;
                                                    }
                                                    else if (posBalance >= 49) {
                                                        if (refDCval < 161) {
                                                            return 0;
                                                        }
                                                        else if (refDCval >= 161) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (Rs >= 999) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 1612) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 93) {
                            if (RsDiff < 3) {
                                return 0;
                            }
                            else if (RsDiff >= 3) {
                                return 1;
                            }
                        }
                    }
                }
                else if (TSC >= 32) {
                    if (Rs < 830) {
                        return 0;
                    }
                    else if (Rs >= 830) {
                        if (AFD < 51) {
                            return 0;
                        }
                        else if (AFD >= 51) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffAFD >= 8) {
                if (ssDCval < 42) {
                    if (gchDC < 10) {
                        if (ssDCval < 41) {
                            if (refDCval < 25) {
                                if (Rs < 156) {
                                    return 1;
                                }
                                else if (Rs >= 156) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 25) {
                                if (SC < 75) {
                                    if (Rs < 30) {
                                        return 1;
                                    }
                                    else if (Rs >= 30) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 75) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 41) {
                            return 0;
                        }
                    }
                    else if (gchDC >= 10) {
                        if (Cs < 184) {
                            if (CsDiff < 1) {
                                return 0;
                            }
                            else if (CsDiff >= 1) {
                                if (RsDiff < 13) {
                                    return 1;
                                }
                                else if (RsDiff >= 13) {
                                    if (Cs < 45) {
                                        return 1;
                                    }
                                    else if (Cs >= 45) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 184) {
                            return 0;
                        }
                    }
                }
                else if (ssDCval >= 42) {
                    if (refDCval < 80) {
                        if (Scindex < 3) {
                            if (diffRsCsdiff < 31) {
                                if (AFD < 14) {
                                    if (ssDCval < 54) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 54) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 14) {
                                    if (diffRsCsdiff < 6) {
                                        if (posBalance < 114) {
                                            return 0;
                                        }
                                        else if (posBalance >= 114) {
                                            return 1;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 6) {
                                        if (ssDCval < 50) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 50) {
                                            if (Rs < 90) {
                                                return 1;
                                            }
                                            else if (Rs >= 90) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 31) {
                                return 1;
                            }
                        }
                        else if (Scindex >= 3) {
                            if (CsDiff < 7) {
                                return 0;
                            }
                            else if (CsDiff >= 7) {
                                if (SC < 1769) {
                                    if (TSC < 9) {
                                        return 0;
                                    }
                                    else if (TSC >= 9) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 1769) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 80) {
                        if (ssDCval < 133) {
                            return 0;
                        }
                        else if (ssDCval >= 133) {
                            if (AFD < 38) {
                                return 0;
                            }
                            else if (AFD >= 38) {
                                if (CsDiff < 9) {
                                    return 0;
                                }
                                else if (CsDiff >= 9) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 13) {
            if (diffRsCsdiff < 66) {
                if (ssDCval < 66) {
                    if (negBalance < 17) {
                        if (ssDCval < 43) {
                            return 1;
                        }
                        else if (ssDCval >= 43) {
                            return 0;
                        }
                    }
                    else if (negBalance >= 17) {
                        if (RsDiff < 6) {
                            if (AFD < 39) {
                                if (diffTSC < 7) {
                                    if (refDCval < 59) {
                                        if (TSC < 6) {
                                            if (Rs < 79) {
                                                return 1;
                                            }
                                            else if (Rs >= 79) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 6) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 59) {
                                        if (CsDiff < 6) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 6) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffTSC >= 7) {
                                    if (SC < 590) {
                                        return 1;
                                    }
                                    else if (SC >= 590) {
                                        if (refDCval < 38) {
                                            if (RsCsDiff < 29) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 29) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 38) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 39) {
                                if (CsDiff < 3) {
                                    if (SC < 374) {
                                        return 0;
                                    }
                                    else if (SC >= 374) {
                                        return 1;
                                    }
                                }
                                else if (CsDiff >= 3) {
                                    if (MVDiff < 164) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 164) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 6) {
                            if (Rs < 273) {
                                return 1;
                            }
                            else if (Rs >= 273) {
                                if (SC < 1216) {
                                    return 0;
                                }
                                else if (SC >= 1216) {
                                    if (Rs < 1024) {
                                        return 0;
                                    }
                                    else if (Rs >= 1024) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 66) {
                    if (Cs < 728) {
                        if (diffTSC < 7) {
                            if (diffAFD < 24) {
                                return 0;
                            }
                            else if (diffAFD >= 24) {
                                return 1;
                            }
                        }
                        else if (diffTSC >= 7) {
                            if (negBalance < 14) {
                                if (TSC < 13) {
                                    if (Rs < 238) {
                                        return 1;
                                    }
                                    else if (Rs >= 238) {
                                        return 0;
                                    }
                                }
                                else if (TSC >= 13) {
                                    if (gchDC < 57) {
                                        if (gchDC < 21) {
                                            return 1;
                                        }
                                        else if (gchDC >= 21) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 57) {
                                        if (Cs < 24) {
                                            return 0;
                                        }
                                        else if (Cs >= 24) {
                                            if (TSC < 14) {
                                                return 0;
                                            }
                                            else if (TSC >= 14) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 14) {
                                if (TSC < 33) {
                                    if (SC < 1076) {
                                        if (diffTSC < 8) {
                                            if (SC < 476) {
                                                if (Rs < 42) {
                                                    return 0;
                                                }
                                                else if (Rs >= 42) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 476) {
                                                return 0;
                                            }
                                        }
                                        else if (diffTSC >= 8) {
                                            if (Cs < 616) {
                                                if (diffMVdiffVal < 95) {
                                                    if (CsDiff < 3) {
                                                        return 0;
                                                    }
                                                    else if (CsDiff >= 3) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 95) {
                                                    return 1;
                                                }
                                            }
                                            else if (Cs >= 616) {
                                                if (Cs < 659) {
                                                    return 0;
                                                }
                                                else if (Cs >= 659) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 1076) {
                                        if (diffTSC < 15) {
                                            if (refDCval < 127) {
                                                return 0;
                                            }
                                            else if (refDCval >= 127) {
                                                return 1;
                                            }
                                        }
                                        else if (diffTSC >= 15) {
                                            if (SC < 1115) {
                                                return 0;
                                            }
                                            else if (SC >= 1115) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 33) {
                                    if (Rs < 872) {
                                        if (CsDiff < 2) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 2) {
                                            if (AFD < 74) {
                                                return 0;
                                            }
                                            else if (AFD >= 74) {
                                                if (Cs < 498) {
                                                    return 1;
                                                }
                                                else if (Cs >= 498) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (Rs >= 872) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 728) {
                        if (CsDiff < 5) {
                            if (Rs < 1194) {
                                return 0;
                            }
                            else if (Rs >= 1194) {
                                return 1;
                            }
                        }
                        else if (CsDiff >= 5) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 66) {
                if (diffTSC < 12) {
                    if (Rs < 1258) {
                        if (SC < 824) {
                            if (MVDiff < 741) {
                                return 1;
                            }
                            else if (MVDiff >= 741) {
                                return 0;
                            }
                        }
                        else if (SC >= 824) {
                            if (negBalance < 33) {
                                if (diffRsCsdiff < 184) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 184) {
                                    if (Rs < 894) {
                                        return 1;
                                    }
                                    else if (Rs >= 894) {
                                        return 0;
                                    }
                                }
                            }
                            else if (negBalance >= 33) {
                                if (diffMVdiffVal < 66) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 66) {
                                    if (Cs < 241) {
                                        return 0;
                                    }
                                    else if (Cs >= 241) {
                                        if (TSC < 10) {
                                            if (ssDCval < 36) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 36) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 10) {
                                            if (refDCval < 41) {
                                                if (diffTSC < 9) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 9) {
                                                    return 1;
                                                }
                                            }
                                            else if (refDCval >= 41) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 1258) {
                        if (diffMVdiffVal < 416) {
                            if (posBalance < 60) {
                                if (refDCval < 77) {
                                    if (MVDiff < 236) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 236) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 77) {
                                    if (posBalance < 52) {
                                        if (diffAFD < 25) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 25) {
                                            if (Rs < 1934) {
                                                return 1;
                                            }
                                            else if (Rs >= 1934) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 52) {
                                        if (TSC < 17) {
                                            return 0;
                                        }
                                        else if (TSC >= 17) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 60) {
                                return 0;
                            }
                        }
                        else if (diffMVdiffVal >= 416) {
                            return 1;
                        }
                    }
                }
                else if (diffTSC >= 12) {
                    if (negBalance < 26) {
                        if (diffAFD < 37) {
                            if (Rs < 819) {
                                return 1;
                            }
                            else if (Rs >= 819) {
                                if (RsCsDiff < 1029) {
                                    if (refDCval < 43) {
                                        if (MVDiff < 353) {
                                            if (negBalance < 8) {
                                                if (SC < 2932) {
                                                    return 0;
                                                }
                                                else if (SC >= 2932) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 8) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 353) {
                                            return 0;
                                        }
                                    }
                                    else if (refDCval >= 43) {
                                        if (diffAFD < 32) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 32) {
                                            if (TSC < 32) {
                                                return 1;
                                            }
                                            else if (TSC >= 32) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 1029) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffAFD >= 37) {
                            if (negBalance < 1) {
                                if (SC < 2996) {
                                    return 1;
                                }
                                else if (SC >= 2996) {
                                    return 0;
                                }
                            }
                            else if (negBalance >= 1) {
                                if (RsCsDiff < 188) {
                                    if (SC < 2190) {
                                        return 1;
                                    }
                                    else if (SC >= 2190) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 188) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 26) {
                        if (diffAFD < 26) {
                            if (gchDC < 19) {
                                return 1;
                            }
                            else if (gchDC >= 19) {
                                if (ssDCval < 123) {
                                    if (negBalance < 75) {
                                        if (diffTSC < 19) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 19) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 75) {
                                        if (Cs < 4619) {
                                            return 1;
                                        }
                                        else if (Cs >= 4619) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (ssDCval >= 123) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffAFD >= 26) {
                            if (refDCval < 65) {
                                if (diffMVdiffVal < 703) {
                                    if (diffTSC < 18) {
                                        if (AFD < 40) {
                                            return 1;
                                        }
                                        else if (AFD >= 40) {
                                            if (diffMVdiffVal < 214) {
                                                if (MVDiff < 134) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 134) {
                                                    if (ssDCval < 83) {
                                                        return 0;
                                                    }
                                                    else if (ssDCval >= 83) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= 214) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 18) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 703) {
                                    if (Cs < 2582) {
                                        return 0;
                                    }
                                    else if (Cs >= 2582) {
                                        return 1;
                                    }
                                }
                            }
                            else if (refDCval >= 65) {
                                if (SC < 16399) {
                                    if (ssDCval < 55) {
                                        if (Rs < 1324) {
                                            return 1;
                                        }
                                        else if (Rs >= 1324) {
                                            if (Rs < 1381) {
                                                return 0;
                                            }
                                            else if (Rs >= 1381) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 55) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 16399) {
                                    if (SC < 17009) {
                                        return 0;
                                    }
                                    else if (SC >= 17009) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect7(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 14) {
        if (MVDiff < 82) {
            if (diffRsCsdiff < 248) {
                if (ssDCval < 18) {
                    if (RsCsDiff < 7) {
                        return 0;
                    }
                    else if (RsCsDiff >= 7) {
                        if (Cs < 570) {
                            return 0;
                        }
                        else if (Cs >= 570) {
                            if (CsDiff < 4) {
                                if (MVDiff < 4) {
                                    return 1;
                                }
                                else if (MVDiff >= 4) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 4) {
                                return 1;
                            }
                        }
                    }
                }
                else if (ssDCval >= 18) {
                    if (diffMVdiffVal < 67) {
                        if (refDCval < 226) {
                            if (ssDCval < 229) {
                                return 0;
                            }
                            else if (ssDCval >= 229) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 226) {
                            if (ssDCval < 226) {
                                return 1;
                            }
                            else if (ssDCval >= 226) {
                                return 0;
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 67) {
                        if (Rs < 302) {
                            return 0;
                        }
                        else if (Rs >= 302) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 248) {
                if (refDCval < 36) {
                    if (gchDC < 1) {
                        if (Rs < 2749) {
                            return 1;
                        }
                        else if (Rs >= 2749) {
                            return 0;
                        }
                    }
                    else if (gchDC >= 1) {
                        return 1;
                    }
                }
                else if (refDCval >= 36) {
                    if (diffMVdiffVal < 14) {
                        return 0;
                    }
                    else if (diffMVdiffVal >= 14) {
                        if (refDCval < 80) {
                            if (gchDC < 7) {
                                if (Rs < 5571) {
                                    return 0;
                                }
                                else if (Rs >= 5571) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 7) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 80) {
                            if (TSC < 4) {
                                if (Rs < 2111) {
                                    return 1;
                                }
                                else if (Rs >= 2111) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 4) {
                                return 0;
                            }
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 82) {
            if (diffMVdiffVal < 76) {
                if (AFD < 59) {
                    if (diffTSC < 6) {
                        if (refDCval < 49) {
                            if (AFD < 16) {
                                if (MVDiff < 1538) {
                                    if (RsCsDiff < 5) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 5) {
                                        if (refDCval < 29) {
                                            if (negBalance < 39) {
                                                if (AFD < 1) {
                                                    return 1;
                                                }
                                                else if (AFD >= 1) {
                                                    return 0;
                                                }
                                            }
                                            else if (negBalance >= 39) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 29) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (MVDiff >= 1538) {
                                    if (AFD < 1) {
                                        return 0;
                                    }
                                    else if (AFD >= 1) {
                                        if (Rs < 41) {
                                            return 1;
                                        }
                                        else if (Rs >= 41) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 16) {
                                if (diffRsCsdiff < 1) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 1) {
                                    if (Cs < 257) {
                                        if (Rs < 269) {
                                            return 1;
                                        }
                                        else if (Rs >= 269) {
                                            return 0;
                                        }
                                    }
                                    else if (Cs >= 257) {
                                        if (Cs < 747) {
                                            return 0;
                                        }
                                        else if (Cs >= 747) {
                                            if (AFD < 23) {
                                                return 0;
                                            }
                                            else if (AFD >= 23) {
                                                if (diffTSC < 2) {
                                                    return 1;
                                                }
                                                else if (diffTSC >= 2) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 49) {
                            if (AFD < 43) {
                                if (RsCsDiff < 4482) {
                                    if (ssDCval < 37) {
                                        if (Rs < 47) {
                                            if (Rs < 11) {
                                                return 0;
                                            }
                                            else if (Rs >= 11) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 47) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 37) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 4482) {
                                    if (Rs < 3227) {
                                        return 1;
                                    }
                                    else if (Rs >= 3227) {
                                        return 0;
                                    }
                                }
                            }
                            else if (AFD >= 43) {
                                if (ssDCval < 76) {
                                    if (ssDCval < 73) {
                                        if (ssDCval < 27) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 27) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 73) {
                                        if (Rs < 1696) {
                                            return 1;
                                        }
                                        else if (Rs >= 1696) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (ssDCval >= 76) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 6) {
                        if (diffAFD < 6) {
                            if (ssDCval < 92) {
                                if (posBalance < 97) {
                                    if (diffMVdiffVal < -24) {
                                        if (Cs < 367) {
                                            return 0;
                                        }
                                        else if (Cs >= 367) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= -24) {
                                        if (RsDiff < 17) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 17) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (posBalance >= 97) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 92) {
                                return 0;
                            }
                        }
                        else if (diffAFD >= 6) {
                            return 0;
                        }
                    }
                }
                else if (AFD >= 59) {
                    if (SC < 1203) {
                        return 0;
                    }
                    else if (SC >= 1203) {
                        if (Rs < 2800) {
                            return 1;
                        }
                        else if (Rs >= 2800) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 76) {
                if (diffRsCsdiff < 1) {
                    if (diffAFD < 8) {
                        if (ssDCval < 31) {
                            if (AFD < 6) {
                                return 0;
                            }
                            else if (AFD >= 6) {
                                if (AFD < 26) {
                                    if (gchDC < 56) {
                                        if (RsCsDiff < 43) {
                                            if (refDCval < 23) {
                                                return 1;
                                            }
                                            else if (refDCval >= 23) {
                                                return 0;
                                            }
                                        }
                                        else if (RsCsDiff >= 43) {
                                            return 1;
                                        }
                                    }
                                    else if (gchDC >= 56) {
                                        if (gchDC < 177) {
                                            return 1;
                                        }
                                        else if (gchDC >= 177) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (AFD >= 26) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 31) {
                            if (AFD < 61) {
                                if (TSC < 25) {
                                    return 0;
                                }
                                else if (TSC >= 25) {
                                    if (gchDC < 13) {
                                        if (Cs < 1658) {
                                            return 1;
                                        }
                                        else if (Cs >= 1658) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 13) {
                                        return 0;
                                    }
                                }
                            }
                            else if (AFD >= 61) {
                                return 1;
                            }
                        }
                    }
                    else if (diffAFD >= 8) {
                        if (ssDCval < 42) {
                            if (diffAFD < 10) {
                                return 1;
                            }
                            else if (diffAFD >= 10) {
                                if (Cs < 39) {
                                    return 0;
                                }
                                else if (Cs >= 39) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 42) {
                            if (Rs < 3839) {
                                if (negBalance < 11) {
                                    if (Rs < 61) {
                                        return 1;
                                    }
                                    else if (Rs >= 61) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 11) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 3839) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 1) {
                    if (refDCval < 41) {
                        if (MVDiff < 113) {
                            return 1;
                        }
                        else if (MVDiff >= 113) {
                            if (gchDC < 8) {
                                if (posBalance < 60) {
                                    if (ssDCval < 26) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 26) {
                                        if (posBalance < 51) {
                                            if (TSC < 5) {
                                                return 0;
                                            }
                                            else if (TSC >= 5) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 51) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (posBalance >= 60) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 8) {
                                if (SC < 226) {
                                    if (TSC < 5) {
                                        return 0;
                                    }
                                    else if (TSC >= 5) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 226) {
                                    if (Rs < 2259) {
                                        return 0;
                                    }
                                    else if (Rs >= 2259) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 41) {
                        if (TSC < 34) {
                            if (diffAFD < 11) {
                                if (ssDCval < 48) {
                                    if (Cs < 650) {
                                        if (diffTSC < 3) {
                                            if (diffRsCsdiff < 35) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 35) {
                                                if (AFD < 18) {
                                                    return 0;
                                                }
                                                else if (AFD >= 18) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffTSC >= 3) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 650) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 48) {
                                    if (Rs < 5) {
                                        return 1;
                                    }
                                    else if (Rs >= 5) {
                                        if (TSC < 14) {
                                            return 0;
                                        }
                                        else if (TSC >= 14) {
                                            if (Scindex < 4) {
                                                if (Cs < 518) {
                                                    return 0;
                                                }
                                                else if (Cs >= 518) {
                                                    if (SC < 1253) {
                                                        if (refDCval < 101) {
                                                            return 0;
                                                        }
                                                        else if (refDCval >= 101) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (SC >= 1253) {
                                                        if (Cs < 552) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 552) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Scindex >= 4) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 11) {
                                if (ssDCval < 86) {
                                    if (negBalance < 30) {
                                        if (SC < 427) {
                                            return 1;
                                        }
                                        else if (SC >= 427) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 30) {
                                        if (negBalance < 101) {
                                            if (diffMVdiffVal < 94) {
                                                if (Cs < 30) {
                                                    return 1;
                                                }
                                                else if (Cs >= 30) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 94) {
                                                if (ssDCval < 42) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 42) {
                                                    if (Rs < 956) {
                                                        if (diffRsCsdiff < 23) {
                                                            if (Rs < 179) {
                                                                if (posBalance < 33) {
                                                                    return 0;
                                                                }
                                                                else if (posBalance >= 33) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (Rs >= 179) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (diffRsCsdiff >= 23) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Rs >= 956) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 101) {
                                            if (ssDCval < 39) {
                                                if (refDCval < 52) {
                                                    return 0;
                                                }
                                                else if (refDCval >= 52) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 39) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 86) {
                                    return 0;
                                }
                            }
                        }
                        else if (TSC >= 34) {
                            return 1;
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 14) {
        if (diffAFD < 23) {
            if (diffMVdiffVal < 54) {
                if (ssDCval < 68) {
                    if (Cs < 707) {
                        if (negBalance < 48) {
                            if (diffTSC < 16) {
                                if (Rs < 1137) {
                                    return 0;
                                }
                                else if (Rs >= 1137) {
                                    return 1;
                                }
                            }
                            else if (diffTSC >= 16) {
                                if (Rs < 534) {
                                    return 1;
                                }
                                else if (Rs >= 534) {
                                    return 0;
                                }
                            }
                        }
                        else if (negBalance >= 48) {
                            if (refDCval < 76) {
                                if (CsDiff < 8) {
                                    return 1;
                                }
                                else if (CsDiff >= 8) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 76) {
                                return 0;
                            }
                        }
                    }
                    else if (Cs >= 707) {
                        return 0;
                    }
                }
                else if (ssDCval >= 68) {
                    if (diffTSC < 12) {
                        if (gchDC < 46) {
                            return 0;
                        }
                        else if (gchDC >= 46) {
                            if (Rs < 237) {
                                return 1;
                            }
                            else if (Rs >= 237) {
                                return 0;
                            }
                        }
                    }
                    else if (diffTSC >= 12) {
                        if (diffMVdiffVal < -62) {
                            if (gchDC < 27) {
                                return 1;
                            }
                            else if (gchDC >= 27) {
                                if (Cs < 1738) {
                                    return 0;
                                }
                                else if (Cs >= 1738) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= -62) {
                            if (RsDiff < 99) {
                                if (Rs < 254) {
                                    if (Rs < 165) {
                                        return 0;
                                    }
                                    else if (Rs >= 165) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 254) {
                                    return 0;
                                }
                            }
                            else if (RsDiff >= 99) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 54) {
                if (AFD < 25) {
                    if (refDCval < 102) {
                        if (Rs < 233) {
                            if (diffMVdiffVal < 135) {
                                if (ssDCval < 83) {
                                    if (TSC < 6) {
                                        return 0;
                                    }
                                    else if (TSC >= 6) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 83) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 135) {
                                return 1;
                            }
                        }
                        else if (Rs >= 233) {
                            if (posBalance < 92) {
                                if (TSC < 10) {
                                    if (Rs < 293) {
                                        return 0;
                                    }
                                    else if (Rs >= 293) {
                                        if (negBalance < 38) {
                                            if (Rs < 380) {
                                                return 1;
                                            }
                                            else if (Rs >= 380) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 38) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (TSC >= 10) {
                                    return 1;
                                }
                            }
                            else if (posBalance >= 92) {
                                if (refDCval < 35) {
                                    return 1;
                                }
                                else if (refDCval >= 35) {
                                    if (MVDiff < 785) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 785) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (refDCval >= 102) {
                        if (ssDCval < 88) {
                            return 1;
                        }
                        else if (ssDCval >= 88) {
                            if (gchDC < 2) {
                                return 1;
                            }
                            else if (gchDC >= 2) {
                                return 0;
                            }
                        }
                    }
                }
                else if (AFD >= 25) {
                    if (ssDCval < 65) {
                        if (diffRsCsdiff < 58) {
                            if (Cs < 154) {
                                if (diffAFD < 22) {
                                    if (diffRsCsdiff < -11) {
                                        if (SC < 4) {
                                            return 1;
                                        }
                                        else if (SC >= 4) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= -11) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= 22) {
                                    return 0;
                                }
                            }
                            else if (Cs >= 154) {
                                if (refDCval < 68) {
                                    return 0;
                                }
                                else if (refDCval >= 68) {
                                    if (Rs < 1516) {
                                        return 1;
                                    }
                                    else if (Rs >= 1516) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 58) {
                            if (Rs < 2094) {
                                if (posBalance < 66) {
                                    return 1;
                                }
                                else if (posBalance >= 66) {
                                    if (Rs < 531) {
                                        return 1;
                                    }
                                    else if (Rs >= 531) {
                                        if (Cs < 783) {
                                            return 0;
                                        }
                                        else if (Cs >= 783) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 2094) {
                                return 0;
                            }
                        }
                    }
                    else if (ssDCval >= 65) {
                        if (diffTSC < 6) {
                            if (Cs < 1) {
                                return 1;
                            }
                            else if (Cs >= 1) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 6) {
                            if (diffRsCsdiff < 166) {
                                if (negBalance < 36) {
                                    if (diffMVdiffVal < 674) {
                                        if (diffRsCsdiff < 33) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 33) {
                                            if (Cs < 182) {
                                                return 1;
                                            }
                                            else if (Cs >= 182) {
                                                if (posBalance < 86) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 86) {
                                                    if (RsCsDiff < 47) {
                                                        if (AFD < 27) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 27) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (RsCsDiff >= 47) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 674) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 36) {
                                    if (gchDC < 7) {
                                        return 1;
                                    }
                                    else if (gchDC >= 7) {
                                        if (posBalance < 79) {
                                            if (RsDiff < 9) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 9) {
                                                if (Rs < 1344) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1344) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (posBalance >= 79) {
                                            if (TSC < 27) {
                                                return 1;
                                            }
                                            else if (TSC >= 27) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 166) {
                                if (SC < 3212) {
                                    if (ssDCval < 159) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 159) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 3212) {
                                    if (diffRsCsdiff < 784) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 784) {
                                        if (ssDCval < 137) {
                                            if (AFD < 50) {
                                                return 1;
                                            }
                                            else if (AFD >= 50) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 137) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 23) {
            if (diffMVdiffVal < 48) {
                if (diffTSC < 22) {
                    if (MVDiff < 107) {
                        return 0;
                    }
                    else if (MVDiff >= 107) {
                        if (negBalance < 59) {
                            if (Cs < 468) {
                                if (Cs < 192) {
                                    return 0;
                                }
                                else if (Cs >= 192) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 468) {
                                return 0;
                            }
                        }
                        else if (negBalance >= 59) {
                            if (RsDiff < 5) {
                                return 0;
                            }
                            else if (RsDiff >= 5) {
                                if (AFD < 66) {
                                    return 1;
                                }
                                else if (AFD >= 66) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffTSC >= 22) {
                    if (RsCsDiff < 214) {
                        return 0;
                    }
                    else if (RsCsDiff >= 214) {
                        if (diffTSC < 24) {
                            if (MVDiff < 154) {
                                return 0;
                            }
                            else if (MVDiff >= 154) {
                                return 1;
                            }
                        }
                        else if (diffTSC >= 24) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 48) {
                if (posBalance < 98) {
                    if (diffTSC < 11) {
                        if (Rs < 443) {
                            if (AFD < 74) {
                                if (AFD < 24) {
                                    if (Rs < 198) {
                                        return 1;
                                    }
                                    else if (Rs >= 198) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 24) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 74) {
                                return 0;
                            }
                        }
                        else if (Rs >= 443) {
                            if (MVDiff < 165) {
                                return 0;
                            }
                            else if (MVDiff >= 165) {
                                if (CsDiff < 17) {
                                    if (SC < 1838) {
                                        if (TSC < 20) {
                                            return 1;
                                        }
                                        else if (TSC >= 20) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 1838) {
                                        if (MVDiff < 237) {
                                            if (SC < 2591) {
                                                return 0;
                                            }
                                            else if (SC >= 2591) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 237) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (CsDiff >= 17) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 11) {
                        if (diffRsCsdiff < 5) {
                            if (Cs < 185) {
                                if (CsDiff < 1) {
                                    return 0;
                                }
                                else if (CsDiff >= 1) {
                                    if (MVDiff < 114) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 114) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 185) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 5) {
                            if (diffRsCsdiff < 410) {
                                if (gchDC < 21) {
                                    if (SC < 3100) {
                                        if (diffMVdiffVal < 77) {
                                            if (RsDiff < 11) {
                                                if (Rs < 464) {
                                                    return 1;
                                                }
                                                else if (Rs >= 464) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsDiff >= 11) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 77) {
                                            if (negBalance < 41) {
                                                if (Rs < 298) {
                                                    if (SC < 532) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 532) {
                                                        if (CsDiff < 5) {
                                                            return 1;
                                                        }
                                                        else if (CsDiff >= 5) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (Rs >= 298) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 41) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 3100) {
                                        if (diffMVdiffVal < 172) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 172) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (gchDC >= 21) {
                                    if (diffAFD < 28) {
                                        if (negBalance < 79) {
                                            return 0;
                                        }
                                        else if (negBalance >= 79) {
                                            if (ssDCval < 114) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 114) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffAFD >= 28) {
                                        if (negBalance < 69) {
                                            if (refDCval < 35) {
                                                return 0;
                                            }
                                            else if (refDCval >= 35) {
                                                if (ssDCval < 56) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 56) {
                                                    if (AFD < 50) {
                                                        if (refDCval < 45) {
                                                            if (TSC < 29) {
                                                                return 1;
                                                            }
                                                            else if (TSC >= 29) {
                                                                if (AFD < 47) {
                                                                    return 0;
                                                                }
                                                                else if (AFD >= 47) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (refDCval >= 45) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (AFD >= 50) {
                                                        if (diffAFD < 38) {
                                                            return 0;
                                                        }
                                                        else if (diffAFD >= 38) {
                                                            if (ssDCval < 99) {
                                                                return 0;
                                                            }
                                                            else if (ssDCval >= 99) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 69) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 410) {
                                if (gchDC < 84) {
                                    if (diffMVdiffVal < 55) {
                                        if (refDCval < 100) {
                                            return 1;
                                        }
                                        else if (refDCval >= 100) {
                                            if (Cs < 3532) {
                                                return 0;
                                            }
                                            else if (Cs >= 3532) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 55) {
                                        if (Rs < 3753) {
                                            return 1;
                                        }
                                        else if (Rs >= 3753) {
                                            if (diffRsCsdiff < 847) {
                                                if (Rs < 4489) {
                                                    return 0;
                                                }
                                                else if (Rs >= 4489) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffRsCsdiff >= 847) {
                                                if (Rs < 9666) {
                                                    return 1;
                                                }
                                                else if (Rs >= 9666) {
                                                    if (ssDCval < 146) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 146) {
                                                        if (Rs < 10400) {
                                                            return 0;
                                                        }
                                                        else if (Rs >= 10400) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (gchDC >= 84) {
                                    if (SC < 3522) {
                                        return 1;
                                    }
                                    else if (SC >= 3522) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (posBalance >= 98) {
                    if (refDCval < 61) {
                        if (negBalance < 2) {
                            if (Rs < 1) {
                                return 1;
                            }
                            else if (Rs >= 1) {
                                if (TSC < 46) {
                                    return 0;
                                }
                                else if (TSC >= 46) {
                                    return 1;
                                }
                            }
                        }
                        else if (negBalance >= 2) {
                            if (AFD < 61) {
                                if (Cs < 372) {
                                    return 1;
                                }
                                else if (Cs >= 372) {
                                    if (RsDiff < 6) {
                                        if (diffMVdiffVal < 187) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 187) {
                                            if (posBalance < 114) {
                                                return 1;
                                            }
                                            else if (posBalance >= 114) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 6) {
                                        if (gchDC < 34) {
                                            return 0;
                                        }
                                        else if (gchDC >= 34) {
                                            if (MVDiff < 151) {
                                                if (AFD < 37) {
                                                    return 1;
                                                }
                                                else if (AFD >= 37) {
                                                    return 0;
                                                }
                                            }
                                            else if (MVDiff >= 151) {
                                                if (refDCval < 51) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 51) {
                                                    if (TSC < 33) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 33) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 61) {
                                if (TSC < 55) {
                                    return 0;
                                }
                                else if (TSC >= 55) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (refDCval >= 61) {
                        if (diffAFD < 37) {
                            if (diffMVdiffVal < 95) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 95) {
                                if (Rs < 4606) {
                                    if (negBalance < 7) {
                                        if (gchDC < 106) {
                                            return 0;
                                        }
                                        else if (gchDC >= 106) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 7) {
                                        if (ssDCval < 107) {
                                            if (Rs < 422) {
                                                return 1;
                                            }
                                            else if (Rs >= 422) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 107) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Rs >= 4606) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffAFD >= 37) {
                            return 1;
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect8(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffTSC < 7) {
        if (MVDiff < 100) {
            if (ssDCval < 20) {
                if (diffRsCsdiff < 7) {
                    if (CsDiff < 3) {
                        return 0;
                    }
                    else if (CsDiff >= 3) {
                        if (MVDiff < 9) {
                            return 1;
                        }
                        else if (MVDiff >= 9) {
                            return 0;
                        }
                    }
                }
                else if (diffRsCsdiff >= 7) {
                    if (posBalance < 25) {
                        return 1;
                    }
                    else if (posBalance >= 25) {
                        return 0;
                    }
                }
            }
            else if (ssDCval >= 20) {
                if (diffMVdiffVal < 62) {
                    if (RsDiff < 111) {
                        if (refDCval < 228) {
                            if (diffRsCsdiff < 1076) {
                                if (ssDCval < 228) {
                                    if (diffMVdiffVal < 43) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 43) {
                                        if (diffRsCsdiff < 247) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 247) {
                                            if (SC < 2984) {
                                                return 1;
                                            }
                                            else if (SC >= 2984) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 228) {
                                    return 1;
                                }
                            }
                            else if (diffRsCsdiff >= 1076) {
                                if (refDCval < 80) {
                                    if (SC < 6844) {
                                        return 1;
                                    }
                                    else if (SC >= 6844) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 80) {
                                    return 0;
                                }
                            }
                        }
                        else if (refDCval >= 228) {
                            if (ssDCval < 224) {
                                return 1;
                            }
                            else if (ssDCval >= 224) {
                                return 0;
                            }
                        }
                    }
                    else if (RsDiff >= 111) {
                        return 1;
                    }
                }
                else if (diffMVdiffVal >= 62) {
                    if (diffAFD < 9) {
                        return 0;
                    }
                    else if (diffAFD >= 9) {
                        if (ssDCval < 77) {
                            return 1;
                        }
                        else if (ssDCval >= 77) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 100) {
            if (diffAFD < 8) {
                if (gchDC < 23) {
                    if (ssDCval < 38) {
                        if (diffMVdiffVal < 223) {
                            if (diffRsCsdiff < 12) {
                                if (TSC < 8) {
                                    if (MVDiff < 1517) {
                                        if (gchDC < 5) {
                                            return 0;
                                        }
                                        else if (gchDC >= 5) {
                                            if (ssDCval < 24) {
                                                if (SC < 56) {
                                                    return 0;
                                                }
                                                else if (SC >= 56) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 24) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 1517) {
                                        if (Rs < 3) {
                                            return 0;
                                        }
                                        else if (Rs >= 3) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (TSC >= 8) {
                                    if (negBalance < 72) {
                                        return 1;
                                    }
                                    else if (negBalance >= 72) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 12) {
                                if (refDCval < 32) {
                                    if (Rs < 483) {
                                        return 1;
                                    }
                                    else if (Rs >= 483) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 32) {
                                    if (negBalance < 79) {
                                        return 0;
                                    }
                                    else if (negBalance >= 79) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 223) {
                            if (CsDiff < 5) {
                                if (posBalance < 43) {
                                    return 0;
                                }
                                else if (posBalance >= 43) {
                                    if (MVDiff < 1022) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 1022) {
                                        return 0;
                                    }
                                }
                            }
                            else if (CsDiff >= 5) {
                                return 1;
                            }
                        }
                    }
                    else if (ssDCval >= 38) {
                        if (diffRsCsdiff < 254) {
                            if (AFD < 34) {
                                if (TSC < 16) {
                                    if (Rs < 257) {
                                        if (Rs < 256) {
                                            if (diffTSC < 4) {
                                                if (posBalance < 1) {
                                                    if (MVDiff < 673) {
                                                        return 0;
                                                    }
                                                    else if (MVDiff >= 673) {
                                                        if (MVDiff < 757) {
                                                            if (TSC < 3) {
                                                                return 0;
                                                            }
                                                            else if (TSC >= 3) {
                                                                return 1;
                                                            }
                                                        }
                                                        else if (MVDiff >= 757) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 1) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffTSC >= 4) {
                                                if (gchDC < 3) {
                                                    if (ssDCval < 46) {
                                                        return 1;
                                                    }
                                                    else if (ssDCval >= 46) {
                                                        return 0;
                                                    }
                                                }
                                                else if (gchDC >= 3) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (Rs >= 256) {
                                            if (RsDiff < 2) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 2) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 257) {
                                        return 0;
                                    }
                                }
                                else if (TSC >= 16) {
                                    if (RsCsDiff < 3219) {
                                        if (refDCval < 47) {
                                            if (ssDCval < 59) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 59) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 47) {
                                            return 0;
                                        }
                                    }
                                    else if (RsCsDiff >= 3219) {
                                        return 1;
                                    }
                                }
                            }
                            else if (AFD >= 34) {
                                if (diffMVdiffVal < 53) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 53) {
                                    if (TSC < 15) {
                                        return 0;
                                    }
                                    else if (TSC >= 15) {
                                        if (MVDiff < 227) {
                                            if (Rs < 1887) {
                                                return 1;
                                            }
                                            else if (Rs >= 1887) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 227) {
                                            if (AFD < 35) {
                                                return 1;
                                            }
                                            else if (AFD >= 35) {
                                                if (RsDiff < 16) {
                                                    return 0;
                                                }
                                                else if (RsDiff >= 16) {
                                                    if (RsCsDiff < 445) {
                                                        return 1;
                                                    }
                                                    else if (RsCsDiff >= 445) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 254) {
                            if (refDCval < 55) {
                                return 1;
                            }
                            else if (refDCval >= 55) {
                                return 0;
                            }
                        }
                    }
                }
                else if (gchDC >= 23) {
                    if (TSC < 31) {
                        if (posBalance < 128) {
                            if (gchDC < 91) {
                                if (CsDiff < 24) {
                                    if (diffAFD < 7) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 7) {
                                        if (ssDCval < 35) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 35) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (CsDiff >= 24) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 91) {
                                if (TSC < 15) {
                                    return 0;
                                }
                                else if (TSC >= 15) {
                                    if (refDCval < 136) {
                                        if (Rs < 1028) {
                                            return 0;
                                        }
                                        else if (Rs >= 1028) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 136) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 128) {
                            if (SC < 84) {
                                return 1;
                            }
                            else if (SC >= 84) {
                                return 0;
                            }
                        }
                    }
                    else if (TSC >= 31) {
                        if (refDCval < 137) {
                            return 1;
                        }
                        else if (refDCval >= 137) {
                            if (Rs < 7069) {
                                return 0;
                            }
                            else if (Rs >= 7069) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 8) {
                if (diffMVdiffVal < 94) {
                    if (ssDCval < 40) {
                        if (MVDiff < 173) {
                            if (refDCval < 37) {
                                return 1;
                            }
                            else if (refDCval >= 37) {
                                if (Rs < 59) {
                                    if (Rs < 30) {
                                        return 0;
                                    }
                                    else if (Rs >= 30) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 59) {
                                    return 0;
                                }
                            }
                        }
                        else if (MVDiff >= 173) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 40) {
                        if (posBalance < 61) {
                            if (diffAFD < 11) {
                                return 0;
                            }
                            else if (diffAFD >= 11) {
                                if (posBalance < 18) {
                                    return 1;
                                }
                                else if (posBalance >= 18) {
                                    if (SC < 36) {
                                        if (Cs < 8) {
                                            return 0;
                                        }
                                        else if (Cs >= 8) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 36) {
                                        if (AFD < 22) {
                                            if (Rs < 221) {
                                                return 0;
                                            }
                                            else if (Rs >= 221) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 22) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 61) {
                            return 0;
                        }
                    }
                }
                else if (diffMVdiffVal >= 94) {
                    if (ssDCval < 53) {
                        if (gchDC < 8) {
                            if (refDCval < 35) {
                                if (ssDCval < 30) {
                                    return 1;
                                }
                                else if (ssDCval >= 30) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 35) {
                                return 1;
                            }
                        }
                        else if (gchDC >= 8) {
                            if (diffAFD < 14) {
                                if (ssDCval < 40) {
                                    if (diffMVdiffVal < 150) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 150) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 40) {
                                    if (gchDC < 21) {
                                        return 0;
                                    }
                                    else if (gchDC >= 21) {
                                        if (TSC < 10) {
                                            return 1;
                                        }
                                        else if (TSC >= 10) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 14) {
                                if (AFD < 27) {
                                    return 1;
                                }
                                else if (AFD >= 27) {
                                    if (ssDCval < 45) {
                                        if (Rs < 2) {
                                            return 1;
                                        }
                                        else if (Rs >= 2) {
                                            if (refDCval < 140) {
                                                return 0;
                                            }
                                            else if (refDCval >= 140) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 45) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 53) {
                        if (diffRsCsdiff < 41) {
                            if (diffMVdiffVal < 467) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 467) {
                                if (AFD < 26) {
                                    if (Cs < 57) {
                                        return 0;
                                    }
                                    else if (Cs >= 57) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 26) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 41) {
                            if (Rs < 488) {
                                return 1;
                            }
                            else if (Rs >= 488) {
                                if (diffAFD < 23) {
                                    if (diffMVdiffVal < 517) {
                                        if (ssDCval < 68) {
                                            if (CsDiff < 7) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 7) {
                                                if (Rs < 724) {
                                                    return 0;
                                                }
                                                else if (Rs >= 724) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 68) {
                                            return 0;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 517) {
                                        if (SC < 2260) {
                                            return 1;
                                        }
                                        else if (SC >= 2260) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffAFD >= 23) {
                                    if (RsDiff < 19) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 19) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffTSC >= 7) {
        if (diffMVdiffVal < 58) {
            if (AFD < 34) {
                if (refDCval < 34) {
                    if (RsCsDiff < 108) {
                        if (gchDC < 102) {
                            return 0;
                        }
                        else if (gchDC >= 102) {
                            return 1;
                        }
                    }
                    else if (RsCsDiff >= 108) {
                        if (CsDiff < 6) {
                            return 0;
                        }
                        else if (CsDiff >= 6) {
                            return 1;
                        }
                    }
                }
                else if (refDCval >= 34) {
                    if (Cs < 929) {
                        if (RsCsDiff < 94) {
                            if (negBalance < 40) {
                                return 0;
                            }
                            else if (negBalance >= 40) {
                                if (ssDCval < 69) {
                                    if (posBalance < 72) {
                                        if (diffMVdiffVal < 41) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 41) {
                                            if (Cs < 210) {
                                                return 1;
                                            }
                                            else if (Cs >= 210) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 72) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 69) {
                                    return 0;
                                }
                            }
                        }
                        else if (RsCsDiff >= 94) {
                            if (diffTSC < 10) {
                                return 0;
                            }
                            else if (diffTSC >= 10) {
                                if (AFD < 31) {
                                    return 1;
                                }
                                else if (AFD >= 31) {
                                    if (Cs < 601) {
                                        return 1;
                                    }
                                    else if (Cs >= 601) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (Cs >= 929) {
                        if (diffTSC < 16) {
                            return 0;
                        }
                        else if (diffTSC >= 16) {
                            if (TSC < 17) {
                                return 1;
                            }
                            else if (TSC >= 17) {
                                if (ssDCval < 65) {
                                    if (AFD < 29) {
                                        return 1;
                                    }
                                    else if (AFD >= 29) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 65) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (AFD >= 34) {
                if (diffAFD < 20) {
                    if (diffTSC < 24) {
                        if (refDCval < 40) {
                            if (Cs < 708) {
                                return 1;
                            }
                            else if (Cs >= 708) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 40) {
                            if (ssDCval < 76) {
                                if (diffMVdiffVal < -166) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= -166) {
                                    return 0;
                                }
                            }
                            else if (ssDCval >= 76) {
                                return 0;
                            }
                        }
                    }
                    else if (diffTSC >= 24) {
                        if (refDCval < 99) {
                            return 1;
                        }
                        else if (refDCval >= 99) {
                            return 0;
                        }
                    }
                }
                else if (diffAFD >= 20) {
                    if (ssDCval < 103) {
                        if (MVDiff < 226) {
                            if (RsCsDiff < 63) {
                                return 0;
                            }
                            else if (RsCsDiff >= 63) {
                                if (TSC < 15) {
                                    return 0;
                                }
                                else if (TSC >= 15) {
                                    if (gchDC < 44) {
                                        return 1;
                                    }
                                    else if (gchDC >= 44) {
                                        if (gchDC < 58) {
                                            return 0;
                                        }
                                        else if (gchDC >= 58) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 226) {
                            if (AFD < 54) {
                                return 0;
                            }
                            else if (AFD >= 54) {
                                if (Rs < 1894) {
                                    if (Cs < 1642) {
                                        return 0;
                                    }
                                    else if (Cs >= 1642) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 1894) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 103) {
                        if (diffAFD < 35) {
                            if (diffRsCsdiff < -4) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= -4) {
                                if (Rs < 8909) {
                                    return 0;
                                }
                                else if (Rs >= 8909) {
                                    if (Rs < 9790) {
                                        return 1;
                                    }
                                    else if (Rs >= 9790) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 35) {
                            if (diffMVdiffVal < -36) {
                                return 1;
                            }
                            else if (diffMVdiffVal >= -36) {
                                if (RsDiff < 21) {
                                    if (Cs < 281) {
                                        return 1;
                                    }
                                    else if (Cs >= 281) {
                                        return 0;
                                    }
                                }
                                else if (RsDiff >= 21) {
                                    if (MVDiff < 46) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 46) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 58) {
            if (negBalance < 30) {
                if (diffAFD < 21) {
                    if (refDCval < 34) {
                        if (MVDiff < 438) {
                            if (Rs < 681) {
                                if (RsCsDiff < 1) {
                                    if (SC < 52) {
                                        return 1;
                                    }
                                    else if (SC >= 52) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 1) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 681) {
                                return 0;
                            }
                        }
                        else if (MVDiff >= 438) {
                            return 0;
                        }
                    }
                    else if (refDCval >= 34) {
                        if (SC < 170) {
                            if (TSC < 12) {
                                return 1;
                            }
                            else if (TSC >= 12) {
                                if (CsDiff < 6) {
                                    if (MVDiff < 311) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 311) {
                                        if (posBalance < 112) {
                                            return 0;
                                        }
                                        else if (posBalance >= 112) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (CsDiff >= 6) {
                                    return 1;
                                }
                            }
                        }
                        else if (SC >= 170) {
                            if (diffRsCsdiff < 369) {
                                if (SC < 423) {
                                    if (Cs < 222) {
                                        return 0;
                                    }
                                    else if (Cs >= 222) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 423) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 369) {
                                if (diffAFD < 18) {
                                    return 0;
                                }
                                else if (diffAFD >= 18) {
                                    if (refDCval < 48) {
                                        return 0;
                                    }
                                    else if (refDCval >= 48) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 21) {
                    if (diffRsCsdiff < 123) {
                        if (TSC < 22) {
                            if (negBalance < 5) {
                                if (Rs < 1) {
                                    return 1;
                                }
                                else if (Rs >= 1) {
                                    return 0;
                                }
                            }
                            else if (negBalance >= 5) {
                                if (Cs < 857) {
                                    if (CsDiff < 1) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 1) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 857) {
                                    return 0;
                                }
                            }
                        }
                        else if (TSC >= 22) {
                            if (ssDCval < 152) {
                                if (Rs < 787) {
                                    if (diffMVdiffVal < 174) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 174) {
                                        if (MVDiff < 280) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 280) {
                                            if (refDCval < 35) {
                                                return 1;
                                            }
                                            else if (refDCval >= 35) {
                                                if (TSC < 24) {
                                                    return 1;
                                                }
                                                else if (TSC >= 24) {
                                                    if (TSC < 37) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 37) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 787) {
                                    return 0;
                                }
                            }
                            else if (ssDCval >= 152) {
                                return 1;
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 123) {
                        if (negBalance < 3) {
                            if (gchDC < 104) {
                                return 0;
                            }
                            else if (gchDC >= 104) {
                                if (Cs < 1565) {
                                    return 1;
                                }
                                else if (Cs >= 1565) {
                                    return 0;
                                }
                            }
                        }
                        else if (negBalance >= 3) {
                            if (posBalance < 105) {
                                if (Cs < 1024) {
                                    if (CsDiff < 7) {
                                        if (MVDiff < 174) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 174) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 7) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 1024) {
                                    if (diffRsCsdiff < 636) {
                                        if (TSC < 31) {
                                            return 0;
                                        }
                                        else if (TSC >= 31) {
                                            if (Cs < 1774) {
                                                return 1;
                                            }
                                            else if (Cs >= 1774) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 636) {
                                        if (MVDiff < 134) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 134) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 105) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 30) {
                if (diffRsCsdiff < 2) {
                    if (ssDCval < 119) {
                        if (AFD < 69) {
                            if (negBalance < 48) {
                                return 0;
                            }
                            else if (negBalance >= 48) {
                                if (posBalance < 54) {
                                    if (TSC < 37) {
                                        if (Cs < 169) {
                                            if (MVDiff < 458) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 458) {
                                                return 0;
                                            }
                                        }
                                        else if (Cs >= 169) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 37) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 54) {
                                    return 1;
                                }
                            }
                        }
                        else if (AFD >= 69) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 119) {
                        return 0;
                    }
                }
                else if (diffRsCsdiff >= 2) {
                    if (diffAFD < 18) {
                        if (ssDCval < 57) {
                            if (refDCval < 25) {
                                return 0;
                            }
                            else if (refDCval >= 25) {
                                if (Rs < 262) {
                                    return 1;
                                }
                                else if (Rs >= 262) {
                                    if (negBalance < 35) {
                                        return 0;
                                    }
                                    else if (negBalance >= 35) {
                                        if (Rs < 1227) {
                                            return 1;
                                        }
                                        else if (Rs >= 1227) {
                                            if (Rs < 1446) {
                                                return 0;
                                            }
                                            else if (Rs >= 1446) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 57) {
                            if (gchDC < 4) {
                                return 1;
                            }
                            else if (gchDC >= 4) {
                                if (RsDiff < 34) {
                                    if (diffAFD < 13) {
                                        if (posBalance < 84) {
                                            return 0;
                                        }
                                        else if (posBalance >= 84) {
                                            return 1;
                                        }
                                    }
                                    else if (diffAFD >= 13) {
                                        if (AFD < 32) {
                                            if (RsDiff < 8) {
                                                if (Cs < 137) {
                                                    return 1;
                                                }
                                                else if (Cs >= 137) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsDiff >= 8) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 32) {
                                            if (diffMVdiffVal < 142) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 142) {
                                                if (Rs < 307) {
                                                    return 0;
                                                }
                                                else if (Rs >= 307) {
                                                    if (SC < 3197) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 3197) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 34) {
                                    if (Rs < 6129) {
                                        return 1;
                                    }
                                    else if (Rs >= 6129) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 18) {
                        if (gchDC < 39) {
                            if (diffTSC < 11) {
                                if (SC < 2902) {
                                    if (diffMVdiffVal < 161) {
                                        if (TSC < 9) {
                                            return 0;
                                        }
                                        else if (TSC >= 9) {
                                            if (refDCval < 97) {
                                                return 1;
                                            }
                                            else if (refDCval >= 97) {
                                                if (AFD < 34) {
                                                    return 1;
                                                }
                                                else if (AFD >= 34) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 161) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 2902) {
                                    if (refDCval < 69) {
                                        return 1;
                                    }
                                    else if (refDCval >= 69) {
                                        if (negBalance < 73) {
                                            if (posBalance < 85) {
                                                return 0;
                                            }
                                            else if (posBalance >= 85) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 73) {
                                            if (SC < 3933) {
                                                return 1;
                                            }
                                            else if (SC >= 3933) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 11) {
                                if (diffAFD < 41) {
                                    if (TSC < 33) {
                                        if (diffMVdiffVal < 80) {
                                            if (posBalance < 78) {
                                                if (RsDiff < 18) {
                                                    if (CsDiff < 21) {
                                                        if (diffAFD < 29) {
                                                            return 1;
                                                        }
                                                        else if (diffAFD >= 29) {
                                                            if (Rs < 1274) {
                                                                return 0;
                                                            }
                                                            else if (Rs >= 1274) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (CsDiff >= 21) {
                                                        return 0;
                                                    }
                                                }
                                                else if (RsDiff >= 18) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 78) {
                                                return 0;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 80) {
                                            if (TSC < 19) {
                                                return 1;
                                            }
                                            else if (TSC >= 19) {
                                                if (gchDC < 20) {
                                                    return 1;
                                                }
                                                else if (gchDC >= 20) {
                                                    if (diffTSC < 19) {
                                                        if (posBalance < 60) {
                                                            if (negBalance < 97) {
                                                                return 1;
                                                            }
                                                            else if (negBalance >= 97) {
                                                                if (Rs < 2228) {
                                                                    return 1;
                                                                }
                                                                else if (Rs >= 2228) {
                                                                    return 0;
                                                                }
                                                            }
                                                        }
                                                        else if (posBalance >= 60) {
                                                            if (posBalance < 81) {
                                                                return 0;
                                                            }
                                                            else if (posBalance >= 81) {
                                                                if (AFD < 58) {
                                                                    return 1;
                                                                }
                                                                else if (AFD >= 58) {
                                                                    return 0;
                                                                }
                                                            }
                                                        }
                                                    }
                                                    else if (diffTSC >= 19) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 33) {
                                        if (RsDiff < 26) {
                                            if (AFD < 54) {
                                                return 1;
                                            }
                                            else if (AFD >= 54) {
                                                if (ssDCval < 91) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 91) {
                                                    if (RsCsDiff < 622) {
                                                        return 0;
                                                    }
                                                    else if (RsCsDiff >= 622) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 26) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= 41) {
                                    if (MVDiff < 1241) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 1241) {
                                        if (TSC < 26) {
                                            return 0;
                                        }
                                        else if (TSC >= 26) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 39) {
                            if (refDCval < 162) {
                                if (CsDiff < 4) {
                                    if (CsDiff < 3) {
                                        if (Rs < 74) {
                                            return 0;
                                        }
                                        else if (Rs >= 74) {
                                            return 1;
                                        }
                                    }
                                    else if (CsDiff >= 3) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 4) {
                                    if (diffAFD < 36) {
                                        if (TSC < 28) {
                                            if (Rs < 2633) {
                                                return 1;
                                            }
                                            else if (Rs >= 2633) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 28) {
                                            return 0;
                                        }
                                    }
                                    else if (diffAFD >= 36) {
                                        if (AFD < 86) {
                                            return 1;
                                        }
                                        else if (AFD >= 86) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 162) {
                                if (Cs < 529) {
                                    if (Cs < 226) {
                                        return 0;
                                    }
                                    else if (Cs >= 226) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 529) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect9(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 13) {
        if (MVDiff < 125) {
            if (diffMVdiffVal < 31) {
                if (refDCval < 24) {
                    if (diffAFD < 2) {
                        if (refDCval < 18) {
                            if (RsCsDiff < 7) {
                                return 0;
                            }
                            else if (RsCsDiff >= 7) {
                                if (Cs < 718) {
                                    return 0;
                                }
                                else if (Cs >= 718) {
                                    if (MVDiff < 7) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 7) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 18) {
                            if (Rs < 1) {
                                if (gchDC < 108) {
                                    return 0;
                                }
                                else if (gchDC >= 108) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 1) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 2) {
                        if (Cs < 282) {
                            if (AFD < 9) {
                                return 0;
                            }
                            else if (AFD >= 9) {
                                return 1;
                            }
                        }
                        else if (Cs >= 282) {
                            return 1;
                        }
                    }
                }
                else if (refDCval >= 24) {
                    if (diffRsCsdiff < 1513) {
                        return 0;
                    }
                    else if (diffRsCsdiff >= 1513) {
                        if (Rs < 2808) {
                            return 1;
                        }
                        else if (Rs >= 2808) {
                            if (RsDiff < 73) {
                                return 0;
                            }
                            else if (RsDiff >= 73) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 31) {
                if (ssDCval < 44) {
                    if (diffAFD < 9) {
                        if (RsCsDiff < 418) {
                            if (Cs < 27) {
                                if (diffAFD < 8) {
                                    return 0;
                                }
                                else if (diffAFD >= 8) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 27) {
                                return 0;
                            }
                        }
                        else if (RsCsDiff >= 418) {
                            if (AFD < 13) {
                                return 1;
                            }
                            else if (AFD >= 13) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 9) {
                        if (MVDiff < 114) {
                            return 1;
                        }
                        else if (MVDiff >= 114) {
                            return 0;
                        }
                    }
                }
                else if (ssDCval >= 44) {
                    if (RsDiff < 32) {
                        if (Cs < 9) {
                            if (TSC < 9) {
                                return 0;
                            }
                            else if (TSC >= 9) {
                                return 1;
                            }
                        }
                        else if (Cs >= 9) {
                            if (diffMVdiffVal < 102) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 102) {
                                if (RsDiff < 8) {
                                    return 0;
                                }
                                else if (RsDiff >= 8) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (RsDiff >= 32) {
                        if (Cs < 3730) {
                            if (Cs < 2366) {
                                if (ssDCval < 122) {
                                    return 0;
                                }
                                else if (ssDCval >= 122) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 2366) {
                                if (SC < 6466) {
                                    return 0;
                                }
                                else if (SC >= 6466) {
                                    return 1;
                                }
                            }
                        }
                        else if (Cs >= 3730) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 125) {
            if (diffTSC < 4) {
                if (diffRsCsdiff < -1) {
                    if (diffAFD < -8) {
                        if (gchDC < 155) {
                            return 0;
                        }
                        else if (gchDC >= 155) {
                            return 1;
                        }
                    }
                    else if (diffAFD >= -8) {
                        if (Rs < 115) {
                            if (CsDiff < 6) {
                                if (diffRsCsdiff < -29) {
                                    if (ssDCval < 37) {
                                        if (RsDiff < 2) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 2) {
                                            if (Cs < 12) {
                                                return 0;
                                            }
                                            else if (Cs >= 12) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 37) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= -29) {
                                    if (AFD < 36) {
                                        return 0;
                                    }
                                    else if (AFD >= 36) {
                                        if (SC < 81) {
                                            return 1;
                                        }
                                        else if (SC >= 81) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 6) {
                                if (diffAFD < -1) {
                                    return 0;
                                }
                                else if (diffAFD >= -1) {
                                    return 1;
                                }
                            }
                        }
                        else if (Rs >= 115) {
                            if (TSC < 15) {
                                return 0;
                            }
                            else if (TSC >= 15) {
                                if (diffRsCsdiff < -4) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= -4) {
                                    if (RsDiff < 7) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 7) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= -1) {
                    if (RsDiff < 12) {
                        if (gchDC < 98) {
                            if (ssDCval < 42) {
                                if (diffAFD < 6) {
                                    if (diffRsCsdiff < 12) {
                                        if (diffAFD < -21) {
                                            if (posBalance < 37) {
                                                return 0;
                                            }
                                            else if (posBalance >= 37) {
                                                if (SC < 120) {
                                                    return 1;
                                                }
                                                else if (SC >= 120) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffAFD >= -21) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 12) {
                                        if (ssDCval < 37) {
                                            if (RsCsDiff < 38) {
                                                return 1;
                                            }
                                            else if (RsCsDiff >= 38) {
                                                if (refDCval < 28) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 28) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 37) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffAFD >= 6) {
                                    if (Cs < 134) {
                                        if (AFD < 16) {
                                            return 1;
                                        }
                                        else if (AFD >= 16) {
                                            if (RsDiff < 4) {
                                                if (AFD < 22) {
                                                    return 0;
                                                }
                                                else if (AFD >= 22) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsDiff >= 4) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 134) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 42) {
                                if (TSC < 16) {
                                    if (diffRsCsdiff < 13) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 13) {
                                        if (diffRsCsdiff < 14) {
                                            if (AFD < 18) {
                                                return 1;
                                            }
                                            else if (AFD >= 18) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 14) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (TSC >= 16) {
                                    if (gchDC < 16) {
                                        if (refDCval < 77) {
                                            if (CsDiff < 7) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 7) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 77) {
                                            if (RsCsDiff < 53) {
                                                if (Rs < 401) {
                                                    return 0;
                                                }
                                                else if (Rs >= 401) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 53) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 16) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 98) {
                            if (Rs < 23) {
                                return 1;
                            }
                            else if (Rs >= 23) {
                                if (RsDiff < 7) {
                                    return 0;
                                }
                                else if (RsDiff >= 7) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (RsDiff >= 12) {
                        if (gchDC < 23) {
                            if (ssDCval < 46) {
                                if (CsDiff < 10) {
                                    return 0;
                                }
                                else if (CsDiff >= 10) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 46) {
                                if (diffMVdiffVal < 111) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 111) {
                                    if (Rs < 1577) {
                                        if (RsDiff < 17) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 17) {
                                            if (TSC < 15) {
                                                return 0;
                                            }
                                            else if (TSC >= 15) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 1577) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 23) {
                            if (AFD < 56) {
                                if (ssDCval < 37) {
                                    return 1;
                                }
                                else if (ssDCval >= 37) {
                                    if (MVDiff < 144) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 144) {
                                        return 0;
                                    }
                                }
                            }
                            else if (AFD >= 56) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 4) {
                if (TSCindex < 8) {
                    if (diffRsCsdiff < 1) {
                        if (Rs < 96) {
                            if (refDCval < 97) {
                                return 1;
                            }
                            else if (refDCval >= 97) {
                                return 0;
                            }
                        }
                        else if (Rs >= 96) {
                            return 0;
                        }
                    }
                    else if (diffRsCsdiff >= 1) {
                        return 1;
                    }
                }
                else if (TSCindex >= 8) {
                    if (gchDC < 8) {
                        if (refDCval < 48) {
                            if (diffMVdiffVal < 96) {
                                if (ssDCval < 35) {
                                    return 1;
                                }
                                else if (ssDCval >= 35) {
                                    if (Rs < 39) {
                                        return 1;
                                    }
                                    else if (Rs >= 39) {
                                        if (MVDiff < 381) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 381) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 96) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 48) {
                            if (AFD < 28) {
                                if (TSC < 8) {
                                    return 1;
                                }
                                else if (TSC >= 8) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 28) {
                                if (AFD < 29) {
                                    return 1;
                                }
                                else if (AFD >= 29) {
                                    if (RsDiff < 19) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 19) {
                                        if (TSC < 16) {
                                            return 0;
                                        }
                                        else if (TSC >= 16) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (gchDC >= 8) {
                        if (diffMVdiffVal < 100) {
                            if (negBalance < 1) {
                                if (MVDiff < 360) {
                                    return 0;
                                }
                                else if (MVDiff >= 360) {
                                    if (refDCval < 30) {
                                        return 1;
                                    }
                                    else if (refDCval >= 30) {
                                        return 0;
                                    }
                                }
                            }
                            else if (negBalance >= 1) {
                                if (RsCsDiff < 73) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 73) {
                                    if (TSC < 30) {
                                        return 0;
                                    }
                                    else if (TSC >= 30) {
                                        if (diffTSC < 11) {
                                            if (Rs < 2496) {
                                                return 1;
                                            }
                                            else if (Rs >= 2496) {
                                                return 0;
                                            }
                                        }
                                        else if (diffTSC >= 11) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 100) {
                            if (RsCsDiff < 23) {
                                if (SC < 24) {
                                    if (Cs < 10) {
                                        return 1;
                                    }
                                    else if (Cs >= 10) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 24) {
                                    return 0;
                                }
                            }
                            else if (RsCsDiff >= 23) {
                                if (ssDCval < 46) {
                                    return 1;
                                }
                                else if (ssDCval >= 46) {
                                    if (TSC < 33) {
                                        if (Rs < 46) {
                                            return 1;
                                        }
                                        else if (Rs >= 46) {
                                            if (diffTSC < 11) {
                                                if (posBalance < 7) {
                                                    if (Rs < 369) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 369) {
                                                        return 1;
                                                    }
                                                }
                                                else if (posBalance >= 7) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffTSC >= 11) {
                                                if (Rs < 709) {
                                                    return 0;
                                                }
                                                else if (Rs >= 709) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 33) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 13) {
        if (diffMVdiffVal < 65) {
            if (TSC < 16) {
                if (refDCval < 62) {
                    if (CsDiff < 2) {
                        if (TSC < 14) {
                            return 1;
                        }
                        else if (TSC >= 14) {
                            return 0;
                        }
                    }
                    else if (CsDiff >= 2) {
                        if (ssDCval < 33) {
                            if (Rs < 261) {
                                return 0;
                            }
                            else if (Rs >= 261) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 33) {
                            if (TSC < 14) {
                                return 0;
                            }
                            else if (TSC >= 14) {
                                if (posBalance < 59) {
                                    return 1;
                                }
                                else if (posBalance >= 59) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (refDCval >= 62) {
                    if (Cs < 435) {
                        if (negBalance < 78) {
                            return 0;
                        }
                        else if (negBalance >= 78) {
                            if (TSC < 15) {
                                return 1;
                            }
                            else if (TSC >= 15) {
                                return 0;
                            }
                        }
                    }
                    else if (Cs >= 435) {
                        return 0;
                    }
                }
            }
            else if (TSC >= 16) {
                if (diffTSC < 21) {
                    if (diffAFD < 23) {
                        if (Rs < 1588) {
                            if (diffRsCsdiff < 123) {
                                if (refDCval < 22) {
                                    return 1;
                                }
                                else if (refDCval >= 22) {
                                    if (diffAFD < 20) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 20) {
                                        if (negBalance < 69) {
                                            if (Rs < 240) {
                                                return 1;
                                            }
                                            else if (Rs >= 240) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 69) {
                                            if (Rs < 206) {
                                                return 0;
                                            }
                                            else if (Rs >= 206) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 123) {
                                if (MVDiff < 246) {
                                    if (TSC < 20) {
                                        if (Cs < 1361) {
                                            return 1;
                                        }
                                        else if (Cs >= 1361) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 20) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 246) {
                                    return 1;
                                }
                            }
                        }
                        else if (Rs >= 1588) {
                            if (Cs < 7358) {
                                return 0;
                            }
                            else if (Cs >= 7358) {
                                return 1;
                            }
                        }
                    }
                    else if (diffAFD >= 23) {
                        if (negBalance < 34) {
                            return 0;
                        }
                        else if (negBalance >= 34) {
                            if (SC < 6908) {
                                if (diffMVdiffVal < 45) {
                                    if (diffMVdiffVal < -30) {
                                        return 1;
                                    }
                                    else if (diffMVdiffVal >= -30) {
                                        if (posBalance < 54) {
                                            return 0;
                                        }
                                        else if (posBalance >= 54) {
                                            if (SC < 1183) {
                                                return 0;
                                            }
                                            else if (SC >= 1183) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 45) {
                                    return 1;
                                }
                            }
                            else if (SC >= 6908) {
                                return 0;
                            }
                        }
                    }
                }
                else if (diffTSC >= 21) {
                    if (RsCsDiff < 275) {
                        if (ssDCval < 60) {
                            return 1;
                        }
                        else if (ssDCval >= 60) {
                            if (diffRsCsdiff < -62) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= -62) {
                                return 0;
                            }
                        }
                    }
                    else if (RsCsDiff >= 275) {
                        if (RsDiff < 25) {
                            if (diffMVdiffVal < 21) {
                                if (TSC < 27) {
                                    return 0;
                                }
                                else if (TSC >= 27) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 21) {
                                return 0;
                            }
                        }
                        else if (RsDiff >= 25) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 65) {
            if (diffAFD < 22) {
                if (ssDCval < 70) {
                    if (ssDCval < 41) {
                        if (TSC < 6) {
                            return 0;
                        }
                        else if (TSC >= 6) {
                            if (diffRsCsdiff < -11) {
                                if (Rs < 3) {
                                    return 1;
                                }
                                else if (Rs >= 3) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= -11) {
                                return 1;
                            }
                        }
                    }
                    else if (ssDCval >= 41) {
                        if (diffAFD < 17) {
                            if (gchDC < 9) {
                                if (AFD < 32) {
                                    return 1;
                                }
                                else if (AFD >= 32) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 9) {
                                if (gchDC < 16) {
                                    return 0;
                                }
                                else if (gchDC >= 16) {
                                    if (ssDCval < 51) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 51) {
                                        if (Rs < 445) {
                                            if (Rs < 29) {
                                                return 1;
                                            }
                                            else if (Rs >= 29) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 445) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 17) {
                            if (negBalance < 22) {
                                if (posBalance < 85) {
                                    return 1;
                                }
                                else if (posBalance >= 85) {
                                    return 0;
                                }
                            }
                            else if (negBalance >= 22) {
                                return 1;
                            }
                        }
                    }
                }
                else if (ssDCval >= 70) {
                    if (diffTSC < 7) {
                        if (AFD < 19) {
                            return 1;
                        }
                        else if (AFD >= 19) {
                            if (diffMVdiffVal < 419) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 419) {
                                if (Rs < 1256) {
                                    return 0;
                                }
                                else if (Rs >= 1256) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 7) {
                        if (refDCval < 143) {
                            if (posBalance < 78) {
                                if (Cs < 1418) {
                                    if (AFD < 44) {
                                        return 1;
                                    }
                                    else if (AFD >= 44) {
                                        if (TSC < 23) {
                                            return 0;
                                        }
                                        else if (TSC >= 23) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Cs >= 1418) {
                                    if (SC < 5906) {
                                        return 0;
                                    }
                                    else if (SC >= 5906) {
                                        return 1;
                                    }
                                }
                            }
                            else if (posBalance >= 78) {
                                if (Cs < 392) {
                                    if (diffMVdiffVal < 97) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 97) {
                                        if (AFD < 22) {
                                            return 1;
                                        }
                                        else if (AFD >= 22) {
                                            if (ssDCval < 74) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 74) {
                                                if (refDCval < 79) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 79) {
                                                    if (Cs < 2) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 392) {
                                    if (diffAFD < 19) {
                                        if (Rs < 271) {
                                            return 1;
                                        }
                                        else if (Rs >= 271) {
                                            return 0;
                                        }
                                    }
                                    else if (diffAFD >= 19) {
                                        if (diffRsCsdiff < 20) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 20) {
                                            if (AFD < 46) {
                                                if (ssDCval < 95) {
                                                    if (gchDC < 31) {
                                                        if (Cs < 541) {
                                                            return 0;
                                                        }
                                                        else if (Cs >= 541) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (gchDC >= 31) {
                                                        return 0;
                                                    }
                                                }
                                                else if (ssDCval >= 95) {
                                                    return 1;
                                                }
                                            }
                                            else if (AFD >= 46) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 143) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffAFD >= 22) {
                if (negBalance < 26) {
                    if (diffRsCsdiff < 157) {
                        if (diffMVdiffVal < 165) {
                            if (AFD < 25) {
                                return 1;
                            }
                            else if (AFD >= 25) {
                                if (RsCsDiff < 27) {
                                    if (CsDiff < 2) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 2) {
                                        return 1;
                                    }
                                }
                                else if (RsCsDiff >= 27) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 165) {
                            if (posBalance < 97) {
                                return 0;
                            }
                            else if (posBalance >= 97) {
                                if (TSC < 27) {
                                    if (RsCsDiff < 23) {
                                        if (diffAFD < 31) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 31) {
                                            return 1;
                                        }
                                    }
                                    else if (RsCsDiff >= 23) {
                                        if (diffAFD < 23) {
                                            if (Rs < 660) {
                                                return 1;
                                            }
                                            else if (Rs >= 660) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= 23) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (TSC >= 27) {
                                    if (ssDCval < 152) {
                                        if (MVDiff < 244) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 244) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 152) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 157) {
                        if (posBalance < 125) {
                            if (ssDCval < 98) {
                                if (MVDiff < 131) {
                                    return 1;
                                }
                                else if (MVDiff >= 131) {
                                    if (refDCval < 35) {
                                        if (refDCval < 23) {
                                            return 0;
                                        }
                                        else if (refDCval >= 23) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 35) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 98) {
                                if (negBalance < 25) {
                                    return 1;
                                }
                                else if (negBalance >= 25) {
                                    if (refDCval < 69) {
                                        return 1;
                                    }
                                    else if (refDCval >= 69) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 125) {
                            if (AFD < 30) {
                                return 1;
                            }
                            else if (AFD >= 30) {
                                if (CsDiff < 9) {
                                    return 1;
                                }
                                else if (CsDiff >= 9) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 26) {
                    if (RsCsDiff < 306) {
                        if (ssDCval < 92) {
                            if (ssDCval < 55) {
                                if (ssDCval < 54) {
                                    if (AFD < 39) {
                                        if (refDCval < 105) {
                                            return 1;
                                        }
                                        else if (refDCval >= 105) {
                                            return 0;
                                        }
                                    }
                                    else if (AFD >= 39) {
                                        if (RsCsDiff < 38) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 38) {
                                            if (TSC < 32) {
                                                if (refDCval < 60) {
                                                    return 0;
                                                }
                                                else if (refDCval >= 60) {
                                                    return 1;
                                                }
                                            }
                                            else if (TSC >= 32) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (ssDCval >= 54) {
                                    if (MVDiff < 204) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 204) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 55) {
                                if (CsDiff < 1) {
                                    if (SC < 196) {
                                        return 1;
                                    }
                                    else if (SC >= 196) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 1) {
                                    if (AFD < 46) {
                                        return 1;
                                    }
                                    else if (AFD >= 46) {
                                        if (SC < 2044) {
                                            return 1;
                                        }
                                        else if (SC >= 2044) {
                                            if (negBalance < 70) {
                                                return 0;
                                            }
                                            else if (negBalance >= 70) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 92) {
                            if (diffTSC < 8) {
                                if (Cs < 456) {
                                    if (Rs < 36) {
                                        return 0;
                                    }
                                    else if (Rs >= 36) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 456) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= 8) {
                                if (negBalance < 110) {
                                    if (diffAFD < 26) {
                                        if (ssDCval < 112) {
                                            if (SC < 1077) {
                                                return 1;
                                            }
                                            else if (SC >= 1077) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 112) {
                                            return 1;
                                        }
                                    }
                                    else if (diffAFD >= 26) {
                                        if (MVDiff < 1116) {
                                            if (refDCval < 90) {
                                                if (diffRsCsdiff < 46) {
                                                    if (Cs < 459) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 459) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffRsCsdiff >= 46) {
                                                    if (diffMVdiffVal < 88) {
                                                        return 0;
                                                    }
                                                    else if (diffMVdiffVal >= 88) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (refDCval >= 90) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 1116) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (negBalance >= 110) {
                                    if (Rs < 840) {
                                        return 0;
                                    }
                                    else if (Rs >= 840) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 306) {
                        if (diffTSC < 16) {
                            if (TSC < 30) {
                                if (Rs < 1663) {
                                    return 1;
                                }
                                else if (Rs >= 1663) {
                                    if (TSC < 20) {
                                        if (negBalance < 53) {
                                            return 1;
                                        }
                                        else if (negBalance >= 53) {
                                            if (ssDCval < 62) {
                                                if (MVDiff < 218) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 218) {
                                                    return 0;
                                                }
                                            }
                                            else if (ssDCval >= 62) {
                                                if (TSC < 8) {
                                                    return 1;
                                                }
                                                else if (TSC >= 8) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 20) {
                                        return 1;
                                    }
                                }
                            }
                            else if (TSC >= 30) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 16) {
                            if (Rs < 9642) {
                                return 1;
                            }
                            else if (Rs >= 9642) {
                                if (Rs < 10400) {
                                    return 0;
                                }
                                else if (Rs >= 10400) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectA(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffMVdiffVal < 65) {
        if (diffTSC < 13) {
            if (TSC < 25) {
                if (MVDiff < 54) {
                    if (refDCval < 24) {
                        if (diffAFD < 3) {
                            if (RsDiff < 2) {
                                if (ssDCval < 130) {
                                    return 0;
                                }
                                else if (ssDCval >= 130) {
                                    return 1;
                                }
                            }
                            else if (RsDiff >= 2) {
                                if (posBalance < 2) {
                                    if (diffRsCsdiff < 3) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 3) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 2) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffAFD >= 3) {
                            if (Rs < 54) {
                                return 0;
                            }
                            else if (Rs >= 54) {
                                return 1;
                            }
                        }
                    }
                    else if (refDCval >= 24) {
                        if (diffRsCsdiff < 577) {
                            if (refDCval < 229) {
                                return 0;
                            }
                            else if (refDCval >= 229) {
                                if (AFD < 4) {
                                    return 0;
                                }
                                else if (AFD >= 4) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 577) {
                            if (Cs < 344) {
                                return 1;
                            }
                            else if (Cs >= 344) {
                                if (AFD < 12) {
                                    if (TSC < 11) {
                                        return 0;
                                    }
                                    else if (TSC >= 11) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 12) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (MVDiff >= 54) {
                    if (MVDiff < 173) {
                        if (diffAFD < 11) {
                            if (RsCsDiff < 649) {
                                if (diffAFD < 9) {
                                    return 0;
                                }
                                else if (diffAFD >= 9) {
                                    if (Rs < 15) {
                                        return 1;
                                    }
                                    else if (Rs >= 15) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 649) {
                                if (refDCval < 30) {
                                    return 1;
                                }
                                else if (refDCval >= 30) {
                                    if (posBalance < 103) {
                                        if (CsDiff < 6) {
                                            if (RsDiff < 61) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 61) {
                                                return 1;
                                            }
                                        }
                                        else if (CsDiff >= 6) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 103) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 11) {
                            if (ssDCval < 49) {
                                if (diffRsCsdiff < 150) {
                                    if (MVDiff < 144) {
                                        if (gchDC < 2) {
                                            if (Rs < 354) {
                                                return 0;
                                            }
                                            else if (Rs >= 354) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 2) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 144) {
                                        return 1;
                                    }
                                }
                                else if (diffRsCsdiff >= 150) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 49) {
                                if (diffTSC < 9) {
                                    return 0;
                                }
                                else if (diffTSC >= 9) {
                                    if (gchDC < 15) {
                                        if (AFD < 37) {
                                            if (gchDC < 13) {
                                                return 0;
                                            }
                                            else if (gchDC >= 13) {
                                                if (SC < 1739) {
                                                    return 0;
                                                }
                                                else if (SC >= 1739) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (AFD >= 37) {
                                            return 1;
                                        }
                                    }
                                    else if (gchDC >= 15) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (MVDiff >= 173) {
                        if (diffTSC < 6) {
                            if (diffRsCsdiff < -30) {
                                if (gchDC < 19) {
                                    return 0;
                                }
                                else if (gchDC >= 19) {
                                    if (refDCval < 58) {
                                        if (Rs < 105) {
                                            return 1;
                                        }
                                        else if (Rs >= 105) {
                                            return 0;
                                        }
                                    }
                                    else if (refDCval >= 58) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= -30) {
                                if (diffRsCsdiff < -25) {
                                    if (diffTSC < -3) {
                                        return 0;
                                    }
                                    else if (diffTSC >= -3) {
                                        if (SC < 118) {
                                            return 1;
                                        }
                                        else if (SC >= 118) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= -25) {
                                    if (gchDC < 57) {
                                        if (ssDCval < 78) {
                                            if (TSC < 13) {
                                                if (MVDiff < 1610) {
                                                    if (diffRsCsdiff < 17) {
                                                        return 0;
                                                    }
                                                    else if (diffRsCsdiff >= 17) {
                                                        if (TSC < 2) {
                                                            return 1;
                                                        }
                                                        else if (TSC >= 2) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 1610) {
                                                    if (diffTSC < -3) {
                                                        return 0;
                                                    }
                                                    else if (diffTSC >= -3) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (TSC >= 13) {
                                                if (Cs < 57) {
                                                    if (CsDiff < 1) {
                                                        return 0;
                                                    }
                                                    else if (CsDiff >= 1) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 57) {
                                                    if (posBalance < 73) {
                                                        return 0;
                                                    }
                                                    else if (posBalance >= 73) {
                                                        if (Rs < 977) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 977) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 78) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 57) {
                                        if (AFD < 47) {
                                            return 0;
                                        }
                                        else if (AFD >= 47) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 6) {
                            if (CsDiff < 5) {
                                if (diffMVdiffVal < 45) {
                                    if (TSC < 9) {
                                        return 1;
                                    }
                                    else if (TSC >= 9) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 45) {
                                    if (TSC < 19) {
                                        return 1;
                                    }
                                    else if (TSC >= 19) {
                                        return 0;
                                    }
                                }
                            }
                            else if (CsDiff >= 5) {
                                if (Cs < 984) {
                                    if (RsDiff < 6) {
                                        if (ssDCval < 47) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 47) {
                                            return 0;
                                        }
                                    }
                                    else if (RsDiff >= 6) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 984) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (TSC >= 25) {
                if (AFD < 58) {
                    if (RsDiff < 80) {
                        if (ssDCval < 98) {
                            if (gchDC < 11) {
                                return 1;
                            }
                            else if (gchDC >= 11) {
                                if (diffAFD < 26) {
                                    if (MVDiff < 317) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 317) {
                                        if (Cs < 705) {
                                            return 1;
                                        }
                                        else if (Cs >= 705) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffAFD >= 26) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 98) {
                            return 0;
                        }
                    }
                    else if (RsDiff >= 80) {
                        return 1;
                    }
                }
                else if (AFD >= 58) {
                    if (diffMVdiffVal < 2) {
                        return 1;
                    }
                    else if (diffMVdiffVal >= 2) {
                        return 0;
                    }
                }
            }
        }
        else if (diffTSC >= 13) {
            if (RsDiff < 25) {
                if (refDCval < 25) {
                    if (ssDCval < 132) {
                        return 1;
                    }
                    else if (ssDCval >= 132) {
                        return 0;
                    }
                }
                else if (refDCval >= 25) {
                    if (Rs < 2771) {
                        if (RsDiff < 12) {
                            if (ssDCval < 91) {
                                if (SC < 916) {
                                    if (CsDiff < 3) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 3) {
                                        if (posBalance < 115) {
                                            if (AFD < 42) {
                                                return 1;
                                            }
                                            else if (AFD >= 42) {
                                                return 0;
                                            }
                                        }
                                        else if (posBalance >= 115) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (SC >= 916) {
                                    if (posBalance < 78) {
                                        if (ssDCval < 55) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 55) {
                                            if (gchDC < 20) {
                                                return 1;
                                            }
                                            else if (gchDC >= 20) {
                                                if (Cs < 1596) {
                                                    return 0;
                                                }
                                                else if (Cs >= 1596) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 78) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 91) {
                                if (Rs < 198) {
                                    if (Cs < 246) {
                                        return 0;
                                    }
                                    else if (Cs >= 246) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 198) {
                                    return 0;
                                }
                            }
                        }
                        else if (RsDiff >= 12) {
                            if (gchDC < 22) {
                                if (diffTSC < 14) {
                                    if (Rs < 1967) {
                                        if (AFD < 28) {
                                            return 1;
                                        }
                                        else if (AFD >= 28) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 1967) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 14) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 22) {
                                if (CsDiff < 16) {
                                    if (RsDiff < 15) {
                                        if (TSC < 24) {
                                            return 0;
                                        }
                                        else if (TSC >= 24) {
                                            if (Rs < 1944) {
                                                return 1;
                                            }
                                            else if (Rs >= 1944) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 15) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 16) {
                                    if (RsDiff < 13) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 13) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (Rs >= 2771) {
                        return 0;
                    }
                }
            }
            else if (RsDiff >= 25) {
                if (Cs < 5022) {
                    return 1;
                }
                else if (Cs >= 5022) {
                    if (MVDiff < 51) {
                        return 0;
                    }
                    else if (MVDiff >= 51) {
                        return 1;
                    }
                }
            }
        }
    }
    else if (diffMVdiffVal >= 65) {
        if (diffAFD < 13) {
            if (diffTSC < 4) {
                if (gchDC < 27) {
                    if (RsCsDiff < 18) {
                        if (MVDiff < 106) {
                            if (SC < 87) {
                                if (Rs < 9) {
                                    return 0;
                                }
                                else if (Rs >= 9) {
                                    return 1;
                                }
                            }
                            else if (SC >= 87) {
                                return 0;
                            }
                        }
                        else if (MVDiff >= 106) {
                            if (diffRsCsdiff < 13) {
                                if (refDCval < 35) {
                                    if (ssDCval < 36) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 36) {
                                        if (ssDCval < 37) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 37) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 35) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 13) {
                                return 1;
                            }
                        }
                    }
                    else if (RsCsDiff >= 18) {
                        if (ssDCval < 42) {
                            if (diffRsCsdiff < 50) {
                                if (Rs < 476) {
                                    if (refDCval < 30) {
                                        return 1;
                                    }
                                    else if (refDCval >= 30) {
                                        if (posBalance < 38) {
                                            if (diffTSC < -11) {
                                                return 1;
                                            }
                                            else if (diffTSC >= -11) {
                                                return 0;
                                            }
                                        }
                                        else if (posBalance >= 38) {
                                            if (posBalance < 52) {
                                                return 1;
                                            }
                                            else if (posBalance >= 52) {
                                                if (AFD < 20) {
                                                    return 1;
                                                }
                                                else if (AFD >= 20) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Rs >= 476) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 50) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 42) {
                            if (AFD < 36) {
                                if (Rs < 101) {
                                    if (Rs < 91) {
                                        return 0;
                                    }
                                    else if (Rs >= 91) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 101) {
                                    return 0;
                                }
                            }
                            else if (AFD >= 36) {
                                if (diffTSC < 2) {
                                    if (RsCsDiff < 197) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 197) {
                                        if (Cs < 860) {
                                            if (CsDiff < 8) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 8) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 860) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffTSC >= 2) {
                                    if (Rs < 1317) {
                                        if (SC < 828) {
                                            return 0;
                                        }
                                        else if (SC >= 828) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 1317) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (gchDC >= 27) {
                    if (ssDCval < 24) {
                        return 1;
                    }
                    else if (ssDCval >= 24) {
                        if (diffRsCsdiff < 45) {
                            if (ssDCval < 228) {
                                return 0;
                            }
                            else if (ssDCval >= 228) {
                                return 1;
                            }
                        }
                        else if (diffRsCsdiff >= 45) {
                            if (CsDiff < 12) {
                                if (diffAFD < -3) {
                                    return 1;
                                }
                                else if (diffAFD >= -3) {
                                    if (SC < 111) {
                                        return 1;
                                    }
                                    else if (SC >= 111) {
                                        return 0;
                                    }
                                }
                            }
                            else if (CsDiff >= 12) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 4) {
                if (ssDCval < 48) {
                    if (negBalance < 17) {
                        if (Rs < 956) {
                            return 0;
                        }
                        else if (Rs >= 956) {
                            return 1;
                        }
                    }
                    else if (negBalance >= 17) {
                        if (negBalance < 113) {
                            if (AFD < 20) {
                                if (refDCval < 25) {
                                    if (Rs < 164) {
                                        return 1;
                                    }
                                    else if (Rs >= 164) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 25) {
                                    if (diffRsCsdiff < 2) {
                                        if (gchDC < 8) {
                                            return 1;
                                        }
                                        else if (gchDC >= 8) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 2) {
                                        return 1;
                                    }
                                }
                            }
                            else if (AFD >= 20) {
                                if (Rs < 173) {
                                    return 0;
                                }
                                else if (Rs >= 173) {
                                    if (AFD < 22) {
                                        if (Cs < 306) {
                                            return 1;
                                        }
                                        else if (Cs >= 306) {
                                            return 0;
                                        }
                                    }
                                    else if (AFD >= 22) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 113) {
                            return 0;
                        }
                    }
                }
                else if (ssDCval >= 48) {
                    if (CsDiff < 5) {
                        if (Rs < 7) {
                            if (SC < 6) {
                                return 0;
                            }
                            else if (SC >= 6) {
                                return 1;
                            }
                        }
                        else if (Rs >= 7) {
                            return 0;
                        }
                    }
                    else if (CsDiff >= 5) {
                        if (AFD < 59) {
                            if (Rs < 396) {
                                if (TSC < 24) {
                                    return 1;
                                }
                                else if (TSC >= 24) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 396) {
                                if (TSC < 24) {
                                    if (diffMVdiffVal < 255) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 255) {
                                        if (Rs < 950) {
                                            return 1;
                                        }
                                        else if (Rs >= 950) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (TSC >= 24) {
                                    if (CsDiff < 9) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 9) {
                                        if (Rs < 1633) {
                                            return 1;
                                        }
                                        else if (Rs >= 1633) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (AFD >= 59) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 13) {
            if (diffRsCsdiff < 66) {
                if (Cs < 601) {
                    if (posBalance < 88) {
                        if (diffRsCsdiff < 2) {
                            if (ssDCval < 62) {
                                if (MVDiff < 186) {
                                    return 0;
                                }
                                else if (MVDiff >= 186) {
                                    if (AFD < 37) {
                                        return 1;
                                    }
                                    else if (AFD >= 37) {
                                        if (TSC < 20) {
                                            if (Cs < 1) {
                                                return 1;
                                            }
                                            else if (Cs >= 1) {
                                                return 0;
                                            }
                                        }
                                        else if (TSC >= 20) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 62) {
                                if (SC < 198) {
                                    if (MVDiff < 417) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 417) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 198) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 2) {
                            if (diffMVdiffVal < 87) {
                                if (Cs < 150) {
                                    return 1;
                                }
                                else if (Cs >= 150) {
                                    if (MVDiff < 88) {
                                        if (Cs < 241) {
                                            return 0;
                                        }
                                        else if (Cs >= 241) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 88) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 87) {
                                if (diffTSC < 7) {
                                    if (refDCval < 65) {
                                        return 1;
                                    }
                                    else if (refDCval >= 65) {
                                        if (diffMVdiffVal < 432) {
                                            if (diffRsCsdiff < 51) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 51) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 432) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffTSC >= 7) {
                                    if (negBalance < 34) {
                                        if (Rs < 396) {
                                            return 1;
                                        }
                                        else if (Rs >= 396) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 34) {
                                        if (Scindex < 1) {
                                            if (MVDiff < 429) {
                                                if (RsCsDiff < 70) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 70) {
                                                    return 0;
                                                }
                                            }
                                            else if (MVDiff >= 429) {
                                                return 0;
                                            }
                                        }
                                        else if (Scindex >= 1) {
                                            if (ssDCval < 167) {
                                                if (negBalance < 43) {
                                                    if (diffAFD < 31) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= 31) {
                                                        return 0;
                                                    }
                                                }
                                                else if (negBalance >= 43) {
                                                    if (MVDiff < 933) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 933) {
                                                        if (TSC < 13) {
                                                            return 0;
                                                        }
                                                        else if (TSC >= 13) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 167) {
                                                if (Rs < 228) {
                                                    return 0;
                                                }
                                                else if (Rs >= 228) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (posBalance >= 88) {
                        if (diffRsCsdiff < 18) {
                            if (TSC < 11) {
                                return 1;
                            }
                            else if (TSC >= 11) {
                                if (diffAFD < 29) {
                                    if (gchDC < 20) {
                                        if (diffRsCsdiff < 2) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 2) {
                                            if (Rs < 306) {
                                                return 1;
                                            }
                                            else if (Rs >= 306) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 20) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 29) {
                                    if (TSC < 27) {
                                        return 1;
                                    }
                                    else if (TSC >= 27) {
                                        if (RsDiff < 4) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 4) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 18) {
                            if (TSC < 32) {
                                if (SC < 872) {
                                    if (RsCsDiff < 49) {
                                        if (negBalance < 2) {
                                            if (MVDiff < 121) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 121) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 2) {
                                            return 1;
                                        }
                                    }
                                    else if (RsCsDiff >= 49) {
                                        if (negBalance < 21) {
                                            if (gchDC < 87) {
                                                return 0;
                                            }
                                            else if (gchDC >= 87) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 21) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (SC >= 872) {
                                    if (RsCsDiff < 41) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 41) {
                                        if (posBalance < 93) {
                                            return 0;
                                        }
                                        else if (posBalance >= 93) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 32) {
                                if (ssDCval < 152) {
                                    return 0;
                                }
                                else if (ssDCval >= 152) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 601) {
                    if (RsCsDiff < 83) {
                        if (Rs < 986) {
                            if (diffMVdiffVal < 569) {
                                if (SC < 1212) {
                                    if (Rs < 461) {
                                        return 0;
                                    }
                                    else if (Rs >= 461) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 1212) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 569) {
                                return 1;
                            }
                        }
                        else if (Rs >= 986) {
                            return 1;
                        }
                    }
                    else if (RsCsDiff >= 83) {
                        return 0;
                    }
                }
            }
            else if (diffRsCsdiff >= 66) {
                if (diffTSC < 10) {
                    if (ssDCval < 67) {
                        if (negBalance < 35) {
                            return 0;
                        }
                        else if (negBalance >= 35) {
                            if (MVDiff < 258) {
                                if (AFD < 34) {
                                    if (diffMVdiffVal < 67) {
                                        if (Rs < 430) {
                                            return 0;
                                        }
                                        else if (Rs >= 430) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 67) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 34) {
                                    if (TSC < 15) {
                                        return 0;
                                    }
                                    else if (TSC >= 15) {
                                        return 1;
                                    }
                                }
                            }
                            else if (MVDiff >= 258) {
                                return 1;
                            }
                        }
                    }
                    else if (ssDCval >= 67) {
                        if (Rs < 1579) {
                            if (Cs < 906) {
                                if (negBalance < 16) {
                                    return 0;
                                }
                                else if (negBalance >= 16) {
                                    if (SC < 1141) {
                                        return 1;
                                    }
                                    else if (SC >= 1141) {
                                        if (RsCsDiff < 199) {
                                            if (AFD < 28) {
                                                return 1;
                                            }
                                            else if (AFD >= 28) {
                                                return 0;
                                            }
                                        }
                                        else if (RsCsDiff >= 199) {
                                            if (Cs < 227) {
                                                return 0;
                                            }
                                            else if (Cs >= 227) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 906) {
                                if (RsCsDiff < 146) {
                                    return 1;
                                }
                                else if (RsCsDiff >= 146) {
                                    return 0;
                                }
                            }
                        }
                        else if (Rs >= 1579) {
                            if (ssDCval < 164) {
                                return 0;
                            }
                            else if (ssDCval >= 164) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffTSC >= 10) {
                    if (negBalance < 21) {
                        if (diffAFD < 46) {
                            if (Rs < 573) {
                                return 1;
                            }
                            else if (Rs >= 573) {
                                if (diffRsCsdiff < 208) {
                                    if (Cs < 809) {
                                        if (Cs < 768) {
                                            return 0;
                                        }
                                        else if (Cs >= 768) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 809) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 208) {
                                    if (AFD < 53) {
                                        if (negBalance < 12) {
                                            if (posBalance < 113) {
                                                return 0;
                                            }
                                            else if (posBalance >= 113) {
                                                if (AFD < 32) {
                                                    return 1;
                                                }
                                                else if (AFD >= 32) {
                                                    if (ssDCval < 132) {
                                                        return 0;
                                                    }
                                                    else if (ssDCval >= 132) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (negBalance >= 12) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 53) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 46) {
                            return 1;
                        }
                    }
                    else if (negBalance >= 21) {
                        if (gchDC < 20) {
                            if (Rs < 9642) {
                                if (diffTSC < 12) {
                                    if (ssDCval < 92) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 92) {
                                        if (posBalance < 63) {
                                            return 1;
                                        }
                                        else if (posBalance >= 63) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffTSC >= 12) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 9642) {
                                return 0;
                            }
                        }
                        else if (gchDC >= 20) {
                            if (AFD < 42) {
                                if (diffAFD < 27) {
                                    if (AFD < 39) {
                                        if (diffMVdiffVal < 175) {
                                            if (RsCsDiff < 386) {
                                                if (diffRsCsdiff < 172) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= 172) {
                                                    return 1;
                                                }
                                            }
                                            else if (RsCsDiff >= 386) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 175) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 39) {
                                        if (Rs < 930) {
                                            return 1;
                                        }
                                        else if (Rs >= 930) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffAFD >= 27) {
                                    if (negBalance < 120) {
                                        return 1;
                                    }
                                    else if (negBalance >= 120) {
                                        if (AFD < 38) {
                                            return 1;
                                        }
                                        else if (AFD >= 38) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 42) {
                                if (diffAFD < 26) {
                                    if (RsDiff < 16) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 16) {
                                        if (AFD < 49) {
                                            if (Rs < 2005) {
                                                return 0;
                                            }
                                            else if (Rs >= 2005) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 49) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffAFD >= 26) {
                                    if (CsDiff < 7) {
                                        if (Cs < 825) {
                                            return 1;
                                        }
                                        else if (Cs >= 825) {
                                            if (Rs < 1670) {
                                                return 0;
                                            }
                                            else if (Rs >= 1670) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 7) {
                                        if (gchDC < 22) {
                                            if (TSC < 40) {
                                                return 1;
                                            }
                                            else if (TSC >= 40) {
                                                if (Rs < 1879) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1879) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 22) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectB(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 13) {
        if (diffMVdiffVal < 43) {
            if (gchDC < 14) {
                if (MVDiff < 107) {
                    if (ssDCval < 21) {
                        if (diffRsCsdiff < 83) {
                            if (RsDiff < 2) {
                                return 0;
                            }
                            else if (RsDiff >= 2) {
                                if (MVDiff < 6) {
                                    return 1;
                                }
                                else if (MVDiff >= 6) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 83) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 21) {
                        if (diffTSC < 10) {
                            if (diffRsCsdiff < 1513) {
                                if (refDCval < 230) {
                                    return 0;
                                }
                                else if (refDCval >= 230) {
                                    if (TSC < 4) {
                                        return 0;
                                    }
                                    else if (TSC >= 4) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 1513) {
                                if (SC < 5352) {
                                    return 1;
                                }
                                else if (SC >= 5352) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffTSC >= 10) {
                            if (ssDCval < 47) {
                                if (SC < 3868) {
                                    return 0;
                                }
                                else if (SC >= 3868) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 47) {
                                return 0;
                            }
                        }
                    }
                }
                else if (MVDiff >= 107) {
                    if (diffTSC < 5) {
                        if (MVDiff < 690) {
                            if (RsDiff < 63) {
                                if (AFD < 18) {
                                    return 0;
                                }
                                else if (AFD >= 18) {
                                    if (MVDiff < 223) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 223) {
                                        if (refDCval < 51) {
                                            if (CsDiff < 4) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 4) {
                                                if (Cs < 450) {
                                                    return 1;
                                                }
                                                else if (Cs >= 450) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 51) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (RsDiff >= 63) {
                                if (SC < 7960) {
                                    return 1;
                                }
                                else if (SC >= 7960) {
                                    return 0;
                                }
                            }
                        }
                        else if (MVDiff >= 690) {
                            if (diffTSC < -3) {
                                return 0;
                            }
                            else if (diffTSC >= -3) {
                                if (diffAFD < -9) {
                                    if (gchDC < 1) {
                                        return 0;
                                    }
                                    else if (gchDC >= 1) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= -9) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 5) {
                        if (ssDCval < 44) {
                            if (Rs < 174) {
                                return 0;
                            }
                            else if (Rs >= 174) {
                                if (Cs < 293) {
                                    return 1;
                                }
                                else if (Cs >= 293) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 44) {
                            return 0;
                        }
                    }
                }
            }
            else if (gchDC >= 14) {
                if (AFD < 59) {
                    if (posBalance < 128) {
                        if (refDCval < 24) {
                            return 1;
                        }
                        else if (refDCval >= 24) {
                            if (diffRsCsdiff < 644) {
                                if (MVDiff < 231) {
                                    return 0;
                                }
                                else if (MVDiff >= 231) {
                                    if (ssDCval < 27) {
                                        if (AFD < 41) {
                                            return 0;
                                        }
                                        else if (AFD >= 41) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 27) {
                                        if (refDCval < 57) {
                                            if (RsDiff < 3) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 3) {
                                                if (negBalance < 41) {
                                                    if (MVDiff < 237) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 237) {
                                                        return 0;
                                                    }
                                                }
                                                else if (negBalance >= 41) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 57) {
                                            if (diffRsCsdiff < 92) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 92) {
                                                if (SC < 1961) {
                                                    return 1;
                                                }
                                                else if (SC >= 1961) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 644) {
                                if (Cs < 1204) {
                                    return 1;
                                }
                                else if (Cs >= 1204) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (posBalance >= 128) {
                        if (SC < 910) {
                            if (refDCval < 90) {
                                return 1;
                            }
                            else if (refDCval >= 90) {
                                return 0;
                            }
                        }
                        else if (SC >= 910) {
                            return 0;
                        }
                    }
                }
                else if (AFD >= 59) {
                    if (Cs < 1102) {
                        return 0;
                    }
                    else if (Cs >= 1102) {
                        return 1;
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 43) {
            if (ssDCval < 50) {
                if (diffRsCsdiff < 1) {
                    if (diffAFD < 7) {
                        if (gchDC < 65) {
                            if (diffRsCsdiff < -246) {
                                if (ssDCval < 29) {
                                    return 1;
                                }
                                else if (ssDCval >= 29) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= -246) {
                                if (diffMVdiffVal < 282) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 282) {
                                    if (MVDiff < 441) {
                                        if (posBalance < 76) {
                                            return 0;
                                        }
                                        else if (posBalance >= 76) {
                                            if (Rs < 4) {
                                                return 0;
                                            }
                                            else if (Rs >= 4) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 441) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 65) {
                            if (Rs < 4) {
                                return 1;
                            }
                            else if (Rs >= 4) {
                                if (AFD < 17) {
                                    return 1;
                                }
                                else if (AFD >= 17) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 7) {
                        if (MVDiff < 218) {
                            if (AFD < 12) {
                                if (Rs < 28) {
                                    return 0;
                                }
                                else if (Rs >= 28) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 12) {
                                return 0;
                            }
                        }
                        else if (MVDiff >= 218) {
                            if (diffAFD < 8) {
                                if (Cs < 244) {
                                    return 0;
                                }
                                else if (Cs >= 244) {
                                    return 1;
                                }
                            }
                            else if (diffAFD >= 8) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 1) {
                    if (diffAFD < 8) {
                        if (AFD < 32) {
                            if (ssDCval < 30) {
                                if (diffRsCsdiff < 9) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 9) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 30) {
                                if (diffAFD < 7) {
                                    if (TSC < 13) {
                                        return 0;
                                    }
                                    else if (TSC >= 13) {
                                        if (Rs < 1059) {
                                            return 1;
                                        }
                                        else if (Rs >= 1059) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffAFD >= 7) {
                                    if (Rs < 228) {
                                        return 1;
                                    }
                                    else if (Rs >= 228) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (AFD >= 32) {
                            return 1;
                        }
                    }
                    else if (diffAFD >= 8) {
                        if (ssDCval < 31) {
                            return 1;
                        }
                        else if (ssDCval >= 31) {
                            if (diffMVdiffVal < 97) {
                                if (MVDiff < 97) {
                                    if (RsDiff < 7) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 7) {
                                        return 1;
                                    }
                                }
                                else if (MVDiff >= 97) {
                                    if (RsDiff < 5) {
                                        if (MVDiff < 196) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 196) {
                                            return 1;
                                        }
                                    }
                                    else if (RsDiff >= 5) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 97) {
                                if (posBalance < 13) {
                                    return 0;
                                }
                                else if (posBalance >= 13) {
                                    if (diffTSC < 9) {
                                        if (TSC < 5) {
                                            return 0;
                                        }
                                        else if (TSC >= 5) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= 9) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (ssDCval >= 50) {
                if (diffTSC < 5) {
                    if (gchDC < 87) {
                        if (TSC < 32) {
                            if (TSC < 14) {
                                if (negBalance < 30) {
                                    if (diffRsCsdiff < 13) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 13) {
                                        if (SC < 588) {
                                            return 1;
                                        }
                                        else if (SC >= 588) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (negBalance >= 30) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 14) {
                                if (Cs < 33) {
                                    if (RsDiff < 9) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 9) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 33) {
                                    if (negBalance < 69) {
                                        return 0;
                                    }
                                    else if (negBalance >= 69) {
                                        if (posBalance < 54) {
                                            if (ssDCval < 52) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 52) {
                                                return 0;
                                            }
                                        }
                                        else if (posBalance >= 54) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (TSC >= 32) {
                            if (Rs < 1676) {
                                if (AFD < 54) {
                                    return 1;
                                }
                                else if (AFD >= 54) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 1676) {
                                return 1;
                            }
                        }
                    }
                    else if (gchDC >= 87) {
                        return 1;
                    }
                }
                else if (diffTSC >= 5) {
                    if (TSC < 36) {
                        if (diffRsCsdiff < 39) {
                            if (Rs < 64) {
                                if (ssDCval < 54) {
                                    return 1;
                                }
                                else if (ssDCval >= 54) {
                                    if (RsDiff < 5) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 5) {
                                        if (TSC < 21) {
                                            return 1;
                                        }
                                        else if (TSC >= 21) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 64) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 39) {
                            if (Rs < 1013) {
                                if (MVDiff < 253) {
                                    if (AFD < 23) {
                                        return 0;
                                    }
                                    else if (AFD >= 23) {
                                        if (Rs < 314) {
                                            return 0;
                                        }
                                        else if (Rs >= 314) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (MVDiff >= 253) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 1013) {
                                if (refDCval < 82) {
                                    if (RsDiff < 18) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 18) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 82) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (TSC >= 36) {
                        return 1;
                    }
                }
            }
        }
    }
    else if (diffAFD >= 13) {
        if (MVDiff < 56) {
            if (TSC < 24) {
                if (gchDC < 7) {
                    if (diffMVdiffVal < 48) {
                        if (CsDiff < 4) {
                            if (diffTSC < 11) {
                                return 0;
                            }
                            else if (diffTSC >= 11) {
                                return 1;
                            }
                        }
                        else if (CsDiff >= 4) {
                            return 0;
                        }
                    }
                    else if (diffMVdiffVal >= 48) {
                        return 1;
                    }
                }
                else if (gchDC >= 7) {
                    if (refDCval < 32) {
                        return 1;
                    }
                    else if (refDCval >= 32) {
                        if (RsDiff < 15) {
                            return 0;
                        }
                        else if (RsDiff >= 15) {
                            if (AFD < 25) {
                                return 1;
                            }
                            else if (AFD >= 25) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (TSC >= 24) {
                if (Cs < 4804) {
                    if (ssDCval < 105) {
                        return 1;
                    }
                    else if (ssDCval >= 105) {
                        if (diffMVdiffVal < -14) {
                            return 1;
                        }
                        else if (diffMVdiffVal >= -14) {
                            return 0;
                        }
                    }
                }
                else if (Cs >= 4804) {
                    return 0;
                }
            }
        }
        else if (MVDiff >= 56) {
            if (diffMVdiffVal < 63) {
                if (RsDiff < 10) {
                    if (diffRsCsdiff < 67) {
                        if (Rs < 270) {
                            if (SC < 359) {
                                if (diffTSC < 3) {
                                    if (Rs < 30) {
                                        return 0;
                                    }
                                    else if (Rs >= 30) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= 3) {
                                    return 0;
                                }
                            }
                            else if (SC >= 359) {
                                if (negBalance < 2) {
                                    return 1;
                                }
                                else if (negBalance >= 2) {
                                    if (RsDiff < 3) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 3) {
                                        if (Cs < 617) {
                                            return 1;
                                        }
                                        else if (Cs >= 617) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 270) {
                            return 0;
                        }
                    }
                    else if (diffRsCsdiff >= 67) {
                        if (MVDiff < 115) {
                            return 0;
                        }
                        else if (MVDiff >= 115) {
                            if (RsCsDiff < 115) {
                                if (AFD < 29) {
                                    if (SC < 1516) {
                                        return 0;
                                    }
                                    else if (SC >= 1516) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 29) {
                                    if (SC < 1823) {
                                        return 1;
                                    }
                                    else if (SC >= 1823) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 115) {
                                return 0;
                            }
                        }
                    }
                }
                else if (RsDiff >= 10) {
                    if (diffTSC < 9) {
                        if (negBalance < 80) {
                            return 0;
                        }
                        else if (negBalance >= 80) {
                            if (Cs < 658) {
                                return 1;
                            }
                            else if (Cs >= 658) {
                                return 0;
                            }
                        }
                    }
                    else if (diffTSC >= 9) {
                        if (posBalance < 84) {
                            if (Cs < 2727) {
                                if (CsDiff < 3) {
                                    return 0;
                                }
                                else if (CsDiff >= 3) {
                                    if (negBalance < 80) {
                                        return 1;
                                    }
                                    else if (negBalance >= 80) {
                                        if (gchDC < 29) {
                                            if (Cs < 1581) {
                                                return 0;
                                            }
                                            else if (Cs >= 1581) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 29) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 2727) {
                                if (Rs < 8691) {
                                    if (MVDiff < 139) {
                                        if (diffRsCsdiff < 1913) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 1913) {
                                            if (Rs < 6452) {
                                                return 1;
                                            }
                                            else if (Rs >= 6452) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 139) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 8691) {
                                    return 1;
                                }
                            }
                        }
                        else if (posBalance >= 84) {
                            if (TSC < 48) {
                                if (RsDiff < 27) {
                                    if (AFD < 30) {
                                        if (SC < 2873) {
                                            return 1;
                                        }
                                        else if (SC >= 2873) {
                                            return 0;
                                        }
                                    }
                                    else if (AFD >= 30) {
                                        return 0;
                                    }
                                }
                                else if (RsDiff >= 27) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 48) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 63) {
                if (negBalance < 27) {
                    if (diffRsCsdiff < 60) {
                        if (TSC < 11) {
                            return 1;
                        }
                        else if (TSC >= 11) {
                            if (diffAFD < 33) {
                                if (diffMVdiffVal < 116) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 116) {
                                    if (MVDiff < 209) {
                                        if (diffMVdiffVal < 173) {
                                            if (Cs < 65) {
                                                return 1;
                                            }
                                            else if (Cs >= 65) {
                                                if (Rs < 1156) {
                                                    return 0;
                                                }
                                                else if (Rs >= 1156) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 173) {
                                            if (RsCsDiff < 19) {
                                                return 0;
                                            }
                                            else if (RsCsDiff >= 19) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 209) {
                                        if (diffAFD < 18) {
                                            if (refDCval < 41) {
                                                return 1;
                                            }
                                            else if (refDCval >= 41) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= 18) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 33) {
                                if (AFD < 42) {
                                    if (CsDiff < 4) {
                                        return 1;
                                    }
                                    else if (CsDiff >= 4) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 42) {
                                    if (ssDCval < 152) {
                                        if (AFD < 51) {
                                            if (ssDCval < 105) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 105) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 51) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 152) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 60) {
                        if (diffAFD < 31) {
                            if (Cs < 694) {
                                if (diffAFD < 16) {
                                    return 0;
                                }
                                else if (diffAFD >= 16) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 694) {
                                if (diffRsCsdiff < 916) {
                                    if (Rs < 1622) {
                                        if (diffRsCsdiff < 339) {
                                            if (Cs < 847) {
                                                return 1;
                                            }
                                            else if (Cs >= 847) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 339) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 1622) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 916) {
                                    if (AFD < 26) {
                                        return 0;
                                    }
                                    else if (AFD >= 26) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 31) {
                            if (negBalance < 1) {
                                if (CsDiff < 9) {
                                    return 1;
                                }
                                else if (CsDiff >= 9) {
                                    return 0;
                                }
                            }
                            else if (negBalance >= 1) {
                                if (ssDCval < 86) {
                                    if (Rs < 1174) {
                                        return 1;
                                    }
                                    else if (Rs >= 1174) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 86) {
                                    if (Cs < 816) {
                                        return 1;
                                    }
                                    else if (Cs >= 816) {
                                        if (RsCsDiff < 207) {
                                            if (TSC < 59) {
                                                return 0;
                                            }
                                            else if (TSC >= 59) {
                                                return 1;
                                            }
                                        }
                                        else if (RsCsDiff >= 207) {
                                            if (diffAFD < 35) {
                                                if (Rs < 2141) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2141) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffAFD >= 35) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 27) {
                    if (diffTSC < 10) {
                        if (AFD < 27) {
                            if (ssDCval < 36) {
                                return 1;
                            }
                            else if (ssDCval >= 36) {
                                if (diffMVdiffVal < 108) {
                                    if (Rs < 479) {
                                        return 0;
                                    }
                                    else if (Rs >= 479) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 108) {
                                    if (posBalance < 12) {
                                        if (Rs < 534) {
                                            return 1;
                                        }
                                        else if (Rs >= 534) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 12) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (AFD >= 27) {
                            if (diffAFD < 18) {
                                if (negBalance < 98) {
                                    if (ssDCval < 78) {
                                        if (diffRsCsdiff < 78) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 78) {
                                            if (CsDiff < 7) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 7) {
                                                if (posBalance < 60) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 60) {
                                                    if (Rs < 733) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 733) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 78) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 98) {
                                    return 1;
                                }
                            }
                            else if (diffAFD >= 18) {
                                if (diffRsCsdiff < 48) {
                                    if (diffTSC < 8) {
                                        if (posBalance < 70) {
                                            return 0;
                                        }
                                        else if (posBalance >= 70) {
                                            if (Rs < 498) {
                                                return 1;
                                            }
                                            else if (Rs >= 498) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 8) {
                                        return 1;
                                    }
                                }
                                else if (diffRsCsdiff >= 48) {
                                    if (SC < 1703) {
                                        return 1;
                                    }
                                    else if (SC >= 1703) {
                                        if (Rs < 1633) {
                                            if (diffTSC < 5) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 5) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 1633) {
                                            if (SC < 2722) {
                                                return 1;
                                            }
                                            else if (SC >= 2722) {
                                                if (SC < 3950) {
                                                    return 0;
                                                }
                                                else if (SC >= 3950) {
                                                    if (diffTSC < 7) {
                                                        return 0;
                                                    }
                                                    else if (diffTSC >= 7) {
                                                        if (Cs < 1917) {
                                                            return 0;
                                                        }
                                                        else if (Cs >= 1917) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 10) {
                        if (gchDC < 21) {
                            if (negBalance < 43) {
                                if (diffRsCsdiff < 135) {
                                    if (AFD < 31) {
                                        return 1;
                                    }
                                    else if (AFD >= 31) {
                                        if (diffTSC < 13) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 13) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 135) {
                                    if (Rs < 8314) {
                                        return 1;
                                    }
                                    else if (Rs >= 8314) {
                                        return 0;
                                    }
                                }
                            }
                            else if (negBalance >= 43) {
                                if (diffTSC < 12) {
                                    if (AFD < 49) {
                                        if (MVDiff < 80) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 80) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 49) {
                                        if (diffMVdiffVal < 163) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 163) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffTSC >= 12) {
                                    if (MVDiff < 1206) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 1206) {
                                        if (Rs < 513) {
                                            return 0;
                                        }
                                        else if (Rs >= 513) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 21) {
                            if (diffAFD < 29) {
                                if (TSC < 28) {
                                    if (ssDCval < 63) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 63) {
                                        if (posBalance < 91) {
                                            if (SC < 2277) {
                                                if (ssDCval < 90) {
                                                    if (diffRsCsdiff < -72) {
                                                        return 0;
                                                    }
                                                    else if (diffRsCsdiff >= -72) {
                                                        return 1;
                                                    }
                                                }
                                                else if (ssDCval >= 90) {
                                                    if (Rs < 519) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 519) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (SC >= 2277) {
                                                return 0;
                                            }
                                        }
                                        else if (posBalance >= 91) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (TSC >= 28) {
                                    if (diffTSC < 41) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 41) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffAFD >= 29) {
                                if (AFD < 87) {
                                    if (CsDiff < 7) {
                                        if (Cs < 622) {
                                            if (diffRsCsdiff < 41) {
                                                if (diffMVdiffVal < 590) {
                                                    if (diffMVdiffVal < 111) {
                                                        return 0;
                                                    }
                                                    else if (diffMVdiffVal >= 111) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 590) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffRsCsdiff >= 41) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 622) {
                                            if (SC < 2542) {
                                                return 0;
                                            }
                                            else if (SC >= 2542) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (CsDiff >= 7) {
                                        if (gchDC < 22) {
                                            if (negBalance < 47) {
                                                if (AFD < 50) {
                                                    return 1;
                                                }
                                                else if (AFD >= 50) {
                                                    return 0;
                                                }
                                            }
                                            else if (negBalance >= 47) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 22) {
                                            if (diffTSC < 16) {
                                                if (ssDCval < 113) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 113) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffTSC >= 16) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (AFD >= 87) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectC(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 13) {
        if (diffMVdiffVal < 52) {
            if (MVDiff < 231) {
                if (ssDCval < 20) {
                    if (CsDiff < 6) {
                        if (RsDiff < 2) {
                            return 0;
                        }
                        else if (RsDiff >= 2) {
                            if (MVDiff < 6) {
                                return 1;
                            }
                            else if (MVDiff >= 6) {
                                return 0;
                            }
                        }
                    }
                    else if (CsDiff >= 6) {
                        return 1;
                    }
                }
                else if (ssDCval >= 20) {
                    if (RsDiff < 66) {
                        if (diffAFD < 9) {
                            if (refDCval < 228) {
                                if (diffRsCsdiff < 260) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 260) {
                                    if (Scindex < 5) {
                                        if (TSCindex < 6) {
                                            if (Cs < 1923) {
                                                return 1;
                                            }
                                            else if (Cs >= 1923) {
                                                return 0;
                                            }
                                        }
                                        else if (TSCindex >= 6) {
                                            if (refDCval < 53) {
                                                if (SC < 3378) {
                                                    return 1;
                                                }
                                                else if (SC >= 3378) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 53) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Scindex >= 5) {
                                        return 0;
                                    }
                                }
                            }
                            else if (refDCval >= 228) {
                                if (ssDCval < 207) {
                                    return 1;
                                }
                                else if (ssDCval >= 207) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffAFD >= 9) {
                            if (Cs < 158) {
                                if (CsDiff < 4) {
                                    if (Rs < 15) {
                                        if (SC < 50) {
                                            return 0;
                                        }
                                        else if (SC >= 50) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 15) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 4) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 158) {
                                if (TSC < 27) {
                                    return 0;
                                }
                                else if (TSC >= 27) {
                                    if (ssDCval < 93) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 93) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (RsDiff >= 66) {
                        if (Cs < 853) {
                            return 1;
                        }
                        else if (Cs >= 853) {
                            if (TSC < 32) {
                                return 0;
                            }
                            else if (TSC >= 32) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (MVDiff >= 231) {
                if (AFD < 62) {
                    if (diffRsCsdiff < 12) {
                        if (gchDC < 6) {
                            if (diffTSC < -3) {
                                return 0;
                            }
                            else if (diffTSC >= -3) {
                                if (diffAFD < -18) {
                                    if (Rs < 2) {
                                        return 0;
                                    }
                                    else if (Rs >= 2) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= -18) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 6) {
                            if (refDCval < 26) {
                                return 1;
                            }
                            else if (refDCval >= 26) {
                                if (refDCval < 230) {
                                    if (refDCval < 72) {
                                        if (Rs < 24) {
                                            if (diffMVdiffVal < -20) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= -20) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 24) {
                                            if (diffRsCsdiff < 5) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 5) {
                                                if (Cs < 707) {
                                                    return 0;
                                                }
                                                else if (Cs >= 707) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (refDCval >= 72) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 230) {
                                    if (MVDiff < 468) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 468) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 12) {
                        if (negBalance < 1) {
                            if (Cs < 287) {
                                return 0;
                            }
                            else if (Cs >= 287) {
                                return 1;
                            }
                        }
                        else if (negBalance >= 1) {
                            if (ssDCval < 46) {
                                if (AFD < 18) {
                                    return 0;
                                }
                                else if (AFD >= 18) {
                                    if (diffMVdiffVal < -52) {
                                        return 1;
                                    }
                                    else if (diffMVdiffVal >= -52) {
                                        if (Cs < 1509) {
                                            return 0;
                                        }
                                        else if (Cs >= 1509) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 46) {
                                if (MVDiff < 232) {
                                    return 1;
                                }
                                else if (MVDiff >= 232) {
                                    if (diffRsCsdiff < 304) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 304) {
                                        if (SC < 3249) {
                                            return 1;
                                        }
                                        else if (SC >= 3249) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (AFD >= 62) {
                    return 1;
                }
            }
        }
        else if (diffMVdiffVal >= 52) {
            if (diffRsCsdiff < 1) {
                if (diffMVdiffVal < 219) {
                    if (AFD < 22) {
                        return 0;
                    }
                    else if (AFD >= 22) {
                        if (MVDiff < 263) {
                            if (Cs < 4432) {
                                return 0;
                            }
                            else if (Cs >= 4432) {
                                return 1;
                            }
                        }
                        else if (MVDiff >= 263) {
                            if (MVDiff < 270) {
                                if (gchDC < 10) {
                                    return 1;
                                }
                                else if (gchDC >= 10) {
                                    return 0;
                                }
                            }
                            else if (MVDiff >= 270) {
                                if (MVDiff < 393) {
                                    if (gchDC < 9) {
                                        return 0;
                                    }
                                    else if (gchDC >= 9) {
                                        if (TSC < 21) {
                                            if (SC < 2) {
                                                return 1;
                                            }
                                            else if (SC >= 2) {
                                                if (diffAFD < 12) {
                                                    return 0;
                                                }
                                                else if (diffAFD >= 12) {
                                                    if (refDCval < 45) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 45) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (TSC >= 21) {
                                            if (negBalance < 81) {
                                                return 1;
                                            }
                                            else if (negBalance >= 81) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (MVDiff >= 393) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 219) {
                    if (ssDCval < 41) {
                        if (diffAFD < 3) {
                            if (Rs < 7) {
                                if (Rs < 3) {
                                    if (AFD < 23) {
                                        return 0;
                                    }
                                    else if (AFD >= 23) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 3) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 7) {
                                return 0;
                            }
                        }
                        else if (diffAFD >= 3) {
                            if (posBalance < 88) {
                                if (Scindex < 3) {
                                    return 1;
                                }
                                else if (Scindex >= 3) {
                                    if (Rs < 429) {
                                        return 0;
                                    }
                                    else if (Rs >= 429) {
                                        return 1;
                                    }
                                }
                            }
                            else if (posBalance >= 88) {
                                return 0;
                            }
                        }
                    }
                    else if (ssDCval >= 41) {
                        if (gchDC < 162) {
                            if (diffTSC < 5) {
                                return 0;
                            }
                            else if (diffTSC >= 5) {
                                if (Cs < 3) {
                                    return 1;
                                }
                                else if (Cs >= 3) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 162) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 1) {
                if (ssDCval < 44) {
                    if (CsDiff < 3) {
                        if (AFD < 11) {
                            return 0;
                        }
                        else if (AFD >= 11) {
                            if (ssDCval < 33) {
                                return 1;
                            }
                            else if (ssDCval >= 33) {
                                if (diffTSC < 6) {
                                    if (Rs < 336) {
                                        if (TSCindex < 6) {
                                            return 1;
                                        }
                                        else if (TSCindex >= 6) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 336) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= 6) {
                                    if (Rs < 189) {
                                        return 1;
                                    }
                                    else if (Rs >= 189) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 3) {
                        if (diffTSC < 5) {
                            if (diffRsCsdiff < 149) {
                                if (Cs < 651) {
                                    if (diffRsCsdiff < 17) {
                                        if (refDCval < 31) {
                                            return 1;
                                        }
                                        else if (refDCval >= 31) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 17) {
                                        if (ssDCval < 40) {
                                            if (MVDiff < 238) {
                                                if (negBalance < 117) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 117) {
                                                    if (Rs < 155) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 155) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 238) {
                                                if (RsCsDiff < 67) {
                                                    if (Rs < 266) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 266) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsCsDiff >= 67) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 40) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (Cs >= 651) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 149) {
                                return 1;
                            }
                        }
                        else if (diffTSC >= 5) {
                            return 1;
                        }
                    }
                }
                else if (ssDCval >= 44) {
                    if (diffTSC < 4) {
                        if (RsCsDiff < 1883) {
                            if (TSC < 32) {
                                if (diffMVdiffVal < 391) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 391) {
                                    if (MVDiff < 610) {
                                        if (Rs < 797) {
                                            return 1;
                                        }
                                        else if (Rs >= 797) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 610) {
                                        return 0;
                                    }
                                }
                            }
                            else if (TSC >= 32) {
                                if (RsDiff < 10) {
                                    return 0;
                                }
                                else if (RsDiff >= 10) {
                                    return 1;
                                }
                            }
                        }
                        else if (RsCsDiff >= 1883) {
                            if (posBalance < 39) {
                                return 1;
                            }
                            else if (posBalance >= 39) {
                                if (negBalance < 23) {
                                    if (Rs < 770) {
                                        return 0;
                                    }
                                    else if (Rs >= 770) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 23) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 4) {
                        if (Rs < 111) {
                            if (CsDiff < 2) {
                                if (MVDiff < 140) {
                                    return 1;
                                }
                                else if (MVDiff >= 140) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 2) {
                                return 1;
                            }
                        }
                        else if (Rs >= 111) {
                            if (negBalance < 34) {
                                return 0;
                            }
                            else if (negBalance >= 34) {
                                if (Cs < 1712) {
                                    if (diffRsCsdiff < 25) {
                                        if (refDCval < 45) {
                                            return 1;
                                        }
                                        else if (refDCval >= 45) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 25) {
                                        if (diffTSC < 8) {
                                            if (diffMVdiffVal < 101) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 101) {
                                                if (posBalance < 44) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 44) {
                                                    if (MVDiff < 378) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 378) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (diffTSC >= 8) {
                                            if (Cs < 1269) {
                                                return 1;
                                            }
                                            else if (Cs >= 1269) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1712) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 13) {
        if (diffMVdiffVal < 62) {
            if (diffTSC < 13) {
                if (MVDiff < 56) {
                    return 0;
                }
                else if (MVDiff >= 56) {
                    if (posBalance < 47) {
                        if (diffTSC < 3) {
                            return 1;
                        }
                        else if (diffTSC >= 3) {
                            if (diffMVdiffVal < 51) {
                                if (diffMVdiffVal < -103) {
                                    if (RsDiff < 2) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 2) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= -103) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 51) {
                                if (Rs < 3795) {
                                    return 1;
                                }
                                else if (Rs >= 3795) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (posBalance >= 47) {
                        if (diffRsCsdiff < 76) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 76) {
                            if (MVDiff < 388) {
                                if (diffRsCsdiff < 84) {
                                    if (Rs < 793) {
                                        return 1;
                                    }
                                    else if (Rs >= 793) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 84) {
                                    if (refDCval < 79) {
                                        if (TSC < 27) {
                                            if (diffMVdiffVal < 53) {
                                                if (Cs < 425) {
                                                    if (Rs < 1896) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 1896) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Cs >= 425) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 53) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 27) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 79) {
                                        return 0;
                                    }
                                }
                            }
                            else if (MVDiff >= 388) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 13) {
                if (RsDiff < 9) {
                    if (Scindex < 4) {
                        if (diffRsCsdiff < 60) {
                            if (MVDiff < 24) {
                                return 1;
                            }
                            else if (MVDiff >= 24) {
                                if (diffRsCsdiff < -114) {
                                    if (AFD < 47) {
                                        return 1;
                                    }
                                    else if (AFD >= 47) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= -114) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 60) {
                            return 1;
                        }
                    }
                    else if (Scindex >= 4) {
                        return 0;
                    }
                }
                else if (RsDiff >= 9) {
                    if (diffTSC < 42) {
                        if (Cs < 3538) {
                            if (RsCsDiff < 733) {
                                if (posBalance < 91) {
                                    if (Cs < 2220) {
                                        if (diffRsCsdiff < 200) {
                                            if (ssDCval < 69) {
                                                if (Cs < 790) {
                                                    return 1;
                                                }
                                                else if (Cs >= 790) {
                                                    return 0;
                                                }
                                            }
                                            else if (ssDCval >= 69) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 200) {
                                            if (refDCval < 132) {
                                                return 1;
                                            }
                                            else if (refDCval >= 132) {
                                                if (ssDCval < 112) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 112) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (Cs >= 2220) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 91) {
                                    if (AFD < 29) {
                                        if (Rs < 1867) {
                                            return 1;
                                        }
                                        else if (Rs >= 1867) {
                                            return 0;
                                        }
                                    }
                                    else if (AFD >= 29) {
                                        return 0;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 733) {
                                if (SC < 12216) {
                                    return 1;
                                }
                                else if (SC >= 12216) {
                                    if (SC < 17545) {
                                        return 0;
                                    }
                                    else if (SC >= 17545) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 3538) {
                            if (gchDC < 9) {
                                if (Cs < 12050) {
                                    return 0;
                                }
                                else if (Cs >= 12050) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 9) {
                                if (RsCsDiff < 6143) {
                                    if (AFD < 41) {
                                        if (CsDiff < 23) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 23) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 41) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 6143) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 42) {
                        return 1;
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 62) {
            if (diffAFD < 21) {
                if (refDCval < 82) {
                    if (negBalance < 42) {
                        if (TSC < 13) {
                            if (diffTSC < 9) {
                                if (SC < 406) {
                                    return 1;
                                }
                                else if (SC >= 406) {
                                    if (AFD < 17) {
                                        return 1;
                                    }
                                    else if (AFD >= 17) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= 9) {
                                if (MVDiff < 505) {
                                    return 1;
                                }
                                else if (MVDiff >= 505) {
                                    if (Cs < 422) {
                                        return 1;
                                    }
                                    else if (Cs >= 422) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 13) {
                            if (CsDiff < 16) {
                                if (refDCval < 25) {
                                    return 1;
                                }
                                else if (refDCval >= 25) {
                                    if (negBalance < 17) {
                                        return 0;
                                    }
                                    else if (negBalance >= 17) {
                                        if (AFD < 37) {
                                            if (diffMVdiffVal < 451) {
                                                if (diffRsCsdiff < 4) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= 4) {
                                                    if (Cs < 376) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 376) {
                                                        if (MVDiff < 172) {
                                                            return 1;
                                                        }
                                                        else if (MVDiff >= 172) {
                                                            if (diffTSC < 14) {
                                                                return 0;
                                                            }
                                                            else if (diffTSC >= 14) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                            else if (diffMVdiffVal >= 451) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 37) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 16) {
                                if (AFD < 46) {
                                    return 1;
                                }
                                else if (AFD >= 46) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 42) {
                        if (ssDCval < 41) {
                            return 1;
                        }
                        else if (ssDCval >= 41) {
                            if (diffTSC < 8) {
                                if (Cs < 369) {
                                    if (CsDiff < 2) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 2) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 369) {
                                    if (AFD < 26) {
                                        return 1;
                                    }
                                    else if (AFD >= 26) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= 8) {
                                if (gchDC < 13) {
                                    if (refDCval < 42) {
                                        if (Cs < 429) {
                                            return 1;
                                        }
                                        else if (Cs >= 429) {
                                            return 0;
                                        }
                                    }
                                    else if (refDCval >= 42) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 13) {
                                    if (AFD < 37) {
                                        return 0;
                                    }
                                    else if (AFD >= 37) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (refDCval >= 82) {
                    if (ssDCval < 53) {
                        return 1;
                    }
                    else if (ssDCval >= 53) {
                        if (diffTSC < 7) {
                            return 0;
                        }
                        else if (diffTSC >= 7) {
                            if (ssDCval < 116) {
                                if (refDCval < 107) {
                                    if (Rs < 5104) {
                                        if (Cs < 835) {
                                            if (SC < 1155) {
                                                return 0;
                                            }
                                            else if (SC >= 1155) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 835) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 5104) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 107) {
                                    if (posBalance < 3) {
                                        return 0;
                                    }
                                    else if (posBalance >= 3) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 116) {
                                if (Rs < 32) {
                                    return 1;
                                }
                                else if (Rs >= 32) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 21) {
                if (diffRsCsdiff < 59) {
                    if (negBalance < 21) {
                        if (SC < 1013) {
                            if (gchDC < 77) {
                                if (AFD < 36) {
                                    if (RsDiff < 5) {
                                        if (posBalance < 107) {
                                            return 0;
                                        }
                                        else if (posBalance >= 107) {
                                            if (Cs < 493) {
                                                return 1;
                                            }
                                            else if (Cs >= 493) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 5) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 36) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 77) {
                                if (Cs < 34) {
                                    if (Rs < 1) {
                                        return 1;
                                    }
                                    else if (Rs >= 1) {
                                        return 0;
                                    }
                                }
                                else if (Cs >= 34) {
                                    return 1;
                                }
                            }
                        }
                        else if (SC >= 1013) {
                            return 0;
                        }
                    }
                    else if (negBalance >= 21) {
                        if (RsDiff < 7) {
                            if (gchDC < 20) {
                                if (MVDiff < 107) {
                                    if (Cs < 241) {
                                        return 0;
                                    }
                                    else if (Cs >= 241) {
                                        return 1;
                                    }
                                }
                                else if (MVDiff >= 107) {
                                    if (refDCval < 93) {
                                        return 1;
                                    }
                                    else if (refDCval >= 93) {
                                        if (refDCval < 99) {
                                            return 0;
                                        }
                                        else if (refDCval >= 99) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (gchDC >= 20) {
                                if (MVDiff < 543) {
                                    if (Cs < 419) {
                                        if (gchDC < 78) {
                                            return 1;
                                        }
                                        else if (gchDC >= 78) {
                                            return 0;
                                        }
                                    }
                                    else if (Cs >= 419) {
                                        if (gchDC < 31) {
                                            if (Rs < 221) {
                                                return 0;
                                            }
                                            else if (Rs >= 221) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 31) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (MVDiff >= 543) {
                                    if (diffRsCsdiff < 40) {
                                        if (Rs < 219) {
                                            return 0;
                                        }
                                        else if (Rs >= 219) {
                                            if (AFD < 70) {
                                                return 0;
                                            }
                                            else if (AFD >= 70) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= 40) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 7) {
                            if (MVDiff < 96) {
                                return 1;
                            }
                            else if (MVDiff >= 96) {
                                return 0;
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 59) {
                    if (diffAFD < 31) {
                        if (negBalance < 5) {
                            if (ssDCval < 173) {
                                return 0;
                            }
                            else if (ssDCval >= 173) {
                                return 1;
                            }
                        }
                        else if (negBalance >= 5) {
                            if (diffTSC < 11) {
                                if (SC < 1182) {
                                    if (refDCval < 45) {
                                        if (Rs < 384) {
                                            return 0;
                                        }
                                        else if (Rs >= 384) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 45) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 1182) {
                                    if (posBalance < 48) {
                                        if (Rs < 1961) {
                                            return 1;
                                        }
                                        else if (Rs >= 1961) {
                                            if (SC < 4409) {
                                                return 0;
                                            }
                                            else if (SC >= 4409) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 48) {
                                        if (negBalance < 33) {
                                            return 1;
                                        }
                                        else if (negBalance >= 33) {
                                            if (ssDCval < 71) {
                                                if (MVDiff < 134) {
                                                    return 0;
                                                }
                                                else if (MVDiff >= 134) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 71) {
                                                if (RsDiff < 11) {
                                                    if (Rs < 815) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 815) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 11) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 11) {
                                if (negBalance < 30) {
                                    if (SC < 2439) {
                                        if (posBalance < 73) {
                                            return 0;
                                        }
                                        else if (posBalance >= 73) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 2439) {
                                        if (diffRsCsdiff < 916) {
                                            if (ssDCval < 128) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 128) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 916) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (negBalance >= 30) {
                                    if (TSC < 30) {
                                        if (ssDCval < 36) {
                                            if (RsDiff < 6) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 6) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 36) {
                                            if (RsDiff < 27) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 27) {
                                                if (diffRsCsdiff < 920) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= 920) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 30) {
                                        if (AFD < 60) {
                                            return 0;
                                        }
                                        else if (AFD >= 60) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 31) {
                        if (posBalance < 95) {
                            if (negBalance < 123) {
                                if (CsDiff < 4) {
                                    if (Cs < 501) {
                                        return 1;
                                    }
                                    else if (Cs >= 501) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 4) {
                                    if (MVDiff < 1246) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 1246) {
                                        if (MVDiff < 1319) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 1319) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 123) {
                                return 0;
                            }
                        }
                        else if (posBalance >= 95) {
                            if (TSC < 35) {
                                if (refDCval < 46) {
                                    if (Rs < 1474) {
                                        if (Cs < 813) {
                                            return 1;
                                        }
                                        else if (Cs >= 813) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 1474) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 46) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 35) {
                                if (diffMVdiffVal < 218) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 218) {
                                    if (MVDiff < 1033) {
                                        if (AFD < 66) {
                                            return 1;
                                        }
                                        else if (AFD >= 66) {
                                            if (Rs < 930) {
                                                return 1;
                                            }
                                            else if (Rs >= 930) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 1033) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectD(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffTSC < 6) {
        if (diffAFD < 7) {
            if (AFD < 47) {
                if (diffMVdiffVal < 23) {
                    if (refDCval < 24) {
                        if (CsDiff < 12) {
                            if (ssDCval < 127) {
                                if (RsCsDiff < 7) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 7) {
                                    if (refDCval < 18) {
                                        if (MVDiff < 9) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 9) {
                                            return 0;
                                        }
                                    }
                                    else if (refDCval >= 18) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 127) {
                                return 1;
                            }
                        }
                        else if (CsDiff >= 12) {
                            return 1;
                        }
                    }
                    else if (refDCval >= 24) {
                        if (gchDC < 4) {
                            return 0;
                        }
                        else if (gchDC >= 4) {
                            if (SC < 87) {
                                if (CsDiff < 4) {
                                    if (diffRsCsdiff < -1334) {
                                        return 1;
                                    }
                                    else if (diffRsCsdiff >= -1334) {
                                        if (SC < 74) {
                                            return 0;
                                        }
                                        else if (SC >= 74) {
                                            if (refDCval < 69) {
                                                return 0;
                                            }
                                            else if (refDCval >= 69) {
                                                if (Cs < 39) {
                                                    return 0;
                                                }
                                                else if (Cs >= 39) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 4) {
                                    return 1;
                                }
                            }
                            else if (SC >= 87) {
                                if (TSC < 14) {
                                    return 0;
                                }
                                else if (TSC >= 14) {
                                    if (ssDCval < 62) {
                                        if (gchDC < 17) {
                                            if (Rs < 885) {
                                                if (SC < 955) {
                                                    return 0;
                                                }
                                                else if (SC >= 955) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 885) {
                                                return 0;
                                            }
                                        }
                                        else if (gchDC >= 17) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 62) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 23) {
                    if (gchDC < 117) {
                        if (ssDCval < 37) {
                            if (RsCsDiff < 18) {
                                if (diffTSC < 4) {
                                    if (ssDCval < 36) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 36) {
                                        if (diffMVdiffVal < 264) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 264) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffTSC >= 4) {
                                    if (MVDiff < 244) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 244) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 18) {
                                if (diffRsCsdiff < 60) {
                                    if (Scindex < 1) {
                                        return 1;
                                    }
                                    else if (Scindex >= 1) {
                                        if (diffMVdiffVal < 97) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 97) {
                                            if (MVDiff < 300) {
                                                if (TSC < 5) {
                                                    return 0;
                                                }
                                                else if (TSC >= 5) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 300) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 60) {
                                    if (RsCsDiff < 1965) {
                                        return 1;
                                    }
                                    else if (RsCsDiff >= 1965) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 37) {
                            if (TSC < 15) {
                                if (diffRsCsdiff < 1500) {
                                    if (diffMVdiffVal < 100) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 100) {
                                        if (posBalance < 95) {
                                            return 0;
                                        }
                                        else if (posBalance >= 95) {
                                            if (MVDiff < 176) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 176) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 1500) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 15) {
                                if (diffTSC < -3) {
                                    if (gchDC < 12) {
                                        if (diffAFD < -9) {
                                            return 0;
                                        }
                                        else if (diffAFD >= -9) {
                                            if (MVDiff < 308) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 308) {
                                                if (Rs < 1939) {
                                                    if (Rs < 793) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 793) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 1939) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (gchDC >= 12) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= -3) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 117) {
                        if (TSCindex < 9) {
                            return 1;
                        }
                        else if (TSCindex >= 9) {
                            if (Rs < 6) {
                                return 1;
                            }
                            else if (Rs >= 6) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (AFD >= 47) {
                if (CsDiff < 9) {
                    return 0;
                }
                else if (CsDiff >= 9) {
                    if (gchDC < 22) {
                        return 0;
                    }
                    else if (gchDC >= 22) {
                        if (Rs < 2189) {
                            if (posBalance < 109) {
                                if (refDCval < 137) {
                                    return 1;
                                }
                                else if (refDCval >= 137) {
                                    return 0;
                                }
                            }
                            else if (posBalance >= 109) {
                                return 0;
                            }
                        }
                        else if (Rs >= 2189) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 7) {
            if (ssDCval < 60) {
                if (diffMVdiffVal < 94) {
                    if (Rs < 21) {
                        if (Rs < 5) {
                            return 0;
                        }
                        else if (Rs >= 5) {
                            return 1;
                        }
                    }
                    else if (Rs >= 21) {
                        if (refDCval < 34) {
                            if (ssDCval < 32) {
                                return 1;
                            }
                            else if (ssDCval >= 32) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 34) {
                            if (MVDiff < 119) {
                                return 0;
                            }
                            else if (MVDiff >= 119) {
                                if (SC < 524) {
                                    if (diffAFD < 10) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 10) {
                                        if (AFD < 19) {
                                            return 0;
                                        }
                                        else if (AFD >= 19) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (SC >= 524) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 94) {
                    if (negBalance < 78) {
                        if (posBalance < 80) {
                            if (Rs < 61) {
                                if (ssDCval < 40) {
                                    return 1;
                                }
                                else if (ssDCval >= 40) {
                                    if (CsDiff < 1) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 1) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Rs >= 61) {
                                return 1;
                            }
                        }
                        else if (posBalance >= 80) {
                            if (Rs < 238) {
                                if (Rs < 35) {
                                    return 0;
                                }
                                else if (Rs >= 35) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 238) {
                                return 0;
                            }
                        }
                    }
                    else if (negBalance >= 78) {
                        if (gchDC < 21) {
                            if (Rs < 169) {
                                if (posBalance < 28) {
                                    if (TSC < 3) {
                                        return 1;
                                    }
                                    else if (TSC >= 3) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 28) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 169) {
                                return 0;
                            }
                        }
                        else if (gchDC >= 21) {
                            if (TSC < 14) {
                                if (gchDC < 25) {
                                    if (Cs < 100) {
                                        return 1;
                                    }
                                    else if (Cs >= 100) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 25) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 14) {
                                if (gchDC < 36) {
                                    if (AFD < 30) {
                                        return 0;
                                    }
                                    else if (AFD >= 30) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 36) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (ssDCval >= 60) {
                if (gchDC < 28) {
                    if (MVDiff < 70) {
                        return 0;
                    }
                    else if (MVDiff >= 70) {
                        if (diffMVdiffVal < 784) {
                            if (AFD < 11) {
                                if (Rs < 2301) {
                                    return 0;
                                }
                                else if (Rs >= 2301) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 11) {
                                if (ssDCval < 76) {
                                    if (Rs < 101) {
                                        if (TSC < 9) {
                                            return 0;
                                        }
                                        else if (TSC >= 9) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 101) {
                                        if (TSC < 27) {
                                            if (negBalance < 68) {
                                                return 0;
                                            }
                                            else if (negBalance >= 68) {
                                                if (posBalance < 49) {
                                                    return 0;
                                                }
                                                else if (posBalance >= 49) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (TSC >= 27) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (ssDCval >= 76) {
                                    if (AFD < 67) {
                                        return 0;
                                    }
                                    else if (AFD >= 67) {
                                        if (TSC < 25) {
                                            return 1;
                                        }
                                        else if (TSC >= 25) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 784) {
                            if (ssDCval < 111) {
                                return 1;
                            }
                            else if (ssDCval >= 111) {
                                return 0;
                            }
                        }
                    }
                }
                else if (gchDC >= 28) {
                    if (Cs < 292) {
                        if (diffTSC < 4) {
                            if (Cs < 210) {
                                return 0;
                            }
                            else if (Cs >= 210) {
                                return 1;
                            }
                        }
                        else if (diffTSC >= 4) {
                            if (refDCval < 179) {
                                return 1;
                            }
                            else if (refDCval >= 179) {
                                return 0;
                            }
                        }
                    }
                    else if (Cs >= 292) {
                        return 0;
                    }
                }
            }
        }
    }
    else if (diffTSC >= 6) {
        if (diffAFD < 22) {
            if (diffMVdiffVal < 54) {
                if (SC < 1002) {
                    if (MVDiff < 482) {
                        if (ssDCval < 52) {
                            if (TSC < 13) {
                                if (gchDC < 5) {
                                    if (Cs < 214) {
                                        return 1;
                                    }
                                    else if (Cs >= 214) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 5) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 13) {
                                if (SC < 781) {
                                    if (gchDC < 8) {
                                        return 1;
                                    }
                                    else if (gchDC >= 8) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 781) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 52) {
                            if (diffRsCsdiff < -19) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= -19) {
                                if (AFD < 11) {
                                    if (TSC < 9) {
                                        return 0;
                                    }
                                    else if (TSC >= 9) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 11) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (MVDiff >= 482) {
                        if (RsCsDiff < 4) {
                            return 0;
                        }
                        else if (RsCsDiff >= 4) {
                            return 1;
                        }
                    }
                }
                else if (SC >= 1002) {
                    if (diffTSC < 12) {
                        if (MVDiff < 60) {
                            return 0;
                        }
                        else if (MVDiff >= 60) {
                            if (MVDiff < 61) {
                                if (Rs < 1201) {
                                    return 1;
                                }
                                else if (Rs >= 1201) {
                                    return 0;
                                }
                            }
                            else if (MVDiff >= 61) {
                                if (AFD < 28) {
                                    return 0;
                                }
                                else if (AFD >= 28) {
                                    if (ssDCval < 88) {
                                        if (gchDC < 14) {
                                            if (CsDiff < 10) {
                                                return 1;
                                            }
                                            else if (CsDiff >= 10) {
                                                if (Rs < 3418) {
                                                    return 0;
                                                }
                                                else if (Rs >= 3418) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 14) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 88) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 12) {
                        if (RsDiff < 53) {
                            if (CsDiff < 20) {
                                if (RsCsDiff < 115) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 115) {
                                    if (diffMVdiffVal < -168) {
                                        return 1;
                                    }
                                    else if (diffMVdiffVal >= -168) {
                                        if (Cs < 953) {
                                            if (RsDiff < 13) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 13) {
                                                if (Rs < 2999) {
                                                    return 0;
                                                }
                                                else if (Rs >= 2999) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (Cs >= 953) {
                                            if (diffTSC < 24) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 24) {
                                                if (AFD < 63) {
                                                    return 1;
                                                }
                                                else if (AFD >= 63) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 20) {
                                if (SC < 3973) {
                                    return 1;
                                }
                                else if (SC >= 3973) {
                                    if (MVDiff < 188) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 188) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 53) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 54) {
                if (ssDCval < 49) {
                    if (diffRsCsdiff < 1) {
                        if (diffMVdiffVal < 197) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 197) {
                            if (TSC < 11) {
                                return 1;
                            }
                            else if (TSC >= 11) {
                                if (AFD < 43) {
                                    return 0;
                                }
                                else if (AFD >= 43) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 1) {
                        if (RsDiff < 3) {
                            return 1;
                        }
                        else if (RsDiff >= 3) {
                            if (diffRsCsdiff < 175) {
                                if (diffMVdiffVal < 76) {
                                    if (AFD < 20) {
                                        return 0;
                                    }
                                    else if (AFD >= 20) {
                                        if (SC < 1845) {
                                            return 1;
                                        }
                                        else if (SC >= 1845) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 76) {
                                    if (AFD < 31) {
                                        return 1;
                                    }
                                    else if (AFD >= 31) {
                                        if (gchDC < 15) {
                                            return 0;
                                        }
                                        else if (gchDC >= 15) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 175) {
                                return 1;
                            }
                        }
                    }
                }
                else if (ssDCval >= 49) {
                    if (diffAFD < 17) {
                        if (diffMVdiffVal < 485) {
                            if (SC < 175) {
                                if (diffRsCsdiff < -1) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= -1) {
                                    if (refDCval < 119) {
                                        return 1;
                                    }
                                    else if (refDCval >= 119) {
                                        return 0;
                                    }
                                }
                            }
                            else if (SC >= 175) {
                                if (RsCsDiff < 41) {
                                    if (refDCval < 34) {
                                        return 1;
                                    }
                                    else if (refDCval >= 34) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 41) {
                                    if (Rs < 1436) {
                                        if (gchDC < 10) {
                                            if (diffMVdiffVal < 84) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 84) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 10) {
                                            if (Rs < 1271) {
                                                if (diffRsCsdiff < 41) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= 41) {
                                                    if (ssDCval < 88) {
                                                        if (MVDiff < 184) {
                                                            if (Cs < 433) {
                                                                return 1;
                                                            }
                                                            else if (Cs >= 433) {
                                                                if (SC < 1523) {
                                                                    return 1;
                                                                }
                                                                else if (SC >= 1523) {
                                                                    return 0;
                                                                }
                                                            }
                                                        }
                                                        else if (MVDiff >= 184) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (ssDCval >= 88) {
                                                        if (MVDiff < 482) {
                                                            return 0;
                                                        }
                                                        else if (MVDiff >= 482) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (Rs >= 1271) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Rs >= 1436) {
                                        if (TSC < 39) {
                                            return 0;
                                        }
                                        else if (TSC >= 39) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 485) {
                            return 1;
                        }
                    }
                    else if (diffAFD >= 17) {
                        if (gchDC < 16) {
                            if (ssDCval < 121) {
                                return 1;
                            }
                            else if (ssDCval >= 121) {
                                if (SC < 1137) {
                                    return 1;
                                }
                                else if (SC >= 1137) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 16) {
                            if (diffMVdiffVal < 212) {
                                if (SC < 14) {
                                    return 1;
                                }
                                else if (SC >= 14) {
                                    if (ssDCval < 58) {
                                        if (Rs < 223) {
                                            return 0;
                                        }
                                        else if (Rs >= 223) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 58) {
                                        if (CsDiff < 30) {
                                            if (MVDiff < 150) {
                                                if (AFD < 32) {
                                                    if (diffMVdiffVal < 128) {
                                                        return 0;
                                                    }
                                                    else if (diffMVdiffVal >= 128) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 32) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 150) {
                                                return 0;
                                            }
                                        }
                                        else if (CsDiff >= 30) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= 212) {
                                if (diffRsCsdiff < 19) {
                                    if (Rs < 141) {
                                        if (Rs < 39) {
                                            return 0;
                                        }
                                        else if (Rs >= 39) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 141) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 19) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 22) {
            if (diffMVdiffVal < 64) {
                if (RsDiff < 6) {
                    if (diffMVdiffVal < 26) {
                        return 0;
                    }
                    else if (diffMVdiffVal >= 26) {
                        if (ssDCval < 59) {
                            return 1;
                        }
                        else if (ssDCval >= 59) {
                            return 0;
                        }
                    }
                }
                else if (RsDiff >= 6) {
                    if (TSC < 16) {
                        if (Cs < 589) {
                            return 1;
                        }
                        else if (Cs >= 589) {
                            if (refDCval < 40) {
                                return 1;
                            }
                            else if (refDCval >= 40) {
                                return 0;
                            }
                        }
                    }
                    else if (TSC >= 16) {
                        if (Cs < 369) {
                            return 1;
                        }
                        else if (Cs >= 369) {
                            if (diffRsCsdiff < 337) {
                                if (gchDC < 25) {
                                    if (diffMVdiffVal < 30) {
                                        if (SC < 1468) {
                                            if (Rs < 677) {
                                                return 1;
                                            }
                                            else if (Rs >= 677) {
                                                return 0;
                                            }
                                        }
                                        else if (SC >= 1468) {
                                            return 1;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 30) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 25) {
                                    if (Cs < 851) {
                                        if (gchDC < 60) {
                                            return 0;
                                        }
                                        else if (gchDC >= 60) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 851) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 337) {
                                if (refDCval < 106) {
                                    if (gchDC < 84) {
                                        return 1;
                                    }
                                    else if (gchDC >= 84) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 106) {
                                    if (MVDiff < 58) {
                                        if (Rs < 22468) {
                                            return 0;
                                        }
                                        else if (Rs >= 22468) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 58) {
                                        if (Rs < 1998) {
                                            if (TSC < 20) {
                                                return 1;
                                            }
                                            else if (TSC >= 20) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 1998) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 64) {
                if (negBalance < 13) {
                    if (ssDCval < 102) {
                        if (refDCval < 24) {
                            return 1;
                        }
                        else if (refDCval >= 24) {
                            return 0;
                        }
                    }
                    else if (ssDCval >= 102) {
                        if (diffTSC < 33) {
                            if (negBalance < 7) {
                                if (AFD < 32) {
                                    if (CsDiff < 5) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 5) {
                                        if (Cs < 4820) {
                                            return 1;
                                        }
                                        else if (Cs >= 4820) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (AFD >= 32) {
                                    return 0;
                                }
                            }
                            else if (negBalance >= 7) {
                                if (AFD < 59) {
                                    return 1;
                                }
                                else if (AFD >= 59) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffTSC >= 33) {
                            if (ssDCval < 111) {
                                return 0;
                            }
                            else if (ssDCval >= 111) {
                                return 1;
                            }
                        }
                    }
                }
                else if (negBalance >= 13) {
                    if (CsDiff < 6) {
                        if (negBalance < 46) {
                            if (Rs < 192) {
                                if (RsCsDiff < 5) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 5) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 192) {
                                if (CsDiff < 3) {
                                    if (CsDiff < 2) {
                                        if (Cs < 78) {
                                            return 0;
                                        }
                                        else if (Cs >= 78) {
                                            return 1;
                                        }
                                    }
                                    else if (CsDiff >= 2) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 3) {
                                    if (diffAFD < 25) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 25) {
                                        if (diffMVdiffVal < 162) {
                                            if (diffAFD < 31) {
                                                if (TSC < 23) {
                                                    return 1;
                                                }
                                                else if (TSC >= 23) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffAFD >= 31) {
                                                if (gchDC < 37) {
                                                    return 0;
                                                }
                                                else if (gchDC >= 37) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 162) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (negBalance >= 46) {
                            if (diffRsCsdiff < -1) {
                                if (AFD < 50) {
                                    return 1;
                                }
                                else if (AFD >= 50) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= -1) {
                                if (posBalance < 26) {
                                    if (MVDiff < 225) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 225) {
                                        if (Cs < 615) {
                                            if (Rs < 22) {
                                                return 0;
                                            }
                                            else if (Rs >= 22) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 615) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (posBalance >= 26) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (CsDiff >= 6) {
                        if (diffRsCsdiff < 34) {
                            if (Cs < 121) {
                                return 1;
                            }
                            else if (Cs >= 121) {
                                return 0;
                            }
                        }
                        else if (diffRsCsdiff >= 34) {
                            if (diffTSC < 16) {
                                if (Rs < 1964) {
                                    if (TSC < 29) {
                                        if (gchDC < 1) {
                                            if (Cs < 352) {
                                                return 0;
                                            }
                                            else if (Cs >= 352) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 1) {
                                            if (diffMVdiffVal < 67) {
                                                if (TSC < 13) {
                                                    return 0;
                                                }
                                                else if (TSC >= 13) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 67) {
                                                if (SC < 3034) {
                                                    if (diffMVdiffVal < 77) {
                                                        if (AFD < 39) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 39) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (diffMVdiffVal >= 77) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 3034) {
                                                    if (Cs < 2207) {
                                                        if (SC < 3093) {
                                                            return 0;
                                                        }
                                                        else if (SC >= 3093) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (Cs >= 2207) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (TSC >= 29) {
                                        if (Cs < 2633) {
                                            return 0;
                                        }
                                        else if (Cs >= 2633) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Rs >= 1964) {
                                    if (RsDiff < 29) {
                                        if (posBalance < 35) {
                                            return 0;
                                        }
                                        else if (posBalance >= 35) {
                                            if (RsDiff < 24) {
                                                return 1;
                                            }
                                            else if (RsDiff >= 24) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 29) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffTSC >= 16) {
                                if (negBalance < 121) {
                                    if (TSC < 39) {
                                        if (SC < 740) {
                                            if (SC < 730) {
                                                return 1;
                                            }
                                            else if (SC >= 730) {
                                                return 0;
                                            }
                                        }
                                        else if (SC >= 740) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 39) {
                                        if (diffAFD < 37) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 37) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (negBalance >= 121) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectE(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 13) {
        if (diffMVdiffVal < 52) {
            if (AFD < 53) {
                if (refDCval < 24) {
                    if (diffAFD < 2) {
                        if (ssDCval < 34) {
                            if (diffRsCsdiff < 5) {
                                if (MVDiff < 1517) {
                                    if (CsDiff < 3) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 3) {
                                        if (refDCval < 18) {
                                            return 1;
                                        }
                                        else if (refDCval >= 18) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (MVDiff >= 1517) {
                                    if (ssDCval < 24) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 24) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 5) {
                                if (posBalance < 2) {
                                    if (MVDiff < 9) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 9) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 2) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 34) {
                            return 1;
                        }
                    }
                    else if (diffAFD >= 2) {
                        if (posBalance < 16) {
                            return 1;
                        }
                        else if (posBalance >= 16) {
                            if (ssDCval < 54) {
                                return 0;
                            }
                            else if (ssDCval >= 54) {
                                return 1;
                            }
                        }
                    }
                }
                else if (refDCval >= 24) {
                    if (diffRsCsdiff < 1513) {
                        if (diffTSC < 5) {
                            if (Cs < 27) {
                                if (diffRsCsdiff < 37) {
                                    if (diffRsCsdiff < -25) {
                                        if (negBalance < 102) {
                                            return 0;
                                        }
                                        else if (negBalance >= 102) {
                                            if (ssDCval < 216) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 216) {
                                                if (AFD < 3) {
                                                    return 0;
                                                }
                                                else if (AFD >= 3) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (diffRsCsdiff >= -25) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 37) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 27) {
                                if (MVDiff < 107) {
                                    return 0;
                                }
                                else if (MVDiff >= 107) {
                                    if (TSC < 14) {
                                        return 0;
                                    }
                                    else if (TSC >= 14) {
                                        if (refDCval < 52) {
                                            if (refDCval < 50) {
                                                return 0;
                                            }
                                            else if (refDCval >= 50) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 52) {
                                            if (RsDiff < 56) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 56) {
                                                if (SC < 8492) {
                                                    return 1;
                                                }
                                                else if (SC >= 8492) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffTSC >= 5) {
                            if (diffMVdiffVal < -9) {
                                if (MVDiff < 230) {
                                    if (gchDC < 2) {
                                        if (MVDiff < 87) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 87) {
                                            return 1;
                                        }
                                    }
                                    else if (gchDC >= 2) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 230) {
                                    if (CsDiff < 3) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 3) {
                                        if (diffAFD < 6) {
                                            if (refDCval < 29) {
                                                return 0;
                                            }
                                            else if (refDCval >= 29) {
                                                return 1;
                                            }
                                        }
                                        else if (diffAFD >= 6) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffMVdiffVal >= -9) {
                                if (AFD < 23) {
                                    return 0;
                                }
                                else if (AFD >= 23) {
                                    if (ssDCval < 87) {
                                        if (diffRsCsdiff < 1090) {
                                            if (refDCval < 35) {
                                                return 1;
                                            }
                                            else if (refDCval >= 35) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 1090) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 87) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 1513) {
                        if (Rs < 2757) {
                            if (AFD < 6) {
                                return 0;
                            }
                            else if (AFD >= 6) {
                                return 1;
                            }
                        }
                        else if (Rs >= 2757) {
                            if (ssDCval < 50) {
                                return 1;
                            }
                            else if (ssDCval >= 50) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (AFD >= 53) {
                if (ssDCval < 114) {
                    return 1;
                }
                else if (ssDCval >= 114) {
                    if (Cs < 1498) {
                        return 0;
                    }
                    else if (Cs >= 1498) {
                        return 1;
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 52) {
            if (diffTSC < 4) {
                if (gchDC < 32) {
                    if (diffAFD < 8) {
                        if (ssDCval < 30) {
                            if (CsDiff < 3) {
                                if (diffRsCsdiff < -21) {
                                    if (Rs < 84) {
                                        return 1;
                                    }
                                    else if (Rs >= 84) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= -21) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 3) {
                                if (diffTSC < -1) {
                                    if (Cs < 404) {
                                        return 0;
                                    }
                                    else if (Cs >= 404) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= -1) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 30) {
                            if (AFD < 40) {
                                if (diffRsCsdiff < 13) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 13) {
                                    if (RsCsDiff < 15) {
                                        return 1;
                                    }
                                    else if (RsCsDiff >= 15) {
                                        if (ssDCval < 38) {
                                            if (AFD < 20) {
                                                return 0;
                                            }
                                            else if (AFD >= 20) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 38) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 40) {
                                if (SC < 1558) {
                                    if (RsDiff < 10) {
                                        if (CsDiff < 12) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 12) {
                                            return 1;
                                        }
                                    }
                                    else if (RsDiff >= 10) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 1558) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 8) {
                        if (ssDCval < 60) {
                            if (posBalance < 26) {
                                return 0;
                            }
                            else if (posBalance >= 26) {
                                if (Rs < 147) {
                                    return 1;
                                }
                                else if (Rs >= 147) {
                                    if (negBalance < 57) {
                                        return 0;
                                    }
                                    else if (negBalance >= 57) {
                                        if (Rs < 321) {
                                            return 0;
                                        }
                                        else if (Rs >= 321) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 60) {
                            return 0;
                        }
                    }
                }
                else if (gchDC >= 32) {
                    if (diffRsCsdiff < 119) {
                        if (MVDiff < 1653) {
                            if (diffMVdiffVal < 110) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 110) {
                                if (RsCsDiff < 110) {
                                    if (refDCval < 63) {
                                        return 1;
                                    }
                                    else if (refDCval >= 63) {
                                        if (diffAFD < -37) {
                                            return 1;
                                        }
                                        else if (diffAFD >= -37) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 110) {
                                    if (RsDiff < 13) {
                                        if (TSC < 18) {
                                            return 1;
                                        }
                                        else if (TSC >= 18) {
                                            if (Cs < 592) {
                                                return 0;
                                            }
                                            else if (Cs >= 592) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 13) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 1653) {
                            return 1;
                        }
                    }
                    else if (diffRsCsdiff >= 119) {
                        if (CsDiff < 12) {
                            return 0;
                        }
                        else if (CsDiff >= 12) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffTSC >= 4) {
                if (ssDCval < 50) {
                    if (diffAFD < 9) {
                        if (negBalance < 39) {
                            return 0;
                        }
                        else if (negBalance >= 39) {
                            if (Rs < 899) {
                                if (MVDiff < 141) {
                                    if (Rs < 212) {
                                        return 0;
                                    }
                                    else if (Rs >= 212) {
                                        return 1;
                                    }
                                }
                                else if (MVDiff >= 141) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 899) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 9) {
                        if (diffMVdiffVal < 55) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 55) {
                            if (gchDC < 13) {
                                if (AFD < 20) {
                                    return 1;
                                }
                                else if (AFD >= 20) {
                                    if (refDCval < 41) {
                                        return 1;
                                    }
                                    else if (refDCval >= 41) {
                                        return 0;
                                    }
                                }
                            }
                            else if (gchDC >= 13) {
                                if (TSC < 9) {
                                    return 1;
                                }
                                else if (TSC >= 9) {
                                    if (Rs < 381) {
                                        return 0;
                                    }
                                    else if (Rs >= 381) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 50) {
                    if (AFD < 59) {
                        if (Rs < 4) {
                            if (MVDiff < 228) {
                                return 0;
                            }
                            else if (MVDiff >= 228) {
                                return 1;
                            }
                        }
                        else if (Rs >= 4) {
                            if (Cs < 1016) {
                                if (RsCsDiff < 39) {
                                    if (posBalance < 116) {
                                        if (diffAFD < 12) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 12) {
                                            if (gchDC < 11) {
                                                return 1;
                                            }
                                            else if (gchDC >= 11) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (posBalance >= 116) {
                                        if (negBalance < 6) {
                                            return 0;
                                        }
                                        else if (negBalance >= 6) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 39) {
                                    if (negBalance < 32) {
                                        if (TSC < 9) {
                                            return 1;
                                        }
                                        else if (TSC >= 9) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 32) {
                                        if (diffRsCsdiff < 34) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 34) {
                                            if (SC < 1331) {
                                                return 1;
                                            }
                                            else if (SC >= 1331) {
                                                if (diffTSC < 8) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 8) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 1016) {
                                return 0;
                            }
                        }
                    }
                    else if (AFD >= 59) {
                        if (Rs < 1114) {
                            return 0;
                        }
                        else if (Rs >= 1114) {
                            return 1;
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 13) {
        if (diffMVdiffVal < 65) {
            if (diffTSC < 13) {
                if (MVDiff < 143) {
                    if (ssDCval < 69) {
                        if (ssDCval < 66) {
                            if (diffMVdiffVal < 50) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 50) {
                                if (RsCsDiff < 156) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 156) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 66) {
                            if (Rs < 1439) {
                                return 0;
                            }
                            else if (Rs >= 1439) {
                                return 1;
                            }
                        }
                    }
                    else if (ssDCval >= 69) {
                        return 0;
                    }
                }
                else if (MVDiff >= 143) {
                    if (AFD < 30) {
                        if (diffMVdiffVal < -84) {
                            return 1;
                        }
                        else if (diffMVdiffVal >= -84) {
                            if (SC < 128) {
                                return 1;
                            }
                            else if (SC >= 128) {
                                if (gchDC < 6) {
                                    return 1;
                                }
                                else if (gchDC >= 6) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (AFD >= 30) {
                        if (diffAFD < 31) {
                            if (AFD < 60) {
                                if (MVDiff < 158) {
                                    if (AFD < 39) {
                                        return 1;
                                    }
                                    else if (AFD >= 39) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 158) {
                                    if (posBalance < 109) {
                                        return 0;
                                    }
                                    else if (posBalance >= 109) {
                                        if (TSC < 20) {
                                            return 1;
                                        }
                                        else if (TSC >= 20) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 60) {
                                return 1;
                            }
                        }
                        else if (diffAFD >= 31) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffTSC >= 13) {
                if (RsDiff < 11) {
                    if (SC < 1535) {
                        if (diffRsCsdiff < 60) {
                            if (refDCval < 24) {
                                if (AFD < 40) {
                                    return 1;
                                }
                                else if (AFD >= 40) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 24) {
                                if (Rs < 224) {
                                    if (RsDiff < 3) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 3) {
                                        if (Cs < 214) {
                                            return 0;
                                        }
                                        else if (Cs >= 214) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Rs >= 224) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 60) {
                            if (gchDC < 2) {
                                if (RsDiff < 8) {
                                    return 1;
                                }
                                else if (RsDiff >= 8) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 2) {
                                return 1;
                            }
                        }
                    }
                    else if (SC >= 1535) {
                        return 0;
                    }
                }
                else if (RsDiff >= 11) {
                    if (Cs < 3054) {
                        if (RsCsDiff < 725) {
                            if (SC < 2534) {
                                if (posBalance < 92) {
                                    return 1;
                                }
                                else if (posBalance >= 92) {
                                    if (SC < 1837) {
                                        return 0;
                                    }
                                    else if (SC >= 1837) {
                                        return 1;
                                    }
                                }
                            }
                            else if (SC >= 2534) {
                                if (gchDC < 25) {
                                    if (diffTSC < 14) {
                                        if (refDCval < 38) {
                                            return 1;
                                        }
                                        else if (refDCval >= 38) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 14) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 25) {
                                    if (Rs < 1579) {
                                        if (Cs < 1381) {
                                            return 0;
                                        }
                                        else if (Cs >= 1381) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 1579) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 725) {
                            if (gchDC < 2) {
                                if (Rs < 6382) {
                                    return 1;
                                }
                                else if (Rs >= 6382) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 2) {
                                return 1;
                            }
                        }
                    }
                    else if (Cs >= 3054) {
                        if (posBalance < 75) {
                            if (MVDiff < 46) {
                                return 0;
                            }
                            else if (MVDiff >= 46) {
                                return 1;
                            }
                        }
                        else if (posBalance >= 75) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 65) {
            if (diffAFD < 23) {
                if (AFD < 27) {
                    if (ssDCval < 44) {
                        if (MVDiff < 94) {
                            if (ssDCval < 36) {
                                return 1;
                            }
                            else if (ssDCval >= 36) {
                                if (AFD < 18) {
                                    return 0;
                                }
                                else if (AFD >= 18) {
                                    return 1;
                                }
                            }
                        }
                        else if (MVDiff >= 94) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 44) {
                        if (diffRsCsdiff < 2) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 2) {
                            if (negBalance < 16) {
                                if (Rs < 163) {
                                    return 1;
                                }
                                else if (Rs >= 163) {
                                    if (diffAFD < 21) {
                                        if (ssDCval < 69) {
                                            if (diffMVdiffVal < 80) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 80) {
                                                if (Cs < 883) {
                                                    return 1;
                                                }
                                                else if (Cs >= 883) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 69) {
                                            return 0;
                                        }
                                    }
                                    else if (diffAFD >= 21) {
                                        return 1;
                                    }
                                }
                            }
                            else if (negBalance >= 16) {
                                if (diffTSC < 5) {
                                    return 0;
                                }
                                else if (diffTSC >= 5) {
                                    if (ssDCval < 97) {
                                        if (MVDiff < 813) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 813) {
                                            if (Cs < 231) {
                                                return 0;
                                            }
                                            else if (Cs >= 231) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 97) {
                                        if (Cs < 842) {
                                            return 1;
                                        }
                                        else if (Cs >= 842) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (AFD >= 27) {
                    if (posBalance < 69) {
                        if (ssDCval < 114) {
                            if (AFD < 53) {
                                if (diffTSC < 8) {
                                    if (SC < 1600) {
                                        if (refDCval < 72) {
                                            return 0;
                                        }
                                        else if (refDCval >= 72) {
                                            if (Rs < 1220) {
                                                if (Rs < 20) {
                                                    if (AFD < 35) {
                                                        return 0;
                                                    }
                                                    else if (AFD >= 35) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 20) {
                                                    return 1;
                                                }
                                            }
                                            else if (Rs >= 1220) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (SC >= 1600) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 8) {
                                    if (gchDC < 10) {
                                        return 1;
                                    }
                                    else if (gchDC >= 10) {
                                        if (diffAFD < 19) {
                                            if (gchDC < 31) {
                                                return 0;
                                            }
                                            else if (gchDC >= 31) {
                                                if (Rs < 2808) {
                                                    if (SC < 92) {
                                                        return 0;
                                                    }
                                                    else if (SC >= 92) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 2808) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffAFD >= 19) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 53) {
                                return 0;
                            }
                        }
                        else if (ssDCval >= 114) {
                            return 0;
                        }
                    }
                    else if (posBalance >= 69) {
                        if (CsDiff < 7) {
                            if (refDCval < 44) {
                                if (ssDCval < 55) {
                                    return 1;
                                }
                                else if (ssDCval >= 55) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 44) {
                                if (MVDiff < 411) {
                                    return 0;
                                }
                                else if (MVDiff >= 411) {
                                    if (MVDiff < 425) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 425) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 7) {
                            if (refDCval < 74) {
                                if (diffRsCsdiff < 74) {
                                    if (Rs < 401) {
                                        return 1;
                                    }
                                    else if (Rs >= 401) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 74) {
                                    if (Rs < 2430) {
                                        return 1;
                                    }
                                    else if (Rs >= 2430) {
                                        if (Rs < 4084) {
                                            return 0;
                                        }
                                        else if (Rs >= 4084) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 74) {
                                if (Cs < 140) {
                                    return 1;
                                }
                                else if (Cs >= 140) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 23) {
                if (negBalance < 22) {
                    if (negBalance < 5) {
                        if (Cs < 1861) {
                            if (diffAFD < 37) {
                                return 0;
                            }
                            else if (diffAFD >= 37) {
                                if (Cs < 924) {
                                    return 1;
                                }
                                else if (Cs >= 924) {
                                    return 0;
                                }
                            }
                        }
                        else if (Cs >= 1861) {
                            if (Rs < 5452) {
                                return 1;
                            }
                            else if (Rs >= 5452) {
                                return 0;
                            }
                        }
                    }
                    else if (negBalance >= 5) {
                        if (TSC < 24) {
                            if (CsDiff < 1) {
                                return 0;
                            }
                            else if (CsDiff >= 1) {
                                return 1;
                            }
                        }
                        else if (TSC >= 24) {
                            if (ssDCval < 112) {
                                if (posBalance < 88) {
                                    return 1;
                                }
                                else if (posBalance >= 88) {
                                    if (diffMVdiffVal < 154) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 154) {
                                        if (RsDiff < 11) {
                                            if (gchDC < 57) {
                                                return 0;
                                            }
                                            else if (gchDC >= 57) {
                                                return 1;
                                            }
                                        }
                                        else if (RsDiff >= 11) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 112) {
                                if (gchDC < 68) {
                                    if (Cs < 786) {
                                        return 1;
                                    }
                                    else if (Cs >= 786) {
                                        if (Cs < 2134) {
                                            return 0;
                                        }
                                        else if (Cs >= 2134) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (gchDC >= 68) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 22) {
                    if (diffTSC < 12) {
                        if (MVDiff < 80) {
                            return 0;
                        }
                        else if (MVDiff >= 80) {
                            if (AFD < 69) {
                                if (diffRsCsdiff < 28) {
                                    if (AFD < 38) {
                                        return 1;
                                    }
                                    else if (AFD >= 38) {
                                        if (diffMVdiffVal < 979) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 979) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 28) {
                                    if (Rs < 1961) {
                                        if (diffMVdiffVal < 119) {
                                            if (RsDiff < 7) {
                                                if (MVDiff < 135) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 135) {
                                                    return 0;
                                                }
                                            }
                                            else if (RsDiff >= 7) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 119) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 1961) {
                                        if (RsCsDiff < 822) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 822) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 69) {
                                return 0;
                            }
                        }
                    }
                    else if (diffTSC >= 12) {
                        if (diffRsCsdiff < -16) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= -16) {
                            if (RsCsDiff < 245) {
                                if (AFD < 41) {
                                    if (negBalance < 38) {
                                        if (posBalance < 85) {
                                            if (posBalance < 78) {
                                                if (Rs < 1410) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1410) {
                                                    return 0;
                                                }
                                            }
                                            else if (posBalance >= 78) {
                                                return 0;
                                            }
                                        }
                                        else if (posBalance >= 85) {
                                            if (ssDCval < 136) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 136) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 38) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 41) {
                                    if (diffMVdiffVal < 85) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 85) {
                                        if (refDCval < 57) {
                                            if (CsDiff < 7) {
                                                if (RsCsDiff < 25) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 25) {
                                                    return 0;
                                                }
                                            }
                                            else if (CsDiff >= 7) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 57) {
                                            if (diffAFD < 35) {
                                                if (Rs < 499) {
                                                    return 1;
                                                }
                                                else if (Rs >= 499) {
                                                    if (Cs < 806) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 806) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 35) {
                                                if (RsCsDiff < 11) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 11) {
                                                    if (gchDC < 3) {
                                                        if (Rs < 364) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 364) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (gchDC >= 3) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (RsCsDiff >= 245) {
                                if (posBalance < 21) {
                                    if (diffAFD < 26) {
                                        if (AFD < 33) {
                                            return 1;
                                        }
                                        else if (AFD >= 33) {
                                            return 0;
                                        }
                                    }
                                    else if (diffAFD >= 26) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 21) {
                                    if (SC < 4988) {
                                        return 1;
                                    }
                                    else if (SC >= 4988) {
                                        if (diffRsCsdiff < 496) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 496) {
                                            if (negBalance < 43) {
                                                if (negBalance < 41) {
                                                    return 1;
                                                }
                                                else if (negBalance >= 41) {
                                                    if (Cs < 2641) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 2641) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 43) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetectF(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffMVdiffVal < 65) {
        if (diffTSC < 13) {
            if (MVDiff < 55) {
                if (diffRsCsdiff < 577) {
                    if (refDCval < 24) {
                        if (diffRsCsdiff < 5) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 5) {
                            if (posBalance < 13) {
                                if (CsDiff < 4) {
                                    return 0;
                                }
                                else if (CsDiff >= 4) {
                                    return 1;
                                }
                            }
                            else if (posBalance >= 13) {
                                return 0;
                            }
                        }
                    }
                    else if (refDCval >= 24) {
                        return 0;
                    }
                }
                else if (diffRsCsdiff >= 577) {
                    if (ssDCval < 42) {
                        if (diffTSC < 1) {
                            return 0;
                        }
                        else if (diffTSC >= 1) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 42) {
                        if (Cs < 374) {
                            return 1;
                        }
                        else if (Cs >= 374) {
                            return 0;
                        }
                    }
                }
            }
            else if (MVDiff >= 55) {
                if (AFD < 52) {
                    if (diffAFD < 10) {
                        if (AFD < 28) {
                            if (diffTSC < 6) {
                                if (RsDiff < 37) {
                                    if (MVDiff < 690) {
                                        if (diffTSC < 4) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 4) {
                                            if (Rs < 17) {
                                                return 1;
                                            }
                                            else if (Rs >= 17) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 690) {
                                        if (ssDCval < 211) {
                                            if (diffTSC < -3) {
                                                return 0;
                                            }
                                            else if (diffTSC >= -3) {
                                                if (MVDiff < 1610) {
                                                    return 0;
                                                }
                                                else if (MVDiff >= 1610) {
                                                    if (Rs < 41) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 41) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 211) {
                                            if (AFD < 2) {
                                                return 0;
                                            }
                                            else if (AFD >= 2) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (RsDiff >= 37) {
                                    if (Rs < 2335) {
                                        if (SC < 1971) {
                                            return 0;
                                        }
                                        else if (SC >= 1971) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 2335) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= 6) {
                                if (MVDiff < 387) {
                                    if (MVDiff < 230) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 230) {
                                        if (MVDiff < 243) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 243) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (MVDiff >= 387) {
                                    if (TSC < 14) {
                                        return 0;
                                    }
                                    else if (TSC >= 14) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (AFD >= 28) {
                            if (ssDCval < 78) {
                                if (diffTSC < 8) {
                                    if (refDCval < 56) {
                                        if (diffAFD < -7) {
                                            if (gchDC < 11) {
                                                return 1;
                                            }
                                            else if (gchDC >= 11) {
                                                if (Cs < 45) {
                                                    return 1;
                                                }
                                                else if (Cs >= 45) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffAFD >= -7) {
                                            return 0;
                                        }
                                    }
                                    else if (refDCval >= 56) {
                                        if (RsDiff < 63) {
                                            if (diffMVdiffVal < 53) {
                                                if (Rs < 25) {
                                                    if (Rs < 14) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 14) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 25) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 53) {
                                                if (diffAFD < -4) {
                                                    if (Rs < 518) {
                                                        return 0;
                                                    }
                                                    else if (Rs >= 518) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffAFD >= -4) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 63) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffTSC >= 8) {
                                    if (negBalance < 45) {
                                        if (AFD < 42) {
                                            return 0;
                                        }
                                        else if (AFD >= 42) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 45) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 78) {
                                if (RsDiff < 112) {
                                    return 0;
                                }
                                else if (RsDiff >= 112) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 10) {
                        if (negBalance < 39) {
                            if (Cs < 2839) {
                                return 0;
                            }
                            else if (Cs >= 2839) {
                                if (Rs < 4052) {
                                    return 1;
                                }
                                else if (Rs >= 4052) {
                                    return 0;
                                }
                            }
                        }
                        else if (negBalance >= 39) {
                            if (ssDCval < 87) {
                                if (diffTSC < 5) {
                                    if (Rs < 66) {
                                        return 1;
                                    }
                                    else if (Rs >= 66) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 5) {
                                    if (gchDC < 7) {
                                        if (AFD < 23) {
                                            if (TSC < 10) {
                                                if (Rs < 609) {
                                                    return 0;
                                                }
                                                else if (Rs >= 609) {
                                                    return 1;
                                                }
                                            }
                                            else if (TSC >= 10) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 23) {
                                            if (MVDiff < 155) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 155) {
                                                if (AFD < 42) {
                                                    return 1;
                                                }
                                                else if (AFD >= 42) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (gchDC >= 7) {
                                        if (diffMVdiffVal < -178) {
                                            if (gchDC < 12) {
                                                return 0;
                                            }
                                            else if (gchDC >= 12) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= -178) {
                                            if (Rs < 3382) {
                                                return 0;
                                            }
                                            else if (Rs >= 3382) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (ssDCval >= 87) {
                                if (gchDC < 49) {
                                    if (diffMVdiffVal < 56) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 56) {
                                        if (Rs < 1432) {
                                            return 1;
                                        }
                                        else if (Rs >= 1432) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (gchDC >= 49) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (AFD >= 52) {
                    if (CsDiff < 13) {
                        if (diffMVdiffVal < 57) {
                            if (diffRsCsdiff < -49) {
                                return 1;
                            }
                            else if (diffRsCsdiff >= -49) {
                                return 0;
                            }
                        }
                        else if (diffMVdiffVal >= 57) {
                            return 1;
                        }
                    }
                    else if (CsDiff >= 13) {
                        if (Rs < 2496) {
                            return 1;
                        }
                        else if (Rs >= 2496) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (diffTSC >= 13) {
            if (RsCsDiff < 487) {
                if (gchDC < 21) {
                    if (Cs < 2165) {
                        if (negBalance < 51) {
                            if (diffTSC < 16) {
                                if (Rs < 1310) {
                                    return 0;
                                }
                                else if (Rs >= 1310) {
                                    return 1;
                                }
                            }
                            else if (diffTSC >= 16) {
                                return 1;
                            }
                        }
                        else if (negBalance >= 51) {
                            if (CsDiff < 2) {
                                return 0;
                            }
                            else if (CsDiff >= 2) {
                                if (diffAFD < 23) {
                                    if (Cs < 652) {
                                        return 1;
                                    }
                                    else if (Cs >= 652) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 23) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (Cs >= 2165) {
                        return 0;
                    }
                }
                else if (gchDC >= 21) {
                    if (refDCval < 25) {
                        if (ssDCval < 147) {
                            if (SC < 1899) {
                                return 1;
                            }
                            else if (SC >= 1899) {
                                if (SC < 2725) {
                                    if (RsDiff < 9) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 9) {
                                        return 0;
                                    }
                                }
                                else if (SC >= 2725) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 147) {
                            return 0;
                        }
                    }
                    else if (refDCval >= 25) {
                        if (Cs < 103) {
                            if (gchDC < 55) {
                                return 0;
                            }
                            else if (gchDC >= 55) {
                                if (Rs < 260) {
                                    return 1;
                                }
                                else if (Rs >= 260) {
                                    return 0;
                                }
                            }
                        }
                        else if (Cs >= 103) {
                            if (TSC < 45) {
                                if (Rs < 198) {
                                    if (Cs < 269) {
                                        return 0;
                                    }
                                    else if (Cs >= 269) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 198) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 45) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (RsCsDiff >= 487) {
                if (Cs < 3637) {
                    if (refDCval < 106) {
                        if (diffTSC < 17) {
                            if (ssDCval < 57) {
                                return 1;
                            }
                            else if (ssDCval >= 57) {
                                if (refDCval < 70) {
                                    return 0;
                                }
                                else if (refDCval >= 70) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffTSC >= 17) {
                            return 1;
                        }
                    }
                    else if (refDCval >= 106) {
                        if (refDCval < 112) {
                            return 0;
                        }
                        else if (refDCval >= 112) {
                            if (TSC < 16) {
                                return 0;
                            }
                            else if (TSC >= 16) {
                                if (MVDiff < 536) {
                                    return 1;
                                }
                                else if (MVDiff >= 536) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (Cs >= 3637) {
                    if (MVDiff < 46) {
                        if (refDCval < 33) {
                            return 1;
                        }
                        else if (refDCval >= 33) {
                            return 0;
                        }
                    }
                    else if (MVDiff >= 46) {
                        return 1;
                    }
                }
            }
        }
    }
    else if (diffMVdiffVal >= 65) {
        if (diffAFD < 13) {
            if (diffRsCsdiff < 1) {
                if (gchDC < 23) {
                    if (diffAFD < 8) {
                        if (AFD < 40) {
                            if (refDCval < 23) {
                                if (Cs < 23) {
                                    return 0;
                                }
                                else if (Cs >= 23) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 23) {
                                return 0;
                            }
                        }
                        else if (AFD >= 40) {
                            if (MVDiff < 330) {
                                if (Cs < 211) {
                                    return 0;
                                }
                                else if (Cs >= 211) {
                                    return 1;
                                }
                            }
                            else if (MVDiff >= 330) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 8) {
                        if (AFD < 16) {
                            if (posBalance < 70) {
                                return 1;
                            }
                            else if (posBalance >= 70) {
                                return 0;
                            }
                        }
                        else if (AFD >= 16) {
                            if (ssDCval < 55) {
                                if (AFD < 24) {
                                    return 0;
                                }
                                else if (AFD >= 24) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 55) {
                                return 0;
                            }
                        }
                    }
                }
                else if (gchDC >= 23) {
                    if (RsCsDiff < 189) {
                        if (ssDCval < 24) {
                            return 1;
                        }
                        else if (ssDCval >= 24) {
                            if (refDCval < 64) {
                                if (Rs < 101) {
                                    return 1;
                                }
                                else if (Rs >= 101) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 64) {
                                if (diffTSC < -44) {
                                    return 1;
                                }
                                else if (diffTSC >= -44) {
                                    if (SC < 2) {
                                        if (Cs < 1) {
                                            return 0;
                                        }
                                        else if (Cs >= 1) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 2) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (RsCsDiff >= 189) {
                        if (MVDiff < 269) {
                            if (Rs < 3761) {
                                return 0;
                            }
                            else if (Rs >= 3761) {
                                return 1;
                            }
                        }
                        else if (MVDiff >= 269) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffRsCsdiff >= 1) {
                if (refDCval < 32) {
                    if (refDCval < 25) {
                        if (ssDCval < 28) {
                            return 1;
                        }
                        else if (ssDCval >= 28) {
                            return 0;
                        }
                    }
                    else if (refDCval >= 25) {
                        return 1;
                    }
                }
                else if (refDCval >= 32) {
                    if (ssDCval < 48) {
                        if (diffTSC < 3) {
                            if (Cs < 384) {
                                if (RsDiff < 6) {
                                    return 0;
                                }
                                else if (RsDiff >= 6) {
                                    if (diffAFD < 3) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 3) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 384) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 3) {
                            if (gchDC < 8) {
                                if (SC < 975) {
                                    return 1;
                                }
                                else if (SC >= 975) {
                                    if (Rs < 466) {
                                        return 0;
                                    }
                                    else if (Rs >= 466) {
                                        return 1;
                                    }
                                }
                            }
                            else if (gchDC >= 8) {
                                if (ssDCval < 31) {
                                    return 1;
                                }
                                else if (ssDCval >= 31) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 48) {
                        if (TSC < 20) {
                            if (diffMVdiffVal < 100) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 100) {
                                if (diffRsCsdiff < 12) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 12) {
                                    if (SC < 2446) {
                                        if (diffMVdiffVal < 113) {
                                            if (CsDiff < 7) {
                                                if (MVDiff < 166) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 166) {
                                                    return 0;
                                                }
                                            }
                                            else if (CsDiff >= 7) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 113) {
                                            if (diffAFD < 11) {
                                                if (ssDCval < 136) {
                                                    if (MVDiff < 197) {
                                                        if (MVDiff < 180) {
                                                            return 0;
                                                        }
                                                        else if (MVDiff >= 180) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (MVDiff >= 197) {
                                                        return 0;
                                                    }
                                                }
                                                else if (ssDCval >= 136) {
                                                    if (SC < 1755) {
                                                        return 0;
                                                    }
                                                    else if (SC >= 1755) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 11) {
                                                if (MVDiff < 423) {
                                                    if (posBalance < 85) {
                                                        return 0;
                                                    }
                                                    else if (posBalance >= 85) {
                                                        if (Cs < 386) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 386) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 423) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 2446) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 20) {
                            if (diffRsCsdiff < 38) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 38) {
                                if (Rs < 2164) {
                                    if (RsCsDiff < 531) {
                                        if (MVDiff < 439) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 439) {
                                            return 1;
                                        }
                                    }
                                    else if (RsCsDiff >= 531) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 2164) {
                                    if (Rs < 6613) {
                                        return 0;
                                    }
                                    else if (Rs >= 6613) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 13) {
            if (negBalance < 30) {
                if (diffRsCsdiff < 141) {
                    if (diffMVdiffVal < 99) {
                        if (posBalance < 84) {
                            if (SC < 2080) {
                                if (MVDiff < 85) {
                                    return 1;
                                }
                                else if (MVDiff >= 85) {
                                    return 0;
                                }
                            }
                            else if (SC >= 2080) {
                                return 1;
                            }
                        }
                        else if (posBalance >= 84) {
                            return 0;
                        }
                    }
                    else if (diffMVdiffVal >= 99) {
                        if (TSC < 12) {
                            if (Rs < 311) {
                                return 1;
                            }
                            else if (Rs >= 311) {
                                return 0;
                            }
                        }
                        else if (TSC >= 12) {
                            if (gchDC < 76) {
                                if (diffAFD < 28) {
                                    if (AFD < 25) {
                                        if (refDCval < 37) {
                                            if (Rs < 183) {
                                                return 0;
                                            }
                                            else if (Rs >= 183) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 37) {
                                            if (diffAFD < 18) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 18) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (AFD >= 25) {
                                        if (ssDCval < 77) {
                                            if (diffRsCsdiff < 32) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 32) {
                                                if (Cs < 172) {
                                                    return 0;
                                                }
                                                else if (Cs >= 172) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 77) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffAFD >= 28) {
                                    if (negBalance < 12) {
                                        if (gchDC < 62) {
                                            return 0;
                                        }
                                        else if (gchDC >= 62) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 12) {
                                        if (MVDiff < 700) {
                                            if (TSC < 39) {
                                                return 1;
                                            }
                                            else if (TSC >= 39) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 700) {
                                            if (CsDiff < 6) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 6) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (gchDC >= 76) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 141) {
                    if (posBalance < 86) {
                        return 1;
                    }
                    else if (posBalance >= 86) {
                        if (Rs < 1069) {
                            if (diffTSC < 10) {
                                if (AFD < 31) {
                                    return 1;
                                }
                                else if (AFD >= 31) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= 10) {
                                return 1;
                            }
                        }
                        else if (Rs >= 1069) {
                            if (negBalance < 3) {
                                return 0;
                            }
                            else if (negBalance >= 3) {
                                if (diffRsCsdiff < 999) {
                                    if (diffTSC < 16) {
                                        if (Cs < 353) {
                                            return 1;
                                        }
                                        else if (Cs >= 353) {
                                            if (refDCval < 49) {
                                                if (Rs < 3011) {
                                                    return 1;
                                                }
                                                else if (Rs >= 3011) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 49) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 16) {
                                        if (RsCsDiff < 207) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 207) {
                                            if (TSC < 28) {
                                                return 1;
                                            }
                                            else if (TSC >= 28) {
                                                if (Rs < 1848) {
                                                    if (MVDiff < 168) {
                                                        return 0;
                                                    }
                                                    else if (MVDiff >= 168) {
                                                        return 1;
                                                    }
                                                }
                                                else if (Rs >= 1848) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 999) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (negBalance >= 30) {
                if (diffTSC < 10) {
                    if (TSC < 25) {
                        if (ssDCval < 78) {
                            if (refDCval < 65) {
                                if (diffMVdiffVal < 132) {
                                    if (diffRsCsdiff < 98) {
                                        if (AFD < 26) {
                                            if (TSC < 6) {
                                                return 0;
                                            }
                                            else if (TSC >= 6) {
                                                if (SC < 1186) {
                                                    return 1;
                                                }
                                                else if (SC >= 1186) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (AFD >= 26) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= 98) {
                                        return 1;
                                    }
                                }
                                else if (diffMVdiffVal >= 132) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 65) {
                                if (diffTSC < 9) {
                                    if (refDCval < 72) {
                                        if (Rs < 583) {
                                            if (MVDiff < 1539) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 1539) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 583) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 72) {
                                        if (SC < 2395) {
                                            if (diffRsCsdiff < 43) {
                                                if (RsCsDiff < 39) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 39) {
                                                    if (MVDiff < 257) {
                                                        return 1;
                                                    }
                                                    else if (MVDiff >= 257) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (diffRsCsdiff >= 43) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 2395) {
                                            if (CsDiff < 13) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 13) {
                                                if (Rs < 1314) {
                                                    return 0;
                                                }
                                                else if (Rs >= 1314) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (diffTSC >= 9) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 78) {
                            if (diffMVdiffVal < 207) {
                                if (Scindex < 4) {
                                    if (MVDiff < 272) {
                                        if (Cs < 282) {
                                            if (Rs < 205) {
                                                return 1;
                                            }
                                            else if (Rs >= 205) {
                                                return 0;
                                            }
                                        }
                                        else if (Cs >= 282) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 272) {
                                        return 0;
                                    }
                                }
                                else if (Scindex >= 4) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 207) {
                                if (refDCval < 108) {
                                    if (CsDiff < 4) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 4) {
                                        if (AFD < 53) {
                                            if (Cs < 1892) {
                                                return 1;
                                            }
                                            else if (Cs >= 1892) {
                                                return 0;
                                            }
                                        }
                                        else if (AFD >= 53) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (refDCval >= 108) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (TSC >= 25) {
                        if (gchDC < 4) {
                            return 1;
                        }
                        else if (gchDC >= 4) {
                            return 0;
                        }
                    }
                }
                else if (diffTSC >= 10) {
                    if (posBalance < 14) {
                        if (diffRsCsdiff < 31) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 31) {
                            if (SC < 5050) {
                                if (TSC < 32) {
                                    return 1;
                                }
                                else if (TSC >= 32) {
                                    if (SC < 1401) {
                                        return 1;
                                    }
                                    else if (SC >= 1401) {
                                        return 0;
                                    }
                                }
                            }
                            else if (SC >= 5050) {
                                if (AFD < 45) {
                                    return 0;
                                }
                                else if (AFD >= 45) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (posBalance >= 14) {
                        if (diffRsCsdiff < 239) {
                            if (TSC < 24) {
                                if (diffTSC < 12) {
                                    if (diffMVdiffVal < 134) {
                                        if (AFD < 34) {
                                            if (Cs < 710) {
                                                return 1;
                                            }
                                            else if (Cs >= 710) {
                                                return 0;
                                            }
                                        }
                                        else if (AFD >= 34) {
                                            return 0;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 134) {
                                        return 1;
                                    }
                                }
                                else if (diffTSC >= 12) {
                                    if (RsCsDiff < 181) {
                                        return 1;
                                    }
                                    else if (RsCsDiff >= 181) {
                                        if (refDCval < 152) {
                                            if (diffMVdiffVal < 78) {
                                                if (Rs < 892) {
                                                    return 0;
                                                }
                                                else if (Rs >= 892) {
                                                    return 1;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 78) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 152) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 24) {
                                if (diffAFD < 42) {
                                    if (AFD < 55) {
                                        if (negBalance < 46) {
                                            if (AFD < 44) {
                                                if (Rs < 1413) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1413) {
                                                    return 0;
                                                }
                                            }
                                            else if (AFD >= 44) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 46) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 55) {
                                        if (RsCsDiff < 222) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 222) {
                                            if (SC < 2757) {
                                                return 1;
                                            }
                                            else if (SC >= 2757) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 42) {
                                    if (TSC < 59) {
                                        return 1;
                                    }
                                    else if (TSC >= 59) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 239) {
                            if (diffTSC < 14) {
                                if (diffAFD < 24) {
                                    if (SC < 3709) {
                                        return 1;
                                    }
                                    else if (SC >= 3709) {
                                        if (SC < 6538) {
                                            return 0;
                                        }
                                        else if (SC >= 6538) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= 24) {
                                    return 1;
                                }
                            }
                            else if (diffTSC >= 14) {
                                if (Rs < 8722) {
                                    return 1;
                                }
                                else if (Rs >= 8722) {
                                    if (refDCval < 156) {
                                        return 1;
                                    }
                                    else if (refDCval >= 156) {
                                        if (TSC < 34) {
                                            return 0;
                                        }
                                        else if (TSC >= 34) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect10(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 13) {
        if (diffMVdiffVal < 54) {
            if (MVDiff < 233) {
                if (refDCval < 24) {
                    if (diffAFD < 3) {
                        if (ssDCval < 130) {
                            if (RsCsDiff < 7) {
                                return 0;
                            }
                            else if (RsCsDiff >= 7) {
                                if (posBalance < 4) {
                                    if (SC < 1400) {
                                        if (MVDiff < 17) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 17) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 1400) {
                                        if (RsDiff < 4) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 4) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (posBalance >= 4) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 130) {
                            return 1;
                        }
                    }
                    else if (diffAFD >= 3) {
                        if (Rs < 52) {
                            return 0;
                        }
                        else if (Rs >= 52) {
                            return 1;
                        }
                    }
                }
                else if (refDCval >= 24) {
                    if (gchDC < 7) {
                        if (MVDiff < 60) {
                            return 0;
                        }
                        else if (MVDiff >= 60) {
                            if (TSC < 8) {
                                return 0;
                            }
                            else if (TSC >= 8) {
                                if (ssDCval < 39) {
                                    if (TSC < 10) {
                                        return 1;
                                    }
                                    else if (TSC >= 10) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 39) {
                                    if (RsCsDiff < 4482) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 4482) {
                                        if (Rs < 3227) {
                                            return 1;
                                        }
                                        else if (Rs >= 3227) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (gchDC >= 7) {
                        if (RsDiff < 53) {
                            if (refDCval < 228) {
                                if (RsDiff < 30) {
                                    return 0;
                                }
                                else if (RsDiff >= 30) {
                                    if (diffMVdiffVal < 33) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 33) {
                                        if (Rs < 3191) {
                                            return 0;
                                        }
                                        else if (Rs >= 3191) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (refDCval >= 228) {
                                if (Cs < 73) {
                                    return 1;
                                }
                                else if (Cs >= 73) {
                                    return 0;
                                }
                            }
                        }
                        else if (RsDiff >= 53) {
                            if (Rs < 1351) {
                                return 0;
                            }
                            else if (Rs >= 1351) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (MVDiff >= 233) {
                if (CsDiff < 15) {
                    if (diffAFD < -23) {
                        return 0;
                    }
                    else if (diffAFD >= -23) {
                        if (diffTSC < 6) {
                            if (MVDiff < 1661) {
                                if (AFD < 28) {
                                    if (negBalance < 122) {
                                        if (diffRsCsdiff < 16) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 16) {
                                            if (TSC < 3) {
                                                return 1;
                                            }
                                            else if (TSC >= 3) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (negBalance >= 122) {
                                        if (ssDCval < 219) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 219) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (AFD >= 28) {
                                    if (ssDCval < 62) {
                                        if (gchDC < 21) {
                                            if (TSC < 10) {
                                                return 0;
                                            }
                                            else if (TSC >= 10) {
                                                if (Rs < 1505) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1505) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 21) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 62) {
                                        return 0;
                                    }
                                }
                            }
                            else if (MVDiff >= 1661) {
                                if (posBalance < 34) {
                                    return 0;
                                }
                                else if (posBalance >= 34) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffTSC >= 6) {
                            if (ssDCval < 97) {
                                if (TSC < 15) {
                                    return 1;
                                }
                                else if (TSC >= 15) {
                                    if (ssDCval < 71) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 71) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 97) {
                                return 0;
                            }
                        }
                    }
                }
                else if (CsDiff >= 15) {
                    if (SC < 4381) {
                        if (diffRsCsdiff < -21) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= -21) {
                            if (Rs < 2125) {
                                return 1;
                            }
                            else if (Rs >= 2125) {
                                return 0;
                            }
                        }
                    }
                    else if (SC >= 4381) {
                        return 0;
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 54) {
            if (diffAFD < 8) {
                if (TSC < 33) {
                    if (ssDCval < 39) {
                        if (CsDiff < 5) {
                            if (gchDC < 65) {
                                if (AFD < 10) {
                                    return 0;
                                }
                                else if (AFD >= 10) {
                                    if (gchDC < 6) {
                                        if (MVDiff < 174) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 174) {
                                            if (Rs < 3) {
                                                return 0;
                                            }
                                            else if (Rs >= 3) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (gchDC >= 6) {
                                        return 0;
                                    }
                                }
                            }
                            else if (gchDC >= 65) {
                                if (Rs < 5) {
                                    return 1;
                                }
                                else if (Rs >= 5) {
                                    return 0;
                                }
                            }
                        }
                        else if (CsDiff >= 5) {
                            if (diffRsCsdiff < 32) {
                                if (diffRsCsdiff < -64) {
                                    if (TSC < 12) {
                                        return 1;
                                    }
                                    else if (TSC >= 12) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= -64) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 32) {
                                if (diffAFD < 6) {
                                    return 1;
                                }
                                else if (diffAFD >= 6) {
                                    if (Rs < 348) {
                                        return 1;
                                    }
                                    else if (Rs >= 348) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 39) {
                        if (AFD < 34) {
                            if (posBalance < 95) {
                                if (diffTSC < 4) {
                                    return 0;
                                }
                                else if (diffTSC >= 4) {
                                    if (TSC < 7) {
                                        return 1;
                                    }
                                    else if (TSC >= 7) {
                                        return 0;
                                    }
                                }
                            }
                            else if (posBalance >= 95) {
                                if (negBalance < 28) {
                                    if (gchDC < 132) {
                                        return 0;
                                    }
                                    else if (gchDC >= 132) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 28) {
                                    return 1;
                                }
                            }
                        }
                        else if (AFD >= 34) {
                            if (diffAFD < -4) {
                                if (MVDiff < 304) {
                                    if (posBalance < 46) {
                                        return 0;
                                    }
                                    else if (posBalance >= 46) {
                                        if (ssDCval < 174) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 174) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (MVDiff >= 304) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= -4) {
                                if (diffTSC < 5) {
                                    return 0;
                                }
                                else if (diffTSC >= 5) {
                                    if (diffAFD < -1) {
                                        return 1;
                                    }
                                    else if (diffAFD >= -1) {
                                        if (Rs < 312) {
                                            if (CsDiff < 4) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 4) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 312) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (TSC >= 33) {
                    if (SC < 2198) {
                        return 0;
                    }
                    else if (SC >= 2198) {
                        return 1;
                    }
                }
            }
            else if (diffAFD >= 8) {
                if (refDCval < 48) {
                    if (ssDCval < 50) {
                        if (refDCval < 31) {
                            if (Rs < 260) {
                                return 1;
                            }
                            else if (Rs >= 260) {
                                if (AFD < 11) {
                                    return 0;
                                }
                                else if (AFD >= 11) {
                                    return 1;
                                }
                            }
                        }
                        else if (refDCval >= 31) {
                            if (refDCval < 35) {
                                if (ssDCval < 38) {
                                    if (RsDiff < 5) {
                                        if (AFD < 13) {
                                            return 1;
                                        }
                                        else if (AFD >= 13) {
                                            return 0;
                                        }
                                    }
                                    else if (RsDiff >= 5) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 38) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 35) {
                                if (Cs < 180) {
                                    return 1;
                                }
                                else if (Cs >= 180) {
                                    if (diffTSC < 4) {
                                        if (Rs < 828) {
                                            return 0;
                                        }
                                        else if (Rs >= 828) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= 4) {
                                        if (Rs < 185) {
                                            return 0;
                                        }
                                        else if (Rs >= 185) {
                                            if (AFD < 26) {
                                                return 1;
                                            }
                                            else if (AFD >= 26) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 50) {
                        if (negBalance < 35) {
                            if (Cs < 107) {
                                return 1;
                            }
                            else if (Cs >= 107) {
                                return 0;
                            }
                        }
                        else if (negBalance >= 35) {
                            return 1;
                        }
                    }
                }
                else if (refDCval >= 48) {
                    if (diffTSC < 16) {
                        if (MVDiff < 574) {
                            if (negBalance < 70) {
                                if (diffTSC < 9) {
                                    return 0;
                                }
                                else if (diffTSC >= 9) {
                                    if (CsDiff < 9) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 9) {
                                        if (Rs < 2041) {
                                            return 1;
                                        }
                                        else if (Rs >= 2041) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 70) {
                                if (posBalance < 42) {
                                    if (diffMVdiffVal < 217) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 217) {
                                        if (MVDiff < 317) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 317) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (posBalance >= 42) {
                                    if (ssDCval < 79) {
                                        if (CsDiff < 5) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 5) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 79) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 574) {
                            if (AFD < 28) {
                                return 1;
                            }
                            else if (AFD >= 28) {
                                if (Cs < 856) {
                                    return 0;
                                }
                                else if (Cs >= 856) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 16) {
                        return 1;
                    }
                }
            }
        }
    }
    else if (diffAFD >= 13) {
        if (diffTSC < 14) {
            if (MVDiff < 56) {
                return 0;
            }
            else if (MVDiff >= 56) {
                if (diffMVdiffVal < 80) {
                    if (diffRsCsdiff < 122) {
                        if (posBalance < 59) {
                            if (Rs < 619) {
                                if (MVDiff < 216) {
                                    if (MVDiff < 130) {
                                        if (refDCval < 46) {
                                            if (posBalance < 47) {
                                                return 1;
                                            }
                                            else if (posBalance >= 47) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 46) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 130) {
                                        if (CsDiff < 8) {
                                            if (negBalance < 59) {
                                                return 0;
                                            }
                                            else if (negBalance >= 59) {
                                                return 1;
                                            }
                                        }
                                        else if (CsDiff >= 8) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (MVDiff >= 216) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 619) {
                                return 0;
                            }
                        }
                        else if (posBalance >= 59) {
                            if (negBalance < 58) {
                                return 0;
                            }
                            else if (negBalance >= 58) {
                                if (TSC < 12) {
                                    return 0;
                                }
                                else if (TSC >= 12) {
                                    if (Rs < 204) {
                                        return 0;
                                    }
                                    else if (Rs >= 204) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 122) {
                        if (diffTSC < 7) {
                            if (posBalance < 23) {
                                return 1;
                            }
                            else if (posBalance >= 23) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 7) {
                            if (Rs < 711) {
                                return 1;
                            }
                            else if (Rs >= 711) {
                                if (Cs < 1262) {
                                    if (SC < 1634) {
                                        return 0;
                                    }
                                    else if (SC >= 1634) {
                                        if (gchDC < 19) {
                                            if (TSC < 11) {
                                                return 0;
                                            }
                                            else if (TSC >= 11) {
                                                return 1;
                                            }
                                        }
                                        else if (gchDC >= 19) {
                                            if (Rs < 2965) {
                                                return 0;
                                            }
                                            else if (Rs >= 2965) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 1262) {
                                    if (TSC < 22) {
                                        return 0;
                                    }
                                    else if (TSC >= 22) {
                                        if (posBalance < 58) {
                                            return 1;
                                        }
                                        else if (posBalance >= 58) {
                                            if (Rs < 1819) {
                                                return 1;
                                            }
                                            else if (Rs >= 1819) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 80) {
                    if (posBalance < 75) {
                        if (ssDCval < 78) {
                            if (diffAFD < 19) {
                                if (refDCval < 53) {
                                    if (negBalance < 50) {
                                        if (Rs < 599) {
                                            if (TSC < 6) {
                                                return 0;
                                            }
                                            else if (TSC >= 6) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 599) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 50) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 53) {
                                    if (SC < 3058) {
                                        if (gchDC < 26) {
                                            if (TSC < 15) {
                                                if (ssDCval < 39) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 39) {
                                                    return 0;
                                                }
                                            }
                                            else if (TSC >= 15) {
                                                if (TSC < 21) {
                                                    return 1;
                                                }
                                                else if (TSC >= 21) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (gchDC >= 26) {
                                            if (ssDCval < 55) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 55) {
                                                if (Rs < 48) {
                                                    return 0;
                                                }
                                                else if (Rs >= 48) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (SC >= 3058) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffAFD >= 19) {
                                if (gchDC < 84) {
                                    if (Cs < 1529) {
                                        if (negBalance < 41) {
                                            if (Rs < 223) {
                                                return 1;
                                            }
                                            else if (Rs >= 223) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 41) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 1529) {
                                        if (CsDiff < 20) {
                                            if (MVDiff < 210) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 210) {
                                                return 0;
                                            }
                                        }
                                        else if (CsDiff >= 20) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (gchDC >= 84) {
                                    if (Rs < 20) {
                                        return 0;
                                    }
                                    else if (Rs >= 20) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 78) {
                            if (diffAFD < 24) {
                                if (CsDiff < 4) {
                                    return 0;
                                }
                                else if (CsDiff >= 4) {
                                    if (diffTSC < 9) {
                                        if (Rs < 211) {
                                            return 1;
                                        }
                                        else if (Rs >= 211) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 9) {
                                        if (gchDC < 10) {
                                            return 1;
                                        }
                                        else if (gchDC >= 10) {
                                            if (SC < 1697) {
                                                return 1;
                                            }
                                            else if (SC >= 1697) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 24) {
                                if (AFD < 69) {
                                    if (gchDC < 36) {
                                        return 1;
                                    }
                                    else if (gchDC >= 36) {
                                        if (Cs < 591) {
                                            return 1;
                                        }
                                        else if (Cs >= 591) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (AFD >= 69) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (posBalance >= 75) {
                        if (TSC < 11) {
                            if (Cs < 891) {
                                return 1;
                            }
                            else if (Cs >= 891) {
                                return 0;
                            }
                        }
                        else if (TSC >= 11) {
                            if (diffTSC < 9) {
                                if (negBalance < 42) {
                                    if (SC < 6923) {
                                        return 0;
                                    }
                                    else if (SC >= 6923) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 42) {
                                    if (diffRsCsdiff < 13) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 13) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffTSC >= 9) {
                                if (negBalance < 35) {
                                    if (TSC < 16) {
                                        if (negBalance < 7) {
                                            if (refDCval < 84) {
                                                if (SC < 793) {
                                                    if (MVDiff < 119) {
                                                        return 0;
                                                    }
                                                    else if (MVDiff >= 119) {
                                                        return 1;
                                                    }
                                                }
                                                else if (SC >= 793) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 84) {
                                                return 0;
                                            }
                                        }
                                        else if (negBalance >= 7) {
                                            if (diffAFD < 15) {
                                                if (Rs < 159) {
                                                    return 1;
                                                }
                                                else if (Rs >= 159) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffAFD >= 15) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (TSC >= 16) {
                                        if (ssDCval < 137) {
                                            if (MVDiff < 174) {
                                                if (gchDC < 22) {
                                                    return 0;
                                                }
                                                else if (gchDC >= 22) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 174) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 137) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (negBalance >= 35) {
                                    if (TSC < 24) {
                                        return 1;
                                    }
                                    else if (TSC >= 24) {
                                        if (Rs < 1316) {
                                            return 1;
                                        }
                                        else if (Rs >= 1316) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffTSC >= 14) {
            if (negBalance < 26) {
                if (CsDiff < 13) {
                    if (posBalance < 85) {
                        if (diffMVdiffVal < 46) {
                            return 0;
                        }
                        else if (diffMVdiffVal >= 46) {
                            return 1;
                        }
                    }
                    else if (posBalance >= 85) {
                        if (diffMVdiffVal < 142) {
                            if (Rs < 116) {
                                if (TSC < 18) {
                                    return 0;
                                }
                                else if (TSC >= 18) {
                                    return 1;
                                }
                            }
                            else if (Rs >= 116) {
                                if (refDCval < 18) {
                                    if (Rs < 1284) {
                                        return 1;
                                    }
                                    else if (Rs >= 1284) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 18) {
                                    if (diffRsCsdiff < 350) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 350) {
                                        if (negBalance < 21) {
                                            return 0;
                                        }
                                        else if (negBalance >= 21) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 142) {
                            if (Cs < 809) {
                                if (CsDiff < 3) {
                                    if (TSC < 19) {
                                        return 1;
                                    }
                                    else if (TSC >= 19) {
                                        return 0;
                                    }
                                }
                                else if (CsDiff >= 3) {
                                    if (TSC < 25) {
                                        if (negBalance < 10) {
                                            if (RsDiff < 5) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 5) {
                                                return 1;
                                            }
                                        }
                                        else if (negBalance >= 10) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 25) {
                                        if (AFD < 51) {
                                            if (SC < 691) {
                                                return 1;
                                            }
                                            else if (SC >= 691) {
                                                return 0;
                                            }
                                        }
                                        else if (AFD >= 51) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (Cs >= 809) {
                                if (diffAFD < 53) {
                                    if (refDCval < 33) {
                                        return 1;
                                    }
                                    else if (refDCval >= 33) {
                                        return 0;
                                    }
                                }
                                else if (diffAFD >= 53) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (CsDiff >= 13) {
                    if (Cs < 5227) {
                        if (diffMVdiffVal < 293) {
                            return 1;
                        }
                        else if (diffMVdiffVal >= 293) {
                            if (ssDCval < 102) {
                                return 0;
                            }
                            else if (ssDCval >= 102) {
                                if (posBalance < 126) {
                                    return 1;
                                }
                                else if (posBalance >= 126) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (Cs >= 5227) {
                        if (Rs < 3183) {
                            return 1;
                        }
                        else if (Rs >= 3183) {
                            return 0;
                        }
                    }
                }
            }
            else if (negBalance >= 26) {
                if (diffAFD < 26) {
                    if (ssDCval < 82) {
                        if (gchDC < 19) {
                            if (diffMVdiffVal < 49) {
                                if (Rs < 689) {
                                    return 0;
                                }
                                else if (Rs >= 689) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 49) {
                                return 1;
                            }
                        }
                        else if (gchDC >= 19) {
                            if (RsCsDiff < 73) {
                                return 0;
                            }
                            else if (RsCsDiff >= 73) {
                                if (diffTSC < 16) {
                                    if (diffTSC < 15) {
                                        if (RsDiff < 7) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 7) {
                                            return 1;
                                        }
                                    }
                                    else if (diffTSC >= 15) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 16) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 82) {
                        if (diffMVdiffVal < 69) {
                            if (SC < 17052) {
                                if (TSC < 52) {
                                    if (CsDiff < 24) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 24) {
                                        if (Cs < 4506) {
                                            return 1;
                                        }
                                        else if (Cs >= 4506) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (TSC >= 52) {
                                    return 1;
                                }
                            }
                            else if (SC >= 17052) {
                                return 1;
                            }
                        }
                        else if (diffMVdiffVal >= 69) {
                            if (posBalance < 17) {
                                return 0;
                            }
                            else if (posBalance >= 17) {
                                if (gchDC < 19) {
                                    return 1;
                                }
                                else if (gchDC >= 19) {
                                    if (gchDC < 22) {
                                        return 0;
                                    }
                                    else if (gchDC >= 22) {
                                        if (MVDiff < 420) {
                                            if (CsDiff < 9) {
                                                if (Cs < 404) {
                                                    return 1;
                                                }
                                                else if (Cs >= 404) {
                                                    return 0;
                                                }
                                            }
                                            else if (CsDiff >= 9) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 420) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 26) {
                    if (refDCval < 124) {
                        if (diffRsCsdiff < 44) {
                            if (diffMVdiffVal < 96) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 96) {
                                if (ssDCval < 102) {
                                    if (ssDCval < 36) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 36) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 102) {
                                    if (refDCval < 99) {
                                        return 0;
                                    }
                                    else if (refDCval >= 99) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 44) {
                            if (diffAFD < 43) {
                                if (diffMVdiffVal < 77) {
                                    if (RsDiff < 16) {
                                        if (RsCsDiff < 155) {
                                            return 1;
                                        }
                                        else if (RsCsDiff >= 155) {
                                            if (TSC < 32) {
                                                return 0;
                                            }
                                            else if (TSC >= 32) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 16) {
                                        if (diffAFD < 42) {
                                            return 1;
                                        }
                                        else if (diffAFD >= 42) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffMVdiffVal >= 77) {
                                    if (MVDiff < 377) {
                                        if (CsDiff < 26) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 26) {
                                            if (diffTSC < 15) {
                                                if (Rs < 1300) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1300) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffTSC >= 15) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 377) {
                                        if (negBalance < 47) {
                                            if (posBalance < 76) {
                                                return 0;
                                            }
                                            else if (posBalance >= 76) {
                                                if (AFD < 55) {
                                                    return 1;
                                                }
                                                else if (AFD >= 55) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 47) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 43) {
                                return 1;
                            }
                        }
                    }
                    else if (refDCval >= 124) {
                        if (diffMVdiffVal < 56) {
                            if (Cs < 2318) {
                                if (Cs < 1265) {
                                    if (RsDiff < 3) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 3) {
                                        return 0;
                                    }
                                }
                                else if (Cs >= 1265) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 2318) {
                                return 0;
                            }
                        }
                        else if (diffMVdiffVal >= 56) {
                            if (diffRsCsdiff < 1) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 1) {
                                if (Rs < 9642) {
                                    if (RsDiff < 10) {
                                        if (diffMVdiffVal < 84) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 84) {
                                            if (MVDiff < 1201) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 1201) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 10) {
                                        return 1;
                                    }
                                }
                                else if (Rs >= 9642) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect11(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffTSC < 7) {
        if (diffMVdiffVal < 66) {
            if (AFD < 30) {
                if (refDCval < 24) {
                    if (CsDiff < 12) {
                        if (ssDCval < 34) {
                            if (diffRsCsdiff < 5) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 5) {
                                if (ssDCval < 18) {
                                    if (diffMVdiffVal < 3) {
                                        return 1;
                                    }
                                    else if (diffMVdiffVal >= 3) {
                                        return 0;
                                    }
                                }
                                else if (ssDCval >= 18) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 34) {
                            return 1;
                        }
                    }
                    else if (CsDiff >= 12) {
                        return 1;
                    }
                }
                else if (refDCval >= 24) {
                    if (MVDiff < 97) {
                        if (refDCval < 228) {
                            if (diffRsCsdiff < 248) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 248) {
                                if (SC < 3906) {
                                    if (SC < 3900) {
                                        if (diffRsCsdiff < 249) {
                                            if (TSC < 9) {
                                                return 1;
                                            }
                                            else if (TSC >= 9) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 249) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 3900) {
                                        return 1;
                                    }
                                }
                                else if (SC >= 3906) {
                                    return 0;
                                }
                            }
                        }
                        else if (refDCval >= 228) {
                            if (ssDCval < 224) {
                                return 1;
                            }
                            else if (ssDCval >= 224) {
                                return 0;
                            }
                        }
                    }
                    else if (MVDiff >= 97) {
                        if (diffAFD < 8) {
                            if (diffRsCsdiff < 17) {
                                if (RsDiff < 62) {
                                    if (negBalance < 123) {
                                        return 0;
                                    }
                                    else if (negBalance >= 123) {
                                        if (negBalance < 124) {
                                            return 1;
                                        }
                                        else if (negBalance >= 124) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (RsDiff >= 62) {
                                    return 1;
                                }
                            }
                            else if (diffRsCsdiff >= 17) {
                                if (Rs < 121) {
                                    return 1;
                                }
                                else if (Rs >= 121) {
                                    if (TSC < 17) {
                                        return 0;
                                    }
                                    else if (TSC >= 17) {
                                        if (ssDCval < 49) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 49) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 8) {
                            if (ssDCval < 50) {
                                if (TSC < 10) {
                                    if (ssDCval < 38) {
                                        if (diffAFD < 15) {
                                            return 1;
                                        }
                                        else if (diffAFD >= 15) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 38) {
                                        return 0;
                                    }
                                }
                                else if (TSC >= 10) {
                                    if (CsDiff < 4) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 4) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 50) {
                                if (gchDC < 41) {
                                    return 0;
                                }
                                else if (gchDC >= 41) {
                                    if (SC < 72) {
                                        return 0;
                                    }
                                    else if (SC >= 72) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (AFD >= 30) {
                if (TSC < 24) {
                    if (MVDiff < 182) {
                        return 0;
                    }
                    else if (MVDiff >= 182) {
                        if (ssDCval < 57) {
                            if (diffTSC < -7) {
                                return 0;
                            }
                            else if (diffTSC >= -7) {
                                if (diffMVdiffVal < -95) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= -95) {
                                    if (diffRsCsdiff < -2) {
                                        if (Rs < 952) {
                                            return 1;
                                        }
                                        else if (Rs >= 952) {
                                            return 0;
                                        }
                                    }
                                    else if (diffRsCsdiff >= -2) {
                                        if (AFD < 44) {
                                            return 0;
                                        }
                                        else if (AFD >= 44) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 57) {
                            if (MVDiff < 183) {
                                return 1;
                            }
                            else if (MVDiff >= 183) {
                                if (AFD < 47) {
                                    return 0;
                                }
                                else if (AFD >= 47) {
                                    if (diffRsCsdiff < -34) {
                                        return 1;
                                    }
                                    else if (diffRsCsdiff >= -34) {
                                        if (Rs < 1193) {
                                            return 0;
                                        }
                                        else if (Rs >= 1193) {
                                            if (Rs < 2486) {
                                                return 1;
                                            }
                                            else if (Rs >= 2486) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (TSC >= 24) {
                    if (diffTSC < -2) {
                        if (ssDCval < 113) {
                            return 1;
                        }
                        else if (ssDCval >= 113) {
                            return 0;
                        }
                    }
                    else if (diffTSC >= -2) {
                        if (ssDCval < 91) {
                            if (diffAFD < 3) {
                                return 1;
                            }
                            else if (diffAFD >= 3) {
                                if (ssDCval < 86) {
                                    return 0;
                                }
                                else if (ssDCval >= 86) {
                                    if (Cs < 1449) {
                                        return 0;
                                    }
                                    else if (Cs >= 1449) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 91) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 66) {
            if (diffTSC < 3) {
                if (diffRsCsdiff < 12) {
                    if (gchDC < 105) {
                        if (ssDCval < 41) {
                            if (diffTSC < 1) {
                                if (CsDiff < 7) {
                                    return 0;
                                }
                                else if (CsDiff >= 7) {
                                    if (Rs < 587) {
                                        return 1;
                                    }
                                    else if (Rs >= 587) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= 1) {
                                if (ssDCval < 25) {
                                    return 1;
                                }
                                else if (ssDCval >= 25) {
                                    if (diffMVdiffVal < 173) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 173) {
                                        if (Cs < 108) {
                                            if (ssDCval < 36) {
                                                if (Rs < 62) {
                                                    return 0;
                                                }
                                                else if (Rs >= 62) {
                                                    return 1;
                                                }
                                            }
                                            else if (ssDCval >= 36) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 108) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 41) {
                            if (diffMVdiffVal < 70) {
                                if (AFD < 35) {
                                    return 0;
                                }
                                else if (AFD >= 35) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 70) {
                                return 0;
                            }
                        }
                    }
                    else if (gchDC >= 105) {
                        if (TSC < 8) {
                            return 1;
                        }
                        else if (TSC >= 8) {
                            return 0;
                        }
                    }
                }
                else if (diffRsCsdiff >= 12) {
                    if (Rs < 498) {
                        if (SC < 207) {
                            if (RsDiff < 6) {
                                return 0;
                            }
                            else if (RsDiff >= 6) {
                                return 1;
                            }
                        }
                        else if (SC >= 207) {
                            if (posBalance < 55) {
                                if (negBalance < 98) {
                                    if (posBalance < 49) {
                                        if (gchDC < 28) {
                                            if (AFD < 3) {
                                                return 1;
                                            }
                                            else if (AFD >= 3) {
                                                return 0;
                                            }
                                        }
                                        else if (gchDC >= 28) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 49) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 98) {
                                    return 0;
                                }
                            }
                            else if (posBalance >= 55) {
                                if (RsDiff < 3) {
                                    if (gchDC < 5) {
                                        return 0;
                                    }
                                    else if (gchDC >= 5) {
                                        return 1;
                                    }
                                }
                                else if (RsDiff >= 3) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (Rs >= 498) {
                        if (RsCsDiff < 1855) {
                            if (diffAFD < 22) {
                                if (ssDCval < 32) {
                                    return 1;
                                }
                                else if (ssDCval >= 32) {
                                    if (diffTSC < 2) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 2) {
                                        if (posBalance < 50) {
                                            if (refDCval < 75) {
                                                return 1;
                                            }
                                            else if (refDCval >= 75) {
                                                return 0;
                                            }
                                        }
                                        else if (posBalance >= 50) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 22) {
                                return 1;
                            }
                        }
                        else if (RsCsDiff >= 1855) {
                            if (Cs < 2759) {
                                return 1;
                            }
                            else if (Cs >= 2759) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 3) {
                if (Rs < 301) {
                    if (ssDCval < 53) {
                        if (diffRsCsdiff < 4) {
                            if (diffTSC < 6) {
                                if (posBalance < 23) {
                                    if (SC < 19) {
                                        if (TSC < 4) {
                                            return 0;
                                        }
                                        else if (TSC >= 4) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 19) {
                                        return 0;
                                    }
                                }
                                else if (posBalance >= 23) {
                                    if (SC < 277) {
                                        if (Cs < 18) {
                                            if (AFD < 15) {
                                                if (ssDCval < 27) {
                                                    return 1;
                                                }
                                                else if (ssDCval >= 27) {
                                                    return 0;
                                                }
                                            }
                                            else if (AFD >= 15) {
                                                return 0;
                                            }
                                        }
                                        else if (Cs >= 18) {
                                            if (MVDiff < 233) {
                                                if (MVDiff < 230) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 230) {
                                                    return 0;
                                                }
                                            }
                                            else if (MVDiff >= 233) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 277) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffTSC >= 6) {
                                return 1;
                            }
                        }
                        else if (diffRsCsdiff >= 4) {
                            return 1;
                        }
                    }
                    else if (ssDCval >= 53) {
                        if (CsDiff < 4) {
                            if (negBalance < 84) {
                                return 0;
                            }
                            else if (negBalance >= 84) {
                                if (MVDiff < 480) {
                                    return 1;
                                }
                                else if (MVDiff >= 480) {
                                    return 0;
                                }
                            }
                        }
                        else if (CsDiff >= 4) {
                            if (posBalance < 83) {
                                if (RsDiff < 6) {
                                    return 0;
                                }
                                else if (RsDiff >= 6) {
                                    return 1;
                                }
                            }
                            else if (posBalance >= 83) {
                                return 1;
                            }
                        }
                    }
                }
                else if (Rs >= 301) {
                    if (TSC < 33) {
                        if (posBalance < 68) {
                            if (ssDCval < 42) {
                                if (negBalance < 104) {
                                    return 1;
                                }
                                else if (negBalance >= 104) {
                                    if (Rs < 432) {
                                        return 0;
                                    }
                                    else if (Rs >= 432) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 42) {
                                if (diffRsCsdiff < 58) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 58) {
                                    if (ssDCval < 73) {
                                        if (Rs < 2099) {
                                            return 1;
                                        }
                                        else if (Rs >= 2099) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 73) {
                                        if (MVDiff < 534) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 534) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (posBalance >= 68) {
                            if (diffAFD < 21) {
                                return 0;
                            }
                            else if (diffAFD >= 21) {
                                return 1;
                            }
                        }
                    }
                    else if (TSC >= 33) {
                        return 1;
                    }
                }
            }
        }
    }
    else if (diffTSC >= 7) {
        if (AFD < 29) {
            if (MVDiff < 46) {
                if (gchDC < 7) {
                    return 0;
                }
                else if (gchDC >= 7) {
                    if (diffTSC < 12) {
                        return 0;
                    }
                    else if (diffTSC >= 12) {
                        if (Cs < 558) {
                            if (diffAFD < 13) {
                                return 0;
                            }
                            else if (diffAFD >= 13) {
                                return 1;
                            }
                        }
                        else if (Cs >= 558) {
                            if (refDCval < 32) {
                                return 1;
                            }
                            else if (refDCval >= 32) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (MVDiff >= 46) {
                if (diffMVdiffVal < 54) {
                    if (CsDiff < 8) {
                        if (RsCsDiff < 182) {
                            if (refDCval < 19) {
                                return 1;
                            }
                            else if (refDCval >= 19) {
                                if (diffMVdiffVal < 46) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 46) {
                                    if (SC < 384) {
                                        return 1;
                                    }
                                    else if (SC >= 384) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 182) {
                            if (diffTSC < 10) {
                                return 0;
                            }
                            else if (diffTSC >= 10) {
                                return 1;
                            }
                        }
                    }
                    else if (CsDiff >= 8) {
                        if (diffTSC < 13) {
                            if (Rs < 224) {
                                return 1;
                            }
                            else if (Rs >= 224) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 13) {
                            if (refDCval < 93) {
                                return 1;
                            }
                            else if (refDCval >= 93) {
                                return 0;
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 54) {
                    if (negBalance < 22) {
                        if (diffRsCsdiff < 297) {
                            if (TSC < 11) {
                                if (SC < 1304) {
                                    return 1;
                                }
                                else if (SC >= 1304) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 11) {
                                if (ssDCval < 41) {
                                    return 1;
                                }
                                else if (ssDCval >= 41) {
                                    if (refDCval < 38) {
                                        if (ssDCval < 72) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 72) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 38) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 297) {
                            if (Rs < 2293) {
                                return 1;
                            }
                            else if (Rs >= 2293) {
                                if (AFD < 27) {
                                    return 1;
                                }
                                else if (AFD >= 27) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 22) {
                        if (ssDCval < 49) {
                            if (diffMVdiffVal < 67) {
                                if (Rs < 327) {
                                    return 0;
                                }
                                else if (Rs >= 327) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 67) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 49) {
                            if (diffTSC < 10) {
                                if (diffAFD < 13) {
                                    if (posBalance < 84) {
                                        return 0;
                                    }
                                    else if (posBalance >= 84) {
                                        if (SC < 600) {
                                            return 1;
                                        }
                                        else if (SC >= 600) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffAFD >= 13) {
                                    if (diffMVdiffVal < 82) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 82) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffTSC >= 10) {
                                if (diffAFD < 14) {
                                    return 0;
                                }
                                else if (diffAFD >= 14) {
                                    if (gchDC < 59) {
                                        if (diffMVdiffVal < 75) {
                                            if (negBalance < 33) {
                                                return 0;
                                            }
                                            else if (negBalance >= 33) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 75) {
                                            return 1;
                                        }
                                    }
                                    else if (gchDC >= 59) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (AFD >= 29) {
            if (diffMVdiffVal < 68) {
                if (diffTSC < 31) {
                    if (diffRsCsdiff < 69) {
                        if (gchDC < 4) {
                            if (CsDiff < 1) {
                                return 0;
                            }
                            else if (CsDiff >= 1) {
                                return 1;
                            }
                        }
                        else if (gchDC >= 4) {
                            if (SC < 1260) {
                                if (SC < 1209) {
                                    if (diffTSC < 16) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 16) {
                                        if (diffAFD < 21) {
                                            if (Rs < 211) {
                                                return 1;
                                            }
                                            else if (Rs >= 211) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= 21) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (SC >= 1209) {
                                    return 1;
                                }
                            }
                            else if (SC >= 1260) {
                                return 0;
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 69) {
                        if (MVDiff < 49) {
                            return 0;
                        }
                        else if (MVDiff >= 49) {
                            if (diffTSC < 11) {
                                if (Rs < 677) {
                                    return 1;
                                }
                                else if (Rs >= 677) {
                                    if (AFD < 54) {
                                        if (diffMVdiffVal < 56) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 56) {
                                            if (Cs < 803) {
                                                return 1;
                                            }
                                            else if (Cs >= 803) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (AFD >= 54) {
                                        return 1;
                                    }
                                }
                            }
                            else if (diffTSC >= 11) {
                                if (negBalance < 21) {
                                    if (RsDiff < 7) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 7) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 21) {
                                    if (posBalance < 36) {
                                        if (TSC < 20) {
                                            if (TSC < 18) {
                                                return 0;
                                            }
                                            else if (TSC >= 18) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 20) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 36) {
                                        if (ssDCval < 115) {
                                            if (diffAFD < 19) {
                                                if (posBalance < 70) {
                                                    return 1;
                                                }
                                                else if (posBalance >= 70) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffAFD >= 19) {
                                                if (ssDCval < 50) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 50) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 115) {
                                            if (diffRsCsdiff < 304) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 304) {
                                                if (diffTSC < 13) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 13) {
                                                    if (diffAFD < 50) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= 50) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffTSC >= 31) {
                    if (MVDiff < 90) {
                        return 1;
                    }
                    else if (MVDiff >= 90) {
                        if (SC < 3051) {
                            return 0;
                        }
                        else if (SC >= 3051) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 68) {
                if (diffRsCsdiff < 64) {
                    if (gchDC < 20) {
                        if (diffTSC < 9) {
                            if (TSC < 13) {
                                if (MVDiff < 160) {
                                    return 0;
                                }
                                else if (MVDiff >= 160) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 13) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 9) {
                            if (AFD < 49) {
                                return 1;
                            }
                            else if (AFD >= 49) {
                                if (TSC < 33) {
                                    return 1;
                                }
                                else if (TSC >= 33) {
                                    if (Rs < 204) {
                                        return 1;
                                    }
                                    else if (Rs >= 204) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (gchDC >= 20) {
                        if (diffAFD < 25) {
                            if (ssDCval < 123) {
                                if (refDCval < 104) {
                                    if (Rs < 4708) {
                                        if (SC < 1033) {
                                            if (diffRsCsdiff < 21) {
                                                return 0;
                                            }
                                            else if (diffRsCsdiff >= 21) {
                                                if (refDCval < 40) {
                                                    return 0;
                                                }
                                                else if (refDCval >= 40) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (SC >= 1033) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 4708) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 104) {
                                    if (MVDiff < 485) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 485) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 123) {
                                return 0;
                            }
                        }
                        else if (diffAFD >= 25) {
                            if (TSC < 33) {
                                if (Cs < 683) {
                                    if (CsDiff < 1) {
                                        if (refDCval < 44) {
                                            return 1;
                                        }
                                        else if (refDCval >= 44) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 1) {
                                        if (ssDCval < 36) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 36) {
                                            if (posBalance < 92) {
                                                return 1;
                                            }
                                            else if (posBalance >= 92) {
                                                if (posBalance < 99) {
                                                    if (SC < 252) {
                                                        return 1;
                                                    }
                                                    else if (SC >= 252) {
                                                        if (AFD < 35) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 35) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (posBalance >= 99) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 683) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 33) {
                                if (AFD < 71) {
                                    return 0;
                                }
                                else if (AFD >= 71) {
                                    if (TSC < 61) {
                                        return 1;
                                    }
                                    else if (TSC >= 61) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffRsCsdiff >= 64) {
                    if (negBalance < 32) {
                        if (diffRsCsdiff < 451) {
                            if (Rs < 941) {
                                if (diffAFD < 16) {
                                    return 0;
                                }
                                else if (diffAFD >= 16) {
                                    if (negBalance < 4) {
                                        if (gchDC < 104) {
                                            return 0;
                                        }
                                        else if (gchDC >= 104) {
                                            return 1;
                                        }
                                    }
                                    else if (negBalance >= 4) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Rs >= 941) {
                                if (diffRsCsdiff < 194) {
                                    if (MVDiff < 164) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 164) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 194) {
                                    if (posBalance < 123) {
                                        if (Rs < 2430) {
                                            if (refDCval < 74) {
                                                if (RsDiff < 11) {
                                                    if (Rs < 1275) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1275) {
                                                        return 0;
                                                    }
                                                }
                                                else if (RsDiff >= 11) {
                                                    return 1;
                                                }
                                            }
                                            else if (refDCval >= 74) {
                                                return 0;
                                            }
                                        }
                                        else if (Rs >= 2430) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 123) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 451) {
                            if (MVDiff < 254) {
                                return 1;
                            }
                            else if (MVDiff >= 254) {
                                if (diffRsCsdiff < 939) {
                                    if (Rs < 2464) {
                                        if (Rs < 1201) {
                                            return 0;
                                        }
                                        else if (Rs >= 1201) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 2464) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 939) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 32) {
                        if (diffAFD < 15) {
                            if (SC < 2370) {
                                return 1;
                            }
                            else if (SC >= 2370) {
                                if (gchDC < 36) {
                                    return 0;
                                }
                                else if (gchDC >= 36) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffAFD >= 15) {
                            if (diffAFD < 41) {
                                if (posBalance < 69) {
                                    if (diffTSC < 16) {
                                        if (AFD < 36) {
                                            return 1;
                                        }
                                        else if (AFD >= 36) {
                                            if (gchDC < 1) {
                                                return 0;
                                            }
                                            else if (gchDC >= 1) {
                                                if (RsDiff < 23) {
                                                    if (refDCval < 60) {
                                                        if (Cs < 572) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 572) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (refDCval >= 60) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 23) {
                                                    if (TSC < 18) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 18) {
                                                        if (posBalance < 24) {
                                                            return 0;
                                                        }
                                                        else if (posBalance >= 24) {
                                                            return 1;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 16) {
                                        if (AFD < 49) {
                                            return 1;
                                        }
                                        else if (AFD >= 49) {
                                            if (CsDiff < 7) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 7) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 69) {
                                    if (TSC < 25) {
                                        return 1;
                                    }
                                    else if (TSC >= 25) {
                                        if (RsCsDiff < 1314) {
                                            if (diffTSC < 19) {
                                                if (SC < 1975) {
                                                    return 1;
                                                }
                                                else if (SC >= 1975) {
                                                    if (gchDC < 28) {
                                                        if (gchDC < 17) {
                                                            if (AFD < 63) {
                                                                return 1;
                                                            }
                                                            else if (AFD >= 63) {
                                                                return 0;
                                                            }
                                                        }
                                                        else if (gchDC >= 17) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (gchDC >= 28) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffTSC >= 19) {
                                                if (CsDiff < 8) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 8) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (RsCsDiff >= 1314) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 41) {
                                if (diffMVdiffVal < 948) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= 948) {
                                    if (Rs < 513) {
                                        return 0;
                                    }
                                    else if (Rs >= 513) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect12(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 13) {
        if (diffMVdiffVal < 52) {
            if (MVDiff < 224) {
                if (refDCval < 24) {
                    if (diffAFD < 3) {
                        if (posBalance < 128) {
                            if (RsDiff < 2) {
                                return 0;
                            }
                            else if (RsDiff >= 2) {
                                if (refDCval < 18) {
                                    if (MVDiff < 5) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 5) {
                                        if (RsDiff < 3) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 3) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 18) {
                                    return 0;
                                }
                            }
                        }
                        else if (posBalance >= 128) {
                            return 1;
                        }
                    }
                    else if (diffAFD >= 3) {
                        if (RsDiff < 3) {
                            return 0;
                        }
                        else if (RsDiff >= 3) {
                            return 1;
                        }
                    }
                }
                else if (refDCval >= 24) {
                    if (RsDiff < 55) {
                        if (gchDC < 8) {
                            if (diffAFD < 9) {
                                if (MVDiff < 200) {
                                    return 0;
                                }
                                else if (MVDiff >= 200) {
                                    if (diffRsCsdiff < 256) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 256) {
                                        if (Cs < 924) {
                                            return 0;
                                        }
                                        else if (Cs >= 924) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 9) {
                                if (refDCval < 44) {
                                    if (negBalance < 36) {
                                        return 0;
                                    }
                                    else if (negBalance >= 36) {
                                        if (TSC < 8) {
                                            return 0;
                                        }
                                        else if (TSC >= 8) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (refDCval >= 44) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 8) {
                            if (SC < 1) {
                                if (ssDCval < 220) {
                                    if (CsDiff < 9) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 9) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 220) {
                                    return 1;
                                }
                            }
                            else if (SC >= 1) {
                                if (AFD < 1) {
                                    return 1;
                                }
                                else if (AFD >= 1) {
                                    if (diffRsCsdiff < 1058) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 1058) {
                                        if (Rs < 4349) {
                                            return 1;
                                        }
                                        else if (Rs >= 4349) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (RsDiff >= 55) {
                        if (Cs < 1931) {
                            if (CsDiff < 4) {
                                return 0;
                            }
                            else if (CsDiff >= 4) {
                                return 1;
                            }
                        }
                        else if (Cs >= 1931) {
                            return 0;
                        }
                    }
                }
            }
            else if (MVDiff >= 224) {
                if (AFD < 62) {
                    if (diffRsCsdiff < 12) {
                        if (TSC < 10) {
                            if (diffAFD < -12) {
                                return 0;
                            }
                            else if (diffAFD >= -12) {
                                if (diffTSC < 6) {
                                    if (negBalance < 122) {
                                        return 0;
                                    }
                                    else if (negBalance >= 122) {
                                        if (negBalance < 124) {
                                            return 1;
                                        }
                                        else if (negBalance >= 124) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffTSC >= 6) {
                                    if (Rs < 569) {
                                        return 1;
                                    }
                                    else if (Rs >= 569) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (TSC >= 10) {
                            if (ssDCval < 58) {
                                if (diffTSC < -6) {
                                    return 0;
                                }
                                else if (diffTSC >= -6) {
                                    if (diffRsCsdiff < -17) {
                                        if (posBalance < 28) {
                                            return 0;
                                        }
                                        else if (posBalance >= 28) {
                                            return 1;
                                        }
                                    }
                                    else if (diffRsCsdiff >= -17) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 58) {
                                if (diffTSC < 16) {
                                    return 0;
                                }
                                else if (diffTSC >= 16) {
                                    if (Rs < 114) {
                                        return 1;
                                    }
                                    else if (Rs >= 114) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 12) {
                        if (refDCval < 24) {
                            return 1;
                        }
                        else if (refDCval >= 24) {
                            if (posBalance < 21) {
                                if (ssDCval < 98) {
                                    if (Rs < 561) {
                                        return 0;
                                    }
                                    else if (Rs >= 561) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 98) {
                                    return 0;
                                }
                            }
                            else if (posBalance >= 21) {
                                if (ssDCval < 46) {
                                    if (AFD < 18) {
                                        return 0;
                                    }
                                    else if (AFD >= 18) {
                                        return 1;
                                    }
                                }
                                else if (ssDCval >= 46) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (AFD >= 62) {
                    if (Rs < 2800) {
                        return 1;
                    }
                    else if (Rs >= 2800) {
                        return 0;
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 52) {
            if (ssDCval < 40) {
                if (diffTSC < 3) {
                    if (gchDC < 16) {
                        if (diffRsCsdiff < 10) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 10) {
                            if (CsDiff < 5) {
                                if (Cs < 560) {
                                    return 1;
                                }
                                else if (Cs >= 560) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 5) {
                                return 0;
                            }
                        }
                    }
                    else if (gchDC >= 16) {
                        if (diffRsCsdiff < 35) {
                            if (diffTSC < -17) {
                                return 1;
                            }
                            else if (diffTSC >= -17) {
                                if (MVDiff < 1691) {
                                    if (AFD < 12) {
                                        return 1;
                                    }
                                    else if (AFD >= 12) {
                                        if (diffRsCsdiff < -189) {
                                            if (Cs < 901) {
                                                return 1;
                                            }
                                            else if (Cs >= 901) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -189) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (MVDiff >= 1691) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 35) {
                            if (CsDiff < 5) {
                                return 0;
                            }
                            else if (CsDiff >= 5) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffTSC >= 3) {
                    if (Rs < 102) {
                        if (diffAFD < 8) {
                            if (RsDiff < 11) {
                                return 0;
                            }
                            else if (RsDiff >= 11) {
                                return 1;
                            }
                        }
                        else if (diffAFD >= 8) {
                            if (AFD < 21) {
                                return 1;
                            }
                            else if (AFD >= 21) {
                                if (Rs < 67) {
                                    return 0;
                                }
                                else if (Rs >= 67) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (Rs >= 102) {
                        if (RsDiff < 6) {
                            if (diffMVdiffVal < 129) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 129) {
                                if (AFD < 11) {
                                    return 0;
                                }
                                else if (AFD >= 11) {
                                    if (Cs < 470) {
                                        return 1;
                                    }
                                    else if (Cs >= 470) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 6) {
                            if (Rs < 426) {
                                return 0;
                            }
                            else if (Rs >= 426) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (ssDCval >= 40) {
                if (diffAFD < 8) {
                    if (TSC < 32) {
                        if (gchDC < 93) {
                            if (TSC < 16) {
                                if (CsDiff < 41) {
                                    if (posBalance < 95) {
                                        return 0;
                                    }
                                    else if (posBalance >= 95) {
                                        if (Cs < 276) {
                                            return 0;
                                        }
                                        else if (Cs >= 276) {
                                            if (negBalance < 26) {
                                                return 0;
                                            }
                                            else if (negBalance >= 26) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (CsDiff >= 41) {
                                    return 1;
                                }
                            }
                            else if (TSC >= 16) {
                                if (diffMVdiffVal < 53) {
                                    return 1;
                                }
                                else if (diffMVdiffVal >= 53) {
                                    if (AFD < 36) {
                                        return 0;
                                    }
                                    else if (AFD >= 36) {
                                        if (Cs < 860) {
                                            if (CsDiff < 10) {
                                                if (diffTSC < 4) {
                                                    return 0;
                                                }
                                                else if (diffTSC >= 4) {
                                                    if (diffMVdiffVal < 188) {
                                                        return 0;
                                                    }
                                                    else if (diffMVdiffVal >= 188) {
                                                        if (AFD < 45) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 45) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 10) {
                                                if (TSC < 23) {
                                                    return 1;
                                                }
                                                else if (TSC >= 23) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (Cs >= 860) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 93) {
                            if (Cs < 772) {
                                return 1;
                            }
                            else if (Cs >= 772) {
                                return 0;
                            }
                        }
                    }
                    else if (TSC >= 32) {
                        if (Rs < 779) {
                            return 0;
                        }
                        else if (Rs >= 779) {
                            if (RsDiff < 10) {
                                return 0;
                            }
                            else if (RsDiff >= 10) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffAFD >= 8) {
                    if (SC < 1304) {
                        if (gchDC < 8) {
                            if (negBalance < 61) {
                                if (refDCval < 41) {
                                    if (Rs < 295) {
                                        return 0;
                                    }
                                    else if (Rs >= 295) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 41) {
                                    if (MVDiff < 118) {
                                        if (RsDiff < 3) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 3) {
                                            return 1;
                                        }
                                    }
                                    else if (MVDiff >= 118) {
                                        if (AFD < 15) {
                                            return 1;
                                        }
                                        else if (AFD >= 15) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 61) {
                                if (MVDiff < 452) {
                                    return 1;
                                }
                                else if (MVDiff >= 452) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 8) {
                            if (ssDCval < 227) {
                                if (diffRsCsdiff < 25) {
                                    if (Rs < 90) {
                                        if (ssDCval < 75) {
                                            if (ssDCval < 67) {
                                                if (diffMVdiffVal < 211) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 211) {
                                                    if (RsDiff < 1) {
                                                        return 0;
                                                    }
                                                    else if (RsDiff >= 1) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 67) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 75) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 90) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 25) {
                                    if (Rs < 513) {
                                        if (MVDiff < 445) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 445) {
                                            return 0;
                                        }
                                    }
                                    else if (Rs >= 513) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 227) {
                                return 1;
                            }
                        }
                    }
                    else if (SC >= 1304) {
                        if (AFD < 46) {
                            return 0;
                        }
                        else if (AFD >= 46) {
                            if (CsDiff < 9) {
                                return 0;
                            }
                            else if (CsDiff >= 9) {
                                if (gchDC < 27) {
                                    return 0;
                                }
                                else if (gchDC >= 27) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 13) {
        if (diffAFD < 23) {
            if (ssDCval < 75) {
                if (diffMVdiffVal < 74) {
                    if (negBalance < 69) {
                        if (RsCsDiff < 163) {
                            if (refDCval < 24) {
                                if (Rs < 337) {
                                    return 0;
                                }
                                else if (Rs >= 337) {
                                    return 1;
                                }
                            }
                            else if (refDCval >= 24) {
                                return 0;
                            }
                        }
                        else if (RsCsDiff >= 163) {
                            if (SC < 3849) {
                                if (AFD < 28) {
                                    if (diffMVdiffVal < 12) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 12) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 28) {
                                    return 0;
                                }
                            }
                            else if (SC >= 3849) {
                                if (posBalance < 33) {
                                    if (RsCsDiff < 501) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 501) {
                                        return 1;
                                    }
                                }
                                else if (posBalance >= 33) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 69) {
                        if (Cs < 1014) {
                            if (refDCval < 75) {
                                return 1;
                            }
                            else if (refDCval >= 75) {
                                if (SC < 1960) {
                                    return 0;
                                }
                                else if (SC >= 1960) {
                                    return 1;
                                }
                            }
                        }
                        else if (Cs >= 1014) {
                            return 0;
                        }
                    }
                }
                else if (diffMVdiffVal >= 74) {
                    if (ssDCval < 44) {
                        if (TSCindex < 7) {
                            if (Cs < 119) {
                                return 1;
                            }
                            else if (Cs >= 119) {
                                return 0;
                            }
                        }
                        else if (TSCindex >= 7) {
                            if (diffRsCsdiff < -11) {
                                if (Cs < 1) {
                                    return 1;
                                }
                                else if (Cs >= 1) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= -11) {
                                return 1;
                            }
                        }
                    }
                    else if (ssDCval >= 44) {
                        if (diffAFD < 18) {
                            if (gchDC < 4) {
                                return 1;
                            }
                            else if (gchDC >= 4) {
                                if (diffAFD < 14) {
                                    if (AFD < 32) {
                                        return 0;
                                    }
                                    else if (AFD >= 32) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= 14) {
                                    if (AFD < 22) {
                                        if (CsDiff < 1) {
                                            if (TSC < 16) {
                                                return 0;
                                            }
                                            else if (TSC >= 16) {
                                                return 1;
                                            }
                                        }
                                        else if (CsDiff >= 1) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 22) {
                                        if (diffTSC < 9) {
                                            if (refDCval < 78) {
                                                return 0;
                                            }
                                            else if (refDCval >= 78) {
                                                if (Rs < 197) {
                                                    return 0;
                                                }
                                                else if (Rs >= 197) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (diffTSC >= 9) {
                                            if (diffMVdiffVal < 111) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 111) {
                                                if (TSC < 19) {
                                                    return 1;
                                                }
                                                else if (TSC >= 19) {
                                                    if (TSC < 38) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 38) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 18) {
                            if (gchDC < 19) {
                                if (TSCindex < 9) {
                                    return 0;
                                }
                                else if (TSCindex >= 9) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 19) {
                                if (diffMVdiffVal < 92) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 92) {
                                    if (SC < 3942) {
                                        if (refDCval < 44) {
                                            if (Rs < 1288) {
                                                return 0;
                                            }
                                            else if (Rs >= 1288) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 44) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 3942) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (ssDCval >= 75) {
                if (diffMVdiffVal < 83) {
                    if (gchDC < 7) {
                        if (TSC < 23) {
                            return 0;
                        }
                        else if (TSC >= 23) {
                            if (MVDiff < 45) {
                                return 0;
                            }
                            else if (MVDiff >= 45) {
                                return 1;
                            }
                        }
                    }
                    else if (gchDC >= 7) {
                        if (MVDiff < 402) {
                            if (Rs < 13162) {
                                if (diffMVdiffVal < -115) {
                                    if (AFD < 38) {
                                        if (Rs < 358) {
                                            return 0;
                                        }
                                        else if (Rs >= 358) {
                                            return 1;
                                        }
                                    }
                                    else if (AFD >= 38) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= -115) {
                                    if (negBalance < 111) {
                                        if (ssDCval < 134) {
                                            return 0;
                                        }
                                        else if (ssDCval >= 134) {
                                            if (ssDCval < 136) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 136) {
                                                if (diffRsCsdiff < -469) {
                                                    return 1;
                                                }
                                                else if (diffRsCsdiff >= -469) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (negBalance >= 111) {
                                        if (SC < 3222) {
                                            return 1;
                                        }
                                        else if (SC >= 3222) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (Rs >= 13162) {
                                return 1;
                            }
                        }
                        else if (MVDiff >= 402) {
                            if (ssDCval < 130) {
                                return 1;
                            }
                            else if (ssDCval >= 130) {
                                return 0;
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 83) {
                    if (diffTSC < 6) {
                        return 0;
                    }
                    else if (diffTSC >= 6) {
                        if (ssDCval < 82) {
                            if (diffMVdiffVal < 162) {
                                return 1;
                            }
                            else if (diffMVdiffVal >= 162) {
                                if (AFD < 30) {
                                    return 1;
                                }
                                else if (AFD >= 30) {
                                    return 0;
                                }
                            }
                        }
                        else if (ssDCval >= 82) {
                            if (TSC < 11) {
                                return 1;
                            }
                            else if (TSC >= 11) {
                                if (RsCsDiff < 70) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 70) {
                                    if (Rs < 1855) {
                                        if (diffRsCsdiff < 41) {
                                            if (Cs < 565) {
                                                return 1;
                                            }
                                            else if (Cs >= 565) {
                                                return 0;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 41) {
                                            if (TSC < 33) {
                                                if (diffAFD < 19) {
                                                    if (RsDiff < 16) {
                                                        if (gchDC < 26) {
                                                            return 0;
                                                        }
                                                        else if (gchDC >= 26) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsDiff >= 16) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffAFD >= 19) {
                                                    return 1;
                                                }
                                            }
                                            else if (TSC >= 33) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Rs >= 1855) {
                                        if (MVDiff < 122) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 122) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 23) {
            if (diffMVdiffVal < 55) {
                if (refDCval < 39) {
                    if (MVDiff < 163) {
                        return 1;
                    }
                    else if (MVDiff >= 163) {
                        if (SC < 2193) {
                            return 0;
                        }
                        else if (SC >= 2193) {
                            return 1;
                        }
                    }
                }
                else if (refDCval >= 39) {
                    if (negBalance < 65) {
                        if (RsCsDiff < 19315) {
                            if (TSC < 16) {
                                return 0;
                            }
                            else if (TSC >= 16) {
                                if (RsCsDiff < 545) {
                                    if (TSC < 18) {
                                        if (Rs < 251) {
                                            return 0;
                                        }
                                        else if (Rs >= 251) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 18) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 545) {
                                    if (ssDCval < 105) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 105) {
                                        if (CsDiff < 11) {
                                            return 1;
                                        }
                                        else if (CsDiff >= 11) {
                                            if (AFD < 69) {
                                                return 0;
                                            }
                                            else if (AFD >= 69) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsCsDiff >= 19315) {
                            return 1;
                        }
                    }
                    else if (negBalance >= 65) {
                        if (MVDiff < 105) {
                            if (Rs < 463) {
                                return 1;
                            }
                            else if (Rs >= 463) {
                                return 0;
                            }
                        }
                        else if (MVDiff >= 105) {
                            if (posBalance < 22) {
                                if (RsDiff < 12) {
                                    return 0;
                                }
                                else if (RsDiff >= 12) {
                                    return 1;
                                }
                            }
                            else if (posBalance >= 22) {
                                return 1;
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 55) {
                if (negBalance < 28) {
                    if (diffAFD < 38) {
                        if (negBalance < 5) {
                            if (posBalance < 128) {
                                return 0;
                            }
                            else if (posBalance >= 128) {
                                if (SC < 2) {
                                    return 1;
                                }
                                else if (SC >= 2) {
                                    return 0;
                                }
                            }
                        }
                        else if (negBalance >= 5) {
                            if (AFD < 37) {
                                if (Cs < 224) {
                                    return 1;
                                }
                                else if (Cs >= 224) {
                                    if (diffMVdiffVal < 90) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 90) {
                                        if (RsCsDiff < 94) {
                                            if (ssDCval < 89) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 89) {
                                                return 1;
                                            }
                                        }
                                        else if (RsCsDiff >= 94) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 37) {
                                if (diffRsCsdiff < 30) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 30) {
                                    if (Rs < 1020) {
                                        return 1;
                                    }
                                    else if (Rs >= 1020) {
                                        if (SC < 7548) {
                                            if (refDCval < 37) {
                                                return 1;
                                            }
                                            else if (refDCval >= 37) {
                                                if (MVDiff < 104) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 104) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (SC >= 7548) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffAFD >= 38) {
                        if (refDCval < 57) {
                            if (gchDC < 74) {
                                if (RsCsDiff < 224) {
                                    if (TSC < 33) {
                                        if (Cs < 813) {
                                            return 1;
                                        }
                                        else if (Cs >= 813) {
                                            return 0;
                                        }
                                    }
                                    else if (TSC >= 33) {
                                        return 0;
                                    }
                                }
                                else if (RsCsDiff >= 224) {
                                    return 1;
                                }
                            }
                            else if (gchDC >= 74) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 57) {
                            return 1;
                        }
                    }
                }
                else if (negBalance >= 28) {
                    if (diffRsCsdiff < 80) {
                        if (AFD < 39) {
                            if (MVDiff < 80) {
                                return 0;
                            }
                            else if (MVDiff >= 80) {
                                if (negBalance < 38) {
                                    if (Cs < 419) {
                                        return 1;
                                    }
                                    else if (Cs >= 419) {
                                        return 0;
                                    }
                                }
                                else if (negBalance >= 38) {
                                    return 1;
                                }
                            }
                        }
                        else if (AFD >= 39) {
                            if (diffTSC < 11) {
                                if (MVDiff < 1329) {
                                    if (Rs < 329) {
                                        if (Rs < 246) {
                                            return 0;
                                        }
                                        else if (Rs >= 246) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 329) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 1329) {
                                    return 1;
                                }
                            }
                            else if (diffTSC >= 11) {
                                if (diffRsCsdiff < -17) {
                                    if (AFD < 70) {
                                        return 0;
                                    }
                                    else if (AFD >= 70) {
                                        return 1;
                                    }
                                }
                                else if (diffRsCsdiff >= -17) {
                                    if (posBalance < 19) {
                                        if (AFD < 64) {
                                            return 0;
                                        }
                                        else if (AFD >= 64) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 19) {
                                        if (negBalance < 45) {
                                            if (ssDCval < 81) {
                                                return 0;
                                            }
                                            else if (ssDCval >= 81) {
                                                if (diffTSC < 17) {
                                                    if (Rs < 352) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 352) {
                                                        return 0;
                                                    }
                                                }
                                                else if (diffTSC >= 17) {
                                                    return 1;
                                                }
                                            }
                                        }
                                        else if (negBalance >= 45) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 80) {
                        if (diffAFD < 39) {
                            if (AFD < 52) {
                                if (diffTSC < 11) {
                                    if (gchDC < 2) {
                                        return 0;
                                    }
                                    else if (gchDC >= 2) {
                                        if (Rs < 1663) {
                                            return 1;
                                        }
                                        else if (Rs >= 1663) {
                                            if (posBalance < 31) {
                                                return 1;
                                            }
                                            else if (posBalance >= 31) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffTSC >= 11) {
                                    if (gchDC < 13) {
                                        return 1;
                                    }
                                    else if (gchDC >= 13) {
                                        if (CsDiff < 5) {
                                            if (Cs < 536) {
                                                return 1;
                                            }
                                            else if (Cs >= 536) {
                                                return 0;
                                            }
                                        }
                                        else if (CsDiff >= 5) {
                                            if (ssDCval < 47) {
                                                if (ssDCval < 46) {
                                                    if (diffTSC < 18) {
                                                        return 1;
                                                    }
                                                    else if (diffTSC >= 18) {
                                                        if (gchDC < 26) {
                                                            return 1;
                                                        }
                                                        else if (gchDC >= 26) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (ssDCval >= 46) {
                                                    return 0;
                                                }
                                            }
                                            else if (ssDCval >= 47) {
                                                if (diffAFD < 24) {
                                                    if (TSC < 22) {
                                                        return 0;
                                                    }
                                                    else if (TSC >= 22) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffAFD >= 24) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (AFD >= 52) {
                                if (RsDiff < 6) {
                                    return 0;
                                }
                                else if (RsDiff >= 6) {
                                    if (diffMVdiffVal < 77) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 77) {
                                        if (AFD < 69) {
                                            if (CsDiff < 9) {
                                                return 0;
                                            }
                                            else if (CsDiff >= 9) {
                                                if (TSC < 39) {
                                                    return 1;
                                                }
                                                else if (TSC >= 39) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (AFD >= 69) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffAFD >= 39) {
                            if (MVDiff < 75) {
                                if (Cs < 5148) {
                                    return 1;
                                }
                                else if (Cs >= 5148) {
                                    return 0;
                                }
                            }
                            else if (MVDiff >= 75) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect13(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 13) {
        if (MVDiff < 125) {
            if (diffMVdiffVal < 31) {
                if (ssDCval < 21) {
                    if (diffAFD < 3) {
                        if (RsCsDiff < 7) {
                            return 0;
                        }
                        else if (RsCsDiff >= 7) {
                            if (MVDiff < 6) {
                                return 1;
                            }
                            else if (MVDiff >= 6) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 3) {
                        return 1;
                    }
                }
                else if (ssDCval >= 21) {
                    if (RsDiff < 75) {
                        if (ssDCval < 197) {
                            if (MVDiff < 55) {
                                return 0;
                            }
                            else if (MVDiff >= 55) {
                                if (RsCsDiff < 2455) {
                                    if (diffAFD < 11) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 11) {
                                        if (MVDiff < 120) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 120) {
                                            if (Rs < 462) {
                                                return 1;
                                            }
                                            else if (Rs >= 462) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (RsCsDiff >= 2455) {
                                    if (SC < 5093) {
                                        return 1;
                                    }
                                    else if (SC >= 5093) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (ssDCval >= 197) {
                            if (gchDC < 28) {
                                if (negBalance < 93) {
                                    return 0;
                                }
                                else if (negBalance >= 93) {
                                    if (AFD < 5) {
                                        return 0;
                                    }
                                    else if (AFD >= 5) {
                                        return 1;
                                    }
                                }
                            }
                            else if (gchDC >= 28) {
                                return 1;
                            }
                        }
                    }
                    else if (RsDiff >= 75) {
                        if (Rs < 1541) {
                            return 0;
                        }
                        else if (Rs >= 1541) {
                            return 1;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 31) {
                if (ssDCval < 46) {
                    if (diffAFD < 8) {
                        if (RsDiff < 17) {
                            return 0;
                        }
                        else if (RsDiff >= 17) {
                            if (Rs < 680) {
                                return 1;
                            }
                            else if (Rs >= 680) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 8) {
                        if (ssDCval < 36) {
                            return 1;
                        }
                        else if (ssDCval >= 36) {
                            if (gchDC < 9) {
                                return 1;
                            }
                            else if (gchDC >= 9) {
                                return 0;
                            }
                        }
                    }
                }
                else if (ssDCval >= 46) {
                    if (RsDiff < 30) {
                        return 0;
                    }
                    else if (RsDiff >= 30) {
                        if (posBalance < 69) {
                            if (ssDCval < 126) {
                                return 0;
                            }
                            else if (ssDCval >= 126) {
                                if (Rs < 2452) {
                                    return 1;
                                }
                                else if (Rs >= 2452) {
                                    return 0;
                                }
                            }
                        }
                        else if (posBalance >= 69) {
                            return 1;
                        }
                    }
                }
            }
        }
        else if (MVDiff >= 125) {
            if (diffMVdiffVal < 99) {
                if (TSC < 14) {
                    if (ssDCval < 38) {
                        if (diffRsCsdiff < 2) {
                            if (MVDiff < 1517) {
                                if (negBalance < 82) {
                                    return 0;
                                }
                                else if (negBalance >= 82) {
                                    if (ssDCval < 24) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 24) {
                                        return 0;
                                    }
                                }
                            }
                            else if (MVDiff >= 1517) {
                                if (AFD < 1) {
                                    return 0;
                                }
                                else if (AFD >= 1) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 2) {
                            if (TSCindex < 7) {
                                if (ssDCval < 17) {
                                    return 1;
                                }
                                else if (ssDCval >= 17) {
                                    return 0;
                                }
                            }
                            else if (TSCindex >= 7) {
                                if (diffAFD < 3) {
                                    if (Rs < 1865) {
                                        return 0;
                                    }
                                    else if (Rs >= 1865) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= 3) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 38) {
                        if (MVDiff < 376) {
                            return 0;
                        }
                        else if (MVDiff >= 376) {
                            if (diffAFD < 10) {
                                if (MVDiff < 377) {
                                    if (Rs < 53) {
                                        return 1;
                                    }
                                    else if (Rs >= 53) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 377) {
                                    return 0;
                                }
                            }
                            else if (diffAFD >= 10) {
                                return 1;
                            }
                        }
                    }
                }
                else if (TSC >= 14) {
                    if (MVDiff < 263) {
                        if (Rs < 14) {
                            if (Cs < 50) {
                                return 0;
                            }
                            else if (Cs >= 50) {
                                return 1;
                            }
                        }
                        else if (Rs >= 14) {
                            if (refDCval < 51) {
                                if (diffAFD < -2) {
                                    if (Rs < 628) {
                                        return 0;
                                    }
                                    else if (Rs >= 628) {
                                        return 1;
                                    }
                                }
                                else if (diffAFD >= -2) {
                                    return 0;
                                }
                            }
                            else if (refDCval >= 51) {
                                return 0;
                            }
                        }
                    }
                    else if (MVDiff >= 263) {
                        if (CsDiff < 19) {
                            if (refDCval < 26) {
                                return 1;
                            }
                            else if (refDCval >= 26) {
                                if (MVDiff < 265) {
                                    return 1;
                                }
                                else if (MVDiff >= 265) {
                                    if (ssDCval < 72) {
                                        if (diffTSC < -3) {
                                            return 0;
                                        }
                                        else if (diffTSC >= -3) {
                                            if (SC < 1988) {
                                                if (RsCsDiff < 26) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 26) {
                                                    if (diffMVdiffVal < 33) {
                                                        return 1;
                                                    }
                                                    else if (diffMVdiffVal >= 33) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (SC >= 1988) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 72) {
                                        if (CsDiff < 10) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 10) {
                                            if (Rs < 1076) {
                                                if (Cs < 1223) {
                                                    return 1;
                                                }
                                                else if (Cs >= 1223) {
                                                    return 0;
                                                }
                                            }
                                            else if (Rs >= 1076) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 19) {
                            if (Rs < 2256) {
                                return 1;
                            }
                            else if (Rs >= 2256) {
                                return 0;
                            }
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 99) {
                if (diffTSC < 4) {
                    if (ssDCval < 41) {
                        if (AFD < 8) {
                            return 0;
                        }
                        else if (AFD >= 8) {
                            if (RsCsDiff < 35) {
                                if (ssDCval < 24) {
                                    if (MVDiff < 939) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 939) {
                                        if (AFD < 21) {
                                            return 0;
                                        }
                                        else if (AFD >= 21) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (ssDCval >= 24) {
                                    if (diffAFD < 12) {
                                        if (MVDiff < 230) {
                                            if (MVDiff < 214) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 214) {
                                                return 1;
                                            }
                                        }
                                        else if (MVDiff >= 230) {
                                            return 0;
                                        }
                                    }
                                    else if (diffAFD >= 12) {
                                        return 1;
                                    }
                                }
                            }
                            else if (RsCsDiff >= 35) {
                                if (Cs < 747) {
                                    if (Rs < 105) {
                                        return 1;
                                    }
                                    else if (Rs >= 105) {
                                        if (diffTSC < -6) {
                                            return 1;
                                        }
                                        else if (diffTSC >= -6) {
                                            if (negBalance < 117) {
                                                if (ssDCval < 33) {
                                                    return 0;
                                                }
                                                else if (ssDCval >= 33) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 117) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (Cs >= 747) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 41) {
                        if (TSC < 19) {
                            if (posBalance < 128) {
                                if (diffRsCsdiff < 277) {
                                    if (diffMVdiffVal < 101) {
                                        if (RsDiff < 3) {
                                            if (Rs < 205) {
                                                return 0;
                                            }
                                            else if (Rs >= 205) {
                                                return 1;
                                            }
                                        }
                                        else if (RsDiff >= 3) {
                                            return 0;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 101) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 277) {
                                    if (diffAFD < 11) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 11) {
                                        return 1;
                                    }
                                }
                            }
                            else if (posBalance >= 128) {
                                return 1;
                            }
                        }
                        else if (TSC >= 19) {
                            if (diffRsCsdiff < 897) {
                                if (AFD < 40) {
                                    if (TSC < 21) {
                                        if (AFD < 35) {
                                            return 0;
                                        }
                                        else if (AFD >= 35) {
                                            return 1;
                                        }
                                    }
                                    else if (TSC >= 21) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 40) {
                                    return 0;
                                }
                            }
                            else if (diffRsCsdiff >= 897) {
                                return 1;
                            }
                        }
                    }
                }
                else if (diffTSC >= 4) {
                    if (negBalance < 34) {
                        if (Rs < 60) {
                            if (diffAFD < 6) {
                                return 0;
                            }
                            else if (diffAFD >= 6) {
                                return 1;
                            }
                        }
                        else if (Rs >= 60) {
                            if (gchDC < 27) {
                                return 0;
                            }
                            else if (gchDC >= 27) {
                                if (SC < 239) {
                                    return 1;
                                }
                                else if (SC >= 239) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 34) {
                        if (refDCval < 51) {
                            if (gchDC < 7) {
                                return 1;
                            }
                            else if (gchDC >= 7) {
                                if (refDCval < 43) {
                                    if (diffAFD < 11) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 11) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 43) {
                                    return 1;
                                }
                            }
                        }
                        else if (refDCval >= 51) {
                            if (RsCsDiff < 19) {
                                if (AFD < 15) {
                                    if (Rs < 108) {
                                        return 1;
                                    }
                                    else if (Rs >= 108) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 15) {
                                    return 0;
                                }
                            }
                            else if (RsCsDiff >= 19) {
                                if (TSC < 34) {
                                    if (Rs < 271) {
                                        if (Cs < 115) {
                                            return 0;
                                        }
                                        else if (Cs >= 115) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 271) {
                                        if (MVDiff < 201) {
                                            if (SC < 2521) {
                                                if (CsDiff < 7) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 7) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 2521) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 201) {
                                            if (MVDiff < 381) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 381) {
                                                if (CsDiff < 7) {
                                                    return 0;
                                                }
                                                else if (CsDiff >= 7) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (TSC >= 34) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 13) {
        if (diffMVdiffVal < 63) {
            if (diffTSC < 16) {
                if (MVDiff < 158) {
                    if (ssDCval < 68) {
                        if (diffRsCsdiff < 148) {
                            if (negBalance < 69) {
                                return 0;
                            }
                            else if (negBalance >= 69) {
                                if (AFD < 34) {
                                    return 0;
                                }
                                else if (AFD >= 34) {
                                    return 1;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 148) {
                            if (Cs < 1613) {
                                if (TSC < 8) {
                                    return 0;
                                }
                                else if (TSC >= 8) {
                                    if (diffTSC < 11) {
                                        if (AFD < 20) {
                                            return 1;
                                        }
                                        else if (AFD >= 20) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 11) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 1613) {
                                if (ssDCval < 31) {
                                    return 1;
                                }
                                else if (ssDCval >= 31) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (ssDCval >= 68) {
                        if (gchDC < 103) {
                            return 0;
                        }
                        else if (gchDC >= 103) {
                            return 1;
                        }
                    }
                }
                else if (MVDiff >= 158) {
                    if (negBalance < 48) {
                        if (diffRsCsdiff < 188) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 188) {
                            if (RsCsDiff < 282) {
                                return 1;
                            }
                            else if (RsCsDiff >= 282) {
                                if (diffAFD < 27) {
                                    return 0;
                                }
                                else if (diffAFD >= 27) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (negBalance >= 48) {
                        if (gchDC < 5) {
                            return 1;
                        }
                        else if (gchDC >= 5) {
                            if (CsDiff < 7) {
                                if (diffAFD < 15) {
                                    return 1;
                                }
                                else if (diffAFD >= 15) {
                                    return 0;
                                }
                            }
                            else if (CsDiff >= 7) {
                                if (Cs < 955) {
                                    if (negBalance < 63) {
                                        return 0;
                                    }
                                    else if (negBalance >= 63) {
                                        return 1;
                                    }
                                }
                                else if (Cs >= 955) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
            }
            else if (diffTSC >= 16) {
                if (RsDiff < 11) {
                    if (Cs < 575) {
                        if (AFD < 39) {
                            if (Rs < 284) {
                                return 1;
                            }
                            else if (Rs >= 284) {
                                if (gchDC < 21) {
                                    return 1;
                                }
                                else if (gchDC >= 21) {
                                    if (Cs < 561) {
                                        return 0;
                                    }
                                    else if (Cs >= 561) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (AFD >= 39) {
                            if (TSC < 27) {
                                if (Rs < 314) {
                                    return 1;
                                }
                                else if (Rs >= 314) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 27) {
                                return 0;
                            }
                        }
                    }
                    else if (Cs >= 575) {
                        return 0;
                    }
                }
                else if (RsDiff >= 11) {
                    if (refDCval < 36) {
                        if (CsDiff < 12) {
                            return 0;
                        }
                        else if (CsDiff >= 12) {
                            return 1;
                        }
                    }
                    else if (refDCval >= 36) {
                        if (Cs < 3440) {
                            if (ssDCval < 105) {
                                return 1;
                            }
                            else if (ssDCval >= 105) {
                                if (AFD < 54) {
                                    if (CsDiff < 23) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 23) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 54) {
                                    return 1;
                                }
                            }
                        }
                        else if (Cs >= 3440) {
                            if (TSC < 17) {
                                return 1;
                            }
                            else if (TSC >= 17) {
                                return 0;
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 63) {
            if (diffAFD < 25) {
                if (ssDCval < 66) {
                    if (gchDC < 7) {
                        if (refDCval < 35) {
                            if (diffMVdiffVal < 74) {
                                if (Cs < 745) {
                                    return 0;
                                }
                                else if (Cs >= 745) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 74) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 35) {
                            return 1;
                        }
                    }
                    else if (gchDC >= 7) {
                        if (diffMVdiffVal < 109) {
                            if (ssDCval < 36) {
                                if (refDCval < 51) {
                                    return 1;
                                }
                                else if (refDCval >= 51) {
                                    if (RsDiff < 12) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 12) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 36) {
                                if (diffRsCsdiff < 49) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 49) {
                                    if (ssDCval < 51) {
                                        if (AFD < 22) {
                                            if (Rs < 661) {
                                                return 1;
                                            }
                                            else if (Rs >= 661) {
                                                return 0;
                                            }
                                        }
                                        else if (AFD >= 22) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 51) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 109) {
                            if (MVDiff < 1064) {
                                if (Rs < 1891) {
                                    if (gchDC < 79) {
                                        if (diffRsCsdiff < -15) {
                                            if (Rs < 344) {
                                                return 0;
                                            }
                                            else if (Rs >= 344) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= -15) {
                                            if (SC < 824) {
                                                return 1;
                                            }
                                            else if (SC >= 824) {
                                                if (diffRsCsdiff < 38) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= 38) {
                                                    if (TSCindex < 9) {
                                                        return 0;
                                                    }
                                                    else if (TSCindex >= 9) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    else if (gchDC >= 79) {
                                        if (refDCval < 145) {
                                            return 0;
                                        }
                                        else if (refDCval >= 145) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Rs >= 1891) {
                                    if (CsDiff < 21) {
                                        return 0;
                                    }
                                    else if (CsDiff >= 21) {
                                        return 1;
                                    }
                                }
                            }
                            else if (MVDiff >= 1064) {
                                if (SC < 56) {
                                    return 1;
                                }
                                else if (SC >= 56) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 66) {
                    if (gchDC < 14) {
                        if (diffTSC < 7) {
                            if (SC < 240) {
                                return 1;
                            }
                            else if (SC >= 240) {
                                return 0;
                            }
                        }
                        else if (diffTSC >= 7) {
                            if (Cs < 1418) {
                                if (MVDiff < 510) {
                                    return 1;
                                }
                                else if (MVDiff >= 510) {
                                    if (Rs < 460) {
                                        return 0;
                                    }
                                    else if (Rs >= 460) {
                                        return 1;
                                    }
                                }
                            }
                            else if (Cs >= 1418) {
                                if (diffTSC < 12) {
                                    if (RsCsDiff < 1742) {
                                        return 0;
                                    }
                                    else if (RsCsDiff >= 1742) {
                                        if (Rs < 6338) {
                                            return 1;
                                        }
                                        else if (Rs >= 6338) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffTSC >= 12) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 14) {
                        if (diffMVdiffVal < 284) {
                            if (diffTSC < 7) {
                                return 0;
                            }
                            else if (diffTSC >= 7) {
                                if (TSC < 13) {
                                    if (AFD < 16) {
                                        return 0;
                                    }
                                    else if (AFD >= 16) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 13) {
                                    if (Rs < 1040) {
                                        if (CsDiff < 12) {
                                            if (Rs < 38) {
                                                return 1;
                                            }
                                            else if (Rs >= 38) {
                                                if (Rs < 923) {
                                                    return 0;
                                                }
                                                else if (Rs >= 923) {
                                                    if (diffAFD < 14) {
                                                        return 1;
                                                    }
                                                    else if (diffAFD >= 14) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (CsDiff >= 12) {
                                            if (Cs < 2180) {
                                                return 1;
                                            }
                                            else if (Cs >= 2180) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Rs >= 1040) {
                                        if (diffMVdiffVal < 79) {
                                            return 0;
                                        }
                                        else if (diffMVdiffVal >= 79) {
                                            if (ssDCval < 108) {
                                                if (MVDiff < 160) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 160) {
                                                    if (Rs < 1109) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1109) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (ssDCval >= 108) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 284) {
                            if (posBalance < 98) {
                                if (RsDiff < 14) {
                                    if (negBalance < 118) {
                                        if (diffMVdiffVal < 344) {
                                            if (Cs < 257) {
                                                return 1;
                                            }
                                            else if (Cs >= 257) {
                                                return 0;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 344) {
                                            return 0;
                                        }
                                    }
                                    else if (negBalance >= 118) {
                                        return 1;
                                    }
                                }
                                else if (RsDiff >= 14) {
                                    if (Cs < 2265) {
                                        return 1;
                                    }
                                    else if (Cs >= 2265) {
                                        return 0;
                                    }
                                }
                            }
                            else if (posBalance >= 98) {
                                if (negBalance < 3) {
                                    return 0;
                                }
                                else if (negBalance >= 3) {
                                    if (Rs < 479) {
                                        return 1;
                                    }
                                    else if (Rs >= 479) {
                                        if (RsCsDiff < 392) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 392) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 25) {
                if (negBalance < 16) {
                    if (diffAFD < 33) {
                        if (diffMVdiffVal < 439) {
                            if (CsDiff < 17) {
                                if (negBalance < 13) {
                                    return 0;
                                }
                                else if (negBalance >= 13) {
                                    if (Rs < 794) {
                                        return 1;
                                    }
                                    else if (Rs >= 794) {
                                        return 0;
                                    }
                                }
                            }
                            else if (CsDiff >= 17) {
                                return 1;
                            }
                        }
                        else if (diffMVdiffVal >= 439) {
                            return 1;
                        }
                    }
                    else if (diffAFD >= 33) {
                        if (TSC < 32) {
                            if (negBalance < 7) {
                                if (gchDC < 104) {
                                    return 0;
                                }
                                else if (gchDC >= 104) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 7) {
                                return 1;
                            }
                        }
                        else if (TSC >= 32) {
                            if (posBalance < 111) {
                                return 0;
                            }
                            else if (posBalance >= 111) {
                                if (diffAFD < 51) {
                                    if (gchDC < 62) {
                                        return 1;
                                    }
                                    else if (gchDC >= 62) {
                                        if (TSC < 46) {
                                            if (diffMVdiffVal < 671) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 671) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 46) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffAFD >= 51) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (negBalance >= 16) {
                    if (diffRsCsdiff < 218) {
                        if (AFD < 41) {
                            if (MVDiff < 103) {
                                if (TSC < 12) {
                                    return 0;
                                }
                                else if (TSC >= 12) {
                                    if (SC < 2035) {
                                        return 1;
                                    }
                                    else if (SC >= 2035) {
                                        if (Rs < 1868) {
                                            return 0;
                                        }
                                        else if (Rs >= 1868) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (MVDiff >= 103) {
                                if (refDCval < 37) {
                                    if (MVDiff < 161) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 161) {
                                        return 0;
                                    }
                                }
                                else if (refDCval >= 37) {
                                    return 1;
                                }
                            }
                        }
                        else if (AFD >= 41) {
                            if (diffAFD < 37) {
                                if (diffRsCsdiff < 40) {
                                    if (SC < 503) {
                                        if (diffAFD < 33) {
                                            return 1;
                                        }
                                        else if (diffAFD >= 33) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 503) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 40) {
                                    if (Cs < 762) {
                                        if (diffAFD < 28) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 28) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 762) {
                                        if (TSC < 26) {
                                            if (refDCval < 68) {
                                                return 0;
                                            }
                                            else if (refDCval >= 68) {
                                                return 1;
                                            }
                                        }
                                        else if (TSC >= 26) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 37) {
                                if (diffMVdiffVal < 86) {
                                    if (AFD < 44) {
                                        return 1;
                                    }
                                    else if (AFD >= 44) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 86) {
                                    if (posBalance < 16) {
                                        return 0;
                                    }
                                    else if (posBalance >= 16) {
                                        if (refDCval < 55) {
                                            if (MVDiff < 191) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 191) {
                                                if (RsCsDiff < 28) {
                                                    return 1;
                                                }
                                                else if (RsCsDiff >= 28) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (refDCval >= 55) {
                                            if (ssDCval < 146) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 146) {
                                                if (CsDiff < 10) {
                                                    return 1;
                                                }
                                                else if (CsDiff >= 10) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (diffRsCsdiff >= 218) {
                        if (TSC < 11) {
                            return 0;
                        }
                        else if (TSC >= 11) {
                            if (negBalance < 43) {
                                if (SC < 3578) {
                                    return 1;
                                }
                                else if (SC >= 3578) {
                                    if (gchDC < 2) {
                                        return 0;
                                    }
                                    else if (gchDC >= 2) {
                                        if (Rs < 2476) {
                                            if (diffAFD < 39) {
                                                if (Rs < 2075) {
                                                    return 1;
                                                }
                                                else if (Rs >= 2075) {
                                                    if (MVDiff < 432) {
                                                        return 0;
                                                    }
                                                    else if (MVDiff >= 432) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (diffAFD >= 39) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 2476) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 43) {
                                if (TSC < 18) {
                                    if (diffTSC < 11) {
                                        if (AFD < 34) {
                                            return 1;
                                        }
                                        else if (AFD >= 34) {
                                            if (AFD < 40) {
                                                return 0;
                                            }
                                            else if (AFD >= 40) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffTSC >= 11) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 18) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect14(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 /*TSCindex*/, mfxU32 Scindex, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 14) {
        if (gchDC < 3) {
            if (diffMVdiffVal < 52) {
                if (refDCval < 20) {
                    if (RsDiff < 2) {
                        return 0;
                    }
                    else if (RsDiff >= 2) {
                        if (MVDiff < 31) {
                            if (MVDiff < 6) {
                                return 1;
                            }
                            else if (MVDiff >= 6) {
                                if (TSC < 1) {
                                    return 0;
                                }
                                else if (TSC >= 1) {
                                    return 1;
                                }
                            }
                        }
                        else if (MVDiff >= 31) {
                            return 0;
                        }
                    }
                }
                else if (refDCval >= 20) {
                    if (diffTSC < 5) {
                        if (Rs < 5) {
                            if (MVDiff < 1513) {
                                return 0;
                            }
                            else if (MVDiff >= 1513) {
                                if (Rs < 3) {
                                    return 0;
                                }
                                else if (Rs >= 3) {
                                    return 1;
                                }
                            }
                        }
                        else if (Rs >= 5) {
                            return 0;
                        }
                    }
                    else if (diffTSC >= 5) {
                        if (refDCval < 43) {
                            if (Rs < 746) {
                                return 1;
                            }
                            else if (Rs >= 746) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 43) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 52) {
                if (diffRsCsdiff < 1) {
                    if (diffAFD < 8) {
                        if (posBalance < 80) {
                            return 0;
                        }
                        else if (posBalance >= 80) {
                            if (posBalance < 81) {
                                return 1;
                            }
                            else if (posBalance >= 81) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 8) {
                        if (Rs < 105) {
                            return 1;
                        }
                        else if (Rs >= 105) {
                            return 0;
                        }
                    }
                }
                else if (diffRsCsdiff >= 1) {
                    if (Rs < 126) {
                        if (Rs < 13) {
                            return 0;
                        }
                        else if (Rs >= 13) {
                            return 1;
                        }
                    }
                    else if (Rs >= 126) {
                        if (ssDCval < 40) {
                            if (AFD < 13) {
                                if (diffRsCsdiff < 28) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 28) {
                                    return 1;
                                }
                            }
                            else if (AFD >= 13) {
                                return 1;
                            }
                        }
                        else if (ssDCval >= 40) {
                            if (diffMVdiffVal < 348) {
                                if (diffMVdiffVal < 54) {
                                    if (negBalance < 52) {
                                        return 1;
                                    }
                                    else if (negBalance >= 52) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 54) {
                                    return 0;
                                }
                            }
                            else if (diffMVdiffVal >= 348) {
                                return 1;
                            }
                        }
                    }
                }
            }
        }
        else if (gchDC >= 3) {
            if (MVDiff < 164) {
                if (diffMVdiffVal < 66) {
                    if (refDCval < 24) {
                        if (RsDiff < 4) {
                            if (TSC < 1) {
                                return 1;
                            }
                            else if (TSC >= 1) {
                                return 0;
                            }
                        }
                        else if (RsDiff >= 4) {
                            return 1;
                        }
                    }
                    else if (refDCval >= 24) {
                        if (RsDiff < 55) {
                            if (diffRsCsdiff < 575) {
                                if (refDCval < 228) {
                                    return 0;
                                }
                                else if (refDCval >= 228) {
                                    if (Rs < 84) {
                                        return 1;
                                    }
                                    else if (Rs >= 84) {
                                        return 0;
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 575) {
                                if (Cs < 1889) {
                                    if (RsDiff < 32) {
                                        return 1;
                                    }
                                    else if (RsDiff >= 32) {
                                        return 0;
                                    }
                                }
                                else if (Cs >= 1889) {
                                    if (TSC < 25) {
                                        return 0;
                                    }
                                    else if (TSC >= 25) {
                                        if (Rs < 5256) {
                                            return 1;
                                        }
                                        else if (Rs >= 5256) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 55) {
                            if (diffRsCsdiff < -464) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= -464) {
                                if (SC < 11871) {
                                    return 1;
                                }
                                else if (SC >= 11871) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (diffMVdiffVal >= 66) {
                    if (ssDCval < 43) {
                        if (diffAFD < 9) {
                            if (refDCval < 64) {
                                return 0;
                            }
                            else if (refDCval >= 64) {
                                if (SC < 720) {
                                    return 1;
                                }
                                else if (SC >= 720) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffAFD >= 9) {
                            if (TSC < 9) {
                                return 1;
                            }
                            else if (TSC >= 9) {
                                return 0;
                            }
                        }
                    }
                    else if (ssDCval >= 43) {
                        if (gchDC < 7) {
                            if (diffTSC < 5) {
                                if (negBalance < 31) {
                                    if (Rs < 153) {
                                        return 0;
                                    }
                                    else if (Rs >= 153) {
                                        return 1;
                                    }
                                }
                                else if (negBalance >= 31) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= 5) {
                                if (SC < 1914) {
                                    return 1;
                                }
                                else if (SC >= 1914) {
                                    return 0;
                                }
                            }
                        }
                        else if (gchDC >= 7) {
                            return 0;
                        }
                    }
                }
            }
            else if (MVDiff >= 164) {
                if (ssDCval < 31) {
                    if (CsDiff < 1) {
                        if (refDCval < 148) {
                            if (refDCval < 23) {
                                return 1;
                            }
                            else if (refDCval >= 23) {
                                return 0;
                            }
                        }
                        else if (refDCval >= 148) {
                            if (AFD < 23) {
                                return 1;
                            }
                            else if (AFD >= 23) {
                                return 0;
                            }
                        }
                    }
                    else if (CsDiff >= 1) {
                        if (RsDiff < 8) {
                            if (diffTSC < -1) {
                                if (diffTSC < -11) {
                                    if (TSC < 8) {
                                        return 1;
                                    }
                                    else if (TSC >= 8) {
                                        if (MVDiff < 1123) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 1123) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (diffTSC >= -11) {
                                    return 0;
                                }
                            }
                            else if (diffTSC >= -1) {
                                if (Rs < 193) {
                                    return 1;
                                }
                                else if (Rs >= 193) {
                                    if (Rs < 403) {
                                        return 0;
                                    }
                                    else if (Rs >= 403) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 8) {
                            return 1;
                        }
                    }
                }
                else if (ssDCval >= 31) {
                    if (diffTSC < 4) {
                        if (TSC < 32) {
                            if (diffAFD < 7) {
                                if (diffRsCsdiff < 12) {
                                    if (ssDCval < 223) {
                                        if (ssDCval < 62) {
                                            if (AFD < 28) {
                                                return 0;
                                            }
                                            else if (AFD >= 28) {
                                                if (posBalance < 57) {
                                                    if (SC < 48) {
                                                        if (Rs < 17) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 17) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (SC >= 48) {
                                                        return 0;
                                                    }
                                                }
                                                else if (posBalance >= 57) {
                                                    if (negBalance < 44) {
                                                        return 0;
                                                    }
                                                    else if (negBalance >= 44) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                        else if (ssDCval >= 62) {
                                            return 0;
                                        }
                                    }
                                    else if (ssDCval >= 223) {
                                        if (AFD < 7) {
                                            if (TSC < 3) {
                                                return 0;
                                            }
                                            else if (TSC >= 3) {
                                                return 1;
                                            }
                                        }
                                        else if (AFD >= 7) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 12) {
                                    if (Cs < 27) {
                                        if (posBalance < 54) {
                                            return 0;
                                        }
                                        else if (posBalance >= 54) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 27) {
                                        if (ssDCval < 37) {
                                            if (diffMVdiffVal < 168) {
                                                if (Rs < 442) {
                                                    return 1;
                                                }
                                                else if (Rs >= 442) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffMVdiffVal >= 168) {
                                                return 0;
                                            }
                                        }
                                        else if (ssDCval >= 37) {
                                            if (diffTSC < -5) {
                                                if (MVDiff < 516) {
                                                    if (diffTSC < -6) {
                                                        return 0;
                                                    }
                                                    else if (diffTSC >= -6) {
                                                        if (Rs < 465) {
                                                            return 1;
                                                        }
                                                        else if (Rs >= 465) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (MVDiff >= 516) {
                                                    if (MVDiff < 601) {
                                                        if (Cs < 2009) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 2009) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (MVDiff >= 601) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (diffTSC >= -5) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffAFD >= 7) {
                                if (posBalance < 61) {
                                    if (SC < 187) {
                                        if (SC < 63) {
                                            if (AFD < 42) {
                                                return 0;
                                            }
                                            else if (AFD >= 42) {
                                                return 1;
                                            }
                                        }
                                        else if (SC >= 63) {
                                            return 1;
                                        }
                                    }
                                    else if (SC >= 187) {
                                        if (ssDCval < 60) {
                                            if (CsDiff < 15) {
                                                if (diffRsCsdiff < -47) {
                                                    return 1;
                                                }
                                                else if (diffRsCsdiff >= -47) {
                                                    if (TSC < 7) {
                                                        return 1;
                                                    }
                                                    else if (TSC >= 7) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (CsDiff >= 15) {
                                                return 1;
                                            }
                                        }
                                        else if (ssDCval >= 60) {
                                            if (diffAFD < 13) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 13) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (posBalance >= 61) {
                                    return 0;
                                }
                            }
                        }
                        else if (TSC >= 32) {
                            if (negBalance < 53) {
                                if (Rs < 563) {
                                    return 0;
                                }
                                else if (Rs >= 563) {
                                    return 1;
                                }
                            }
                            else if (negBalance >= 53) {
                                if (diffRsCsdiff < 126) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 126) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 4) {
                        if (CsDiff < 9) {
                            if (TSC < 9) {
                                if (Scindex < 2) {
                                    if (MVDiff < 233) {
                                        if (Rs < 14) {
                                            return 1;
                                        }
                                        else if (Rs >= 14) {
                                            return 0;
                                        }
                                    }
                                    else if (MVDiff >= 233) {
                                        if (diffMVdiffVal < 195) {
                                            if (Cs < 31) {
                                                return 0;
                                            }
                                            else if (Cs >= 31) {
                                                return 1;
                                            }
                                        }
                                        else if (diffMVdiffVal >= 195) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (Scindex >= 2) {
                                    return 0;
                                }
                            }
                            else if (TSC >= 9) {
                                if (diffRsCsdiff < 39) {
                                    if (Rs < 173) {
                                        if (RsCsDiff < 10) {
                                            if (refDCval < 27) {
                                                if (Rs < 32) {
                                                    return 1;
                                                }
                                                else if (Rs >= 32) {
                                                    return 0;
                                                }
                                            }
                                            else if (refDCval >= 27) {
                                                return 0;
                                            }
                                        }
                                        else if (RsCsDiff >= 10) {
                                            if (ssDCval < 136) {
                                                if (Cs < 450) {
                                                    if (refDCval < 65) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 65) {
                                                        if (RsCsDiff < 68) {
                                                            return 1;
                                                        }
                                                        else if (RsCsDiff >= 68) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (Cs >= 450) {
                                                    return 0;
                                                }
                                            }
                                            else if (ssDCval >= 136) {
                                                return 0;
                                            }
                                        }
                                    }
                                    else if (Rs >= 173) {
                                        if (refDCval < 25) {
                                            return 1;
                                        }
                                        else if (refDCval >= 25) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (diffRsCsdiff >= 39) {
                                    if (RsDiff < 8) {
                                        if (refDCval < 50) {
                                            if (Rs < 772) {
                                                return 1;
                                            }
                                            else if (Rs >= 772) {
                                                return 0;
                                            }
                                        }
                                        else if (refDCval >= 50) {
                                            if (MVDiff < 221) {
                                                return 0;
                                            }
                                            else if (MVDiff >= 221) {
                                                if (AFD < 24) {
                                                    return 0;
                                                }
                                                else if (AFD >= 24) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                    else if (RsDiff >= 8) {
                                        if (diffTSC < 12) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 12) {
                                            if (MVDiff < 521) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 521) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 9) {
                            if (MVDiff < 248) {
                                if (AFD < 59) {
                                    if (diffTSC < 11) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 11) {
                                        return 1;
                                    }
                                }
                                else if (AFD >= 59) {
                                    return 1;
                                }
                            }
                            else if (MVDiff >= 248) {
                                if (SC < 2190) {
                                    return 1;
                                }
                                else if (SC >= 2190) {
                                    if (diffAFD < 2) {
                                        return 1;
                                    }
                                    else if (diffAFD >= 2) {
                                        if (diffAFD < 13) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 13) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 14) {
        if (diffAFD < 24) {
            if (diffMVdiffVal < 55) {
                if (refDCval < 31) {
                    if (RsCsDiff < 21) {
                        return 0;
                    }
                    else if (RsCsDiff >= 21) {
                        if (ssDCval < 109) {
                            return 1;
                        }
                        else if (ssDCval >= 109) {
                            return 0;
                        }
                    }
                }
                else if (refDCval >= 31) {
                    if (diffTSC < 11) {
                        if (Rs < 83) {
                            if (Cs < 76) {
                                if (Rs < 39) {
                                    return 0;
                                }
                                else if (Rs >= 39) {
                                    return 1;
                                }
                            }
                            else if (Cs >= 76) {
                                return 0;
                            }
                        }
                        else if (Rs >= 83) {
                            if (ssDCval < 36) {
                                if (Rs < 460) {
                                    return 0;
                                }
                                else if (Rs >= 460) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 36) {
                                return 0;
                            }
                        }
                    }
                    else if (diffTSC >= 11) {
                        if (CsDiff < 15) {
                            if (negBalance < 21) {
                                return 0;
                            }
                            else if (negBalance >= 21) {
                                if (Cs < 799) {
                                    if (CsDiff < 3) {
                                        if (posBalance < 41) {
                                            return 1;
                                        }
                                        else if (posBalance >= 41) {
                                            return 0;
                                        }
                                    }
                                    else if (CsDiff >= 3) {
                                        if (AFD < 44) {
                                            return 1;
                                        }
                                        else if (AFD >= 44) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (Cs >= 799) {
                                    if (SC < 17401) {
                                        if (refDCval < 70) {
                                            if (negBalance < 48) {
                                                if (MVDiff < 268) {
                                                    return 0;
                                                }
                                                else if (MVDiff >= 268) {
                                                    if (Cs < 900) {
                                                        return 0;
                                                    }
                                                    else if (Cs >= 900) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (negBalance >= 48) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 70) {
                                            return 0;
                                        }
                                    }
                                    else if (SC >= 17401) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 15) {
                            return 0;
                        }
                    }
                }
            }
            else if (diffMVdiffVal >= 55) {
                if (ssDCval < 70) {
                    if (diffMVdiffVal < 109) {
                        if (gchDC < 15) {
                            if (MVDiff < 110) {
                                if (diffAFD < 23) {
                                    return 1;
                                }
                                else if (diffAFD >= 23) {
                                    if (Rs < 787) {
                                        return 0;
                                    }
                                    else if (Rs >= 787) {
                                        return 1;
                                    }
                                }
                            }
                            else if (MVDiff >= 110) {
                                if (negBalance < 60) {
                                    return 0;
                                }
                                else if (negBalance >= 60) {
                                    return 1;
                                }
                            }
                        }
                        else if (gchDC >= 15) {
                            if (gchDC < 26) {
                                if (gchDC < 17) {
                                    if (Rs < 1596) {
                                        return 1;
                                    }
                                    else if (Rs >= 1596) {
                                        return 0;
                                    }
                                }
                                else if (gchDC >= 17) {
                                    return 0;
                                }
                            }
                            else if (gchDC >= 26) {
                                if (Rs < 1903) {
                                    return 1;
                                }
                                else if (Rs >= 1903) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 109) {
                        if (gchDC < 7) {
                            return 1;
                        }
                        else if (gchDC >= 7) {
                            if (AFD < 27) {
                                return 1;
                            }
                            else if (AFD >= 27) {
                                if (RsCsDiff < 106) {
                                    if (refDCval < 147) {
                                        if (MVDiff < 1672) {
                                            if (SC < 449) {
                                                if (Rs < 69) {
                                                    return 0;
                                                }
                                                else if (Rs >= 69) {
                                                    return 1;
                                                }
                                            }
                                            else if (SC >= 449) {
                                                return 0;
                                            }
                                        }
                                        else if (MVDiff >= 1672) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 147) {
                                        return 1;
                                    }
                                }
                                else if (RsCsDiff >= 106) {
                                    if (Rs < 3009) {
                                        return 1;
                                    }
                                    else if (Rs >= 3009) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (ssDCval >= 70) {
                    if (SC < 586) {
                        if (refDCval < 126) {
                            if (diffAFD < 15) {
                                return 0;
                            }
                            else if (diffAFD >= 15) {
                                if (diffRsCsdiff < 10) {
                                    if (AFD < 32) {
                                        return 1;
                                    }
                                    else if (AFD >= 32) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 10) {
                                    return 1;
                                }
                            }
                        }
                        else if (refDCval >= 126) {
                            if (Rs < 262) {
                                return 0;
                            }
                            else if (Rs >= 262) {
                                return 1;
                            }
                        }
                    }
                    else if (SC >= 586) {
                        if (diffTSC < 7) {
                            if (posBalance < 95) {
                                return 0;
                            }
                            else if (posBalance >= 95) {
                                if (Rs < 1191) {
                                    return 1;
                                }
                                else if (Rs >= 1191) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffTSC >= 7) {
                            if (diffRsCsdiff < 89) {
                                if (diffMVdiffVal < 635) {
                                    if (SC < 891) {
                                        if (diffRsCsdiff < 21) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 21) {
                                            if (Rs < 297) {
                                                return 0;
                                            }
                                            else if (Rs >= 297) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (SC >= 891) {
                                        return 0;
                                    }
                                }
                                else if (diffMVdiffVal >= 635) {
                                    return 1;
                                }
                            }
                            else if (diffRsCsdiff >= 89) {
                                if (gchDC < 16) {
                                    if (Cs < 1448) {
                                        return 1;
                                    }
                                    else if (Cs >= 1448) {
                                        if (diffRsCsdiff < 1095) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 1095) {
                                            if (diffTSC < 10) {
                                                if (Cs < 2387) {
                                                    return 1;
                                                }
                                                else if (Cs >= 2387) {
                                                    return 0;
                                                }
                                            }
                                            else if (diffTSC >= 10) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                                else if (gchDC >= 16) {
                                    if (diffAFD < 18) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 18) {
                                        if (Cs < 1407) {
                                            if (diffMVdiffVal < 95) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 95) {
                                                return 1;
                                            }
                                        }
                                        else if (Cs >= 1407) {
                                            if (RsDiff < 33) {
                                                return 0;
                                            }
                                            else if (RsDiff >= 33) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffAFD >= 24) {
            if (MVDiff < 51) {
                if (diffTSC < 25) {
                    if (diffMVdiffVal < 39) {
                        return 0;
                    }
                    else if (diffMVdiffVal >= 39) {
                        return 1;
                    }
                }
                else if (diffTSC >= 25) {
                    return 1;
                }
            }
            else if (MVDiff >= 51) {
                if (negBalance < 30) {
                    if (diffMVdiffVal < 175) {
                        if (RsDiff < 11) {
                            if (SC < 476) {
                                if (diffMVdiffVal < 50) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 50) {
                                    return 1;
                                }
                            }
                            else if (SC >= 476) {
                                if (negBalance < 17) {
                                    if (ssDCval < 150) {
                                        return 0;
                                    }
                                    else if (ssDCval >= 150) {
                                        if (Rs < 774) {
                                            return 0;
                                        }
                                        else if (Rs >= 774) {
                                            return 1;
                                        }
                                    }
                                }
                                else if (negBalance >= 17) {
                                    if (Rs < 664) {
                                        return 1;
                                    }
                                    else if (Rs >= 664) {
                                        if (diffAFD < 39) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 39) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (RsDiff >= 11) {
                            if (posBalance < 125) {
                                if (diffTSC < 17) {
                                    if (SC < 3185) {
                                        return 1;
                                    }
                                    else if (SC >= 3185) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 17) {
                                    if (diffMVdiffVal < 49) {
                                        if (Rs < 1388) {
                                            return 1;
                                        }
                                        else if (Rs >= 1388) {
                                            return 0;
                                        }
                                    }
                                    else if (diffMVdiffVal >= 49) {
                                        if (RsCsDiff < 173) {
                                            return 0;
                                        }
                                        else if (RsCsDiff >= 173) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (posBalance >= 125) {
                                return 0;
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 175) {
                        if (diffRsCsdiff < 5) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 5) {
                            if (refDCval < 57) {
                                if (RsDiff < 5) {
                                    if (Rs < 335) {
                                        if (Rs < 30) {
                                            return 0;
                                        }
                                        else if (Rs >= 30) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 335) {
                                        return 0;
                                    }
                                }
                                else if (RsDiff >= 5) {
                                    if (posBalance < 126) {
                                        if (MVDiff < 416) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 416) {
                                            if (posBalance < 105) {
                                                if (Rs < 1934) {
                                                    return 0;
                                                }
                                                else if (Rs >= 1934) {
                                                    return 1;
                                                }
                                            }
                                            else if (posBalance >= 105) {
                                                if (refDCval < 52) {
                                                    return 1;
                                                }
                                                else if (refDCval >= 52) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (posBalance >= 126) {
                                        return 0;
                                    }
                                }
                            }
                            else if (refDCval >= 57) {
                                return 1;
                            }
                        }
                    }
                }
                else if (negBalance >= 30) {
                    if (diffMVdiffVal < 55) {
                        if (RsDiff < 6) {
                            if (refDCval < 40) {
                                return 1;
                            }
                            else if (refDCval >= 40) {
                                return 0;
                            }
                        }
                        else if (RsDiff >= 6) {
                            if (ssDCval < 105) {
                                if (AFD < 33) {
                                    return 0;
                                }
                                else if (AFD >= 33) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 105) {
                                if (MVDiff < 136) {
                                    if (negBalance < 34) {
                                        return 1;
                                    }
                                    else if (negBalance >= 34) {
                                        return 0;
                                    }
                                }
                                else if (MVDiff >= 136) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 55) {
                        if (CsDiff < 9) {
                            if (negBalance < 50) {
                                if (Scindex < 3) {
                                    return 1;
                                }
                                else if (Scindex >= 3) {
                                    if (AFD < 55) {
                                        if (Cs < 251) {
                                            return 0;
                                        }
                                        else if (Cs >= 251) {
                                            if (MVDiff < 619) {
                                                if (diffMVdiffVal < 144) {
                                                    if (RsDiff < 10) {
                                                        if (AFD < 31) {
                                                            return 1;
                                                        }
                                                        else if (AFD >= 31) {
                                                            return 0;
                                                        }
                                                    }
                                                    else if (RsDiff >= 10) {
                                                        return 1;
                                                    }
                                                }
                                                else if (diffMVdiffVal >= 144) {
                                                    return 1;
                                                }
                                            }
                                            else if (MVDiff >= 619) {
                                                if (Cs < 638) {
                                                    return 1;
                                                }
                                                else if (Cs >= 638) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (AFD >= 55) {
                                        if (diffTSC < 35) {
                                            return 0;
                                        }
                                        else if (diffTSC >= 35) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 50) {
                                if (gchDC < 22) {
                                    return 1;
                                }
                                else if (gchDC >= 22) {
                                    if (AFD < 41) {
                                        return 1;
                                    }
                                    else if (AFD >= 41) {
                                        if (diffMVdiffVal < 124) {
                                            if (MVDiff < 129) {
                                                return 1;
                                            }
                                            else if (MVDiff >= 129) {
                                                if (TSC < 30) {
                                                    if (Rs < 1130) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 1130) {
                                                        return 0;
                                                    }
                                                }
                                                else if (TSC >= 30) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (diffMVdiffVal >= 124) {
                                            if (negBalance < 68) {
                                                if (SC < 939) {
                                                    return 0;
                                                }
                                                else if (SC >= 939) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 68) {
                                                if (RsCsDiff < 5) {
                                                    return 0;
                                                }
                                                else if (RsCsDiff >= 5) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (CsDiff >= 9) {
                            if (diffRsCsdiff < 62) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 62) {
                                if (diffAFD < 38) {
                                    if (MVDiff < 128) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 128) {
                                        if (RsDiff < 21) {
                                            if (diffMVdiffVal < 87) {
                                                return 0;
                                            }
                                            else if (diffMVdiffVal >= 87) {
                                                if (MVDiff < 1121) {
                                                    return 1;
                                                }
                                                else if (MVDiff >= 1121) {
                                                    if (Cs < 1259) {
                                                        return 1;
                                                    }
                                                    else if (Cs >= 1259) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                        else if (RsDiff >= 21) {
                                            if (gchDC < 79) {
                                                if (diffRsCsdiff < 494) {
                                                    return 0;
                                                }
                                                else if (diffRsCsdiff >= 494) {
                                                    if (AFD < 65) {
                                                        if (MVDiff < 129) {
                                                            return 0;
                                                        }
                                                        else if (MVDiff >= 129) {
                                                            if (diffMVdiffVal < 107) {
                                                                if (Rs < 1994) {
                                                                    return 0;
                                                                }
                                                                else if (Rs >= 1994) {
                                                                    return 1;
                                                                }
                                                            }
                                                            else if (diffMVdiffVal >= 107) {
                                                                return 1;
                                                            }
                                                        }
                                                    }
                                                    else if (AFD >= 65) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 79) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (diffAFD >= 38) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return 0;
}
bool SCDetect15(mfxU32 MVDiff, mfxU32 RsCsDiff, mfxU32 Rs, mfxU32 gchDC, mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 refDCval, mfxU32 TSC, mfxI32 diffRsCsdiff, mfxU32 posBalance, mfxU32 Cs, mfxU32 TSCindex, mfxU32 /*Scindex*/, mfxU32 AFD, mfxU32 SC, mfxU32 RsDiff, mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxI32 diffMVdiffVal) {
    if (diffAFD < 13) {
        if (diffTSC < 4) {
            if (MVDiff < 168) {
                if (ssDCval < 23) {
                    if (diffAFD < 3) {
                        if (diffRsCsdiff < 7) {
                            if (RsCsDiff < 12) {
                                return 0;
                            }
                            else if (RsCsDiff >= 12) {
                                if (MVDiff < 8) {
                                    return 1;
                                }
                                else if (MVDiff >= 8) {
                                    return 0;
                                }
                            }
                        }
                        else if (diffRsCsdiff >= 7) {
                            if (MVDiff < 7) {
                                return 1;
                            }
                            else if (MVDiff >= 7) {
                                return 0;
                            }
                        }
                    }
                    else if (diffAFD >= 3) {
                        if (CsDiff < 33) {
                            return 1;
                        }
                        else if (CsDiff >= 33) {
                            return 0;
                        }
                    }
                }
                else if (ssDCval >= 23) {
                    if (RsCsDiff < 3703) {
                        if (negBalance < 106) {
                            if (diffMVdiffVal < 66) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 66) {
                                if (AFD < 37) {
                                    if (ssDCval < 26) {
                                        if (Rs < 9) {
                                            return 0;
                                        }
                                        else if (Rs >= 9) {
                                            return 1;
                                        }
                                    }
                                    else if (ssDCval >= 26) {
                                        return 0;
                                    }
                                }
                                else if (AFD >= 37) {
                                    return 1;
                                }
                            }
                        }
                        else if (negBalance >= 106) {
                            if (ssDCval < 212) {
                                if (diffMVdiffVal < 108) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 108) {
                                    if (Cs < 210) {
                                        return 1;
                                    }
                                    else if (Cs >= 210) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 212) {
                                return 1;
                            }
                        }
                    }
                    else if (RsCsDiff >= 3703) {
                        if (Cs < 3803) {
                            if (refDCval < 85) {
                                return 0;
                            }
                            else if (refDCval >= 85) {
                                if (MVDiff < 84) {
                                    return 1;
                                }
                                else if (MVDiff >= 84) {
                                    if (Rs < 2530) {
                                        if (Rs < 330) {
                                            return 0;
                                        }
                                        else if (Rs >= 330) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 2530) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 3803) {
                            return 0;
                        }
                    }
                }
            }
            else if (MVDiff >= 168) {
                if (diffAFD < 6) {
                    if (gchDC < 6) {
                        if (diffRsCsdiff < 17) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 17) {
                            if (TSC < 2) {
                                return 1;
                            }
                            else if (TSC >= 2) {
                                if (refDCval < 47) {
                                    if (Rs < 1476) {
                                        return 0;
                                    }
                                    else if (Rs >= 1476) {
                                        return 1;
                                    }
                                }
                                else if (refDCval >= 47) {
                                    return 0;
                                }
                            }
                        }
                    }
                    else if (gchDC >= 6) {
                        if (TSC < 32) {
                            if (gchDC < 173) {
                                if (MVDiff < 217) {
                                    return 0;
                                }
                                else if (MVDiff >= 217) {
                                    if (ssDCval < 37) {
                                        if (diffAFD < -7) {
                                            if (MVDiff < 348) {
                                                if (AFD < 20) {
                                                    if (SC < 3743) {
                                                        return 0;
                                                    }
                                                    else if (SC >= 3743) {
                                                        return 1;
                                                    }
                                                }
                                                else if (AFD >= 20) {
                                                    if (refDCval < 113) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 113) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                            else if (MVDiff >= 348) {
                                                return 0;
                                            }
                                        }
                                        else if (diffAFD >= -7) {
                                            if (diffAFD < 3) {
                                                return 0;
                                            }
                                            else if (diffAFD >= 3) {
                                                if (TSC < 4) {
                                                    return 1;
                                                }
                                                else if (TSC >= 4) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                    else if (ssDCval >= 37) {
                                        if (TSC < 13) {
                                            if (Cs < 1662) {
                                                return 0;
                                            }
                                            else if (Cs >= 1662) {
                                                if (SC < 3496) {
                                                    return 1;
                                                }
                                                else if (SC >= 3496) {
                                                    return 0;
                                                }
                                            }
                                        }
                                        else if (TSC >= 13) {
                                            if (posBalance < 53) {
                                                return 0;
                                            }
                                            else if (posBalance >= 53) {
                                                if (negBalance < 46) {
                                                    return 0;
                                                }
                                                else if (negBalance >= 46) {
                                                    if (diffTSC < 1) {
                                                        if (gchDC < 11) {
                                                            if (gchDC < 8) {
                                                                return 1;
                                                            }
                                                            else if (gchDC >= 8) {
                                                                if (TSC < 19) {
                                                                    return 0;
                                                                }
                                                                else if (TSC >= 19) {
                                                                    return 1;
                                                                }
                                                            }
                                                        }
                                                        else if (gchDC >= 11) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (diffTSC >= 1) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (gchDC >= 173) {
                                return 1;
                            }
                        }
                        else if (TSC >= 32) {
                            if (Rs < 1355) {
                                return 0;
                            }
                            else if (Rs >= 1355) {
                                if (Rs < 3635) {
                                    return 1;
                                }
                                else if (Rs >= 3635) {
                                    if (Cs < 7860) {
                                        return 0;
                                    }
                                    else if (Cs >= 7860) {
                                        return 1;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (diffAFD >= 6) {
                    if (ssDCval < 42) {
                        if (MVDiff < 626) {
                            if (gchDC < 14) {
                                if (gchDC < 8) {
                                    return 1;
                                }
                                else if (gchDC >= 8) {
                                    if (refDCval < 46) {
                                        return 0;
                                    }
                                    else if (refDCval >= 46) {
                                        return 1;
                                    }
                                }
                            }
                            else if (gchDC >= 14) {
                                return 1;
                            }
                        }
                        else if (MVDiff >= 626) {
                            return 0;
                        }
                    }
                    else if (ssDCval >= 42) {
                        if (refDCval < 100) {
                            if (Rs < 36) {
                                if (AFD < 53) {
                                    return 1;
                                }
                                else if (AFD >= 53) {
                                    return 0;
                                }
                            }
                            else if (Rs >= 36) {
                                if (TSC < 20) {
                                    if (diffRsCsdiff < 277) {
                                        return 0;
                                    }
                                    else if (diffRsCsdiff >= 277) {
                                        if (RsDiff < 19) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 19) {
                                            return 0;
                                        }
                                    }
                                }
                                else if (TSC >= 20) {
                                    if (gchDC < 7) {
                                        if (SC < 1262) {
                                            return 0;
                                        }
                                        else if (SC >= 1262) {
                                            return 1;
                                        }
                                    }
                                    else if (gchDC >= 7) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (refDCval >= 100) {
                            return 0;
                        }
                    }
                }
            }
        }
        else if (diffTSC >= 4) {
            if (ssDCval < 53) {
                if (negBalance < 48) {
                    if (diffMVdiffVal < 83) {
                        if (refDCval < 24) {
                            if (SC < 2021) {
                                if (Rs < 282) {
                                    return 0;
                                }
                                else if (Rs >= 282) {
                                    if (Cs < 444) {
                                        return 1;
                                    }
                                    else if (Cs >= 444) {
                                        return 0;
                                    }
                                }
                            }
                            else if (SC >= 2021) {
                                return 1;
                            }
                        }
                        else if (refDCval >= 24) {
                            if (RsDiff < 23) {
                                return 0;
                            }
                            else if (RsDiff >= 23) {
                                return 1;
                            }
                        }
                    }
                    else if (diffMVdiffVal >= 83) {
                        if (gchDC < 7) {
                            return 1;
                        }
                        else if (gchDC >= 7) {
                            if (MVDiff < 190) {
                                if (AFD < 16) {
                                    return 1;
                                }
                                else if (AFD >= 16) {
                                    return 0;
                                }
                            }
                            else if (MVDiff >= 190) {
                                return 0;
                            }
                        }
                    }
                }
                else if (negBalance >= 48) {
                    if (posBalance < 21) {
                        if (AFD < 43) {
                            return 0;
                        }
                        else if (AFD >= 43) {
                            return 1;
                        }
                    }
                    else if (posBalance >= 21) {
                        if (Cs < 219) {
                            if (Cs < 15) {
                                return 0;
                            }
                            else if (Cs >= 15) {
                                if (MVDiff < 78) {
                                    return 0;
                                }
                                else if (MVDiff >= 78) {
                                    return 1;
                                }
                            }
                        }
                        else if (Cs >= 219) {
                            if (diffMVdiffVal < 41) {
                                if (posBalance < 66) {
                                    return 0;
                                }
                                else if (posBalance >= 66) {
                                    return 1;
                                }
                            }
                            else if (diffMVdiffVal >= 41) {
                                if (TSCindex < 9) {
                                    if (MVDiff < 317) {
                                        return 0;
                                    }
                                    else if (MVDiff >= 317) {
                                        return 1;
                                    }
                                }
                                else if (TSCindex >= 9) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
            }
            else if (ssDCval >= 53) {
                if (gchDC < 22) {
                    if (MVDiff < 63) {
                        return 0;
                    }
                    else if (MVDiff >= 63) {
                        if (diffRsCsdiff < 12) {
                            return 0;
                        }
                        else if (diffRsCsdiff >= 12) {
                            if (Rs < 174) {
                                return 1;
                            }
                            else if (Rs >= 174) {
                                if (refDCval < 80) {
                                    if (refDCval < 73) {
                                        return 0;
                                    }
                                    else if (refDCval >= 73) {
                                        if (diffRsCsdiff < 46) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 46) {
                                            if (ssDCval < 91) {
                                                return 1;
                                            }
                                            else if (ssDCval >= 91) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                                else if (refDCval >= 80) {
                                    return 0;
                                }
                            }
                        }
                    }
                }
                else if (gchDC >= 22) {
                    if (AFD < 59) {
                        if (Rs < 200) {
                            if (refDCval < 41) {
                                return 1;
                            }
                            else if (refDCval >= 41) {
                                if (RsDiff < 5) {
                                    return 0;
                                }
                                else if (RsDiff >= 5) {
                                    if (diffMVdiffVal < 230) {
                                        return 0;
                                    }
                                    else if (diffMVdiffVal >= 230) {
                                        return 1;
                                    }
                                }
                            }
                        }
                        else if (Rs >= 200) {
                            if (diffRsCsdiff < 60) {
                                return 0;
                            }
                            else if (diffRsCsdiff >= 60) {
                                if (diffMVdiffVal < 74) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 74) {
                                    if (negBalance < 37) {
                                        return 0;
                                    }
                                    else if (negBalance >= 37) {
                                        if (ssDCval < 149) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 149) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (AFD >= 59) {
                        if (SC < 2129) {
                            return 0;
                        }
                        else if (SC >= 2129) {
                            return 1;
                        }
                    }
                }
            }
        }
    }
    else if (diffAFD >= 13) {
        if (diffMVdiffVal < 57) {
            if (refDCval < 27) {
                if (diffAFD < 17) {
                    return 0;
                }
                else if (diffAFD >= 17) {
                    if (gchDC < 118) {
                        if (CsDiff < 12) {
                            if (Cs < 537) {
                                return 1;
                            }
                            else if (Cs >= 537) {
                                return 0;
                            }
                        }
                        else if (CsDiff >= 12) {
                            return 1;
                        }
                    }
                    else if (gchDC >= 118) {
                        return 0;
                    }
                }
            }
            else if (refDCval >= 27) {
                if (diffAFD < 25) {
                    if (Cs < 206) {
                        if (SC < 2540) {
                            if (MVDiff < 151) {
                                return 0;
                            }
                            else if (MVDiff >= 151) {
                                if (TSC < 13) {
                                    if (Rs < 53) {
                                        return 0;
                                    }
                                    else if (Rs >= 53) {
                                        return 1;
                                    }
                                }
                                else if (TSC >= 13) {
                                    if (AFD < 28) {
                                        return 1;
                                    }
                                    else if (AFD >= 28) {
                                        if (MVDiff < 783) {
                                            return 0;
                                        }
                                        else if (MVDiff >= 783) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (SC >= 2540) {
                            return 1;
                        }
                    }
                    else if (Cs >= 206) {
                        if (MVDiff < 681) {
                            if (gchDC < 6) {
                                if (diffMVdiffVal < 24) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 24) {
                                    if (refDCval < 69) {
                                        if (Rs < 561) {
                                            return 0;
                                        }
                                        else if (Rs >= 561) {
                                            return 1;
                                        }
                                    }
                                    else if (refDCval >= 69) {
                                        return 0;
                                    }
                                }
                            }
                            else if (gchDC >= 6) {
                                if (RsDiff < 10) {
                                    if (gchDC < 7) {
                                        if (Rs < 608) {
                                            return 1;
                                        }
                                        else if (Rs >= 608) {
                                            return 0;
                                        }
                                    }
                                    else if (gchDC >= 7) {
                                        return 0;
                                    }
                                }
                                else if (RsDiff >= 10) {
                                    if (ssDCval < 65) {
                                        return 1;
                                    }
                                    else if (ssDCval >= 65) {
                                        if (TSC < 15) {
                                            return 0;
                                        }
                                        else if (TSC >= 15) {
                                            if (gchDC < 19) {
                                                if (RsDiff < 22) {
                                                    if (RsDiff < 12) {
                                                        if (AFD < 36) {
                                                            return 0;
                                                        }
                                                        else if (AFD >= 36) {
                                                            return 1;
                                                        }
                                                    }
                                                    else if (RsDiff >= 12) {
                                                        return 1;
                                                    }
                                                }
                                                else if (RsDiff >= 22) {
                                                    if (SC < 17420) {
                                                        return 0;
                                                    }
                                                    else if (SC >= 17420) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                            else if (gchDC >= 19) {
                                                if (Cs < 1606) {
                                                    return 0;
                                                }
                                                else if (Cs >= 1606) {
                                                    if (refDCval < 120) {
                                                        return 0;
                                                    }
                                                    else if (refDCval >= 120) {
                                                        return 1;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (MVDiff >= 681) {
                            return 1;
                        }
                    }
                }
                else if (diffAFD >= 25) {
                    if (TSC < 14) {
                        return 0;
                    }
                    else if (TSC >= 14) {
                        if (negBalance < 28) {
                            return 0;
                        }
                        else if (negBalance >= 28) {
                            if (ssDCval < 97) {
                                if (Cs < 191) {
                                    return 0;
                                }
                                else if (Cs >= 191) {
                                    if (posBalance < 36) {
                                        if (Rs < 1144) {
                                            if (Cs < 561) {
                                                return 0;
                                            }
                                            else if (Cs >= 561) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 1144) {
                                            return 1;
                                        }
                                    }
                                    else if (posBalance >= 36) {
                                        return 1;
                                    }
                                }
                            }
                            else if (ssDCval >= 97) {
                                if (RsDiff < 11) {
                                    return 0;
                                }
                                else if (RsDiff >= 11) {
                                    if (Cs < 3677) {
                                        if (Rs < 1694) {
                                            if (Cs < 1113) {
                                                return 0;
                                            }
                                            else if (Cs >= 1113) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 1694) {
                                            return 1;
                                        }
                                    }
                                    else if (Cs >= 3677) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        else if (diffMVdiffVal >= 57) {
            if (diffAFD < 23) {
                if (refDCval < 88) {
                    if (posBalance < 63) {
                        if (MVDiff < 505) {
                            if (diffTSC < 8) {
                                if (ssDCval < 36) {
                                    return 1;
                                }
                                else if (ssDCval >= 36) {
                                    if (negBalance < 50) {
                                        return 0;
                                    }
                                    else if (negBalance >= 50) {
                                        if (diffAFD < 14) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 14) {
                                            if (Cs < 573) {
                                                return 1;
                                            }
                                            else if (Cs >= 573) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (diffTSC >= 8) {
                                return 1;
                            }
                        }
                        else if (MVDiff >= 505) {
                            if (ssDCval < 48) {
                                if (diffRsCsdiff < -11) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= -11) {
                                    return 1;
                                }
                            }
                            else if (ssDCval >= 48) {
                                return 0;
                            }
                        }
                    }
                    else if (posBalance >= 63) {
                        if (Rs < 196) {
                            if (posBalance < 65) {
                                return 0;
                            }
                            else if (posBalance >= 65) {
                                if (diffMVdiffVal < 106) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 106) {
                                    if (TSC < 13) {
                                        return 1;
                                    }
                                    else if (TSC >= 13) {
                                        if (ssDCval < 80) {
                                            return 1;
                                        }
                                        else if (ssDCval >= 80) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                        else if (Rs >= 196) {
                            if (TSC < 20) {
                                if (diffMVdiffVal < 74) {
                                    return 0;
                                }
                                else if (diffMVdiffVal >= 74) {
                                    if (diffTSC < 8) {
                                        if (Rs < 533) {
                                            if (Rs < 379) {
                                                return 0;
                                            }
                                            else if (Rs >= 379) {
                                                return 1;
                                            }
                                        }
                                        else if (Rs >= 533) {
                                            return 0;
                                        }
                                    }
                                    else if (diffTSC >= 8) {
                                        if (CsDiff < 3) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 3) {
                                            if (negBalance < 22) {
                                                if (RsDiff < 9) {
                                                    if (refDCval < 35) {
                                                        return 1;
                                                    }
                                                    else if (refDCval >= 35) {
                                                        if (Cs < 393) {
                                                            return 1;
                                                        }
                                                        else if (Cs >= 393) {
                                                            return 0;
                                                        }
                                                    }
                                                }
                                                else if (RsDiff >= 9) {
                                                    return 1;
                                                }
                                            }
                                            else if (negBalance >= 22) {
                                                if (AFD < 33) {
                                                    return 1;
                                                }
                                                else if (AFD >= 33) {
                                                    if (Rs < 676) {
                                                        return 1;
                                                    }
                                                    else if (Rs >= 676) {
                                                        return 0;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            else if (TSC >= 20) {
                                if (CsDiff < 7) {
                                    return 0;
                                }
                                else if (CsDiff >= 7) {
                                    if (Rs < 2122) {
                                        if (refDCval < 55) {
                                            if (Rs < 1235) {
                                                return 0;
                                            }
                                            else if (Rs >= 1235) {
                                                return 1;
                                            }
                                        }
                                        else if (refDCval >= 55) {
                                            return 1;
                                        }
                                    }
                                    else if (Rs >= 2122) {
                                        return 0;
                                    }
                                }
                            }
                        }
                    }
                }
                else if (refDCval >= 88) {
                    if (diffRsCsdiff < 88) {
                        if (Cs < 601) {
                            if (ssDCval < 89) {
                                if (posBalance < 1) {
                                    return 0;
                                }
                                else if (posBalance >= 1) {
                                    if (AFD < 53) {
                                        return 1;
                                    }
                                    else if (AFD >= 53) {
                                        return 0;
                                    }
                                }
                            }
                            else if (ssDCval >= 89) {
                                if (diffRsCsdiff < 18) {
                                    return 0;
                                }
                                else if (diffRsCsdiff >= 18) {
                                    if (TSC < 17) {
                                        return 1;
                                    }
                                    else if (TSC >= 17) {
                                        return 0;
                                    }
                                }
                            }
                        }
                        else if (Cs >= 601) {
                            return 0;
                        }
                    }
                    else if (diffRsCsdiff >= 88) {
                        if (SC < 1850) {
                            if (posBalance < 100) {
                                return 1;
                            }
                            else if (posBalance >= 100) {
                                return 0;
                            }
                        }
                        else if (SC >= 1850) {
                            if (TSC < 16) {
                                return 0;
                            }
                            else if (TSC >= 16) {
                                if (AFD < 34) {
                                    return 1;
                                }
                                else if (AFD >= 34) {
                                    if (diffTSC < 12) {
                                        return 0;
                                    }
                                    else if (diffTSC >= 12) {
                                        if (Cs < 4720) {
                                            return 1;
                                        }
                                        else if (Cs >= 4720) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            else if (diffAFD >= 23) {
                if (posBalance < 81) {
                    if (diffTSC < 10) {
                        if (ssDCval < 80) {
                            if (Rs < 1936) {
                                return 1;
                            }
                            else if (Rs >= 1936) {
                                if (Cs < 1930) {
                                    return 0;
                                }
                                else if (Cs >= 1930) {
                                    return 1;
                                }
                            }
                        }
                        else if (ssDCval >= 80) {
                            if (diffMVdiffVal < 189) {
                                return 0;
                            }
                            else if (diffMVdiffVal >= 189) {
                                if (diffRsCsdiff < 25) {
                                    if (Rs < 157) {
                                        return 1;
                                    }
                                    else if (Rs >= 157) {
                                        return 0;
                                    }
                                }
                                else if (diffRsCsdiff >= 25) {
                                    return 1;
                                }
                            }
                        }
                    }
                    else if (diffTSC >= 10) {
                        if (gchDC < 38) {
                            if (negBalance < 47) {
                                if (diffRsCsdiff < 79) {
                                    if (posBalance < 73) {
                                        if (MVDiff < 86) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 86) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 73) {
                                        return 1;
                                    }
                                }
                                else if (diffRsCsdiff >= 79) {
                                    if (posBalance < 48) {
                                        if (Rs < 1858) {
                                            return 1;
                                        }
                                        else if (Rs >= 1858) {
                                            return 0;
                                        }
                                    }
                                    else if (posBalance >= 48) {
                                        if (MVDiff < 663) {
                                            return 1;
                                        }
                                        else if (MVDiff >= 663) {
                                            if (AFD < 39) {
                                                return 1;
                                            }
                                            else if (AFD >= 39) {
                                                return 0;
                                            }
                                        }
                                    }
                                }
                            }
                            else if (negBalance >= 47) {
                                if (diffTSC < 12) {
                                    if (AFD < 65) {
                                        if (Cs < 1599) {
                                            return 1;
                                        }
                                        else if (Cs >= 1599) {
                                            if (Rs < 1653) {
                                                return 0;
                                            }
                                            else if (Rs >= 1653) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (AFD >= 65) {
                                        return 0;
                                    }
                                }
                                else if (diffTSC >= 12) {
                                    if (MVDiff < 1241) {
                                        if (TSC < 33) {
                                            return 1;
                                        }
                                        else if (TSC >= 33) {
                                            if (diffTSC < 17) {
                                                return 0;
                                            }
                                            else if (diffTSC >= 17) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (MVDiff >= 1241) {
                                        if (RsDiff < 15) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 15) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                        }
                        else if (gchDC >= 38) {
                            if (diffRsCsdiff < 60) {
                                if (SC < 522) {
                                    return 1;
                                }
                                else if (SC >= 522) {
                                    if (RsDiff < 5) {
                                        return 0;
                                    }
                                    else if (RsDiff >= 5) {
                                        if (RsDiff < 6) {
                                            return 1;
                                        }
                                        else if (RsDiff >= 6) {
                                            return 0;
                                        }
                                    }
                                }
                            }
                            else if (diffRsCsdiff >= 60) {
                                if (gchDC < 44) {
                                    if (diffMVdiffVal < 131) {
                                        if (RsDiff < 16) {
                                            return 0;
                                        }
                                        else if (RsDiff >= 16) {
                                            if (SC < 1462) {
                                                return 0;
                                            }
                                            else if (SC >= 1462) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (diffMVdiffVal >= 131) {
                                        return 1;
                                    }
                                }
                                else if (gchDC >= 44) {
                                    return 1;
                                }
                            }
                        }
                    }
                }
                else if (posBalance >= 81) {
                    if (negBalance < 33) {
                        if (diffMVdiffVal < 175) {
                            if (CsDiff < 9) {
                                if (refDCval < 50) {
                                    return 0;
                                }
                                else if (refDCval >= 50) {
                                    if (Cs < 525) {
                                        if (CsDiff < 3) {
                                            return 0;
                                        }
                                        else if (CsDiff >= 3) {
                                            if (TSC < 17) {
                                                return 0;
                                            }
                                            else if (TSC >= 17) {
                                                return 1;
                                            }
                                        }
                                    }
                                    else if (Cs >= 525) {
                                        if (diffAFD < 41) {
                                            return 0;
                                        }
                                        else if (diffAFD >= 41) {
                                            return 1;
                                        }
                                    }
                                }
                            }
                            else if (CsDiff >= 9) {
                                if (SC < 2468) {
                                    return 1;
                                }
                                else if (SC >= 2468) {
                                    if (RsDiff < 12) {
                                        if (diffRsCsdiff < 361) {
                                            return 0;
                                        }
                                        else if (diffRsCsdiff >= 361) {
                                            return 1;
                                        }
                                    }
                                    else if (RsDiff >= 12) {
                                        if (diffRsCsdiff < 916) {
                                            if (AFD < 44) {
                                                return 0;
                                            }
                                            else if (AFD >= 44) {
                                                return 1;
                                            }
                                        }
                                        else if (diffRsCsdiff >= 916) {
                                            if (negBalance < 2) {
                                                return 0;
                                            }
                                            else if (negBalance >= 2) {
                                                return 1;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        else if (diffMVdiffVal >= 175) {
                            if (CsDiff < 3) {
                                if (RsCsDiff < 5) {
                                    return 0;
                                }
                                else if (RsCsDiff >= 5) {
                                    if (diffAFD < 32) {
                                        return 0;
                                    }
                                    else if (diffAFD >= 32) {
                                        return 1;
                                    }
                                }
                            }
                            else if (CsDiff >= 3) {
                                if (negBalance < 7) {
                                    if (gchDC < 69) {
                                        return 0;
                                    }
                                    else if (gchDC >= 69) {
                                        if (CsDiff < 6) {
                                            if (Rs < 2) {
                                                return 1;
                                            }
                                            else if (Rs >= 2) {
                                                return 0;
                                            }
                                        }
                                        else if (CsDiff >= 6) {
                                            if (diffMVdiffVal < 988) {
                                                return 1;
                                            }
                                            else if (diffMVdiffVal >= 988) {
                                                if (Rs < 1861) {
                                                    return 1;
                                                }
                                                else if (Rs >= 1861) {
                                                    return 0;
                                                }
                                            }
                                        }
                                    }
                                }
                                else if (negBalance >= 7) {
                                    if (MVDiff < 364) {
                                        return 1;
                                    }
                                    else if (MVDiff >= 364) {
                                        if (posBalance < 96) {
                                            if (refDCval < 46) {
                                                return 0;
                                            }
                                            else if (refDCval >= 46) {
                                                return 1;
                                            }
                                        }
                                        else if (posBalance >= 96) {
                                            if (AFD < 57) {
                                                return 1;
                                            }
                                            else if (AFD >= 57) {
                                                if (diffMVdiffVal < 346) {
                                                    return 0;
                                                }
                                                else if (diffMVdiffVal >= 346) {
                                                    return 1;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else if (negBalance >= 33) {
                        if (CsDiff < 8) {
                            if (Cs < 767) {
                                return 1;
                            }
                            else if (Cs >= 767) {
                                if (AFD < 61) {
                                    return 0;
                                }
                                else if (AFD >= 61) {
                                    return 1;
                                }
                            }
                        }
                        else if (CsDiff >= 8) {
                            return 1;
                        }
                    }
                }
            }
        }
    }
    return 0;
}

bool SCDetectRF(
    mfxI32 diffMVdiffVal, mfxU32 RsCsDiff, mfxU32 MVDiff, mfxU32 Rs, mfxU32 AFD,
    mfxU32 CsDiff, mfxI32 diffTSC, mfxU32 TSC, mfxU32 gchDC, mfxI32 diffRsCsdiff,
    mfxU32 posBalance, mfxU32 SC, mfxU32 TSCindex, mfxU32 Scindex, mfxU32 Cs,
    mfxI32 diffAFD, mfxU32 negBalance, mfxU32 ssDCval, mfxU32 refDCval, mfxU32 RsDiff,
    mfxU8 control) {
    mfxU8 sum = 0;
    sum += SCDetect1(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect2(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect3(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect4(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect5(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect6(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect7(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect8(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect9(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectA(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectB(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectC(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectD(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectE(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetectF(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect10(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect11(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect12(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect13(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect14(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    sum += SCDetect15(MVDiff, RsCsDiff, Rs, gchDC, CsDiff, diffTSC, refDCval, TSC, diffRsCsdiff, posBalance, Cs, TSCindex, Scindex, AFD, SC, RsDiff, diffAFD, negBalance, ssDCval, diffMVdiffVal);
    return(sum > RF_DECISION_LEVEL + control);
}
